package org.eso.oca.data;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;

import org.eso.oca.properties.OCAProperties;
import org.eso.oca.properties.OCAPropertiesException;


/**
 * 
 * Class CalibTablesDAO
 * 
 * The ConnectionManager provides connections and manages transactions transparently.
 * <br>
 * It is a singleton, you get its instance with the getInstance() method.
 * All of the DAO classes use the Manager to get database connections.
 * 
 * Before doing any operation, you must pass jdbc driver/url/username/password.
 * 
 * @version $Revision: 1.8 $
 * @author pnunes, Apr 1, 2005 
 * 
 */
public class DbConnManager {
    
    public static final String RAW_DB_ALIAS = "raw";
    public static final String CALIB_DB_ALIAS = "calib";
    public static final String DIC_DB_ALIAS = "dic";
    
    private static DbConnManager manager_instance = new DbConnManager();
    
    /*private static InheritableThreadLocal calib_trans_conn = new InheritableThreadLocal();
    private static InheritableThreadLocal obs_trans_conn = new InheritableThreadLocal();
    private static InheritableThreadLocal dic_trans_conn = new InheritableThreadLocal();*/
    private static Connection calib_trans_conn=null;// = new Connection();
    private static Connection obs_trans_conn=null;// = new Connection();
    private static Connection dic_trans_conn=null;// = new Connection();
               
    /*
     * Database connection properties
     */
    
    // calib
    private String calib_jdbc_driver = null;
    private String calib_jdbc_url = null;
    private String calib_jdbc_username = null;
    private String calib_jdbc_password = null;
    
    // obs
    private String obs_jdbc_driver = null;
    private String obs_jdbc_url = null;
    private String obs_jdbc_username = null;
    private String obs_jdbc_password = null;
    
    // dic
    private String dic_jdbc_driver = null;
    private String dic_jdbc_url = null;
    private String dic_jdbc_username = null;
    private String dic_jdbc_password = null;

    /*
     * Properties to be read from the configuration file
     */
    private static final String calib_jdbc_driver_property = "CALIB_JDBC_DRIVER" ;
    private static final String calib_jdbc_url_property= "CALIB_JDBC_URL";
    private static final String calib_jdbc_username_property= "CALIB_JDBC_USERNAME";
    private static final String calib_jdbc_password_property= "CALIB_JDBC_PASSWORD";
    
    private static final String obs_jdbc_driver_property = "RAW_JDBC_DRIVER" ;
    private static final String obs_jdbc_url_property= "RAW_JDBC_URL";
    private static final String obs_jdbc_username_property= "RAW_JDBC_USERNAME";
    private static final String obs_jdbc_password_property= "RAW_JDBC_PASSWORD";
    
    private static final String dic_jdbc_driver_property = "DIC_JDBC_DRIVER" ;
    private static final String dic_jdbc_url_property= "DIC_JDBC_URL";
    private static final String dic_jdbc_username_property= "DIC_JDBC_USERNAME";
    private static final String dic_jdbc_password_property= "DIC_JDBC_PASSWORD";


    /**
     * Returns the manager singleton instance.
     */
    public static DbConnManager getInstance()
    {
        return manager_instance;
    }


    /**
     * Loads the passed jdbc driver.
     * <br>
     */

    public void setCalibJdbcDriver(String jdbc_driver) throws ClassNotFoundException, InstantiationException, IllegalAccessException
    {
        this.calib_jdbc_driver = jdbc_driver;
        Class.forName(jdbc_driver).newInstance();
    }
    
    /**
     * Loads the passed jdbc driver.
     * <br>
     */

    public void setObsJdbcDriver(String jdbc_driver) throws ClassNotFoundException, InstantiationException, IllegalAccessException
    {
        this.obs_jdbc_driver = jdbc_driver;
        Class.forName(jdbc_driver).newInstance();
    }
    
    /**
     * Loads the passed jdbc driver.
     * <br>
     */

    public void setDicJdbcDriver(String jdbc_driver) throws ClassNotFoundException, InstantiationException, IllegalAccessException
    {
        this.dic_jdbc_driver = jdbc_driver;
        Class.forName(jdbc_driver).newInstance();
    }


    /**
     * Sets the jdbc url.
     * <br>
     */
    public void setCalibJdbcUrl(String jdbc_url)
    {
        this.calib_jdbc_url = jdbc_url;
    }
    
    /**
     * Sets the jdbc url.
     * <br>
     */
    public void setObsJdbcUrl(String jdbc_url)
    {
        this.obs_jdbc_url = jdbc_url;
    }
    
    /**
     * Sets the jdbc url.
     * <br>
     */
    public void setDicJdbcUrl(String jdbc_url)
    {
        this.dic_jdbc_url = jdbc_url;
    }


    /**
     * Sets the username used to access the database.
     * <br>
     */
    public void setCalibJdbcUsername(String jdbc_username)
    {
        this.calib_jdbc_username = jdbc_username;
    }
    
    /**
     * Sets the username used to access the database.
     * <br>
     */
    public void setObsJdbcUsername(String jdbc_username)
    {
        this.obs_jdbc_username = jdbc_username;
    }
    
    /**
     * Sets the username used to access the database.
     * <br>
     */
    public void setDicJdbcUsername(String jdbc_username)
    {
        this.dic_jdbc_username = jdbc_username;
    }


    /**
     * Sets the password used to access the database.
     * <br>
     */
    public void setCalibJdbcPassword(String jdbc_password)
    {
        this.calib_jdbc_password = jdbc_password;
    }
    
    /**
     * Sets the password used to access the database.
     * <br>
     */
    public void setObsJdbcPassword(String jdbc_password)
    {
        this.obs_jdbc_password = jdbc_password;
    }
    
    /**
     * Sets the password used to access the database.
     * <br>
     */
    public void setDicJdbcPassword(String jdbc_password)
    {
        this.dic_jdbc_password = jdbc_password;
    }
    
    /**
     * Gets an auto commit connection.
     * <br>
     *
     * @return an auto commit connection
     * @throws JdbcConnectionException          
     */
    public synchronized Connection getCalibConnection() throws SQLException,  JdbcConnectionException {
//        Connection tc = (Connection)calib_trans_conn.get();
        Connection tc = (Connection)calib_trans_conn;
        if (tc != null) {
            return tc;
        } else {
            if (calib_jdbc_driver == null && calib_jdbc_url == null && calib_jdbc_username == null && calib_jdbc_password == null) 
                loadProperties();             
            
            
            tc = DriverManager.getConnection(calib_jdbc_url, calib_jdbc_username, calib_jdbc_password);
            //calib_trans_conn.set(tc);
            calib_trans_conn=tc;
            
            return (tc);          
        }
    }
    
    /**
     * Gets an auto commit connection.
     * <br>
     *
     * @return an auto commit connection
     * @throws JdbcConnectionException
     */
    public synchronized Connection getObsConnection() throws SQLException,  JdbcConnectionException {
        //Connection tc = (Connection)obs_trans_conn.get();
        Connection tc = (Connection)obs_trans_conn;
        
        if (tc != null) {
            return tc;
        } else {

            if (obs_jdbc_driver == null || obs_jdbc_url == null || obs_jdbc_username == null || obs_jdbc_password == null)             
                loadProperties();
            
            tc = DriverManager.getConnection(obs_jdbc_url, obs_jdbc_username, obs_jdbc_password);
            //obs_trans_conn.set(tc);
            obs_trans_conn=tc;
            
            return(tc); 
        }
    }
    
    
    /**
     * Gets an auto commit connection.
     * <br>
     *
     * @return an auto commit connection
     * @throws JdbcConnectionException
     */
    public synchronized Connection getDicConnection() throws SQLException,  JdbcConnectionException {
        //Connection tc = (Connection)dic_trans_conn.get();
        Connection tc = (Connection)dic_trans_conn;
        
        if (tc != null) {
            return tc;
        } else {

            if (dic_jdbc_driver == null || dic_jdbc_url == null || dic_jdbc_username == null || dic_jdbc_password == null)             
                loadProperties();
            
            tc = DriverManager.getConnection(dic_jdbc_url, dic_jdbc_username, dic_jdbc_password);
            //dic_trans_conn.set(tc);
            dic_trans_conn=tc;
            
            return(tc); 
        }
    }


    /**
     * loads the connection properties from OCAProperties
     * 
     * @throws JdbcConnectionException
     * @throws OCAPropertiesException
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws ClassNotFoundException
     */
    private void loadProperties() throws JdbcConnectionException  {
        OCAProperties props = OCAProperties.getInstance();
        
        // Read the properties
        
        try {
            setObsJdbcDriver(props.getProperty(obs_jdbc_driver_property));
            setObsJdbcUrl(props.getProperty(obs_jdbc_url_property));
            setObsJdbcUsername(props.getProperty(obs_jdbc_username_property));
            setObsJdbcPassword(props.getProperty(obs_jdbc_password_property));
            
            setCalibJdbcDriver(props.getProperty(calib_jdbc_driver_property));
            setCalibJdbcUrl(props.getProperty(calib_jdbc_url_property));
            setCalibJdbcUsername(props.getProperty(calib_jdbc_username_property));
            setCalibJdbcPassword(props.getProperty(calib_jdbc_password_property));
            
            setDicJdbcDriver(props.getProperty(dic_jdbc_driver_property));
            setDicJdbcUrl(props.getProperty(dic_jdbc_url_property));
            setDicJdbcUsername(props.getProperty(dic_jdbc_username_property));
            setDicJdbcPassword(props.getProperty(dic_jdbc_password_property));
            
        } catch (ClassNotFoundException e) {
            throw new JdbcConnectionException( e );
        } catch (InstantiationException e) {
            throw new JdbcConnectionException( e );
        } catch (IllegalAccessException e) {
            throw new JdbcConnectionException( e );
        } catch (OCAPropertiesException e) {
            throw new JdbcConnectionException("Couldn't read a property, perharps properties not initialized");            
        }
    }


    /**
     * Releases the database connection.
     * <br>     
     */
    public synchronized void releaseObsConnection(Connection c) {
        //Connection tc = (Connection)obs_trans_conn.get();
        Connection tc = (Connection)obs_trans_conn;
        if (tc != null) {
            return;
        }

        try { 
            if (c != null) { 
                c.close();
            }
        } catch (SQLException x) { 
            System.out.println(x);
        }
    }


    /**
     * Releases the database connection.
     * <br>
     
     */
    public synchronized void releaseCalibConnection(Connection c) {
        //Connection tc = (Connection)calib_trans_conn.get();
        Connection tc = (Connection)calib_trans_conn;
        if (tc != null) {
            return;
        }

        try { 
            if (c != null) { 
                c.close();
            }
        } catch (SQLException x) { 
            System.out.println(x);
        }
    }
    
    /**
     * Releases the database connection.
     * <br>
     
     */
    public synchronized void releaseDicConnection(Connection c) {
        //Connection tc = (Connection)dic_trans_conn.get();
        Connection tc = (Connection)dic_trans_conn;
        if (tc != null) {
            return;
        }

        try { 
            if (c != null) { 
                c.close();
            }
        } catch (SQLException x) { 
            System.out.println(x);
        }
    }
    
    /**
     * Initiates a database transaction.
     * <br>
     * When working within a transaction, you should invoke this method first.
     * The connection is returned just in case you need to set the isolation level.
     *
     * @return a non-auto commit connection with the default transaction isolation level
     * @throws SQLException
     * @throws JdbcConnectionException
     * @throws OCAPropertiesException
     */
    public Connection beginObsTransaction() throws SQLException, JdbcConnectionException {
        Connection c = getObsConnection();
        c.setAutoCommit(false);
        //obs_trans_conn.set(c);
        obs_trans_conn=c;
        return c;
    }
    
    /**
     * Initiates a database transaction.
     * <br>
     * When working within a transaction, you should invoke this method first.
     * The connection is returned just in case you need to set the isolation level.
     *
     * @return a non-auto commit connection with the default transaction isolation level
     * @throws SQLException
     * @throws JdbcConnectionException
     * @throws OCAPropertiesException
     */
    public Connection beginCalibTransaction() throws SQLException, JdbcConnectionException {
        Connection c = getCalibConnection();
        c.setAutoCommit(false);
        //calib_trans_conn.set(c);
        calib_trans_conn=c;
        return c;
    }
    
    /**
     * Initiates a database transaction.
     * <br>
     * When working within a transaction, you should invoke this method first.
     * The connection is returned just in case you need to set the isolation level.
     *
     * @return a non-auto commit connection with the default transaction isolation level
     * @throws SQLException
     * @throws JdbcConnectionException
     * @throws OCAPropertiesException
     */
    public Connection beginDicTransaction() throws SQLException, JdbcConnectionException {
        Connection c = getDicConnection();
        c.setAutoCommit(false);
        //dic_trans_conn.set(c);
        dic_trans_conn=c;
        return c;
    }


    /**
     * Releases connection used for the transaction and performs a commit or rollback.
     *
     * @param commit tells whether this connection should be committed
     *        true for commit(), false for rollback()
     */
    public void endObsTransaction(boolean commit) throws SQLException {
        //Connection c = (Connection)obs_trans_conn.get();
        Connection c = (Connection)obs_trans_conn;
        if (c == null) {
            return;
        }

        try { 
            if (commit) { 
                c.commit();
            } else {
                c.rollback();
            }
        } finally { 
            c.setAutoCommit(true);
            //obs_trans_conn.set(null);
            //releaseObsConnection(c);
        }
    }
    
    /**
     * Releases connection used for the transaction and performs a commit or rollback.
     *
     * @param commit tells whether this connection should be committed
     *        true for commit(), false for rollback()
     */
    public void endCalibTransaction(boolean commit) throws SQLException {
        //Connection c = (Connection)calib_trans_conn.get();
        Connection c = (Connection)calib_trans_conn;
        if (c == null) {
            return;
        }

        try { 
            if (commit) { 
                c.commit();
            } else {
                c.rollback();
            }
        } finally { 
            c.setAutoCommit(true);
            //calib_trans_conn.set(null);
            //releaseCalibConnection(c);
        }
    }
    
    /**
     * Releases connection used for the transaction and performs a commit or rollback.
     *
     * @param commit tells whether this connection should be committed
     *        true for commit(), false for rollback()
     */
    public void endDicTransaction(boolean commit) throws SQLException {
        //Connection c = (Connection)dic_trans_conn.get();
        Connection c = (Connection)dic_trans_conn;
        if (c == null) {
            return;
        }

        try { 
            if (commit) { 
                c.commit();
            } else {
                c.rollback();
            }
        } finally { 
            c.setAutoCommit(true);
            //dic_trans_conn.set(null);
            //releaseDicConnection(c);
        }
    }




    ////////////////////////////////////////////////////
    // Utils method
    ///////////////////////////////////////////////////



    /**
     * Closes the passed Statement.
     */
    public void close(Statement s) {
        try {
            if (s != null) s.close();
        } catch (SQLException x) {
            System.out.println(x);
        }; 
    }

    /**
     * Closes the passed ResultSet.
     */
    public void close(ResultSet rs) {
        try {
            if (rs != null) rs.close();
        } catch (SQLException x) {
            System.out.println(x);
        }; 
    }

    /**
     * Closes the passed Statement and ResultSet.
     */
    public void close(Statement s, ResultSet rs) {
        close(rs);
        close(s);
    }
    
    
    /**
     * Retrieves an int value from the passed result set as an Integer object.
     */
    public static Integer getInteger(ResultSet rs, int pos) throws SQLException
    {
        int i = rs.getInt(pos);
        return rs.wasNull() ? (Integer)null : new Integer(i);
    }
    
    
    /**
     * Set an Integer object to the passed prepared statement as an int or as null.
     */
    public static void  setInteger(PreparedStatement ps, int pos, Integer i) throws SQLException
    {
        if (i==null) {
            ps.setNull(pos, Types.INTEGER);
        } else {
            ps.setInt(pos, i.intValue());
        }
    }
    
    
    /**
     * Retrieves a float value from the passed result set as a Float object.
     */
    public static Float getFloat(ResultSet rs, int pos) throws SQLException
    {
        float f = rs.getFloat(pos);
        return rs.wasNull() ? (Float)null : new Float(f);
    }
    
    
    /**
     * Set a Float object to the passed prepared statement as a float or as null.
     */
    public static void  setFloat(PreparedStatement ps, int pos, Float f) throws SQLException
    {
        if (f==null) {
            ps.setNull(pos, Types.FLOAT);
        } else {
            ps.setFloat(pos, f.floatValue());
        }
    }
    
    
    /**
     * Retrieves a double value from the passed result set as a Double object.
     */
    public static Double getDouble(ResultSet rs, int pos) throws SQLException
    {
        double d = rs.getDouble(pos);
        return rs.wasNull() ? (Double)null : new Double(d);
    }
    
    
    /**
     * Set a Double object to the passed prepared statement as a double or as null.
     */
    public static void  setDouble(PreparedStatement ps, int pos, Double d) throws SQLException
    {
        if (d==null) {
            ps.setNull(pos, Types.DOUBLE);
        } else {
            ps.setDouble(pos, d.doubleValue());
        }
    }
    
    
    /**
     * Retrieves a long value from the passed result set as a Long object.
     */
    public static Long getLong(ResultSet rs, int pos) throws SQLException
    {
        long l = rs.getLong(pos);
        return rs.wasNull() ? (Long)null : new Long(l);
    }
    
    
    /**
     * Set a Long object to the passed prepared statement as a long or as null.
     */
    public static void  setLong(PreparedStatement ps, int pos, Long l) throws SQLException
    {
        if (l==null) {
            ps.setNull(pos, Types.BIGINT);
        } else {
            ps.setLong(pos, l.longValue());
        }
    }
    
    
    /**
     * Retrieves a boolean value from the passed result set as a Boolean object.
     */
    public static Boolean getBoolean(ResultSet rs, int pos) throws SQLException
    {
        boolean b = rs.getBoolean(pos);
        return rs.wasNull() ? (Boolean)null : new Boolean(b);
    }
    /**
     * Set a Boolean object to the passed prepared statement as a boolean or as null.
     */
    public static void  setBoolean(PreparedStatement ps, int pos, Boolean b) throws SQLException
    {
        if (b==null) {
            ps.setNull(pos, Types.BOOLEAN);
        } else {
            ps.setBoolean(pos, b.booleanValue());
        }
    }
    
 

}
