//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//COPYRIGHT (c) 2003 European Southern Observatory 
//LICENSE 
//
//PROJECT: VLT Data Flow System 
//AUTHOR: Olivier Chuzel - ESO/DMD/DFS 
//SUBSYSTEM: OCA 
//CI NUMBER: DFS-XXX-NNN 
//
//Created on May 30, 2005
//Name:  OCAMetaTableDAO.java
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

package org.eso.oca.data;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eso.oca.fitsmapper.DbFitsMapper;

/**
 * Class OCAMetaTableDAO
 * 
 * This class handles the creation of a table in a database and its manipulation
 * It contains a definition of the table and allow creating it, inserting,
 * updating and selecting
 * 
 */

public class OCAMetaTableDAO extends TableDAO
{
    /**
     * @author ochuzel
     *
     * This inner class ties data about type and size
     */
    static public class NUplet
    {
      String _rawType;
      String _defaultValue;
      public NUplet(String rawType, String defaultValue)
      {
          _rawType=rawType;
          _defaultValue=defaultValue;
      }
      public String rawType() {return _rawType;};
      public String defaultValue() {return _defaultValue;};
      private String getDefaultSql()
      {
          String res = null;
          if(_rawType.compareTo("char(1)") == 0)
          {
              res=" default '"+_defaultValue+"'";
          }
          else
          {
              res=" default "+_defaultValue;
          }
          return res;
      }

    }
    
    //Constant used for some tables
    private static final String dpIdField = "dp_id";

    private static final String dpIdType = "varchar(50)";

    /*
     * Logger
     */
    private static Logger logger = Logger.getLogger(OCAMetaTableDAO.class);

    /**
     * Constructor
     * 
     * @param tableName
     * @throws SQLException
     * @throws JdbcConnectionException
     */
    public OCAMetaTableDAO(String tableName,
                           DbFitsMapper mapper,
                           Connection connexion) throws SQLException,
            JdbcConnectionException
    {        
        super(tableName, mapper, connexion);
    }

    static Map _keywordsTypesMap=null;
    
    public void clearMap()
    {
        if(_keywordsTypesMap!=null)
        {
            _keywordsTypesMap.clear();
        }
    }
    public Map getMap()
    {
        return _keywordsTypesMap;
    }
    public void add(String keyword, String type, String defaultValue)
    {
        String keywordNorm=keywordToColumn(keyword);
        NUplet nup = new NUplet(type, defaultValue);
        if(_keywordsTypesMap==null)
        {
            _keywordsTypesMap=new HashMap();
        }
        _keywordsTypesMap.put(keywordNorm, nup);
    }
    
    /**
     * Creates the temporary table using the table definition
     * 
     * @param keywords -
     *            contains a collection of keywords that will be the table
     *            columns
     * 
     * @throws SQLException
     * @throws JdbcConnectionException
     */
    public void create(String     instrument, 
                       Map        keywordsTypesMap,
                       boolean    withDpId)
            throws SQLException, JdbcConnectionException
    {
        // clear table column definition
        clearTableColumns();

        // Creates the table definition
        // build table definition using the keywords,
        // adding each column to the table definition        
        Iterator it = keywordsTypesMap.entrySet().iterator();
        for (int i = 0; i < keywordsTypesMap.size(); i++)
        {
            Map.Entry entry = (Map.Entry) it.next();
            String keyword = (String) entry.getKey();
            NUplet nup=(NUplet) entry.getValue();
            String type = nup.rawType();

            addTableColumn(keyword, type);
        }

        //Build the SQL
        PreparedStatement ps = null;
        StringBuffer _sql = null;

        _sql = new StringBuffer("create table ");
        _sql.append(getTableName());
        _sql.append(" ( ");
        if(withDpId == true)
        {
           _sql.append(dpIdField + "  " + dpIdType);
           _sql.append(",");
        }
        
        // iterate through the columns in the
        // hashtable and construct the statement
        Enumeration enumer = getColumns().keys();
        for (enumer = getColumns().keys(); enumer.hasMoreElements();)
        {
            String name = (String) enumer.nextElement();
            String type = (String) getColumns().get(name);
            NUplet nup = (NUplet)_keywordsTypesMap.get(name);
            
            _sql.append(name + "  " + type);
            if(nup != null)
            {
	            String defaultValue=nup.defaultValue();
	            if(defaultValue != null)
	            {
	                _sql.append(nup.getDefaultSql());
	            }
	            else
	            {
	                _sql.append(" NULL");
	            }
            }
            
            _sql.append(",");
        }

        if(withDpId == true)
        {
           _sql.append(" primary key(" + dpIdField + ") ");
        }
        
        _sql.append(" ) ");
        logger.debug(_sql);
        ps = getConnection().prepareStatement(_sql.toString());
        ps.executeUpdate();
        logger.debug(_sql.toString());

        //update db fits mapper as well
        getDbFitsMapper().addTable(getTableName(), instrument);
    }

}
