package org.eso.oca.fits;

import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Hashtable;

import org.apache.log4j.Logger;

/**
 * This class represents a file within the OCA process. In particular, an OCA
 * file is defined by its keywords (cards) and classification keywords
 * (meta-cards).
 * 
 * @author szampier
 */
public class OCAFile {
	/**
	 * the cards are keyword value pairs extracted from the FITS header.
	 * Tipically only the keywords used for classification-
	 * organization-association are stored in this class
	 */
	private Hashtable cards = new Hashtable();
    
	/**
	 * The meta-cards are user defined keywords, specified through the OCA
	 * classification rules
	 */
	private Hashtable metaCards = new Hashtable();
	
	// some commonly used FITS keywords
	
	/** Instrument ID */
	public static final String INSTRUME = "INSTRUME";
	
	/** Observation block ID */
	public static final Object OBS_ID = "OBS.ID";

	/** ESO program identification */
	public static final Object OBS_PROG_ID = "OBS.PROG.ID";
	
	/** Observation category */
	public static final String DPR_CATG = "DPR.CATG";
	
	/** Product category */
	public static final String PRO_CATG = "PRO.CATG";

	/** MJD start */
    public static final String MJD_OBS = "MJD-OBS";

    /** Template start time */
    public static final String TPL_START = "TPL.START";

    /** Number of exposures within template */
    public static final String TPL_NEXP = "TPL.NEXP";

    /** Exposure number within template */
    public static final String TPL_EXPNO = "TPL.EXPNO";

    /** Total integration time */
	public static final String EXPTIME = "EXPTIME";

    /** Pipeline filename */
	public static final String PIPEFILE = "PIPEFILE";

    // meta-keywords
	
    /** DataOrganiser category */ 
    public static final String DO_CATG = "DO.CATG";
    
    /** Last Frame DO category */ 
    public static final String LF_DO_CATG = "LF.DO.CATG";

    /** The maximum number of exposures a DO template can hold. */ 
    public static final String 
        DO_MAX_TEMPLATE_EXPOSURES = "DO.MAX.TEMPLATE.EXPOSURES"; 

    /** Prefix for template rule-related metakeywords which have special 
     * meaning to the DO. 
     */
    public static final String DO_TEMPLATE_PREFIX = "DO.TEMPLATE."; 

    /** Infix for template rule-related metakeywords whose values persist 
     * across DO template collations. 
     */
    public static final String DO_TEMPLATE_VAR_INFIX = ".VAR."; 

    /** Suffix for template rule-related metakeywords whose values should 
     * be logged after a DO template collation. 
     */
    public static final String DO_TEMPLATE_LOG_SUFFIX = ".LOG"; 

    /** DataOrganiser collated template names */
    public static final String DO_TEMPLATES ="DO.TEMPLATES"; 

    /** Single frame signal for reduction rule evaluation */
    public static final String SIG_FRAME = "SIG.FRAME"; 

    /** Template signal for reduction rule evaluation */
    public static final String SIG_TEMPLATE = "SIG.TEMPLATE"; 

    /** Name of template for reduction rule evaluation */
    public static final String SIG_TEMPLATE_NAME = "SIG.TEMPLATE.NAME"; 

    /** Daytime mode signal for reduction rule evaluation */
    public static final String SIG_DAYTIME = "SIG.DAYTIME"; 

    /** ABbuilder classification */ 
	public static final String DO_CLASS = "DO.CLASS";
    
    /** ABbuilder raw data type */ 
	public static final String RAW_TYPE = "RAW.TYPE";

    /** ABbuilder category */ 
	public static final String CATG = "CATG";

	/** Modified MJD start */ 
	public static final String MJD_OBS_MOD = "MJD-OBS.MOD";

	/** Filename */
	public static final String FILENAME = "FILENAME";

	/** Symbolic filename */
	public static final String LINKNAME = "LINKNAME";

	/** OLAS ID - five character instrument code (up to the first five 
	 * characters before the first dot in FILENAME). 
	 */
	public static final String OLAS_ID = "OLAS-ID";

	/** Data product ID */
	public static final String DP_ID = "DP.ID";

	/** Product extension */
	public static final String PRO_EXT = "PRO.EXT";

	public static final String PACK_DIR = "PACK.DIR";

	/** Association Block name */
    public static final String AB_NAME = "AB.NAME";

    public static final String ALIAS = "ALIAS";

	public static final String VIRTUAL = "VIRTUAL";

	public static final String UNDEFINED = "UNDEFINED";

	public static final String DATE = "DATE";

	private String dpId = null;

	/** the logger */
	static Logger logger = Logger.getLogger(OCAFile.class);

	/** default constructor */
	public OCAFile() {
	}

	/**
	 * Constructor with param-value pairs
	 * 
	 * @param cards :
	 *            keyword-value pairs from the FITS header
	 */
	public OCAFile(Hashtable keys) {
		if (keys == null)
			throw new IllegalArgumentException("null keywords");
		cards = keys;
	}

	public OCAFile(OCAFile aFile) {
		cards = new Hashtable(aFile.getCards());
		metaCards = new Hashtable(aFile.getMetaCards());
	}

	/**
	 * @return the (FITS) keyword values
	 */
	public Hashtable getCards() {
		return cards;
	}

       /** <p>Associates the specified keyword with the given value.</p> 
        * 
        * <p><b>WARNING:</b> The design philosophy of OCA normally considers 
        * keywords to be immutable and metakeywords to be modifiable. This 
        * method sidesteps that restriction, which means that it can be 
        * dangerous if misused. The {@link #setMetaCard()} method should 
        * always be considered before this one.</p> 
        *
        * @throws NullPointerException if key or val is null. 
        * @throws IllegalArgumentException if val cannot be mapped 
        *                                  to a FITS value type by the 
        *                                  {@link TypedHeaderCard} class. 
        * 
        * @param key the keyword.
        * @param val the value to associate with the keyword. 
        */
        public void setCard(final String key, final Object val) 
        {
            logger.trace("OCAFile::setCard()"); 

            if((key == null) || (val == null))
            {
                final String message = 
                    "OCAFile::setCard()" + " - key [" + key 
                    + "] and val" + (val == null ? " [null]" : "") 
                    + " must not be null."; 

                logger.fatal(message); 
                throw new NullPointerException(message); 
            }

            // Keywords may only be associated with values that can 
            // be mapped to a FITS value type by TypedHeaderCard. 
            if(TypedHeaderCard.findFITSType(val) 
               == TypedHeaderCard.UNKNOWN) 
            {
                final String message = 
                    "OCAFile::setCard()" 
                    + " - could not set keyword [" + key 
                    + "] to a value of unknown type [" 
                    + TypedHeaderCard.findFITSTypeString(val) + "]."; 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
            }
            
            // Associate the metakeyword with the value. 
            logger.debug("OCAFile::setCard()" 
                         + " - setting keyword [" + key 
                         + "] to a value of type [" 
                         + TypedHeaderCard.findFITSTypeString(val) 
                         + "]."); 

            cards.put(key, val); 
        }
    
	/**
	 * @return the (classification) meta-keyword values
	 */
	public Hashtable getMetaCards() {
		return metaCards;
	}

	/** Associates the specified metakeyword with the given value. 
         * 
         * @throws NullPointerException if metaKey or metaVal is null. 
         * @throws IllegalArgumentException if metaVal cannot be mapped 
         *                                  to a FITS value type by the 
         *                                  {@link TypedHeaderCard} class. 
	 * 
	 * @param metaKey the metakeyword.
	 * @param metaVal the value to associate with the metakeyword. 
	 */
        public void setMetaCard(final String metaKey, final Object metaVal) 
        {
            logger.trace("OCAFile::setMetaCard()"); 

            if((metaKey == null) || (metaVal == null))
            {
                final String message = 
                    "OCAFile::setMetaCard()" 
                    + " - metaKey [" + metaKey 
                    + "] and metaVal" + (metaVal == null ? " [null]" : "") 
                    + " must not be null."; 

                logger.fatal(message); 
                throw new NullPointerException(message); 
            }

            // Metakeywords may only be associated with values that can 
            // be mapped to a FITS value type by TypedHeaderCard. 
            if(TypedHeaderCard.findFITSType(metaVal) 
               == TypedHeaderCard.UNKNOWN) 
            {
                final String message = 
                    "OCAFile::setMetaCard()" 
                    + " - could not set metakeyword [" + metaKey 
                    + "] to a value of unknown type [" 
                    + TypedHeaderCard.findFITSTypeString(metaVal) + "]."; 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
            }
            
            // Associate the metakeyword with the value. 
            logger.debug("OCAFile::setMetaCard()" 
                         + " - setting metakeyword [" + metaKey 
                         + "] to a value of type [" 
                         + TypedHeaderCard.findFITSTypeString(metaVal) 
                         + "]."); 

            metaCards.put(metaKey, metaVal); 
        }
    
        /** Adds a new metakeyword and value. 
         * 
         * @throws NullPointerException if metaKey or metaVal is null. 

         * @param metaKey the metakeyword.
	 * @param metaVal the value to associate with the metakeyword. 
	 */
        public void addMetaCard(final String metaKey, final String metaVal) 
        {
            logger.trace("OCAFile::addMetaCard(String, String)"); 

            // Delegate. 
            setMetaCard(metaKey, metaVal); 
        }

        /** Adds a new metakeyword and value. 
         * 
         * @throws NullPointerException if metaKey or metaVal is null. 

         * @param metaKey the metakeyword.
	 * @param metaVal the value to associate with the metakeyword. 
	 */
        public void addMetaCard(final String metaKey, 
                                final BigInteger metaVal) 
        {
            logger.trace("OCAFile::addMetaCard(String, BigInteger)"); 

            // Delegate. 
            setMetaCard(metaKey, metaVal); 
        }

	public String getFilename() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(FILENAME);
	}

	public String getLinkname() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(LINKNAME);
	}

	public String getAlias() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(ALIAS);
	}

	public String getDprCatg() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(DPR_CATG);
	}
	
	public String getProCatg() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(PRO_CATG);
	}
	
	public String getDoClass() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(DO_CLASS);
	}
	
    public String getDoCatg() throws KeywordNotFoundException, UndefinedValueException {
        return getKeywordValue(DO_CATG);
    }
    
    public String getLastFrameDoCatg() throws KeywordNotFoundException, UndefinedValueException {
        return getKeywordValue(LF_DO_CATG);
    }

    public String getTplNexp() throws KeywordNotFoundException, UndefinedValueException {
        return getKeywordValue(TPL_NEXP);
    }

    public String getTplExpno() throws KeywordNotFoundException, UndefinedValueException {
        return getKeywordValue(TPL_EXPNO);
    }

    public String getRawType() throws KeywordNotFoundException, UndefinedValueException {
        return getKeywordValue(RAW_TYPE);
    }

    public String getCatg() throws KeywordNotFoundException, UndefinedValueException {
        return getKeywordValue(CATG);
    }

	public BigDecimal getMjdObs() throws KeywordNotFoundException, UndefinedValueException {
		return new BigDecimal(getKeywordValue(MJD_OBS));
	}

	public String getInstrument() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(INSTRUME);
	}

	public String getPipefile() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(PIPEFILE);
	}

	public String getProductExtension() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(PRO_EXT);
	}

	public String getPackDir() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(PACK_DIR);
	}

	public String getDate() throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(DATE);
	}

	public String getABName() throws KeywordNotFoundException, UndefinedValueException {
		String abName = null;
		String pipefile = null;
		// for "virtual" products AB.NAME is a meta-keyword
		Object value = metaCards.get(AB_NAME);
		if (value == null) {
			pipefile = getKeywordValue(PIPEFILE);
		} else {
			abName = value.toString();
		}
		// for "real" products AB.NAME is derived from PIPEFILE
		if (pipefile != null) {
			abName = pipefile;
			int ix = 0;
			// remove leading "r."
			if (abName.indexOf("r.") == 0) {
				abName = abName.substring(2);
			}
			// remove trailing "_XXXX.fits"
			if ((ix = abName.indexOf("_0")) > 1) {
				abName = abName.substring(0, ix) + ".ab";
			}
		}
		return abName;
	}

	/** whether this file is virtual or real */
	public boolean isVirtual() {
		return cards.containsKey(VIRTUAL) || metaCards.containsKey(VIRTUAL);
	}

	/**
	 * Returns the value of the given keyword.
	 * 
	 * @param keyword :
	 *            the keyword name
	 * @param forSQL :
	 *            this flag indicates if the returned value will be used in an
	 *            SQL query. If so, string values are 'quoted'.
	 * @return a string representing the keyword value. If the value is
	 *         undefined, the string "NULL" is returned.
	 * @throws KeywordNotFoundException
	 * @throws UndefinedValueException 
	 */
	private String getKeywordValue(String kw, boolean forSQL)
			throws KeywordNotFoundException, UndefinedValueException {
		// the return value
		String retVal;
		// log message
		String msg;
		// first try to get the keyword value from the available FITS cards
		Object val = cards.get(kw);
		// if no success, try with the meta-cards
		if (val == null)
			val = metaCards.get(kw);
		// this might indicate a serious problem
		if (val == null) {
			msg = "Requested keyword " + kw + " is not available";
			logger.info(msg);
			throw new KeywordNotFoundException(msg);
			// this might indicate a serious problem
		} else if (val instanceof UntypedFITSValue) {
			msg = "Requested keyword " + kw + " has undefined type";
			logger.error(msg);
			throw new KeywordNotFoundException(msg);
			// a keyword can be undefined, e.g. NULL value in the DB
		} else if (val instanceof UndefinedFITSValue) {
			msg = "Requested keyword " + kw + " has undefined value";
			logger.info(msg);
			throw new UndefinedValueException(msg);
            // strings must be quoted for SQL
		} else if (val instanceof String && forSQL == true) {
			retVal = "'" + val.toString() + "'";
		} else
			retVal = val.toString();

		return retVal;
	}

	/**
	 * Returns the value of the given keyword.
	 * 
	 * @param keyword :
	 *            the keyword name
	 * @return a string representing the keyword value. If the value is
	 *         undefined, the string "NULL" is returned.
	 * @throws KeywordNotFoundException
	 * @throws UndefinedValueException 
	 */
	public String getKeywordValue(String kw) throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(kw, false);
	}

	/**
	 * Returns the value of the given keyword. String values are quoted for SQL.
	 * 
	 * @param keyword :
	 *            the keyword name
	 * @return a string representing the keyword value. If the value is
	 *         undefined, the string "NULL" is returned.
	 * @throws KeywordNotFoundException
	 * @throws UndefinedValueException 
	 */
	public String getKeywordValueForSQL(String kw)
			throws KeywordNotFoundException, UndefinedValueException {
		return getKeywordValue(kw, true);
	}

	public void setDpId(String dpId) {
		this.dpId = dpId;
	}
	
	public String getDpId() throws KeywordNotFoundException, UndefinedValueException {
		if (dpId == null) {
			File f = new File(getFilename());
			String basename = f.getName();
			int indexOfLastDot = basename.lastIndexOf('.');
			if (indexOfLastDot > 0)
				dpId = basename.substring(0, indexOfLastDot);
			else
				dpId = UNDEFINED;
		}
		return dpId;
	}

	/**
	 * This method was borrowed from gasgano Access the filesystem (possibly)
	 * and resolve the file path into its canonical form ( resolving all links,
	 * logical drive names, relat ive paths. etc.)
	 * 
	 * @return this canonical file path, or null if it could not be resolved.
	 */
	public static String getCanonicalPath(String fullPath) {
		String rc = null;
		try {
			File file = new File(fullPath);
			rc = file.getCanonicalPath();
		} catch (IOException e) {
			// error resolving name.
			rc = null;
		}
		return rc;
	}
}