package org.eso.oca.parser;

import org.apache.log4j.Logger;
import org.eso.oca.fits.KeywordNotFoundException;
import org.eso.oca.fits.UndefinedValueException;

/**
 * This class implements the translation of OCA rules to SQL statements.
 * 
 * @author szampier
 * @version $Revision
 * 
 */
public class SqlVisitor implements OcaParserVisitor
{
    /** The Apache Log4j logger */
	static Logger logger = Logger.getLogger(SqlVisitor.class);

    /**
     * visits SimpleNode node
     * @throws OcaVisitorException
     */
    public Object visit(SimpleNode node, Object data)
            throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(SimpleNode) not implemented");
    }

    /**
     * visits ASTStart node
     * @throws OcaVisitorException 
     */
    public Object visit(ASTStart node, Object data)
    		throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(ASTStart) not implemented");
    }

    /**
     * visits ASTIfStatement node
     * @throws OcaVisitorException
     */
    public Object visit(ASTIfStatement node, Object data)
            throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(ASTIfStatement) not implemented");
    }

    /**
     * visits the ASTExecute node
     * @throws OcaVisitorException 
     */
    public Object visit(ASTExecute node, Object data)
    		throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(ASTExecute) not implemented");
    }

    /**
     * visits the ASTActionRule node
     * @throws OcaVisitorException
     */
    public Object visit(ASTActionRule node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	// Store the action name in the state object.
    	state.action = node.name;
    	// Visit the children nodes.
    	node.childrenAccept(this, data);
        return data;
    }

    /**
     * visits the ASTMinRet node
     */
    public Object visit(ASTMinRet node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	// Store the minRet value in the state object.
    	state.minRet = node.value;
    	return data;
    }

    /**
     * visits the ASTMaxRet node
     */
    public Object visit(ASTMaxRet node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	// Store the maxRet value in the state object.
    	state.maxRet = node.value;
    	return data;
    }

    /**
     * visits the ASTRecipe node
     */
    public Object visit(ASTRecipe node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	// Store the recipe name in the state object.
    	state.recipe = node.name;
    	return data;
    }

    /**
     * visits the ASTProduct node
     */
    public Object visit(ASTProduct node, Object data)
    		throws OcaVisitorException
    {
    	// No action performed.
    	return data;
    }

    /**
     * visits the ASTMetaKeywordDefinition node
     * @throws OcaVisitorException
     */
    public Object visit(ASTMetaKeywordDefinition node, Object data)
    		throws OcaVisitorException
    {
    	// Not implemented.
    	throw new OcaVisitorException("SqlVisitor::visit(ASTMetaKeywordDefinition) not implemented");
    }

    /**
     * visits the ASTMetaKeyword node
     * @throws OcaVisitorException 
     */
    public Object visit(ASTMetaKeyword node, Object data)
    		throws OcaVisitorException
    {
    	// Not implemented.
    	throw new OcaVisitorException("SqlVisitor::visit(ASTMetaKeyword) not implemented");
    }

    /**
     * visits the ASTLogicalOr node
     * @throws OcaVisitorException
     */
    public Object visit(ASTLogicalOr node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	// Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" or ");
    	// Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);

        return data;
    }

    /**
     * visits the ASTLogicalAnd node if the precedence of this node is greater
     * than the precedence of its child node, it means that the natural
     * operator's precedence was modified through the use of parentheses, which
     * have to be restored in the SQL expression
     * 
     * @throws OcaVisitorException
     */
    public Object visit(ASTLogicalAnd node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	SimpleNode child0 = (SimpleNode) node.jjtGetChild(0);
        SimpleNode child1 = (SimpleNode) node.jjtGetChild(1);
        // Need parentheses ?
        if (node.precedence() > child0.precedence())
        {
        	selectState.whereClause.append("(");
        	// Visit left operand
            child0.jjtAccept(this, data);
            selectState.whereClause.append(")");
        } 
        else
        {
            child0.jjtAccept(this, data);
        }
        
        selectState.whereClause.append(" and ");

        // Need parentheses ?
        if (node.precedence() > child1.precedence())
        {
        	selectState.whereClause.append("(");
        	// Visit right operand
            child1.jjtAccept(this, data);
            selectState.whereClause.append(")");
        } 
        else
        {
        	child1.jjtAccept(this, data);
        }

        return data;
    }

    /**
     * visits the ASTEquals node
     * @throws OcaVisitorException
     */
    public Object visit(ASTEquals node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" = ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);

        return data;
    }

    /**
     * visits the ASTNotEquals node
     * @throws OcaVisitorException
     */
    public Object visit(ASTNotEquals node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" != ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);
        
        return data;
    }

    /**
     * visits the ASTLessThan node
     * @throws OcaVisitorException
     */
    public Object visit(ASTLessThan node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" < ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);

        return data;
    }

    /**
     * visits the ASTGreaterThan node
     * 
     * @throws OcaVisitorException
     */
    public Object visit(ASTGreaterThan node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" > ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);
        
        return data;
    }

    /**
     * visits the ASTLessThanOrEqual node
     * @throws OcaVisitorException
     */
    public Object visit(ASTLessThanOrEqual node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" <= ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);
        
        return data;
    }

    /**
     * visits the ASTGreaterThanOrEqual node
     * @throws OcaVisitorException
     */
    public Object visit(ASTGreaterThanOrEqual node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	selectState.whereClause.append(" >= ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);
        
        return data;
    }

    /**
     * visits the ASTLike node
     * @throws OcaVisitorException
     */
    public Object visit(ASTLike node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
        selectState.whereClause.append(" like ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);
        
        return data;
    }

    /**
     * visits the ASTAddition node
     * @throws OcaVisitorException
     */
    public Object visit(ASTAddition node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
        selectState.whereClause.append(" + ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);
        
        return data;
    }

    /**
     * visits the ASTSubtraction node
     * @throws OcaVisitorException
     */
    public Object visit(ASTSubtraction node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

        // Visit left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
        selectState.whereClause.append(" - ");
        // Visit right operand
        node.jjtGetChild(1).jjtAccept(this, data);

        return data;
    }

    /**
     * visits the ASTMultiplication node
     * @throws OcaVisitorException
     */
    public Object visit(ASTMultiplication node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	SimpleNode child0 = (SimpleNode) node.jjtGetChild(0);
        SimpleNode child1 = (SimpleNode) node.jjtGetChild(1);
        // Need parentheses ?
        if (node.precedence() > child0.precedence())
        {
        	selectState.whereClause.append("(");
            // Visit left operand
            child0.jjtAccept(this, data);
            selectState.whereClause.append(")");
        } 
        else
        {
            child0.jjtAccept(this, data);
        }

        selectState.whereClause.append(" * ");

        // Need parentheses ?
        if (node.precedence() > child1.precedence())
        {
        	selectState.whereClause.append("(");
            // Visit right operand
            child1.jjtAccept(this, data);
            selectState.whereClause.append(")");
        }
        else
        {
            child1.jjtAccept(this, data);
        }

        return data;
    }

    /**
     * visits the ASTDivision node
     * @throws OcaVisitorException
     */
    public Object visit(ASTDivision node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
    	SimpleNode child0 = (SimpleNode) node.jjtGetChild(0);
        SimpleNode child1 = (SimpleNode) node.jjtGetChild(1);
        // Need parentheses ?
        if (node.precedence() > child0.precedence())
        {
        	selectState.whereClause.append("(");
            // Visit left operand
            child0.jjtAccept(this, data);
            selectState.whereClause.append(")");
        }
        else
        {
            child0.jjtAccept(this, data);
        }

        selectState.whereClause.append(" / ");

        // Need parentheses ?
        if (node.precedence() > child1.precedence())
        {
            selectState.whereClause.append("(");
            // Visit right operand
            child1.jjtAccept(this, data);
            selectState.whereClause.append(")");
        }
        else
        {
            child1.jjtAccept(this, data);
        }

        return data;
    }

    /**
     * visits the ASTRemainder node
     * @throws OcaVisitorException
     */
    public Object visit(ASTRemainder node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	SimpleNode child0 = (SimpleNode) node.jjtGetChild(0);
        SimpleNode child1 = (SimpleNode) node.jjtGetChild(1);
        // Need parentheses ?
        if (node.precedence() > child0.precedence())
        {
            selectState.whereClause.append("(");
            // Visit left operand
            child0.jjtAccept(this, data);
            selectState.whereClause.append(")");
        }
        else
        {
            child0.jjtAccept(this, data);
        }
        
        selectState.whereClause.append(" % ");
        
        // Need parentheses ?
        if (node.precedence() > child1.precedence())
        {
        	selectState.whereClause.append("(");
            // Visit right operand
            child1.jjtAccept(this, data);
            selectState.whereClause.append(")");
        }
        else
        {
            child1.jjtAccept(this, data);
        }

        return data;
    }

    /**
     * visits the ASTKeyword node
     * @throws OcaVisitorException
     */
    public Object visit(ASTKeyword node, Object data)
            throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
    	// The keyword name must be converted to a database field name.
    	String keyword = SqlVisitorHelper.formatKeyForDB(node.name);
    	selectState.whereClause.append(keyword);
        
        return data;
    }

    /**
     * visits the ASTIntConst node
     */
    public Object visit(ASTIntConst node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

        selectState.whereClause.append(node.value.toString());

        return data;
    }

    /**
     * visits the ASTFloatConst node
     */
    public Object visit(ASTFloatConst node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

        selectState.whereClause.append(node.value.toString());

        return data;
    }

    /**
     * visits the ASTStringConst node
     */
    public Object visit(ASTStringConst node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	// Enclose string value in quotes. 
    	selectState.whereClause.append("'" + node.value.toString() + "'");

        return data;
    }

    /**
     * visits the ASTSelectExecuteStatement node
     * @throws OcaVisitorException
     */
    public Object visit(ASTSelectExecuteStatement node, Object data)
            throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(ASTSelectExecuteStatement) not implemented");
    }

    /**
     * visits the ASTSelectAssociateStatement node
     * @throws OcaVisitorException
     */
    public Object visit(ASTSelectAssociateStatement node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	/* Create new SelectAssociateState to store
    	 * the SQL query.
    	 */ 
    	state.selects.add(new SelectAssociateState());
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	// Store query information in the state object.
    	selectState.minRet = state.minRet;
    	selectState.maxRet = state.maxRet;
    	selectState.alias = node.alias;
    	selectState.dataSource = node.dataSource;

    	// Visit the children nodes.
        node.childrenAccept(this,data);
        return data;
    }

    /**
     * visits the ASTProductKeywordDefinition node
     * @throws OcaVisitorException
     */
    public Object visit(ASTProductKeywordDefinition node, Object data)
    		throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(ASTProductKeywordDefinition) not implemented");
    }

    /**
     * visits the ASTProductKeyword node
     * @throws OcaVisitorException
     */
    public Object visit(ASTProductKeyword node, Object data)
    		throws OcaVisitorException
    {
    	throw new OcaVisitorException("SqlVisitor::visit(ASTProductKeyword) not implemented");
    }

    /**
     * visits the ASTInputFileKeyword node
     * @throws OcaVisitorException
     */
    public Object visit(ASTInputFileKeyword node, Object data)
    		throws OcaVisitorException
    {
    	final String thisMethod = "SqlVisitor::visit(ASTInputFileKeyword)";
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);
    	
    	/* The inputFile in the OCA grammar corresponds to the
    	* exemplar file in the Association Block.
    	*/
    	if (state.getExemplar() == null) {
    		throw new OcaVisitorException("Null exemplar");
    	}
    	
    	String value;
    	// Get database column name for this keword.
    	String columnName = SqlVisitorHelper.formatKeyForDB(node.name);

    	try {
    		// Get keyword value in exemplar file.
    		value = state.getExemplar().getKeywordValueForSQL(node.name);
		} catch (KeywordNotFoundException e) {
			logger.warn(thisMethod
					+ " - keyword " + node.name
					+ " not found in exemplar");
			value = state.getNullValueForColumn(columnName, selectState.dataSource);
		} catch (UndefinedValueException e) {
			logger.debug(thisMethod
					+ " - keyword " + node.name
					+ " has undefined value in exemplar");
			value = state.getNullValueForColumn(columnName, selectState.dataSource);
		}
		
    	selectState.whereClause.append(value.toString());

    	return data;
    }

    /**
     * visits the ASTPossiblyEquals node
     * @throws OcaVisitorException
     */
    public Object visit(ASTPossiblyEquals node, Object data)
    		throws OcaVisitorException
    {
    	OcaVisitorState state = (OcaVisitorState) data;
    	SelectAssociateState selectState = state.selects.elementAt(state.selects.size()-1);

    	// Visits left operand
    	node.jjtGetChild(0).jjtAccept(this, data);
    	// NOTE: this operator is treated as the usual "equals"
    	selectState.whereClause.append(" = ");
    	// Visits right operand
        node.jjtGetChild(1).jjtAccept(this, data);

        return data;
    }

    /**
     * visits the ASTIs node
     * @throws OcaVisitorException
     */
	public Object visit(ASTIs node, Object data)
			throws OcaVisitorException
	{
    	throw new OcaVisitorException("SqlVisitor::visit(ASTIs) not implemented");
	}

    /**
     * visits the ASTTypeConst node
     * @throws OcaVisitorException
     */
	public Object visit(ASTTypeConst node, Object data)
			throws OcaVisitorException
	{
    	throw new OcaVisitorException("SqlVisitor::visit(ASTTypeConst) not implemented");
	}

	public Object visit(ASTBooleanConst node, Object data) throws OcaVisitorException {
		throw new OcaVisitorException("SqlVisitor::visit(ASTBooleanConst) not implemented");
	}
}
