package org.eso.oca.data;

import java.io.File;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Map;

import junit.framework.TestCase;

import org.eso.oca.properties.OCAProperties;
import org.eso.oca.properties.OCAPropertiesException;

/**
 * Class TestOCATableDAO
 *
 * Unit tests for <code>OCATableDAO</code> class
 *
 * @version $Revision: 1.3 $
 * @author pnunes, Apr 26, 2005
 */

// $Id: TestOCATableDAO.java,v 1.3 2005/06/06 10:32:17 szampier Exp $ 
public class TestOCATableDAO extends TestCase {
    
    /**
     * Database connection 
     */
    private DbConnManager connManager;       
    private Connection conn;
       
    /*
     * configuration files 
     */
    private final static String configDir = 
        	"org" + File.separator + 
        	"eso" + File.separator + 
        	"oca" + File.separator + 
        	"config" + File.separator;
    
    private final static String propertiesFile =
        configDir + ".ocarc";
    
    private OCATableDAO ocaTableDAO;
    
    
    public void testCreate() throws SQLException, JdbcConnectionException {
        
        /* sets up connection */
        connManager = DbConnManager.getInstance();
        conn = connManager.getObsConnection();        
        
        // create the table 
        
        ocaTableDAO = new OCATableDAO("test_table");
        
        try {
            // drop the table eventually
            ocaTableDAO.dropTable();
        } catch ( SQLException e ) {
            
        }
        
        Collection keywords = new LinkedList();
        keywords.add("COL.1");
        keywords.add("COL.2");               
        ocaTableDAO.create("test_instrument",keywords);
        
        // check if the table is there 
        DatabaseMetaData dbMetaData = conn.getMetaData();        
        ResultSet rs = dbMetaData.getTables(null,null,null,null);
        
        boolean found = false;
        while ( rs.next() ) {
            String tableName = rs.getString("TABLE_NAME");
            if (tableName.equalsIgnoreCase("test_table"))
                found = true;            
        }
        
        assertTrue(found);
        
        // drop the table 
        ocaTableDAO.dropTable();
        
        
        // test with a temporary table
        ocaTableDAO = new OCATableDAO("#test_table");
        keywords.add("COL.1");
        keywords.add("COL.2");               
        ocaTableDAO.create("test_instrument",keywords);
        
        // insert a record to check its there
        PreparedStatement ps = null;
        StringBuffer _sql;
        boolean exception = false;
        try {
            _sql =  new StringBuffer();                         
            _sql.append("insert into #test_table (dp_id,COL_1,COL_2) values ('DP_ID','VALUE1','VALUE2') ");
            ps = conn.prepareStatement(_sql.toString());
            ps.executeUpdate();
        

        } catch (SQLException e) {
            exception = true;
        }
        
        // exception shouldn't have occured
        assertTrue(!exception);
    }

    
    public void testInsertEmptyRecords() throws SQLException, JdbcConnectionException {
        
        // sets up connection 
        connManager = DbConnManager.getInstance();
        conn = connManager.getObsConnection();        
        
        // create the table         
        ocaTableDAO = new OCATableDAO("#test_insert_table");                    
        Collection keywords = new LinkedList();
        keywords.add("COL.1");
        keywords.add("COL.2");               
        ocaTableDAO.create("test_instrument",keywords);
        
        // insert the records
        String [] dpIds = 
        	{"GIRAF.1900-01-01T12:00:00.000",
             "GIRAF.1900-01-02T12:00:00.000",
             "GIRAF.1900-01-03T12:00:00.000"};
        
        //insert first the dpIds in data_products
        PreparedStatement ps = null;
        for (int i = 0; i < dpIds.length; i++) {
            String dpId = dpIds[i];
            String sql = "insert into data_products (dp_id)	values ('" + dpId +  "')";
            ps = conn.prepareStatement(sql);
            ps.executeUpdate();            
        }
                        
        ocaTableDAO.insertEmptyRecords( dpIds );
                                
        // check if they are there         
        for (int i = 0; i < dpIds.length; i++) {
            String dpId = dpIds[i];            
            StringBuffer _sql = new StringBuffer();
            
            _sql.append("select * from #test_insert_table where dp_id = '" + dpId + "'");
            
            int count = 0;
            PreparedStatement stmt = conn.prepareStatement(_sql.toString());
            ResultSet rs = stmt.executeQuery();
            
            while ( rs.next()) {
                count++;
            }
            assertTrue(count==1);
        }
        
      
    }
        

    public void testUpdate() {
    }


    public void testGetRow() throws SQLException, JdbcConnectionException {
        
        //  sets up connection 
        connManager = DbConnManager.getInstance();
        conn = connManager.getObsConnection();        
        
        // create the table         
        ocaTableDAO = new OCATableDAO("#test_get_row_table");                    
        Collection keywords = new LinkedList();
        keywords.add("COL.1");
        keywords.add("COL.2");               
        ocaTableDAO.create("test_instrument",keywords);
        
        // insert the records
        String [] dpIds = 
        	{"GIRAF.1900-01-01T12:00:00.000",
             "GIRAF.1900-01-02T12:00:00.000",
             "GIRAF.1900-01-03T12:00:00.000"};
        
        //insert first the dpIds in data_products
        PreparedStatement ps = null;
        for (int i = 0; i < dpIds.length; i++) {
            String dpId = dpIds[i];
            String sql = "insert into #test_get_row_table (dp_id,COL_1,COL_2)	values ('" + dpId +  "','" +i+ "', '" +i+ "')";
            ps = conn.prepareStatement(sql);
            ps.executeUpdate();            
        }                        
        
        
        // get a row
        String [] metaKw = {"COL.1","COL.2"};
        
        // check if the values are correct
        Map m = ocaTableDAO.getRow("GIRAF.1900-01-01T12:00:00.000",metaKw);
        for (int i = 0; i < metaKw.length; i++) {
            String kw = metaKw[i];           
            String value = (String) m.get(kw);
            assertTrue(value.equals("0"));            
        }
        
        
        
    }

    public void testGetAllRows() throws SQLException, JdbcConnectionException {
        // sets up connection 
        connManager = DbConnManager.getInstance();
        conn = connManager.getObsConnection();    
        
        // create the table         
        ocaTableDAO = new OCATableDAO("#test_get_all_rows_table");                    
        Collection keywords = new LinkedList();
        keywords.add("COL.1");
        keywords.add("COL.2");               
        ocaTableDAO.create("test_instrument",keywords);
        
        // insert the records
        String [] dpIds = 
        	{"GIRAF.1900-01-01T12:00:00.000",
             "GIRAF.1900-01-02T12:00:00.000",
             "GIRAF.1900-01-03T12:00:00.000"};
        
        //  insert first the dpIds in data_products
        PreparedStatement ps = null;
        for (int i = 0; i < dpIds.length; i++) {
            String dpId = dpIds[i];
            String sql = "insert into #test_get_all_rows_table (dp_id,COL_1,COL_2)	values ('" + dpId +  "','" +i+ "', '" +i+ "')";
            ps = conn.prepareStatement(sql);
            ps.executeUpdate();            
        }     
        
        // test if the data is there
        String [] metaKw = {"COL.1","COL.2"};
        String [][] data = ocaTableDAO.getAllRows(metaKw);
        
        for (int i = 0; i < data.length; i++) {
            String[] lines = data[i];
            assertTrue(lines[0].equals(dpIds[i])  );
            for (int j = 1; j < lines.length; j++) {
                String value = lines[j];
                
                assertTrue(value.equals(i+""));
                
            }            
        }                
        
    }


         
	public static void main(String[] args) throws OCAPropertiesException {
	    OCAProperties props = OCAProperties.getInstance();
        props.loadProperties(propertiesFile);
	    junit.textui.TestRunner.run(TestOCATableDAO.class);
	}



}