package org.eso.util.configuration;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.TreeSet;

import jargs.gnu.CmdLineParser;

import org.apache.log4j.Logger;

import org.eso.util.misc.WrappedString;

/** This class provides application configuration code relating to command 
 * lines. However, applications should not use it directly - they should 
 * subclass {@link Configuration} and call through its protected methods 
 * of the same name. 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
final class CommandLineHandler 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(CommandLineHandler.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "CommandLineHandler"; 

    /** Private constructor - this class should not be instantiated. */ 
    private CommandLineHandler() 
    {
        final String methodLogName = classLogName + "::CommandLineHandler()"; 

        logger.trace(methodLogName); 

        // This constructor does not need to do anything. 
    }

    /** Parse the given command line arguments and update the configuration 
     * with any values found. 
     *
     * @throws NullPointerException if arguments or configuration is null. 
     * 
     * @param args the command line parameters to parse. 
     * @param configuration the configuration to update with values. 
     *
     * @return true if parsing was successful, false otherwise. 
     */ 
    static boolean parseCommandLine(final String[] arguments, 
                                    final Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::parseCommandLine()"; 

        logger.trace(methodLogName); 

        if((arguments == null) || (configuration == null)) 
        {
            final String message = 
                methodLogName + " - arguments [" + arguments 
                + "] and configuration [" + configuration 
                + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        logger.debug(methodLogName + " - preparing command line parser."); 

        // Command line parser. 
        final CmdLineParser parser = new CmdLineParser(); 

        // Map of configuration options to JArgs command line options. 
        final HashMap<ConfigurationOption, CmdLineParser.Option> optionMap = 
            new HashMap<ConfigurationOption, CmdLineParser.Option>(); 

        // Iterate though all configuration options held by the configuration 
        // map, generating the appropriate JArgs command line option for each 
        // command line configurable option. 
        final Map<String, ConfigurationOption> configurationMap = 
            configuration.getSelectedOptionMap(
                ConfigurationOption.COMMAND_LINE); 

        final Iterator<ConfigurationOption> 
            configurationMapIterator = configurationMap.values().iterator();

        ConfigurationOption currentOption = null; 
            
        while(configurationMapIterator.hasNext())
        {
            currentOption = configurationMapIterator.next(); 

            switch(currentOption.optionType)
            {
              case ConfigurationOption.STRING:
                optionMap.put(currentOption, 
                              parser.addStringOption(
                                  currentOption.commandLineChar,
                                  currentOption.commandLineName)); 
                break;

              case ConfigurationOption.BOOLEAN:
                optionMap.put(currentOption, 
                              parser.addBooleanOption(
                                  currentOption.commandLineChar,
                                  currentOption.commandLineName)); 
                break;

              case ConfigurationOption.INTEGER:
                optionMap.put(currentOption, 
                              parser.addIntegerOption(
                                  currentOption.commandLineChar,
                                  currentOption.commandLineName)); 
                break;

              default:
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName + "] type [" 
                    + ConfigurationOption.getTypeString(
                        currentOption.optionType) 
                    + "] not recognised."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            logger.debug(methodLogName 
                         + " - added option [" + currentOption.optionName 
                         + "] to the command line parser."); 
        }

        // Parse the command line options. 
        logger.debug(methodLogName + " - parsing command line options."); 

        try
        {
            parser.parse(arguments); 
        }
        catch(Exception e)
        {
            // Parsing failed - print & log the JArgs error message. 
            System.out.print("Error parsing command line: " 
                             + e.getMessage() + "\n"); 

            logger.error(methodLogName + " - error parsing command line: " 
                         + e.getMessage()); 

            return(false); 
        }

        // Update the configuration options with any command line values 
        // extracted by the parser. 
        final Iterator<Map.Entry<ConfigurationOption, CmdLineParser.Option>> 
            optionMapIterator = optionMap.entrySet().iterator(); 

        Map.Entry<ConfigurationOption, CmdLineParser.Option> 
            currentOptionMapEntry = null; 
        Object currentOptionValue = null; 

        while(optionMapIterator.hasNext()) 
        {
            currentOptionMapEntry = optionMapIterator.next(); 
            currentOption = currentOptionMapEntry.getKey(); 

            // Get the command line value of the current option, if any. 
            currentOptionValue = 
                parser.getOptionValue(currentOptionMapEntry.getValue()); 

            if(currentOptionValue == null) 
            {
                logger.debug(methodLogName 
                             + " - the command line does not" 
                             + " provide a value for option [" 
                             + currentOption.optionName + "]."); 
                continue; 
            }

            // This option has been set from the command line. Ensure that 
            // no higher priority value is already present, then update it. 
            if(ConfigurationOption.canOverrideValueSource(
                   ConfigurationOption.COMMAND_LINE, 
                   currentOption.getActualValueSource())) 
            {
                logger.debug(methodLogName + " - setting option [" 
                             + currentOption.optionName
                             + "] from the command line."); 

                currentOption.setActualValue(currentOptionValue, 
                                             ConfigurationOption.COMMAND_LINE, 
                                             null); 
            }
            else 
            {
                logger.debug(
                    methodLogName + " - ignoring option [" 
                    + currentOption.optionName 
                    + "] which already has a higher priority [" 
                    + ConfigurationOption.getValueSourceString(
                        currentOption.getActualValueSource()) 
                    + "] value than [" 
                    + ConfigurationOption.getValueSourceString(
                        ConfigurationOption.COMMAND_LINE) 
                    + "]."); 
                continue; 
            }
        }

        // Store any remaining arguments from the command line which were not 
        // consumed by the parser. 
        configuration.
            setRemainingCommandLineArguments(parser.getRemainingArgs()); 

        // The command line was successfully parsed. 
        logger.debug(methodLogName 
                     + " - command line parsing successfully completed."); 

        return(true); 
    }

    /** Prints a help message describing the specified configuration's command 
     * line options to standard output. 
     * 
     * @param configuration the configuration to print command line help for. 
     * 
     * @throws NullPointerException if configuration is null. 
     */ 
    static void printCommandLineHelpMessage(Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::printCommandLineHelpMessage()"; 

        logger.trace(methodLogName); 

        if(configuration == null) 
        {
            final String message = 
                methodLogName + " - configuration must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        logger.debug(methodLogName 
                     + " - printing command line help message" 
                     + " to standard output."); 

        // Iterate though all configuration options held by the configuration 
        // map, pick out the ones which accept values from the command line, 
        // and generate a help text for each one. 
        final Map<String, ConfigurationOption> optionMap = 
            configuration.getSelectedOptionMap(
                ConfigurationOption.COMMAND_LINE); 

        final Iterator<ConfigurationOption> 
            optionMapIterator = optionMap.values().iterator(); 

        // Some options are mandatory, some not. 
        final TreeSet<String> mandatorySet = new TreeSet<String>(); 
        final TreeSet<String> additionalSet = new TreeSet<String>(); 

        // The current option, and components of help text for it. 
        ConfigurationOption currentOption = null; 
        String currentText = null; 
        String parameterTypeString = null; 
            
        while(optionMapIterator.hasNext())
        {
            currentOption = optionMapIterator.next(); 

            // Reset the current option help text, ready for appending. 
            currentText = ""; 

            // If the option is not BOOLEAN, get its type string to use in 
            // specifying its parameter type (BOOLEAN options do not take 
            // a parameter). 
            switch(currentOption.optionType)
            {
              case ConfigurationOption.STRING:
              case ConfigurationOption.INTEGER:
                parameterTypeString = 
                    ConfigurationOption.getTypeString(
                        currentOption.optionType); 
                break; 

              case ConfigurationOption.BOOLEAN:
                parameterTypeString = ""; 
                break; 

              default: 
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName + "] type [" 
                    + ConfigurationOption.getTypeString(
                        currentOption.optionType)
                    + "] not recognised."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // If this option has a short form, add it. 
            if(currentOption.commandLineChar != 0)
            {
                currentText += 
                    "-" + currentOption.commandLineChar 
                    + " " + parameterTypeString + "\n"; 
            }

            // If this option has a long form, add it. 
            if(currentOption.commandLineName != "")
            {
                currentText += 
                    "--" + currentOption.commandLineName 
                    + (parameterTypeString.equals("")
                       ? "" : " " + parameterTypeString) + "\n"; 
            }

            // The option help text must not be empty at this stage. 
            if(currentText.equals(""))
            {
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName 
                    + "] is marked as accepting source" 
                    + (ConfigurationOption.isSingleDefinedValueSource(
                           currentOption.validValueSources) ? " [" : "s [") 
                    + ConfigurationOption.getValueSourceString(
                        currentOption.validValueSources) 
                    + "] but does not have a short character or long name" 
                    + " to use on the command line."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // If the option is non-boolean & has a default value, show it. 
            // 
            // Boolean options implicitly default to false, so there is 
            // no point showing them. 
            if((currentOption.optionType != ConfigurationOption.BOOLEAN)
               && ConfigurationOption.isValidValueSource(
                   ConfigurationOption.DEFAULT_VALUE, 
                   currentOption.validValueSources)) 
            {
                currentText += 
                    new WrappedString(
                        "  ", 
                        "Default value [" + currentOption.defaultValue + "].")
                        + "\n"; 
            }

            // Add the option's help text, if present. 
            if(!currentOption.helpText.equals(""))
            {
                currentText += 
                    new WrappedString("    ", currentOption.helpText) + "\n"; 
            }

            // The text is complete - store it in the appropriate list. 
            if(currentOption.validValueSources 
               == ConfigurationOption.COMMAND_LINE) 
            {
                // This option can only be set from the command line, 
                // i.e. it is mandatory. 
                mandatorySet.add(currentText); 
            }
            else
            {
                // This option can also be set from sources other than 
                // the command line. 
                additionalSet.add(currentText); 
            }
        }

        // True if the appropriate set contains option help texts. 
        final boolean mandatoryOptions = !mandatorySet.isEmpty(); 
        final boolean additionalOptions = !additionalSet.isEmpty(); 

        // Print the help text. 
        System.out.print("\nUsage: " 
                         + configuration.getString("application.command")); 

        if(mandatoryOptions || additionalOptions) 
        {
            // This application has command line options. 
            System.out.print(" OPTIONS"); 
        }

        System.out.print("\n\n"); 

        if(!configuration
           .getString("application.help.header.text").equals("")) 
        {
            // A command line help header has been provided. 
            System.out.print(
                new WrappedString(
                    "", 
                    configuration.getString("application.help.header.text")) 
                + "\n\n"); 
        }

        if(mandatoryOptions) 
        {
            // Inform the user that the following options are mandatory. 
            System.out.print("Mandatory options:\n\n"); 

            // Print the mandatory option texts. 
            final Iterator<String> 
                mandatoryIterator = mandatorySet.iterator(); 

            while(mandatoryIterator.hasNext()) 
            {
                System.out.print(mandatoryIterator.next() + "\n"); 
            }
        }

        if(additionalOptions) 
        {
            if(mandatoryOptions) 
            {
                // Both mandatory and additional options exist - inform 
                // the user that the following options are additional. 
                System.out.print("Additional options:\n\n"); 
            }

            // Print the additional option texts. 
            final Iterator<String> 
                additionalIterator = additionalSet.iterator(); 

            while(additionalIterator.hasNext()) 
            {
                System.out.print(additionalIterator.next() + "\n"); 
            }
        }

        // System property help is considered to be part of command line 
        // help - print it here. 
        PropertyHandler.printSystemPropertyHelpMessage(configuration); 

        if(!configuration
           .getString("application.help.footer.text").equals("")) 
        {
            // A command line help footer has been provided. 
            System.out.print(
                new WrappedString(
                    "", 
                    configuration.getString("application.help.footer.text")) 
                + "\n\n"); 
        }

        // The full help text has now been printed. 
        logger.debug(methodLogName 
                     + " - successfully printed command line help message" 
                     + " to standard output."); 
    }
}
