package org.eso.util.configuration;

import java.io.File;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List; 
import java.util.Map;
import java.util.Properties;
import java.util.TreeSet;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import org.eso.util.filesystem.FileSystemOperations;

import org.eso.util.misc.WrappedString;

/** <p>The Configuration class is the central element of the configuration 
 * package. Almost everything an application does with this package passes 
 * through a Configuration method; the only exceptions to this rule are the 
 * publicly accessible constants and convenience methods provided by the 
 * {@link ConfigurationOption} class.</p> 
 * 
 * <p>To use the configuration package, an application must have a 
 * meta-configuration property file defining its configuration options. 
 * The file format is defined by the {@link MetaConfigurator} class, and 
 * the file itself must be located within the JVM's classpath.</p> 
 * 
 * <p>Applications are required to subclass Configuration before they can 
 * use the majority of its methods. This is a deliberate design decision 
 * to encapsulate the details of configuring an application.</p> 
 * 
 * <p>However, there are several public methods which allow the rest of 
 * the application to get or set the value of an individual configuration 
 * parameter. It is also possible to check whether a value is available, 
 * where it came from, and whether the application is allowed to change 
 * it.</p> 
 * 
 * <p>When a Configuration subclass is instantiated, its constructor must 
 * pass the location of its meta-configuration file to the Configuration 
 * constructor. This constructor uses {@link MetaConfigurator} methods to 
 * initialise the Configuration according to the meta-configuration file. 
 * Once the Configuration constructor returns, the subclass can make full 
 * use of the Configuration methods.</p> 
 * 
 * <p>It is perfectly acceptable for an application to define more than 
 * one subclass of Configuration and instantiate several of each. For 
 * example, it may have a CoreConfiguration which deals with the command 
 * line and a primary configuration file. That object may in turn hold 
 * several PluginConfigurations, each handling a configuration file for 
 * its plugin.</p> 
 * 
 * <p><b>NOTE:</b> Where possible, call methods which set configuration 
 * parameters in order of value source priority, highest first (see the 
 * list provided by {@link ConfigurationOption}). For example, it is best 
 * to parse the command line first, then read system properties, then read 
 * application properties from a configuration file.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
public class Configuration 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(Configuration.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "Configuration"; 

    /** True if this Configuration has been fully constructed (including 
     * meta-configuration). 
     */ 
    private boolean constructionComplete = false; 

    /** Modifiable map of configuration option names to ConfigurationOption 
     * objects. This is only available at construction time; the constructor 
     * locks the map by dropping the reference before it returns. 
     */
    private HashMap<String, ConfigurationOption> modifiableConfigurationMap = 
        new HashMap<String, ConfigurationOption>(); 

    /** Map of configuration option names to ConfigurationOption objects. 
     * This map is not modifiable. 
     * 
     * <p><b>WARNING:</b> although the Map is unmodifiable, the contents 
     * of each ConfigurationOption can still be changed. Methods which do 
     * so <i>must</i> ensure appropriate synchronisation with any other 
     * thread which might be using this Configuration.</p> 
     */ 
    private final Map<String, ConfigurationOption> configurationMap = 
        Collections.unmodifiableMap(modifiableConfigurationMap); 

    /** True if this Configuration has parsed a command line. */ 
    private boolean commandLineParsed = false; 

    /** Modifiable list of command line arguments which were not consumed 
     * by a command line parser (see {@link #parseCommandLine(String[])}). 
     *
     * The {@link #getRemainingCommandLineArguments()} method locks this 
     * list by dropping the reference before it returns. 
     */ 
    private ArrayList<String> modifiableRemainingCommandLineArgumentList = 
        new ArrayList<String>(); 

    /** List of command line arguments which were not consumed by a 
     * command line parser (see {@link #parseCommandLine(String[])}). 
     * This list is not modifiable. 
     */ 
    private final List<String> remainingCommandLineArgumentList = 
        Collections.unmodifiableList(
            modifiableRemainingCommandLineArgumentList); 

    /** Constructs a new Configuration, using the 
     * {@link MetaConfigurator#metaConfigure(Configuration, String)} method 
     * to initialise itself as directed by the meta-configuration file. 
     *
     * @param metaConfigurationFilePath location of the meta-configuration 
     *                                  file within the JVM's classpath. 
     *
     * @throws NullPointerException if the metaConfigurationFilePath is null. 
     * @throws MetaConfigurationError if the meta-configuration file is 
     *                                missing or unusable. 
     */
    protected Configuration(final String metaConfigurationFilePath) 
        throws NullPointerException, MetaConfigurationError 
    {
        final String methodLogName = classLogName + "::Configuration()"; 

        logger.trace(methodLogName); 

        logger.debug(methodLogName + " - commencing meta-configuration."); 

        MetaConfigurator.metaConfigure(this, metaConfigurationFilePath); 

        // Lock the configuration map by dropping its modifiable reference. 
        modifiableConfigurationMap = null; 

        // Mark this Configuration as having been meta-configured. 
        constructionComplete = true; 

        logger.debug(methodLogName + " - meta-configuration complete."); 
    }

    /** <p>Adds the specified ConfigurationOption to this Configuration.</p> 
     * 
     * <p><b>WARNING:</b> This method is not thread-safe. It is intended 
     * only for use by the MetaConfigurator class, and it cannot be called 
     * after this Configuration's constructor returns.</p> 
     * 
     * @param option the ConfigurationOption to add. 
     * 
     * @throws IllegalStateException if this Configuration's constructor 
     *                               has already returned. 
     * @throws MetaConfigurationError if this option is already present in 
     *                                this Configuration. 
     * @throws NullPointerException if option is null. 
     */
    final void addOption(final ConfigurationOption option) 
        throws IllegalStateException, 
               MetaConfigurationError, 
               NullPointerException 
    {
        final String methodLogName = classLogName + "::addOption()"; 

        logger.trace(methodLogName); 

        if(option == null)
        {
            final String message = 
                methodLogName + " - option must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(constructionComplete) 
        {
            final String message = 
                methodLogName 
                + " - this method can only be called at construction time."; 

            logger.fatal(message); 
            throw new IllegalStateException(message); 
        }

        if(modifiableConfigurationMap.containsKey(option.optionName))
        {
            final String message = 
                methodLogName + " - option [" + option.optionName 
                + "] rejected as a duplicate."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }

        // Add the option to the configuration. 
        modifiableConfigurationMap.put(option.optionName, option); 
    }

    /** Returns true if this Configuration has already been meta-configured, 
     * false otherwise. 
     */
    final boolean constructionComplete() 
    {
        final String methodLogName = 
            classLogName + "::constructionComplete()"; 

        logger.trace(methodLogName); 

        return(constructionComplete); 
    }

    /** <p>Returns the subset of this Configuration's options which can be set 
     * from the specified valueSources.</p> 
     * 
     * <p><b>WARNING:</b> the caller may safely modify the map returned by 
     * this method. However, the contents of the ConfigurationOptions held 
     * by the map may only be modified by a method which is synchronised on 
     * this configuration.</p> 
     * 
     * @param valueSources the value sources for which applicable options 
     *                  should be returned (see {@link ConfigurationOption}). 
     * 
     * @throws IllegalArgumentException if valueSources specifies a source not 
     *                      defined by the {@link ConfigurationOption} class. 
     * 
     * @return a map containing the subset of this Configuration's options 
     *         which can be set from the specified valueSources. 
     */
    final Map<String, ConfigurationOption> 
    getSelectedOptionMap(int valueSources) 
        throws IllegalArgumentException 
    {
        final String methodLogName = 
            classLogName + "::getSelectedOptionMap()"; 

        logger.trace(methodLogName); 

        if(!ConfigurationOption.allValueSourcesDefined(valueSources)) 
        {
            final String message = 
                methodLogName + " - valueSources [" 
                + ConfigurationOption.getValueSourceString(valueSources) 
                + "] must specify defined value sources."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        logger.debug(methodLogName + " - selecting [" 
                     + ConfigurationOption.getValueSourceString(valueSources) 
                     + "] options."); 

        // Loop through the configuration map, noting the options which can 
        // be set from the specified value sources and ignoring the rest. 
        final Map<String, ConfigurationOption> selectedOptions = 
            new HashMap<String, ConfigurationOption>(); 

        final Iterator<ConfigurationOption> 
            configurationMapIterator = configurationMap.values().iterator();

        // The current option. 
        ConfigurationOption currentOption = null; 

        while(configurationMapIterator.hasNext())
        {
            currentOption = configurationMapIterator.next(); 

            if((valueSources & currentOption.validValueSources) 
               == ConfigurationOption.NO_SOURCE) 
            {
                logger.debug(methodLogName 
                             + " - ignoring option [" 
                             + currentOption.optionName + "]."); 
                continue; 
            }
            else
            {
                logger.debug(methodLogName 
                             + " - selecting option [" 
                             + currentOption.optionName + "]."); 

                selectedOptions.put(currentOption.optionName, currentOption); 
            }
        }

        return(selectedOptions); 
    }

    /** Update the configuration with any relevant values that might be 
     * available from the specified property file. Any updated values will be 
     * marked with source {@link ConfigurationOption.APPLICATION_PROPERTY}. 
     * 
     * @param propertyFile the property file to load values from. 
     *
     * @throws NullPointerException if propertyFile is null. 
     * 
     * @return true if the update was successful, false otherwise. 
     */ 
    protected final synchronized boolean 
    loadApplicationProperties(final File propertyFile) 
    {
        final String methodLogName = 
            classLogName + "::loadApplicationProperties()"; 

        logger.trace(methodLogName); 

        // Delegate to the corresponding PropertyHandler method. 
        return(PropertyHandler.loadApplicationProperties(propertyFile, this)); 
    }

    /** Update the configuration with any relevant values that might be 
     * available within the system properties. Any updated values will 
     * be marked with source {@link ConfigurationOption.SYSTEM_PROPERTY}. 
     *
     * @return true if the update was successful, false otherwise. 
     */ 
    protected final synchronized boolean loadSystemProperties() 
    {
        final String methodLogName = classLogName + "::loadSystemProperties()";

        logger.trace(methodLogName); 

        // Delegate to the corresponding PropertyHandler method. 
        return(PropertyHandler.loadSystemProperties(this)); 
    }

    /** Update the configuration with any relevant values that might be 
     * available within the specified properties. Any updated values will be 
     * marked with source {@link ConfigurationOption.INTERNAL_CONFIGURATION}. 
     *
     * @param properties the properties to load option values from. 
     * 
     * @throws NullPointerException if properties is null. 
     * 
     * @return true if the update was successful, false otherwise. 
     */ 
    protected final synchronized boolean 
    loadProperties(final Properties properties) 
    {
        final String methodLogName = classLogName + "::loadProperties()"; 

        logger.trace(methodLogName); 

        // Delegate to the corresponding PropertyHandler method. 
        return(PropertyHandler.loadProperties(properties, this)); 
    }

    /** Returns true if this Configuration has parsed a command line, 
     * false otherwise. 
     */
    final boolean commandLineParsed() 
    {
        final String methodLogName = classLogName + "::commandLineParsed()"; 

        logger.trace(methodLogName); 

        return(commandLineParsed); 
    }

    /** Returns the remaining command line arguments (the arguments which 
     * were not consumed by a command line parser, see 
     * {@link #parseCommandLine(String[])}). 
     * 
     * @return an array of Strings containing the remaining command line 
     *         arguments. If no remaining arguments are available, the 
     *         array will have length zero. The returned array does not 
     *         affect the list held by this Configuration object, so the 
     *         caller can safely modify it at will. 
     */ 
    public final synchronized String[] getRemainingCommandLineArguments() 
    {
        final String methodLogName = 
            classLogName + "::getRemainingCommandLineArguments()"; 

        logger.trace(methodLogName); 

        // Create and return an array containing the remaining command 
        // line arguments. 
        final String[] remainingCommandLineArguments = 
            new String[remainingCommandLineArgumentList.size()]; 

        return(remainingCommandLineArgumentList
               .toArray(remainingCommandLineArguments)); 
    }

    /** <p>Sets the remaining command line arguments (the arguments which 
     * were not consumed by a command line parser, see 
     * {@link #parseCommandLine(String[])}) to the given arguments.</p> 
     * 
     * <p><b>WARNING:</b> This method is intended only for use by 
     * the CommandLineHandler class. It may only be called once per 
     * Configuration instance.</p> 
     * 
     * @param arguments the remaining command line arguments. A shallow 
     *                  copy is taken and stored, so that the caller can 
     *                  safely continue using the original. 
     * 
     * @throws IllegalStateException if this method has already been called 
     *                               on this Configuration instance. 
     * @throws NullPointerException if arguments is null. 
     */ 
    final synchronized void 
    setRemainingCommandLineArguments(final String[] arguments) 
        throws IllegalStateException, NullPointerException
    {
        final String methodLogName = 
            classLogName + "::setRemainingCommandLineArguments()"; 

        logger.trace(methodLogName); 

        if(arguments == null)
        {
            final String message = 
                methodLogName + " - arguments must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(commandLineParsed) 
        {
            final String message = 
                methodLogName + " - this method can only be called once" 
                + " per Configuration instance."; 

            logger.fatal(message); 
            throw new IllegalStateException(message); 
        }

        // Take and store a shallow copy of the argument array, so that the 
        // caller can safely continue using the original. 
        // 
        // It is not necessary to copy the Strings themselves, as these 
        // are immutable. 
        for(int i = 0; i < arguments.length; ++i)
        {
            modifiableRemainingCommandLineArgumentList.add(arguments[i]); 
        }

        // Lock the list of remaining command line arguments by dropping its 
        // modifiable reference. 
        modifiableRemainingCommandLineArgumentList = null; 

        // Mark this Configuration as having parsed a command line. 
        commandLineParsed = true; 
    }

    /** <p>Parse the given command line arguments and update the 
     * configuration with any values found. Any arguments not consumed 
     * by the parser will be made available for retrieval using the 
     * {@link #getRemainingCommandLineArguments()} method.</p> 
     *
     * <p><b>WARNING:</b> This method may only be called once per 
     * Configuration instance.</p> 
     * 
     * @throws IllegalStateException if this method has already been called 
     *                               on this Configuration instance. 
     * @throws NullPointerException if arguments is null. 
     * 
     * @param args the command line parameters to parse. 
     *
     * @return true if parsing was successful, false otherwise. 
     */ 
    protected final synchronized boolean 
    parseCommandLine(final String[] arguments) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::parseCommandLine()"; 

        logger.trace(methodLogName); 

        if(commandLineParsed) 
        {
            final String message = 
                methodLogName + " - this method can only be called once" 
                + " per Configuration instance."; 

            logger.fatal(message); 
            throw new IllegalStateException(message); 
        }

        // Delegate to the corresponding CommandLineHandler method. 
        return(CommandLineHandler.parseCommandLine(arguments, this)); 
    }

    /** Returns true if all configuration options have been properly set, 
     * false otherwise. 
     */ 
    protected synchronized boolean checkConfiguration() 
    {
        final String methodLogName = classLogName + "::checkConfiguration()"; 

        logger.trace(methodLogName); 

        // Non-zero if any configuration option does not have a valid value. 
        int unsetOptions = 0; 

        logger.debug(methodLogName + " - checking configuration options."); 

        // Iterate though all configuration options held by the configuration 
        // map, checking each one for a valid value. 
        final Iterator<ConfigurationOption> 
            configurationMapIterator = configurationMap.values().iterator();

        ConfigurationOption currentOption = null; 
            
        while(configurationMapIterator.hasNext())
        {
            currentOption = configurationMapIterator.next(); 

            if((currentOption.getActualValue() == null)
               ^ (currentOption.getActualValueSource() 
                  == ConfigurationOption.NO_SOURCE)) 
            {
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - value source [" 
                    + ConfigurationOption.getValueSourceString(
                        currentOption.getActualValueSource()) 
                    + "] and value [" + currentOption.getActualValue() 
                    + "] are mismatched for option [" 
                    + currentOption.optionName + "]."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }
            else if(currentOption.getActualValue() == null) 
            {
                // This option does not have a valid value - log a message 
                // describing the problem and what to do about it. 
                logger.error(
                    methodLogName + " - option [" 
                    + currentOption.optionName + "] has not been properly" 
                    + " set. It can be set from the following source" 
                    + (ConfigurationOption.isSingleDefinedValueSource(
                           currentOption.validValueSources) ? ": " : "s: ")
                    + ConfigurationOption.getValueSourceString(
                        currentOption.validValueSources) + "."); 

                ++unsetOptions; 
            }
            else
            {
                logger.debug(methodLogName + " - option [" 
                             + currentOption.optionName 
                             + "] has been set from a valid source [" 
                             + ConfigurationOption.getValueSourceString(
                                 currentOption.getActualValueSource()) + "].");
            }
        }

        if(unsetOptions == 0) 
        {
            logger.debug(methodLogName 
                         + " - all configuration options" 
                         + " successfully checked."); 
        }
        else
        {
            logger.error(methodLogName + " - " + unsetOptions 
                         + " configuration option" 
                         + (unsetOptions == 1 ? " has" : "s have")
                         + " not been properly set."); 
        }

        return(unsetOptions == 0); 
    }

    /** Outputs a logging message at the specified log level giving the 
     * name, value and value source of each configuration option. If the 
     * option value is an application option, the property file name is 
     * also logged. 
     * 
     * @param logLevel the Apache Log4J log level to use. 
     * 
     * @throws NullPointerException if logLevel is null. 
     */ 
    protected synchronized void logConfiguration(final Level logLevel) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::logConfiguration()"; 

        logger.trace(methodLogName); 

        if(logLevel == null) 
        {
            final String message = 
                methodLogName + " - logLevel must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Print an initial log message as a header for the following dump. 
        logger.log(logLevel, 
                   methodLogName 
                   + " - configuration options are set as follows:"); 

        // Take a copy of the configuration map and remove the redundant 
        // 'application.XXX' properties containing meta-configuration data. 
        final HashMap<String, ConfigurationOption> optionMap = 
            new HashMap<String, ConfigurationOption>(configurationMap); 

        final Iterator<String> redundantPropertyIterator = 
            MetaConfigurator.mandatoryMetaConfigurationProperties.iterator(); 

        while(redundantPropertyIterator.hasNext()) 
        {
            optionMap.remove(redundantPropertyIterator.next()); 
        } 

        // Iterate though all configuration options held by the configuration 
        // map and generate a log message text for each one. 
        final Iterator<ConfigurationOption> 
            optionMapIterator = optionMap.values().iterator();

        final TreeSet<String> messageSet = new TreeSet<String>(); 

        ConfigurationOption currentOption = null; 
        String logMessage = null; 
            
        while(optionMapIterator.hasNext())
        {
            currentOption = optionMapIterator.next(); 

            // Prepare the common message text for this option. 
            logMessage = 
                currentOption.optionName + "] type [" 
                + ConfigurationOption.getTypeString(currentOption.optionType) 
                + "] value [" 
                + currentOption.getActualValue() + "] source [" 
                + ConfigurationOption.getValueSourceString(
                    currentOption.getActualValueSource()) + "]"; 

            // Add message text specific to this option's value source. 
            switch(currentOption.getActualValueSource()) 
            {
              case ConfigurationOption.NO_SOURCE: 
                logMessage += "[not set yet]."; 
                break; 

              case ConfigurationOption.DEFAULT_VALUE: 
                logMessage += "."; 
                break; 

              case ConfigurationOption.APPLICATION_PROPERTY: 
                logMessage += 
                    " property [" + currentOption.propertyName 
                    + "] file [" + currentOption.getPropertyFile() + "]."; 
                break; 

              case ConfigurationOption.SYSTEM_PROPERTY: 
                logMessage += 
                    " property [" + currentOption.propertyName + "]."; 
                break; 

              case ConfigurationOption.COMMAND_LINE: 
                logMessage += " parameter ["; 

                // The command line options are a little awkward to format, 
                // because there are two types (short and long), where 
                // either or both can be present. 
                boolean useSeparator = false; 

                if(currentOption.commandLineChar != 0) 
                {
                    // This option has a short command line option - add it. 
                    logMessage += "-" + currentOption.commandLineChar; 
                    useSeparator = true; 
                }

                if(currentOption.commandLineName != "") 
                {
                    // This option has a long command line name - add it, 
                    // using a separator if necessary. 
                    logMessage += (useSeparator ? " / " : "") 
                        + "--" + currentOption.commandLineName; 
                }

                logMessage += "]."; 
                break; 

              case ConfigurationOption.INTERNAL_CONFIGURATION: 
                logMessage += "."; 
                break; 

              default: 
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName + "] value source [" 
                    + ConfigurationOption.getValueSourceString(
                        currentOption.getActualValueSource())
                    + "] not recognised."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // Add this option's message to the list. 
            messageSet.add(logMessage); 
        }

        // Log all the option messages. 
        final Iterator<String> messageIterator = messageSet.iterator(); 

        while(messageIterator.hasNext()) 
        {
            logger.log(logLevel, 
                       methodLogName    
                       + " - option [" + messageIterator.next()); 
        }

        // Print a trailing log message as a footer for the preceding dump. 
        logger.log(logLevel, 
                   methodLogName + " - all configuration options logged."); 
    }

    /** Logs the application name and version held by this configuration 
     * at the specified log level. 
     * 
     * @param logLevel the Apache Log4J log level to use. 
     */ 
    protected void logVersionMessage(final Level logLevel) 
    {
        final String methodLogName = classLogName + "::logVersionMessage()"; 

        logger.trace(methodLogName); 

        logger.log(logLevel, 
                   methodLogName + " - " + getString("application.name") 
                   + " version: " + getString("application.version")); 
    }

    /** Prints the application name and version held by this configuration 
     * to standard output. 
     */ 
    protected void printVersionMessage() 
    {
        final String methodLogName = classLogName + "::printVersionMessage()"; 

        logger.trace(methodLogName); 

        logger.debug(methodLogName 
                     + " - printing version message to standard output."); 

        System.out.print(
            new WrappedString("", 
                              getString("application.name") + " version: " 
                              + getString("application.version")) 
            + "\n"); 
    }

    /** Prints a help message describing this configuration's command line 
     * options to standard output. 
     */ 
    protected void printCommandLineHelpMessage() 
    {
        final String methodLogName = 
            classLogName + "::printCommandLineHelpMessage()"; 

        logger.trace(methodLogName); 

        // Delegate to the corresponding CommandLineHandler method. 
        CommandLineHandler.printCommandLineHelpMessage(this); 
    }

    /** Prints a help message describing this configuration's system 
     * properties to standard output. 
     */ 
    protected void printSystemPropertyHelpMessage() 
    {
        final String methodLogName = 
            classLogName + "::printSystemPropertyHelpMessage()"; 

        logger.trace(methodLogName); 

        // Delegate to the corresponding PropertyHandler method. 
        PropertyHandler.printSystemPropertyHelpMessage(this); 
    }

    /** Generates a template property file for configuring this application 
     * at the specified location. The file must not already exist - if it 
     * does, this method will refuse to write to it. 
     * 
     * @param outputFile the location where the template property file should 
     *                   be created. 
     * 
     * @throws NullPointerException if outputFile is null. 
     * 
     * @return true if the template property file was successfully generated, 
     *         false otherwise. 
     */ 
    protected final boolean createTemplatePropertyFile(final File outputFile) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::createTemplatePropertyFile()"; 

        logger.trace(methodLogName); 

        // Delegate to the corresponding PropertyHandler method. 
        return(PropertyHandler.createTemplatePropertyFile(outputFile, this)); 
    }

    /** <p>Returns the configuration option named by optionName.</p> 
     * 
     * <p><b>WARNING:</b> This method is not thread-safe. Call it only from 
     * methods synchronised on this Configuration object.</p> 
     * 
     * @param optionName the name of the desired configuration option. 
     * 
     * @throws IllegalArgumentException if the option name is invalid. 
     * @throws NullPointerException if optionName is null. 
     * 
     * @return the ConfigurationOption object corresponding to optionName. 
     */ 
    private final ConfigurationOption 
    getConfigurationOption(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::getConfigurationOption()"; 

        logger.trace(methodLogName); 

        if(optionName == null) 
        {
            final String message = 
                methodLogName + " - optionName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }
        
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Retrieve the appropriate configuration option. 
        final ConfigurationOption option = configurationMap.get(optionName); 

        if(option == null) 
        {
            // The specified configuration option is invalid. 
            final String message = 
                methodLogName + " - unknown configuration option [" 
                + optionName + "] requested."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        return(option); 
    }

    /** Returns the value of type T corresponding to the named configuration 
     * option. 
     * 
     * @param optionName the configuration option whose value should be 
     *                   returned. 
     * @param valueType the class of the value type the caller expects this 
     *                  method to return. 
     * 
     * @throws NullPointerException if optionName or valueType is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a value.</li> 
     *     <li>the option is not mapped to a value of type T.</li> 
     *   </ul>
     * 
     * @return the value corresponding to the named configuration option. 
     */ 
    private final synchronized <T> T 
    getConfiguredValue(final String optionName, final Class<T> valueType) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getConfiguredValue()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - value requested for option [" 
                     + optionName + "]."); 

        // The option name and value type must not be null. 
        if((optionName == null) || (valueType == null))
        {
            final String message = 
                methodLogName + " - optionName [" + optionName 
                + "] and valueType [" + valueType + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Retrieve the appropriate configuration option. 
        final ConfigurationOption option = getConfigurationOption(optionName);

        if(option.getActualValueSource() == ConfigurationOption.NO_SOURCE)
        {
            // This configuration option does not have a value. 
            final String message = 
                methodLogName + " - configuration option [" 
                + optionName + "] does not have a value."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Retrieve the option's actual value. 
        final Object value = option.getActualValue(); 

        if(!valueType.isInstance(value)) 
        {
            // This configuration option is not mapped to a value of the 
            // appropriate type. 
            final String message = 
                methodLogName + " - configuration option [" + optionName 
                + "] is not mapped to a value of the specified type [mapped: " 
                + (value == null ? null : value.getClass().getName()) 
                + "][specified: " + valueType.getName() + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        return(valueType.cast(value)); 
    }

    /** Returns the String value corresponding to the named configuration 
     * option. 
     * 
     * @param optionName the configuration option whose value should be 
     *                   returned. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a value.</li> 
     *     <li>the option is mapped to a non-String value.</li> 
     *   </ul>
     * 
     * @return the value corresponding to the named configuration option. 
     */
    public final String getString(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getString()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Delegate to the generic value retrieval method. 
        return(getConfiguredValue(optionName, String.class)); 
    }

    /** Returns the Boolean value corresponding to the named configuration 
     * option. 
     * 
     * @param optionName the configuration option whose value should be 
     *                   returned. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a value.</li> 
     *     <li>the option is mapped to a non-Boolean value.</li> 
     *   </ul>
     * 
     * @return the value corresponding to the named configuration option. 
     */
    public final Boolean getBoolean(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getBoolean()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Delegate to the generic value retrieval method. 
        return(getConfiguredValue(optionName, Boolean.class)); 
    }

    /** Returns the Integer value corresponding to the named configuration 
     * option. 
     * 
     * @param optionName the configuration option whose value should be 
     *                   returned. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a value.</li> 
     *     <li>the option is mapped to a non-Integer value.</li> 
     *   </ul>
     * 
     * @return the value corresponding to the named configuration option. 
     */
    public final Integer getInteger(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getInteger()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Delegate to the generic value retrieval method. 
        return(getConfiguredValue(optionName, Integer.class)); 
    }
    
    /** Returns the default value of type T corresponding to the named 
     * configuration option. 
     * 
     * @param optionName the configuration option whose default value should 
     *                   be returned. 
     * @param valueType the class of the value type the caller expects this 
     *                  method to return. 
     * 
     * @throws NullPointerException if optionName or valueType is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a default value.</li> 
     *     <li>the option is not mapped to a value of type T.</li> 
     *   </ul>
     * 
     * @return the default value corresponding to the named configuration 
     *         option. 
     */ 
    private final synchronized <T> T 
    getDefaultValue(final String optionName, final Class<T> valueType) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getDefaultValue()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - default value requested for option ["
                     + optionName + "]."); 

        // The option name and value type must not be null. 
        if((optionName == null) || (valueType == null))
        {
            final String message = 
                methodLogName + " - optionName [" + optionName 
                + "] and valueType [" + valueType + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Retrieve the appropriate configuration option. 
        final ConfigurationOption option = getConfigurationOption(optionName);

        if(!ConfigurationOption.isValidValueSource(
               ConfigurationOption.DEFAULT_VALUE, 
               option.validValueSources)) 
        {
            // This configuration option does not have a default value. 
            final String message = 
                methodLogName + " - configuration option [" 
                + optionName + "] does not have a default value."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Retrieve the option's default value. 
        final Object value = option.defaultValue; 

        if(!valueType.isInstance(value)) 
        {
            // This configuration option is not mapped to a value of the 
            // appropriate type. 
            final String message = 
                methodLogName + " - configuration option [" + optionName 
                + "] is not mapped to a value of the specified type [mapped: " 
                + (value == null ? null : value.getClass().getName()) 
                + "][specified: " + valueType.getName() + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        return(valueType.cast(value)); 
    }

    /** Returns the default String value corresponding to the named 
     * configuration option. 
     * 
     * @param optionName the configuration option whose default value should 
     *                   be returned. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a default value.</li> 
     *     <li>the option is mapped to a non-String value.</li> 
     *   </ul>
     * 
     * @return the default value corresponding to the named configuration 
     *         option. 
     */
    public final String getDefaultString(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getDefaultString()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Delegate to the generic value retrieval method. 
        return(getDefaultValue(optionName, String.class)); 
    }

    /** Returns the default Boolean value corresponding to the named 
     * configuration option. 
     * 
     * @param optionName the configuration option whose default value should 
     *                   be returned. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a default value.</li> 
     *     <li>the option is mapped to a non-Boolean value.</li> 
     *   </ul>
     * 
     * @return the default value corresponding to the named configuration 
     *         option. 
     */
    public final Boolean getDefaultBoolean(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getDefaultBoolean()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Delegate to the generic value retrieval method. 
        return(getDefaultValue(optionName, Boolean.class)); 
    }

    /** Returns the default Integer value corresponding to the named 
     * configuration option. 
     * 
     * @param optionName the configuration option whose default value should 
     *                   be returned. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul>
     *     <li>the option name is invalid.</li> 
     *     <li>the option does not have a default value.</li> 
     *     <li>the option is mapped to a non-Integer value.</li> 
     *   </ul>
     * 
     * @return the default value corresponding to the named configuration 
     *         option. 
     */
    public final Integer getDefaultInteger(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::getDefaultInteger()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - option [" 
                     + optionName + "] requested."); 

        // Delegate to the generic value retrieval method. 
        return(getDefaultValue(optionName, Integer.class)); 
    }
    
    /** Sets the value of the configuration option with name optionName to 
     * optionValue. The value source will be updated to 
     * {@link ConfigurationOption#INTERNAL_CONFIGURATION}. 
     * 
     * @param optionName the name of the configuration option to set. 
     * @param optionValue the new value. 
     * 
     * @throws NullPointerException if optionName or optionValue is null. 
     * @throws IllegalArgumentException if: 
     *   <ul> 
     *     <li>optionName does not match any configuration option.</li> 
     *     <li>if optionValue has the wrong data type.</li>
     *     <li>if INTERNAL_CONFIGURATION is not allowed for this option.</li> 
     *   </ul>
     */ 
    public final synchronized void setValue(final String optionName, 
                                            final Object optionValue) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::setValue()"; 

        logger.trace(methodLogName); 

        if((optionName == null) || (optionValue == null)) 
        {
            final String message = 
                methodLogName + " - optionName [" + optionName 
                + "] and optionValue [" + optionValue + "] must not be null.";

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Retrieve the appropriate configuration option. 
        final ConfigurationOption option = getConfigurationOption(optionName);

        // Delegate to the option's instance method. 
        option.setActualValue(optionValue, 
                              ConfigurationOption.INTERNAL_CONFIGURATION, 
                              null); 
    }

    /** Sets the value of the named configuration option to its default. 
     * The value source will be updated to 
     * {@link ConfigurationOption#INTERNAL_CONFIGURATION} (<i>not</i> 
     * {@link ConfigurationOption#DEFAULT_VALUE}). 
     * 
     * @param optionName the name of the configuration option to set. 
     * 
     * @throws NullPointerException if optionName is null. 
     * @throws IllegalArgumentException if: 
     *   <ul> 
     *     <li>optionName does not match any configuration option.</li> 
     *     <li>if the option does not have a default value.</li> 
     *     <li>if INTERNAL_CONFIGURATION is not allowed for this option.</li> 
     *   </ul>
     */ 
    public final synchronized void setValueToDefault(final String optionName) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::setValueToDefault()"; 

        logger.trace(methodLogName); 

        if(optionName == null) 
        {
            final String message = 
                methodLogName + " - optionName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Retrieve the appropriate configuration option. 
        final ConfigurationOption option = getConfigurationOption(optionName);

        if(!ConfigurationOption.isValidValueSource(
               ConfigurationOption.DEFAULT_VALUE, 
               option.validValueSources)) 
        {
            // This configuration option does not have a default value. 
            final String message = 
                methodLogName + " - configuration option [" 
                + optionName + "] does not have a default value."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Delegate to the option's instance method. 
        option.setActualValue(option.defaultValue, 
                              ConfigurationOption.INTERNAL_CONFIGURATION, 
                              null); 
    }

    /** Returns a bitmask of valid sources (see {@link ConfigurationOption} 
     * for the given option's value. 
     *
     * @param optionName the name of the configuration option whose valid 
     *                   value sources should be returned. 
     * 
     * @throws IllegalArgumentException if the option name is invalid. 
     * @throws NullPointerException if optionName is null. 
     * 
     * @return a bitmask of option value sources 
     *         (see {@link ConfigurationOption}). 
     */ 
    public final synchronized int getValidValueSources(String optionName) 
        throws NullPointerException, IllegalArgumentException 
    {
        final String methodLogName = 
            classLogName + "::getValidValueSources()"; 

        logger.trace(methodLogName); 

        if(optionName == null) 
        {
            final String message = 
                methodLogName + " - optionName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }
        
        // Retrieve the appropriate configuration option and return its 
        // bitmask of valid sources. 
        return(getConfigurationOption(optionName).validValueSources); 
    }

    /** Returns true if the specified valueSource is a valid source for 
     * the named option's value, false otherwise. 
     * 
     * @param optionName the name of the configuration option to check. 
     * @param valueSource the value source to test the validity of 
     *                    (see {@link ConfigurationOption}). 
     * 
     * @throws IllegalArgumentException if the option name is invalid, or 
     *                                  if valueSource is not a single, 
     *                                  defined value source.
     * @throws NullPointerException if optionName is null. 
     * 
     * @return true if the specified valueSource is a valid source for 
     *         the named option's value, false otherwise. 
     */ 
    public final synchronized boolean 
    isValidValueSource(final String optionName, final int valueSource) 
        throws NullPointerException, IllegalArgumentException 
    {
        final String methodLogName = classLogName + "::isValidValueSource()"; 

        logger.trace(methodLogName); 

        if(optionName == null) 
        {
            final String message = 
                methodLogName + " - optionName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Retrieve the appropriate configuration option, check the validity 
        // of the specified value source, and return the result. 
        return(ConfigurationOption.isValidValueSource(
                   valueSource, 
                   getConfigurationOption(optionName).validValueSources)); 
    }

    /** Returns a constant (see {@link ConfigurationOption} identifying 
     * the source of the named option's current value. 
     * 
     * @param optionName the name of the configuration option whose current 
     *                   value source should be returned. 
     * 
     * @throws IllegalArgumentException if the option name is invalid. 
     * @throws NullPointerException if optionName is null. 
     * 
     * @return a constant identifying the source of the option's value 
     *         (see {@link ConfigurationOption}). 
     */ 
    public final synchronized int 
    getActualValueSource(final String optionName) 
        throws NullPointerException, IllegalArgumentException 
    {
        final String methodLogName = 
            classLogName + "::getActualValueSource()"; 

        logger.trace(methodLogName); 

        if(optionName == null) 
        {
            final String message = 
                methodLogName + " - optionName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }
        
        // Retrieve the appropriate configuration option and return the 
        // source of its current value. 
        return(getConfigurationOption(optionName).getActualValueSource()); 
    }

    /** Returns true if a value is available (has been set) for the named 
     * configuration option, false otherwise. 
     * 
     * @param optionName the name of the configuration option to check 
     *                   for a value. 
     * 
     * @throws IllegalArgumentException if the option name is invalid. 
     * @throws NullPointerException if optionName is null. 
     * 
     * @return true if a value is available for the named option, false 
     *         otherwise. 
     */ 
    public final boolean valueAvailable(final String optionName)
        throws NullPointerException, IllegalArgumentException 
    {
        final String methodLogName = classLogName + "::valueAvailable()"; 

        logger.trace(methodLogName); 

        // Check whether the specified configuration option has had a value 
        // set for it and return the result. 
        return(getActualValueSource(optionName) 
               != ConfigurationOption.NO_SOURCE); 
    }
}
