package org.eso.util.configuration; 

import java.io.InputStream;

import java.util.Arrays; 
import java.util.Collections;
import java.util.Enumeration; 
import java.util.HashMap;
import java.util.Iterator; 
import java.util.List; 
import java.util.Properties;
import java.util.TreeSet;

import org.apache.log4j.Logger;

/** <p>The MetaConfigurator class provides common infrastructure for 
 * determining how to configure an application. However, applications 
 * should not use it directly - they should subclass {@link Configuration} 
 * and call through its constructor.</p> 
 * 
 * <b>THE META-CONFIGURATION FILE FORMAT</b> 
 * 
 * <p><b>NOTE:</b> Meta-configurations are not normally intended for the 
 * end user. It is strongly recommended that the meta-configuration file 
 * be deployed within the application's JAR file.</p> 
 * 
 * <p><a name="Table1">Meta-configuration</a> files are standard Java 
 * property files. They must contain the following properties: 
 *   <table border=1> 
 *     <caption>Table 1: Mandatory Meta-Configuration Properties</caption> 
 *     <thead>
 *       <tr>
 *         <th>Meta-Configuration Property</th> 
 *         <th>Value Type</th> 
 *         <th>Description</th>
 *       </tr> 
 *     </thead> 
 *     <tbody>
 *       <tr>
 *         <td>application.name</td>
 *         <td>STRING</td>
 *         <td>The application name.</td>
 *       </tr> 
 *       <tr>
 *         <td>application.command</td>
 *         <td>STRING</td>
 *         <td>The actual command for invoking the application.</td>
 *       </tr> 
 *       <tr>
 *         <td>application.version</td>
 *         <td>STRING</td>
 *         <td>The application version.</td>
 *       </tr> 
 *       <tr>
 *         <td>application.help.header.text</td>
 *         <td>STRING</td>
 *         <td>The application's help header text.</td>
 *       </tr> 
 *       <tr>
 *         <td>application.help.footer.text</td>
 *         <td>STRING</td>
 *         <td>The application's help footer text.</td>
 *       </tr> 
 *     </tbody>
 *   </table> 
 * </p> 
 *
 * <p>An example set of mandatory meta-configuration properties: 
 *   <pre><code>
 *      # Meta-configuration for the Reduction Block Scheduler.
 *      application.name=Reduction Block Scheduler
 *      application.command=RBScheduler
 *      application.version=1.0beta1
 * 
 *      # General help text.
 *      application.help.header.text=\
 *          The Reduction Block Scheduler accepts reduction blocks in the \
 *          normal DFS format and executes the jobs required to achieve \
 *          those reductions. If neither named-RB nor single-pass mode is \
 *          enabled, the RBS will automatically enter daemon mode.
 *      application.help.footer.text=\
 *          In the event of problems, please contact the author:\
 *        \n    Neil Ferguson - ESO/SDD/DFI/BET - &lt;nferguso@eso.org&gt;.
 *   </code></pre>
 * </p>
 * 
 * <p>These properties are made available to the application as unmodifiable 
 * {@link ConfigurationOption#DEFAULT_VALUE DEFAULT_VALUE} configuration 
 * options with {@link ConfigurationOption#STRING STRING} values. The option 
 * names are the same as the meta-configuration property names, which means 
 * that these names are reserved. Applications may not reuse them to name 
 * other configuration options.</p> 
 * 
 * <p><a name="Table2">The</a> application's own configuration options are 
 * specified using sets of option properties, which look like this: 
 *   <table border=1> 
 *     <caption>Table 2: Meta-Configuring Application Options</caption> 
 *     <thead>
 *       <tr>
 *         <th>Meta-Configuration Property</th> 
 *         <th>Value Type</th> 
 *         <th>Usage</th>
 *         <th>Description</th>
 *       </tr> 
 *     </thead> 
 *     <tbody>
 *       <tr>
 *         <td><i>NAME</i>.type</td>
 *         <td>STRING</td>
 *         <td>REQUIRED</td>
 *         <td>The data type of this option's value: STRING, BOOLEAN or 
 *             INTEGER.</td>
 *       </tr> 
 *       <tr>
 *         <td><i>NAME</i>.sources</td>
 *         <td>STRING[,STRING]*</td>
 *         <td>REQUIRED</td>
 *         <td>A comma-separated list of permissible sources for this 
 *             option's value. Valid sources are: DefaultValue, 
 *             ApplicationProperty, SystemProperty, CommandLine, 
 *             InternalConfiguration or (for convenience) AllSources.</td>
 *       </tr> 
 *       <tr>
 *         <td><i>NAME</i>.command.line.name</td>
 *         <td>STRING</td>
 *         <td>If this option can be configured from the command line.</td>
 *         <td>The long name for this option on the command line.</td>
 *       </tr> 
 *       <tr>
 *         <td><i>NAME</i>.command.line.char</td>
 *         <td>CHARACTER</td>
 *         <td>If this option can be configured from the command line.</td>
 *         <td>The single-character name for this option on the command 
 *             line.</td>
 *       </tr> 
 *       <tr>
 *         <td><i>NAME</i>.property.name</td>
 *         <td>STRING</td>
 *         <td>If this option can be configured as a system property, or 
 *             from an application property file, and the property name 
 *             should be something other than <i>NAME</i>.</td>
 *         <td>The name of this option when given as a Java property.</td>
 *       </tr> 
 *       <tr>
 *         <td><i>NAME</i>.default.value</td>
 *         <td>Matching the data type of this option's value.</td>
 *         <td>If this option can have a default value.</td>
 *         <td>The default value of this option.</td>
 *       </tr> 
 *       <tr>
 *         <td><i>NAME</i>.help.text</td>
 *         <td>STRING</td>
 *         <td>OPTIONAL</td>
 *         <td>A help text describing this option.</td>
 *       </tr> 
 *     </tbody>
 *   </table> 
 * </p> 
 * 
 * <p>For each set of option properties, the text "<i>NAME</i>" should 
 * be replaced with the name that the application will use to retrieve 
 * values for that option. It is recommended that the name conform to the 
 * conventions for Java property names.</p>
 * 
 * <p>Here is an example which meta-configures two application options: 
 *   <pre><code>
 *      # Option: enables daemon mode. This option is internally configured
 *      #         - if neither named-rb nor single-pass mode is enabled, 
 *      #         then the RBS goes into daemon mode.
 *      rbs.daemon.mode.type=BOOLEAN
 *      rbs.daemon.mode.sources=InternalConfiguration
 * 
 *      # Option: specifies the location of the RB directory. 
 *      rbs.rb.directory.type=STRING
 *      rbs.rb.directory.sources=DefaultValue,ApplicationProperty,CommandLine
 *      rbs.rb.directory.command.line.name=rb-directory
 *      rbs.rb.directory.command.line.char=r
 *      rbs.rb.directory.property.name=rb.directory.path
 *      rbs.rb.directory.default.value=
 *      rbs.rb.directory.help.text=\
 *          Specifies the directory to search for RB files in single-pass \
 *          or daemon mode.
 *   </pre></code> 
 * Note that omitting <tt>rbs.rb.directory.property.name</tt> would cause 
 * the property name to default to <tt>rbs.rb.directory</tt>.
 * </p>
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
final class MetaConfigurator 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(MetaConfigurator.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "MetaConfigurator"; 

    // The meta-configuration property file should not normally be modified 
    // by the application, so this class does not retain a reference to it 
    // or the properties it contains. 

    /** List of mandatory meta-configuration property names. These names 
     * correspond to the entries in <a href="#Table1">Table 1</a>, and are 
     * used by the {@link #metaConfigure(Configuration, String)} method. 
     * They are reused to designate the corresponding standard values in 
     * the application configuration. 
     */
    static final List<String> mandatoryMetaConfigurationProperties = 
        Collections.unmodifiableList(
            Arrays.asList(
                // The application name. 
                "application.name", 
                // The actual command for invoking the application. 
                "application.command", 
                // The application version. 
                "application.version", 
                // The application's help header text. 
                "application.help.header.text", 
                // The application's help footer text. 
                "application.help.footer.text")); 

    /** List of option meta-configuration property names. These values are 
     * used by the {@link #metaConfigureOption(Properties, String)} method; 
     * they correspond to the entries in <a href="#Table2">Table 2</a>. 
     */ 
    private static final List<String> optionMetaConfigurationProperties = 
        Collections.unmodifiableList(
            Arrays.asList(
                // The data type of this option's value. 
                "type", 
                // The permissible sources for this option's value. 
                "sources", 
                // The long name for this option on the command line. 
                "command.line.name", 
                // The single-char name for this option on the command line. 
                "command.line.char", 
                // The name of this option when given as a Java property. 
                "property.name", 
                // The default value of this option. 
                "default.value", 
                // The help text describing this option. 
                "help.text")); 

    /** Private constructor - this class should not be instantiated. */ 
    private MetaConfigurator() 
    {
        final String methodLogName = classLogName + "::MetaConfigurator()"; 

        logger.trace(methodLogName); 

        // This constructor does not need to do anything. 
    }

    /** <p>Initialises an instance of the {@link Configuration} class, as 
     * directed by the specified meta-configuration file. The file must be 
     * reachable via the JVM's classpath.</p> 
     *
     * <p>This method uses the mandatory meta-configuration properties 
     * specified in <a href="#Table1">Table 1</a>. It also calls methods 
     * which use the properties specified in <a href="#Table2">Table 2</a>. 
     * </p> 
     * 
     * @param configuration the Configuration instance to meta-configure. 
     * @param metaConfigurationFilePath location of the meta-configuration 
     *                                  file within the JVM's classpath. 
     *
     * @throws NullPointerException if the configuration or 
     *                              metaConfigurationFilePath is null. 
     * @throws MetaConfigurationError if:
     *   <ul>
     *     <li>the configuration has already been meta-configured.</li> 
     *     <li>the meta-configuration file is missing or unusable.</li> 
     *   </ul>
     */
    static void metaConfigure(final Configuration configuration, 
                              final String metaConfigurationFilePath) 
        throws NullPointerException, MetaConfigurationError 
    {
        final String methodLogName = classLogName + "::metaConfigure()"; 

        logger.trace(methodLogName); 

        if((configuration == null) || (metaConfigurationFilePath == null)) 
        {
            final String message = 
                methodLogName + " - configuration [" + configuration 
                + "] and metaConfigurationFilePath [" 
                + metaConfigurationFilePath + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(configuration.constructionComplete() == true) 
        {
            final String message = 
                methodLogName 
                + " - this configuration has already been meta-configured."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }

        logger.debug(methodLogName + " - commencing meta-configuration."); 

        // Find the meta-configuration file and get an InputStream for it. 
        final InputStream metaConfigurationStream = 
            ClassLoader.getSystemResourceAsStream(metaConfigurationFilePath); 

        if(metaConfigurationStream == null)
        {
            final String message = 
                methodLogName + " - meta-configuration file [" 
                + metaConfigurationFilePath 
                + "] could not be found on the JVM's classpath."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }

        logger.debug(methodLogName + " - opened meta-configuration file [" 
                     + metaConfigurationFilePath + "]."); 

        // Load the raw meta-configuration properties. 
        final Properties metaConfigurationProperties = new Properties(); 

        try
        {
            metaConfigurationProperties.load(metaConfigurationStream); 
        }
        catch(Exception e) 
        {
            final String message = 
                methodLogName 
                + " - meta-configuration properties could not be loaded" 
                + " from file [" + metaConfigurationFilePath + "]."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message, e); 
        }

        logger.debug(methodLogName 
                     + " - loaded meta-configuration properties from" 
                     + " file [" + metaConfigurationFilePath + "]."); 

        // Add the mandatory properties to the configuration. 
        final Iterator<String> mandatoryMetaConfigurationPropertyIterator = 
            mandatoryMetaConfigurationProperties.iterator(); 

        String currentMandatoryMetaConfigurationProperty = null; 
        String propertyValue = null; 

        while(mandatoryMetaConfigurationPropertyIterator.hasNext()) 
        {
            currentMandatoryMetaConfigurationProperty = 
                mandatoryMetaConfigurationPropertyIterator.next(); 

            // Get the value string for the current meta-configuration 
            // property. 
            propertyValue = 
                metaConfigurationProperties.getProperty(
                    currentMandatoryMetaConfigurationProperty); 

            if(propertyValue == null)
            {
                final String message = 
                    methodLogName + " - mandatory property [" 
                    + currentMandatoryMetaConfigurationProperty
                    + " is missing from meta-configuration file [" 
                    + metaConfigurationFilePath + "]."; 

                logger.fatal(message); 
                throw new MetaConfigurationError(message); 
            }

            // Add the property to the configuration. 
            configuration.addOption(
                new ConfigurationOption(
                    currentMandatoryMetaConfigurationProperty, 
                    ConfigurationOption.STRING, 
                    ConfigurationOption.DEFAULT_VALUE, 
                    ConfigurationOption.DEFAULT_VALUE, 
                    "", 
                    '\0', 
                    "", 
                    null, 
                    propertyValue,
                    propertyValue, 
                    "")); 

            logger.debug(methodLogName + " - mandatory property [" 
                         + currentMandatoryMetaConfigurationProperty + "][" 
                         + propertyValue + "] stored in configuration."); 
        }

        logger.debug(methodLogName + " - " 
                     + mandatoryMetaConfigurationProperties.size() 
                     + " mandatory meta-configuration propert" 
                     + (mandatoryMetaConfigurationProperties.size() == 1 
                        ? "y" : "ies") + " loaded."); 

        // Discover the list of meta-configured application options. 
        final String[] applicationOptions = 
            findOptionPropertyNames(metaConfigurationProperties); 

        // Add the application options to the configuration. 
        ConfigurationOption currentOption = null; 

        for(int i = 0; i < applicationOptions.length; ++i)
        {
            // Add the current application option to the configuration. 
            currentOption = metaConfigureOption(metaConfigurationProperties, 
                                                applicationOptions[i]); 
            configuration.addOption(currentOption); 

            logger.debug(methodLogName + " - application option [" 
                         + applicationOptions[i] 
                         + "] stored in configuration as [" 
                         + currentOption.optionName + "]."); 
        }

        logger.debug(methodLogName + " - meta-configured " 
                     + applicationOptions.length 
                     + " application option" 
                     + (applicationOptions.length == 1 ? "." : "s.")); 

        logger.debug(methodLogName + " - meta-configuration complete."); 
    }

    /** <p>Search the metaConfigurationProperties for properties with names 
     * matching the first entry in <a href="#Table2">Table 2</a> (see 
     * {@link #optionMetaConfigurationProperties}). Extract the option 
     * property names and return the list of names found. 
     * 
     * @param metaConfigurationProperties the property set containing the 
     *                                    meta-configuration. 
     * 
     * @throws NullPointerException if metaConfigurationProperties is null. 
     * 
     * @return the list of option property names found. 
     */ 
    private static String[] 
    findOptionPropertyNames(final Properties metaConfigurationProperties) 
        throws NullPointerException
    {
        final String methodLogName = 
            classLogName + "::findOptionPropertyNames()"; 

        logger.trace(methodLogName);

        if(metaConfigurationProperties == null)
        {
            final String message = 
                methodLogName 
                + " - metaConfigurationProperties must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Get the full set of meta-configuration property names. 
        Enumeration allPropertyNames = 
            metaConfigurationProperties.propertyNames(); 

        // The suffix of the first option meta-configuration property. 
        String sampleSuffix = "." + optionMetaConfigurationProperties.get(0); 

        // The set of extracted option property names. 
        TreeSet<String> optionPropertyNames = new TreeSet<String>(); 

        // Iterate through the names, extracting those of interest. 
        String currentPropertyName = null; 

        while(allPropertyNames.hasMoreElements()) 
        {
            // Get the next property name. 
            currentPropertyName = (String)allPropertyNames.nextElement(); 

            if(currentPropertyName.endsWith(sampleSuffix)) 
            {
                // This meta-configuration property name has the desired form 
                // - extract the option property name and store it. 
                currentPropertyName = 
                    currentPropertyName
                    .substring(0, 
                               (currentPropertyName.length() 
                                - sampleSuffix.length())); 
                        
                optionPropertyNames.add(currentPropertyName); 

                logger.debug(methodLogName 
                             + " - discovered option property name [" 
                             + currentPropertyName + "]."); 
                    
            }
        }

        logger.debug(methodLogName + " - " + optionPropertyNames.size() 
                     + " option property name" 
                     + (optionPropertyNames.size() == 1 ? "" : "s") 
                     + " discovered."); 

        // The extracted option property names are returned as an array. 
        final String[] optionPropertyNameArray = 
            new String[optionPropertyNames.size()]; 

        return(optionPropertyNames.toArray(optionPropertyNameArray)); 
    }

    /** <p>Search the metaConfigurationProperties for the definition of 
     * an application option specified by optionName, and return a 
     * {@link ConfigurationOption} object corresponding to that option.</p> 
     * 
     * <p>This method uses the meta-configuration properties specified 
     * in <a href="#Table2">Table 2</a> (see 
     * {@link #optionMetaConfigurationProperties}).</p> 
     * 
     * @param metaConfigurationProperties the property set containing the 
     *                                 meta-configuration for this option. 
     * @param optionName the name of the option to meta-configure. 
     * 
     * @throws NullPointerException if metaConfigurationProperties or 
     *                              optionName is null. 
     * @throws MetaConfigurationError if the metaConfigurationProperties 
     *                                are malformed or incomplete. 
     * 
     * @return a ConfigurationOption object. 
     */ 
    private static ConfigurationOption 
    metaConfigureOption(final Properties metaConfigurationProperties, 
                        final String optionName) 
        throws NullPointerException, MetaConfigurationError 
    {
        final String methodLogName = classLogName + "::metaConfigureOption()"; 

        logger.trace(methodLogName); 

        if((metaConfigurationProperties == null) 
           || (optionName == null))
        {
            final String message = 
                methodLogName + " - metaConfigurationProperties [" 
                + metaConfigurationProperties + "] and optionName [" 
                + optionName + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Perform basic sanity checks on the option name. 
        if(optionName.equals(""))
        {
            final String message = 
                methodLogName 
                + " - the application option name must not be blank."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }
        
        if(optionName.matches("\\s"))
        {
            final String message = 
                methodLogName + " - the application option name [" 
                + optionName + "] must not contain whitespace."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }
        
        // Construct the required meta-configuration property names. 
        final HashMap<String, String> propertyNameMap = 
            new HashMap<String, String>(); 

        final Iterator<String> optionMetaConfigurationPropertyIterator = 
            optionMetaConfigurationProperties.iterator(); 

        String currentOptionMetaConfigurationProperty = null; 

        while(optionMetaConfigurationPropertyIterator.hasNext()) 
        {
            currentOptionMetaConfigurationProperty = 
                optionMetaConfigurationPropertyIterator.next(); 

            propertyNameMap.put(currentOptionMetaConfigurationProperty, 
                                optionName + "." 
                                + currentOptionMetaConfigurationProperty); 
        }

        // Attempt to collect all the required meta-configuration property 
        // values. Both leading and trailing whitespace are removed. 
        // 
        // If a value is not specified, it defaults to the empty string - 
        // this allows a single error check to cover both the 'missing' and 
        // 'empty string' cases later on. 
        final String optionTypeString = 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("type"), 
                "").trim(); 
        final String validValueSourcesString = 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("sources"), 
                "").trim(); 
        final String commandLineName = 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("command.line.name"), 
                "").trim(); 
        final String commandLineCharString = 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("command.line.char"), 
                "").trim(); 
        String propertyName =           // This is deliberately not final. 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("property.name"), 
                "").trim(); 
        final String defaultValueString = 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("default.value"), 
                "").trim(); 
        final String helpText = 
            metaConfigurationProperties.getProperty(
                propertyNameMap.get("help.text"), 
                "").trim(); 
        
        // These are set later, based on the properties seen above. 
        int optionType = ConfigurationOption.NO_TYPE;
        int validValueSources = ConfigurationOption.NO_SOURCE; 
        int actualValueSource = ConfigurationOption.NO_SOURCE; 
        char commandLineChar = 0; 
        Object defaultValue = null; 

        // Determine the data type of this option's value. 
        if(optionTypeString.equals(""))
        {
            final String message = 
                methodLogName + " - meta-configuration property [" 
                + propertyNameMap.get("type")
                + "] must specify a valid data type."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }

        if(optionTypeString.equalsIgnoreCase("STRING"))
        {
            optionType = ConfigurationOption.STRING; 
        }
        else if(optionTypeString.equalsIgnoreCase("BOOLEAN"))
        {
            optionType = ConfigurationOption.BOOLEAN; 
        }
        else if(optionTypeString.equalsIgnoreCase("INTEGER"))
        {
            optionType = ConfigurationOption.INTEGER; 
        }
        else
        {
            final String message = 
                methodLogName + " - meta-configuration property [" 
                + propertyNameMap.get("type")
                + "] specified an unknown data type [" 
                + optionTypeString + "]."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }

        logger.debug(methodLogName + " - application option [" 
                     + optionName + "] has data type [" 
                     + optionTypeString.toUpperCase() + "]."); 

        // The list of permissible sources for this option's value is comma 
        // separated. Tokenise it and store the value source names. 
        final String[] validValueSourceArray = 
            validValueSourcesString.split(","); 

        // Assemble the set of permissible sources. 
        for(int i = 0; i < validValueSourceArray.length; ++i)
        {
            // Trim any whitespace from the current value source string. 
            validValueSourceArray[i] = validValueSourceArray[i].trim(); 

            if(validValueSourceArray[i].equalsIgnoreCase("DefaultValue"))
            {
                validValueSources |= ConfigurationOption.DEFAULT_VALUE; 
            }
            else if(validValueSourceArray[i]
                    .equalsIgnoreCase("ApplicationProperty"))
            {
                validValueSources |= 
                    ConfigurationOption.APPLICATION_PROPERTY; 
            }
            else if(validValueSourceArray[i]
                    .equalsIgnoreCase("SystemProperty"))
            {
                validValueSources |= ConfigurationOption.SYSTEM_PROPERTY; 
            }
            else if(validValueSourceArray[i].equalsIgnoreCase("CommandLine")) 
            {
                validValueSources |= ConfigurationOption.COMMAND_LINE; 
            }
            else if(validValueSourceArray[i]
                    .equalsIgnoreCase("InternalConfiguration"))
            {
                validValueSources |= 
                    ConfigurationOption.INTERNAL_CONFIGURATION; 
            }
            else if(validValueSourceArray[i].equalsIgnoreCase("AllSources")) 
            {
                validValueSources |= ConfigurationOption.ALL_SOURCES; 
            }
            else 
            {
                final String message = 
                    methodLogName + " - meta-configuration property [" 
                    + propertyNameMap.get("sources")
                    + "] specified an unknown source [" 
                    + validValueSourceArray[i] 
                    + "] for the option's value."; 

                logger.fatal(message); 
                throw new MetaConfigurationError(message); 
            }
        }
        
        if(validValueSources == ConfigurationOption.NO_SOURCE) 
        {
            final String message = 
                methodLogName + " - meta-configuration property [" 
                + propertyNameMap.get("sources")
                + "] must specify at least one permissible source" 
                + " for the option's value."; 

            logger.fatal(message); 
            throw new MetaConfigurationError(message); 
        }

        logger.debug(methodLogName + " - application option [" 
                     + optionName + "] can be set by [" 
                     + ConfigurationOption
                       .getValueSourceString(validValueSources) + "]."); 

        // If this option can be set from the command line, ensure that either 
        // a long option name or a short option character has been specified. 
        // The meta-configuration property for a short option character may 
        // only contain a single character. 
        if(ConfigurationOption.isValidValueSource(
               ConfigurationOption.COMMAND_LINE, 
               validValueSources)) 
        {
            if(commandLineName.equals("") && commandLineCharString.equals("")) 
            {
                final String message = 
                    methodLogName + " - application option [" 
                    + optionName + "] is a command line option." 
                    + " At least one of the meta-configuration properties [" 
                    + propertyNameMap.get("command.line.name") + "] and [" 
                    + propertyNameMap.get("command.line.char") 
                    + "] must be specified."; 

                logger.fatal(message); 
                throw new MetaConfigurationError(message); 
            }
            
            if(commandLineCharString.length() > 1)
            {
                final String message = 
                    methodLogName + " - meta-configuration property [" 
                    + propertyNameMap.get("command.line.char") 
                    + "] may only specify a single character."; 

                logger.fatal(message); 
                throw new MetaConfigurationError(message); 
            }

            if(commandLineCharString.length() == 1)
            {
                // This option can be set from the command line, and it 
                // has a short option character - extract the character. 
                commandLineChar = commandLineCharString.charAt(0); 
            }

            logger.debug(methodLogName + " - application option [" + optionName
                         + "] is a command line option with long name ["
                         + commandLineName + "] and short name [" 
                         + commandLineChar + "]."); 
        }
        else if(!commandLineName.equals("") 
                || !commandLineCharString.equals("")) 
        {
            logger.warn(methodLogName + " - application option [" + optionName 
                        + "] is not a command line option." 
                        + " The meta-configuration properties [" 
                        + propertyNameMap.get("command.line.name") + "] and [" 
                        + propertyNameMap.get("command.line.char") 
                        + "] are redundant."); 
        }

        // If this option can be set using a system or application property, 
        // ensure that the property name has been specified, or default to 
        // the application option name. 
        if(ConfigurationOption.isValidValueSource(
               ConfigurationOption.APPLICATION_PROPERTY, 
               validValueSources)
           || ConfigurationOption.isValidValueSource(
               ConfigurationOption.SYSTEM_PROPERTY, 
               validValueSources)) 
        {
            if(propertyName.equals(""))
            {
                propertyName = optionName; 
            }

            logger.debug(methodLogName + " - application option [" + optionName
                         + "] is a system or application property with name [" 
                         + propertyName + "]."); 
        }
        else if(!propertyName.equals(""))
        {
            logger.warn(methodLogName + " - application option [" + optionName 
                        + "] is not a system or application property." 
                        + " The meta-configuration property [" 
                        + propertyNameMap.get("property.name") 
                        + "] is redundant."); 
        }

        // If this option has a default value, parse the meta-configuration 
        // property according to the appropriate data type. 
        if(ConfigurationOption.isValidValueSource(
               ConfigurationOption.DEFAULT_VALUE, 
               validValueSources)) 
        {
            if(defaultValueString.equals("")) 
            {
                // The default value for a STRING option is allowed to be 
                // blank, but only if it was explicitly specified as such. 
                // 
                // For any other option type, a blank default value is 
                // always an error. 
                if((optionType != ConfigurationOption.STRING) 
                   || (metaConfigurationProperties.getProperty(
                           propertyNameMap.get("default.value")) == null)) 
                {
                    final String message = 
                        methodLogName + " - application option [" 
                        + optionName + "] should have a default value." 
                        + "The meta-configuration property [" 
                        + propertyNameMap.get("default.value") 
                        + "] must be used to specify one."; 

                    logger.fatal(message); 
                    throw new MetaConfigurationError(message); 
                }
            }

            // Attempt to parse the default value string according to the 
            // desired option type. 
            switch(optionType)
            {
              case ConfigurationOption.STRING: 
                defaultValue = defaultValueString; 
                break; 

              case ConfigurationOption.BOOLEAN: 
                defaultValue = Boolean.valueOf(defaultValueString); 
                break; 

              case ConfigurationOption.INTEGER: 
                try
                {
                    defaultValue = Integer.valueOf(defaultValueString); 
                }
                catch(NumberFormatException nfe)
                {
                    final String message = 
                        methodLogName 
                        + " - the default value [" + defaultValueString 
                        + "] specified by meta-configuration property [" 
                        + propertyNameMap.get("default.value") 
                        + "] could not be parsed as an INTEGER value [" 
                        + nfe.getMessage() + "]."; 

                    logger.fatal(message); 
                    throw new MetaConfigurationError(message); 
                }
                break; 

              default:
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - invalid optionType [" 
                    + ConfigurationOption.getTypeString(optionType) 
                    + "] encountered whilst processing meta-configuration" 
                    + " property [" + propertyNameMap.get("default.value") 
                    + "]."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // At this stage, a fully-parsed default value is present and 
            // will be used. 
            actualValueSource = ConfigurationOption.DEFAULT_VALUE; 

            logger.debug(methodLogName + " - application option [" + optionName
                         + "] has default value [" + defaultValueString 
                         + "], parsed as [" + defaultValue + "]."); 
        }
        else if(!defaultValueString.equals(""))
        {
            logger.warn(methodLogName + " - application option [" + optionName 
                        + "] does not have a default value." 
                        + " The meta-configuration property [" 
                        + propertyNameMap.get("default.value") 
                        + "] is redundant."); 
        }

        // Create and return the ConfigurationOption. 
        // 
        // Note that the propertyFile is always null; even if this option's 
        // value comes from a property file, the actual location of that 
        // file is not known yet. 
        return(new ConfigurationOption(optionName, 
                                       optionType, 
                                       validValueSources, 
                                       actualValueSource, 
                                       commandLineName, 
                                       commandLineChar, 
                                       propertyName, 
                                       null, 
                                       defaultValue, 
                                       defaultValue, 
                                       helpText)); 
    }
}
