package org.eso.util.configuration;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;

import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.log4j.Logger;

import org.eso.util.filesystem.FileSystemOperations;

import org.eso.util.misc.WrappedString;

/** This class provides application configuration code relating to 
 * properties and property files. However, applications should not use 
 * it directly - they should subclass {@link Configuration} and call 
 * through its protected methods of the same name. 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
final class PropertyHandler 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(PropertyHandler.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "PropertyHandler"; 

    /** Private constructor - this class should not be instantiated. */ 
    private PropertyHandler() 
    {
        final String methodLogName = classLogName + "::PropertyHandler()"; 

        logger.trace(methodLogName); 

        // This constructor does not need to do anything. 
    }

    /** Update the configuration with any relevant values that might be 
     * available from the specified property file. Any updated values will be 
     * marked with source {@link ConfigurationOption.APPLICATION_PROPERTY}. 
     * 
     * @param propertyFile the property file to load values from. 
     * @param configuration the configuration to update with property values. 
     *
     * @throws NullPointerException if propertyFile or configuration is null. 
     * 
     * @return true if the update was successful, false otherwise. 
     */ 
    static boolean 
    loadApplicationProperties(final File propertyFile, 
                              final Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::loadApplicationProperties()"; 

        logger.trace(methodLogName); 

        if((propertyFile == null) || (configuration == null)) 
        {
            final String message = 
                methodLogName + " - propertyFile [" + propertyFile 
                + "] and configuration [" + configuration 
                + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Get an InputStream for the property file. 
        FileInputStream inputStream; 

        try
        {
            inputStream = new FileInputStream(propertyFile); 
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - unable to open property file [" 
                         + propertyFile + "]."); 
            return(false); 
        }

        // Load the properties from the property file. 
        final Properties properties = new Properties(); 

        try
        {
            properties.load(inputStream); 
        }
        catch(Exception e) 
        {
            logger.error(methodLogName 
                         + " - unable to load properties from file [" 
                         + propertyFile + "]."); 
            return(false); 
        }

        // Delegate to the generic property loading method. 
        return(loadProperties(properties, 
                              ConfigurationOption.APPLICATION_PROPERTY, 
                              propertyFile, 
                              configuration)); 
    }

    /** Update the configuration with any relevant values that might be 
     * available within the system properties. Any updated values will be 
     * marked with source {@link ConfigurationOption.SYSTEM_PROPERTY}. 
     *
     * @param configuration the configuration to update with property values. 
     *
     * @throws NullPointerException if configuration is null. 
     * 
     * @return true if the update was successful, false otherwise. 
     */ 
    static boolean 
    loadSystemProperties(final Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::loadSystemProperties()"; 

        logger.trace(methodLogName); 

        if(configuration == null) 
        {
            final String message = 
                methodLogName + " - configuration must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Get the system properties. 
        final Properties systemProperties = System.getProperties(); 

        // Delegate to the generic property loading method. 
        return(loadProperties(systemProperties, 
                              ConfigurationOption.SYSTEM_PROPERTY, 
                              null, 
                              configuration)); 
    }

    /** Update the configuration with any relevant values that might be
     * available within the specified properties. Any updated values will be 
     * marked with source {@link ConfigurationOption.INTERNAL_CONFIGURATION}. 
     *
     * @param properties the properties to load option values from. 
     * @param configuration the configuration to update with property values. 
     * 
     * @throws NullPointerException if properties or configuration is null. 
     * 
     * @return true if the update was successful, false otherwise. 
     */ 
    static boolean 
    loadProperties(final Properties properties, 
                   final Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::loadProperties(Properties, Configuration)"; 

        logger.trace(methodLogName); 

        if((properties == null) || (configuration == null)) 
        {
            final String message = 
                methodLogName + " - properties [" + properties 
                + "] and configuration [" + configuration 
                + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Delegate to the generic property loading method. 
        return(loadProperties(properties, 
                              ConfigurationOption.INTERNAL_CONFIGURATION, 
                              null, 
                              configuration)); 
    }

    /** Update the configuration with any relevant values that might be 
     * available within the specified properties. 
     * 
     * @param properties the properties to load option values from. 
     * @param valueSource the source of these property values. 
     * @param propertyFile an optional property file to associate these 
     *                     values with. 
     * @param configuration the configuration to update with property values. 
     *
     * @throws IllegalArgumentException if valueSource is not one of: 
     *   <ul>
     *     <li>{@link ConfigurationOption.APPLICATION_PROPERTY}.</li>
     *     <li>{@link ConfigurationOption.SYSTEM_PROPERTY}.</li>
     *     <li>{@link ConfigurationOption.INTERNAL_CONFIGURATION}.</li>
     *   </ul>
     * @throws NullPointerException if properties or configuration is null. 
     * 
     * @return true if the update was successful, false otherwise. 
     */ 
    private static boolean 
    loadProperties(final Properties properties, 
                   final int valueSource, 
                   final File propertyFile, 
                   final Configuration configuration) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName 
            + "::loadProperties(Properties, int, File, Configuration)"; 

        logger.trace(methodLogName); 

        if((properties == null) || (configuration == null)) 
        {
            final String message = 
                methodLogName + " - properties [" + properties 
                + "] and configuration [" + configuration 
                + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        final int acceptableValueSources = 
            ConfigurationOption.APPLICATION_PROPERTY 
            | ConfigurationOption.SYSTEM_PROPERTY 
            | ConfigurationOption.INTERNAL_CONFIGURATION; 

        if(((valueSource & acceptableValueSources) == 0)
           || !ConfigurationOption.isSingleDefinedValueSource(valueSource)) 
        {
            final String message = 
                methodLogName + " - valueSource [" 
                + ConfigurationOption.getValueSourceString(valueSource) 
                + "] must be one of: " 
                + ConfigurationOption.getValueSourceString(
                    acceptableValueSources) + "."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        logger.debug(methodLogName + " - loading properties."); 

        // Iterate though all configuration options held by the configuration 
        // map, attempting to retrieve values for any options which accept 
        // values from the specified value source. 
        final Map<String, ConfigurationOption> 
            optionMap = configuration.getSelectedOptionMap(valueSource); 

        final Iterator<ConfigurationOption> 
            optionIterator = optionMap.values().iterator(); 

        ConfigurationOption currentOption = null; 
        String propertyValue = null; 
        Object optionValue = null; 
            
        while(optionIterator.hasNext())
        {
            currentOption = optionIterator.next(); 

            // Skip any option whose value has already been set from a higher 
            // priority source. 
            if(!ConfigurationOption
               .canOverrideValueSource(valueSource, 
                                       currentOption.getActualValueSource())) 
            {
                logger.debug(
                    methodLogName + " - ignoring option [" 
                    + currentOption.optionName 
                    + "] which already has a higher priority [" 
                    + ConfigurationOption.getValueSourceString(
                        currentOption.getActualValueSource()) 
                    + "] value than [" 
                    + ConfigurationOption.getValueSourceString(valueSource) 
                    + "]."); 
                continue; 
            }

            // Attempt to retrieve a property for this option. 
            propertyValue = 
                properties.getProperty(currentOption.propertyName); 

            if(propertyValue == null)
            {
                logger.debug(methodLogName 
                             + " - these properties do not provide a value for"
                             + " option [" + currentOption.optionName + "]."); 
                continue; 
            }
            else
            {
                // Remove leading and trailing whitespace. 
                propertyValue = propertyValue.trim(); 
            }

            // Attempt to parse the property value string according to the 
            // desired option type. 
            switch(currentOption.optionType)
            {
              case ConfigurationOption.STRING:
                optionValue = propertyValue; 
                break;

              case ConfigurationOption.BOOLEAN:
                optionValue = Boolean.valueOf(propertyValue); 
                break;

              case ConfigurationOption.INTEGER:
                try
                {
                    optionValue = Integer.valueOf(propertyValue); 
                }
                catch(NumberFormatException nfe)
                {
                    logger.error(
                        methodLogName + " - the value [" + propertyValue 
                        + "] specified by property [" 
                        + currentOption.propertyName + "] from source [" 
                        + ConfigurationOption.getValueSourceString(valueSource)
                        + "] could not be parsed [" 
                        + nfe.getMessage() + "] as a value of type [" 
                        + ConfigurationOption.getTypeString(
                            currentOption.optionType) 
                        + "] for option [" + currentOption.optionName + "]."); 

                    return(false); 
                }
                break;

              default:
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName + "] type [" 
                    + ConfigurationOption.getTypeString(
                        currentOption.optionType) 
                    + "] not recognised."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // The property value is now fully parsed - set the option value. 
            logger.debug(methodLogName + " - setting option [" 
                         + currentOption.optionName 
                         + "] from property [" 
                         + currentOption.propertyName + "]."); 

            currentOption.setActualValue(optionValue, 
                                         valueSource, 
                                         propertyFile); 
        }

        logger.debug(methodLogName 
                     + " - all relevant properties successfully loaded."); 

        return(true); 
    }

    /** Prints a help message describing the specified configuration's system 
     * properties to standard output. 
     * 
     * @param configuration the configuration to print system property help 
     *                      for. 
     * 
     * @throws NullPointerException if configuration is null. 
     */ 
    static void 
    printSystemPropertyHelpMessage(final Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::printSystemPropertyHelpMessage()"; 

        logger.trace(methodLogName); 

        if(configuration == null) 
        {
            final String message = 
                methodLogName + " - configuration must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        logger.debug(methodLogName 
                     + " - printing system property help message" 
                     + " to standard output."); 

        // Iterate though all configuration options held by the configuration 
        // map, pick out the ones which accept values from the command line, 
        // and generate a help text for each one. 
        final Map<String, ConfigurationOption> optionMap = 
            configuration.getSelectedOptionMap(
                ConfigurationOption.SYSTEM_PROPERTY); 

        final Iterator<ConfigurationOption> 
            optionMapIterator = optionMap.values().iterator(); 

        // Some options are mandatory, some not. 
        final TreeSet<String> mandatorySet = new TreeSet<String>(); 
        final TreeSet<String> additionalSet = new TreeSet<String>(); 

        // The current option, and components of help text for it. 
        ConfigurationOption currentOption = null; 
        String currentText = null; 
        String parameterTypeString = null; 
            
        while(optionMapIterator.hasNext())
        {
            currentOption = optionMapIterator.next(); 

            // Reset the current option help text, ready for appending. 
            currentText = ""; 

            // Add the common part of the property line... 
            currentText += currentOption.propertyName + "="; 

            // ...and the type-specific part. 
            switch(currentOption.optionType)
            {
              case ConfigurationOption.STRING:
              case ConfigurationOption.INTEGER:
                currentText += 
                    ConfigurationOption.getTypeString(
                        currentOption.optionType) + "\n"; 
                break; 

              case ConfigurationOption.BOOLEAN:
                currentText += "true|false\n"; 
                break; 

              default: 
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName + "] type [" 
                    + ConfigurationOption.getTypeString(
                        currentOption.optionType)
                    + "] not recognised."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // If the property has a default value, show it. 
            if(ConfigurationOption.isValidValueSource(
                   ConfigurationOption.DEFAULT_VALUE, 
                   currentOption.validValueSources)) 
            {
                currentText += 
                    new WrappedString(
                        "  ", 
                        "Default value [" + currentOption.defaultValue + "].")
                    + "\n"; 
            }

            // Add the property's help text, if present. 
            if(!currentOption.helpText.equals(""))
            {
                currentText += 
                    new WrappedString("    ", currentOption.helpText) + "\n"; 
            }

            // The text is complete - store it in the appropriate list. 
            if(currentOption.validValueSources 
               == ConfigurationOption.SYSTEM_PROPERTY) 
            {
                // This option can only be set from a system property, 
                // i.e. it is mandatory. 
                mandatorySet.add(currentText); 
            }
            else
            {
                // This option can also be set from sources other than 
                // a system property. 
                additionalSet.add(currentText); 
            }
        }

        // True if the appropriate set contains property help texts. 
        final boolean mandatoryProperties = !mandatorySet.isEmpty(); 
        final boolean additionalProperties = !additionalSet.isEmpty(); 

        // Print the help text. 
        System.out.print("SYSTEM PROPERTIES"); 

        if(!(mandatoryProperties || additionalProperties)) 
        {
            // This application has no system properties. 
            System.out.print(": this application has no system properties."); 
        }

        System.out.print("\n\n"); 

        if(mandatoryProperties) 
        {
            // Inform the user that the following properties are mandatory. 
            System.out.print("Mandatory properties:\n\n"); 

            // Print the mandatory property texts. 
            final Iterator<String> 
                mandatoryIterator = mandatorySet.iterator(); 

            while(mandatoryIterator.hasNext()) 
            {
                System.out.print(mandatoryIterator.next() + "\n"); 
            }
        }

        if(additionalProperties) 
        {
            if(mandatoryProperties) 
            {
                // Both mandatory and additional properties exist - inform 
                // the user that the following properties are additional. 
                System.out.print("Additional properties:\n\n"); 
            }

            // Print the additional property texts. 
            final Iterator<String> 
                additionalIterator = additionalSet.iterator(); 

            while(additionalIterator.hasNext()) 
            {
                System.out.print(additionalIterator.next() + "\n"); 
            }
        }

        // The full help text has now been printed. 
        logger.debug(methodLogName 
                     + " - successfully printed system property help message" 
                     + " to standard output."); 
    }
    
    /** Generates a template property file for setting the specified 
     * configuration at a given location. The file must not already 
     * exist - if it does, this method will refuse to write to it. 
     * 
     * @param outputFile the location where the template property file 
     *                   should be created. 
     * @param configuration the configuration to generate a template 
     *                      property file for. 
     * 
     * @throws NullPointerException if outputFile or configuration is null. 
     * 
     * @return true if the template property file was successfully generated, 
     *         false otherwise. 
     */ 
    static boolean 
    createTemplatePropertyFile(final File outputFile, 
                               final Configuration configuration) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::createTemplatePropertyFile()"; 

        logger.trace(methodLogName); 

        if((outputFile == null) || (configuration == null)) 
        {
            final String message = 
                methodLogName + " - outputFile [" + outputFile 
                + "] and configuration [" + configuration 
                + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(FileSystemOperations.exists(outputFile)) 
        {
            // The file must not already exist. 
            logger.error(methodLogName + " - file [" 
                         + outputFile + "] already exists."); 
            return(false); 
        }

        logger.debug(methodLogName + " - generating template property file [" 
                     + outputFile + "]."); 

        // Iterate though all configuration options held by the configuration 
        // map, pick out the ones which accept values from an application 
        // property file, and generate a template entry for each one. 
        final Map<String, ConfigurationOption> optionMap = 
            configuration.getSelectedOptionMap(
                ConfigurationOption.APPLICATION_PROPERTY); 

        final Iterator<ConfigurationOption> 
            optionIterator = optionMap.values().iterator(); 

        // To allow sorting by property name, the property name is used as 
        // a key to the full template entry. 
        // 
        // Some options are mandatory, some not. 
        final TreeMap<String, String> 
            mandatoryMap = new TreeMap<String, String>(); 
        final TreeMap<String, String> 
            additionalMap = new TreeMap<String, String>(); 

        // The current option and entry text. 
        ConfigurationOption currentOption = null; 
        String currentText = null; 
        
        while(optionIterator.hasNext()) 
        {
            currentOption = optionIterator.next(); 
            currentText = ""; 

            if(!ConfigurationOption.isValidValueSource(
                   ConfigurationOption.APPLICATION_PROPERTY, 
                   currentOption.validValueSources)) 
            {
                logger.debug(methodLogName 
                             + " - ignoring non-property file option [" 
                             + currentOption.optionName + "]."); 
                continue; 
            }

            // If the property has a help text, it goes at the start of 
            // the template entry. 
            if(!currentOption.helpText.equals(""))
            {
                currentText += currentOption.helpText + "\n\n"; 
            }

            // If the property has a default value, show it. 
            if(ConfigurationOption.isValidValueSource(
                   ConfigurationOption.DEFAULT_VALUE, 
                   currentOption.validValueSources)) 
            {
                currentText += 
                    "Default value [" + currentOption.defaultValue + "].\n\n";
            }

            // Add the common part of the property line... 
            currentText += currentOption.propertyName + "="; 

            // ...and the type-specific part. 
            switch(currentOption.optionType)
            {
              case ConfigurationOption.STRING:
              case ConfigurationOption.INTEGER:
                currentText += 
                    ConfigurationOption.getTypeString(
                        currentOption.optionType); 
                break; 

              case ConfigurationOption.BOOLEAN:
                currentText += "true|false"; 
                break; 

              default: 
                // This indicates a bug. 
                final String message = 
                    methodLogName + " - configuration option [" 
                    + currentOption.optionName + "] type [" 
                    + ConfigurationOption.getTypeString(
                        currentOption.optionType)
                    + "] not recognised."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // Wrap the template entry as a comment and store it in the 
            // appropriate map. 
            if(currentOption.validValueSources == 
               ConfigurationOption.APPLICATION_PROPERTY) 
            {
                // This option can only be set as an application property, 
                // i.e. it is mandatory. 
                mandatoryMap.put(currentOption.propertyName, 
                                 new WrappedString("# ", 
                                                   currentText) + "\n\n"); 
            }
            else
            {
                // This option can also be set from sources other than 
                // an application property. 
                additionalMap.put(currentOption.propertyName, 
                                  new WrappedString("# ", 
                                                    currentText) + "\n\n"); 
            }
        }

        // True if the appropriate map contains property help texts. 
        final boolean mandatoryProperties = !mandatoryMap.isEmpty(); 
        final boolean additionalProperties = !additionalMap.isEmpty(); 

        // Generate an explanatory header text. 
        // 
        // First, the application name and version. 
        currentText = 
            "TEMPLATE CONFIGURATION FILE\n\n" 
            + configuration.getString("application.name") + " version: " 
            + configuration.getString("application.version") + "\n\n"; 

        // Warn the user that the properties are commented out. 
        currentText += 
            "This file was generated with all properties commented out."; 

        // If the help header and footer are available, add them. 
        if(!configuration
           .getString("application.help.header.text").equals(""))
        {
            // A help header text has been provided. 
            currentText += 
                "\n\n" + 
                configuration.getString("application.help.header.text"); 
        }

        if(!configuration
           .getString("application.help.footer.text").equals("")) 
        {
            // A help footer text has been provided. 
            currentText += 
                "\n\n" + 
                configuration.getString("application.help.footer.text"); 
        }

        // Generate the text of the template property file. 
        StringBuffer templateBuffer = new StringBuffer(); 

        // Start with the explanatory header. 
        templateBuffer.append(new WrappedString("# ", currentText) + "\n\n"); 

        if(mandatoryProperties) 
        {
            // Inform the user that the following properties are 
            // mandatory. 
            templateBuffer.append(
                new WrappedString("# ", "MANDATORY PROPERTIES") + "\n\n"); 

            // Add the mandatory property texts. 
            final Iterator<String> 
                mandatoryIterator = mandatoryMap.values().iterator(); 

            while(mandatoryIterator.hasNext()) 
            {
                templateBuffer.append(mandatoryIterator.next()); 
            }
        }

        if(additionalProperties) 
        {
            if(mandatoryProperties) 
            {
                // Both mandatory and additional properties exist - inform
                // the user that the following properties are additional. 
                templateBuffer.append(
                    new WrappedString("# ", "ADDITIONAL PROPERTIES") 
                    + "\n\n"); 
            }

            // Print the additional property texts. 
            final Iterator<String> 
                additionalIterator = additionalMap.values().iterator(); 

            while(additionalIterator.hasNext()) 
            {
                templateBuffer.append(additionalIterator.next()); 
            }
        }

        if(!FileSystemOperations.writeStringToFile(templateBuffer.toString(), 
                                                   false, 
                                                   outputFile)) 
        {
            logger.error(methodLogName 
                         + " - unable to write template property file [" 
                         + outputFile + "]."); 
            return(false); 
        }
        
        logger.debug(methodLogName 
                     + " - successfully generated template property file [" 
                     + outputFile + "]."); 
        return(true); 
    }
}
