package org.eso.util.dal;

import java.sql.Connection;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.sql.Date;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Iterator;
import java.text.SimpleDateFormat;
import java.lang.Math;
import java.util.TimeZone;
import java.lang.IndexOutOfBoundsException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

public class ArchiveStatisticsDAOSyb implements ArchiveStatisticsDAO {
	
	// initialize logger
    private final static Logger logger = Logger.getLogger(ArchiveStatisticsDAOSyb.class);
    
    // db connection
    private String d_dbUrl;
    private String d_dbUsername;
    private String d_dbPassword;
    
    // db tables
    private String d_dataProducts;
    private String d_dtsFiles;
    private String d_pastageFiles;
    private String d_paFiles;
    private String d_saFiles;
    private String d_transferRequests;
    private String d_usbDiskFiles;
    
    // this is used to abort the record fetching loop in executeQuery which
    // in some occasions may take too long and may therefore need to be stopped;
    // 
    private boolean stopQueries = false;
    
    private int executingQueries = 0;
    private Object executingQueriesLock = new Object();
    
    public static final String FILE_ID = "file_id";
    public static final String FILE_DATE = "file_date";
    public static final String FILE_SIZE = "file_size";
    public static final String FILE_PRIORITY = "file_priority";

	public ArchiveStatisticsDAOSyb(String dbUrl, String dbUsername, String dbPassword, 
			String data_products, String dts_files, String pastage_files, 
			String pa_files, String sa_files, String transfer_requests, String usbDiskFiles) throws DAOException {
		d_dataProducts	= data_products;
		d_dtsFiles		= dts_files;
		d_pastageFiles	= pastage_files;
		d_paFiles		= pa_files;
		d_saFiles		= sa_files;
		d_transferRequests	= transfer_requests;
		d_usbDiskFiles	= usbDiskFiles;
		
		d_dbUrl			= dbUrl;
		d_dbUsername	= dbUsername;
		d_dbPassword	= dbPassword;
	}
	
	public List<Map<String, String>> fromOLASToDT(int days) throws DAOException {
		Double epoch = .001*System.currentTimeMillis();
        Double starting_point = epochToMjd(epoch - days*24*60*60);
        Double ending_point	  = epochToMjd(epoch - 60*60); /* files older than one hour */

		String query ="SELECT dp.dp_id AS " + FILE_ID + ", ";
		query += " dateadd(ss, (dp.mjd_obs- 40587.) * (24*60*60), '01 Jan 1970') AS " + FILE_DATE + " FROM " + d_dtsFiles + " dts ";
		query += " RIGHT OUTER JOIN " + d_dataProducts + " dp ON dp.dp_id = dts.file_id ";
		query += " WHERE dts.file_id = NULL AND dp.dp_id NOT LIKE 'APEX%' AND dp.dp_id NOT LIKE 'WFCAM%'";
		query += " AND dp.mjd_obs BETWEEN " + String.valueOf(starting_point) + " AND " + String.valueOf(ending_point);
		query += " ORDER BY dp.mjd_obs";
		
		logger.debug("ArchiveStatistics::fromOLASToDT - " + query);

		return executeQuery(query);
	}
	
	public List<Map<String, String>> fromDTToPAStage(int days) throws DAOException {
		Long now = System.currentTimeMillis();
		/* avoid using dateadd in the query because it slows it prevents using indexes */
		String expDate = new SimpleDateFormat("dd MMM yyyy").format(new Date(now - (long)days*24*3600*1000));

		String query = "SELECT dts.file_id, dts.ingestion_date AS " + FILE_DATE + ", trq.transfer_priority AS file_priority, dts.file_size AS " + FILE_SIZE + " ";
		query += " FROM " + d_dtsFiles + " dts LEFT OUTER JOIN " + d_pastageFiles + " ";
		query += " pa ON dts.file_id = pa.file_id ";
		query += " LEFT OUTER JOIN " + d_transferRequests + " trq ON dts.file_id = trq.file_id ";
		query += " WHERE pa.file_id = NULL and dts.ingestion_date > '" + expDate + "'";
		query += " ORDER BY dts.ingestion_date";

		logger.debug("ArchiveStatistics::fromDTToPAStage - " + query);
		
		return executeQuery(query);
	}
	
	public List<Map<String, String>> fromPAStageToPA(int days) throws DAOException {
		Long now = System.currentTimeMillis();
		/* avoid using dateadd in the query because it slows it prevents using indexes */
		String expDate = new SimpleDateFormat("dd MMM yyyy").format(new Date(now - (long)days*24*3600*1000));

		String query = "SELECT dts.file_id, dts.ingestion_date AS " + FILE_DATE + ", trq.transfer_priority AS file_priority, dts.file_size AS " + FILE_SIZE + " ";
		query += " FROM " + d_pastageFiles + " dts LEFT OUTER JOIN " + d_paFiles + " ";
		query += " pa ON dts.file_id = pa.file_id ";
		query += " LEFT OUTER JOIN " + d_transferRequests + " trq ON dts.file_id = trq.file_id ";
		query += " WHERE pa.file_id = NULL and dts.ingestion_date > '" + expDate + "'";
		query += " ORDER BY dts.ingestion_date";

		logger.debug("ArchiveStatistics::fromDTToPAStage - " + query);
		
		return executeQuery(query);
	}
	
	public List<Map<String, String>> fromPAToSA(int days) throws DAOException {
		Long now = System.currentTimeMillis();
		/* avoid using dateadd in the query because it slows it prevents using indexes */
		String expDate = new SimpleDateFormat("dd MMM yyyy").format(new Date(now - (long)days*24*3600*1000));

		String query = "SELECT pa.file_id, pa.ingestion_date AS " + FILE_DATE + ", pa.file_size AS " + FILE_SIZE + " ";
		query += " FROM " + d_paFiles + " pa LEFT OUTER JOIN " + d_saFiles + " nfs ON pa.file_id = nfs.file_id ";
		query += " WHERE nfs.file_id = NULL and pa.ingestion_date > '" + expDate + "'";
		query += " ORDER BY pa.ingestion_date";
		
		logger.debug("ArchiveStatistics::fromPAToSA - " + query);

		List<Map<String, String>> res = executeQuery(query);
		
		/* TODO */
		/* workaround to remove files containing underscores (_) because they are not stored in SA with the original name */
		for (Iterator<Map<String, String>> iter = res.iterator(); iter.hasNext(); )
			if (iter.next().get("file_id").contains("_"))
				iter.remove();
		
		return res;
	}

	public SortedMap<Timestamp, Double> bandwidthUsage(Timestamp start, Timestamp stop, int step) throws DAOException {
		SortedMap<Timestamp, Double> ret_res = new TreeMap<Timestamp, Double>();
		String query = "SELECT submission_time, last_update_time, size, uncompressed_size ";
		query += "FROM " + d_transferRequests + " WHERE transfer_status = 6 ";
		
		query += "AND submission_time BETWEEN dateadd(ss, " + (long)(start.getTime()*.001) + ", '01 Jan 1970') ";
		query += "AND dateadd(ss, " + (long)(stop.getTime()*.001) + ", '01 Jan 1970') ";
		query += "AND transfer_catg in ('too', 'rawfile') ORDER BY submission_time";
		
		logger.debug("ArchiveStatistics::bandwidthUsage - " + query);
		
		logger.debug("start time: " + start + ", end time: " + stop);
		
		try {
			/* get connection */
			Connection conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
			/* execute the query */
			ResultSet res	= SybaseConnectionService.executeStatement(conn, query);
			
			/* lower limit of the considered interval */
			//Long startTime	= System.currentTimeMillis() - 3600L * 1000 * interval;
			Long startTime = start.getTime();
			int interval = (int) ((stop.getTime() - start.getTime())/(1000*60));
			
			for (int index = 0; index < interval/step; index++) {
				Timestamp   beginDate = new Timestamp(startTime + index*step*60*1000);
				ret_res.put(beginDate, 0.0);
			}
			
			try {
				while (res.next()) {
					Long update = res.getTimestamp("last_update_time").getTime();
					Timestamp endTransmission = new Timestamp(update + TimeZone.getDefault().getOffset(update));
					Long submission = res.getTimestamp("submission_time").getTime();
					Timestamp beginTransmission = new Timestamp(submission + TimeZone.getDefault().getOffset(submission));
					
					logger.debug("begin transmission: " + beginTransmission + ", end transmission: " + endTransmission);
					
					if (beginTransmission.equals(endTransmission))
						continue;
					
					/* let's consider the interval (interval hours long ) as splitted in a number of segments each 
					 * of duration step minutes: beginIndex is the index of the segment where the transmission started
					 */
					Double beginIndex = 1. * (beginTransmission.getTime() - startTime) / (step * 60 * 1000);
					Timestamp  beginDate = new Timestamp(startTime + beginIndex.longValue()*step*60*1000);
					
					Double endIndex = 1. * (endTransmission.getTime() - startTime) / (step * 60 * 1000);

					/* iterate over the intervals on which the file transmission occurred */
					for (long i = beginIndex.longValue(); i <= endIndex.longValue(); i++)
					{
						Double bytes = 1. * res.getLong("size");
						/* if the transmission spanned over more than one interval the size must be splitted over the considered intervals */
						if (beginIndex.longValue() != endIndex.longValue())
							bytes = 1. * res.getLong("size") * (Math.min(endIndex, i+1) - Math.max(beginIndex, i)) / (endIndex - beginIndex);
						/* initialize variable */
						if (ret_res.get(beginDate) == null)
							ret_res.put(beginDate, 0.0);
						/* increment size of the given interval */
						ret_res.put(beginDate, ret_res.get(beginDate) + bytes);
					}
				}			
			} catch (SQLException e) {
				throw new DAOException("Error while parsing result set. Reason: " + e.getMessage());
			}
			conn.close();
		} catch (SQLException e) {
			throw new DAOException(e.getMessage());
		} catch (ConnectionServiceException e) {
			throw new DAOException(e.getMessage());
		} catch (IllegalArgumentException e) {
			throw new DAOException(e.getMessage());
		}
		/* calculate bandwidth in kB/s */
		for (Timestamp d: ret_res.keySet())
			ret_res.put(d, ret_res.get(d) / (1024 * step * 60));
		return ret_res;
	}
	
	public List<Map<String, String>> transmissionTime(Timestamp start, Timestamp stop) throws DAOException {
		
		String query = "SELECT submission_time, last_update_time, size, uncompressed_size, transfer_priority ";
		query += "FROM " + d_transferRequests + " WHERE transfer_status = 6 AND ";
		query += "submission_time BETWEEN dateadd(ss, " + (long)(start.getTime()*.001) + ", '01 Jan 1970') ";
		query += "AND dateadd(ss, " + (long)(stop.getTime()*.001) + ", '01 Jan 1970') ";
		query += "AND transfer_catg in ('too', 'rawfile') ORDER BY submission_time";
		
		logger.debug("ArchiveStatistics::transmissionTime - " + query);
		
		return executeQuery(query);
	}

	public String orignameToFileId(String origname) throws DAOException {
		String query = "SELECT dp_id from " + d_dataProducts + " WHERE origfile='" + origname + "'";
		String res;
		try {
			logger.debug("ArchiveStatistics::orignameToFileId - " + query);
			res = executeQuery(query).get(0).get("dp_id");
			logger.debug("file_id of " + origname + " is " + res);
		} catch (IndexOutOfBoundsException e) {
			throw new DAOException(e.getMessage());
		}
		return res;
	}

	public boolean searchInOLAS(String fileId) throws DAOException {
		return searchInTable(fileId, d_dataProducts);
	}

	public boolean searchInPA(String fileId) throws DAOException {
		return searchInTable(fileId, d_paFiles);
	}

	public boolean searchInSA(String fileId) throws DAOException {
		return searchInTable(fileId, d_saFiles);
	}

	public boolean searchInStagingGarching(String fileId) throws DAOException {
		String query = "SELECT file_id FROM " + d_transferRequests;
		query += " WHERE (transfer_status = 5 OR transfer_status = 6) ";
		query += " AND file_id = '" + fileId + "'";
		
		logger.debug("ArchiveStatistics::searchInTable - " + query);
		List<Map<String, String>> res = executeQuery(query);
		logger.debug("Found " + res.size() + " records of " + fileId + " in " + d_transferRequests);
		return (res.size() > 0);
	}

	public boolean searchInStagingChile(String fileId) throws DAOException {
		return searchInTable(fileId, d_dtsFiles);
	}
	
	public boolean searchInTable(String fileId, String table) throws DAOException {
		String field = "file_id";
		if (table == d_dataProducts)
			field = "dp_id";
		String query = "SELECT * FROM " + table + " WHERE " + field + "='" + fileId + "'";
		logger.debug("ArchiveStatistics::searchInTable - " + query);
		List<Map<String, String>> res = executeQuery(query);
		logger.debug("Found " + res.size() + " records of " + fileId + " in " + table);
		return (res.size() > 0);
	}
	
	public String searchInUSBDisks(String fileId) throws DAOException {
		String query = "SELECT logical_name FROM " + d_usbDiskFiles + " WHERE file_id = '" + fileId + "'";
		logger.debug("ArchiveStatistics::searchInUSBDisks - " + query);
		List<Map<String, String>> res = executeQuery(query);
		logger.debug("Found " + res.size() + " records of " + fileId + " in " + d_usbDiskFiles);
		if (res.size() == 0)
			return "---";
		else {
			String disks = "";
			for (Map<String, String> disk: res)
				disks += ", " + disk.get("logical_name");
			return disks.substring(2);
		}
	}
	
	/**
	 * Utility method to convert seconds since Epoch to MJD
	 * @param epoch seconds since 01/01/1970
	 * @return time converted to MJD
	 */
	public static Double epochToMjd(Double epoch) {
		return 40587. + epoch/(24*60*60);
	}
	
	/**
	 * Utility method to convert an MJD date in seconds since Epoch
	 * @param mjd MJD date
	 * @return date converted to seconds since 01/01/1970
	 */
	public static Double mjdToEpoch(Double mjd) {
		return (mjd - 40587.) * (24*60*60);
	}

	public void stopAllQueries() {
		
		final String logHead = "stopAllQueries() - ";
		
		stopQueries = true;

		logger.debug(logHead + "Waiting for queries to terminate ...");
		
		// wait until all the queries are terminated
		while( this.executingQueries > 0 ) {
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				logger.error(logHead + "Cannot sleep between pollings: " + e.getMessage());
			}
		}

		logger.debug(logHead + "All queries terminated.");

		stopQueries = false;
	}
	
	private List<Map<String, String>> executeQuery(String query) throws DAOException {

		synchronized(executingQueriesLock) {
			executingQueries++;
		}
		
		final String logHead = "executeQuery() - ";
		
		/* return variable */
		List<Map<String, String>> ret_val = null;
		try {
			/* get connection */
			Connection conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);

			/* execute the query */
			//ResultSet res	= SybaseConnectionService.executeStatement(conn, query);
			
			// this will execute the query asynchronously, so that it is possible to
			// cancel it while iterating over the fetched records
			PreparedStatement sqlStatement = conn.prepareStatement(query);
			sqlStatement.execute();
			ResultSet res = sqlStatement.getResultSet();

			
			/* initialize variable */
			ret_val = new ArrayList<Map<String, String>>();

			try {
				// Get result set meta data
				ResultSetMetaData rsmd = res.getMetaData();
				int numColumns = rsmd.getColumnCount();

				// Get the column names; column indices start from 1
				List<String> columns = new ArrayList<String>();
				for (int i=1; i < numColumns+1; i++) {
					columns.add(rsmd.getColumnName(i));
				}

				/* iterate over fetched records */
				while (res.next()) {
					
					if( stopQueries ) {
						sqlStatement.cancel();
						logger.debug(logHead + "query and records fetching stopped for query [" + query + "].");
						break;
					}

					Map<String, String> temp = new HashMap<String, String>();
					for (String s: columns)
						temp.put(s, res.getString(s));
					ret_val.add(temp);
				}
				
			} catch (SQLException e) {
				throw new DAOException("Error while parsing result set. Reason: " + e.getMessage());
			}
			conn.close();
		} catch (SQLException e) {
			throw new DAOException(e.getMessage());		
		} catch (ConnectionServiceException e) {
			throw new DAOException(e.getMessage());
		} catch (IllegalArgumentException e) {
			throw new DAOException(e.getMessage());
		} finally {
			synchronized(executingQueriesLock) {
				executingQueries--;
			}
		}
		return ret_val;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		BasicConfigurator.configure();
		
		try {
//			ArchiveStatisticsDAO dao = new ArchiveStatisticsDAOSyb("acdbdev.hq.eso.org:6789", "sara", "sara_ps", 
//					"sara..data_products_pl", "sara..dts_files_pl", "sara..dts_files_pastage", 
//					"sara..pa_files", "ngas..ngas_files_safe", "sara..transfer_requests", "sara..usb_disk_files");
//
			ArchiveStatisticsDAO dao = new ArchiveStatisticsDAOSyb("acdb.hq.eso.org:2025", "sara", "sara_ps", 
					"observations..data_products", "sara..dts_files", "sara..dts_files_pastage", 
					"sara..pa_files", "ngas..ngas_files_safe", "sara..transfer_requests", "sara..usb_disk_files");
		
			int days = 40;
			
			List<Map<String, String>> res = null;
			
			res = dao.fromOLASToDT(days);
			if (res != null) {
//				for (Map<String, String> map: res) {
//					for (String k: map.keySet()) {
//						System.out.print(k + ": " + map.get(k) + " ");
//					}
//					System.out.println();
//				}
				System.out.println("Not in DT: " + res.size());
			}
			
			SortedMap<Integer, Integer> files = new TreeMap<Integer, Integer>();
			files.put(10, 0);
			files.put(30, 0);
			files.put(40, 0);
			files.put(50, 0);
			files.put(100, 0);
			
			res = null;
			res = dao.fromDTToPAStage(days);
			if (res != null) {
				for (Map<String, String> map: res) {
					if (map.get("file_priority") == null)
						map.put("file_priority", "60");
					
					Integer tp = Integer.valueOf(map.get("file_priority"));
					for (Integer p: files.keySet())
						if (tp <= p) {
							files.put(p, files.get(p)+1);
							break;
						}
				}
				System.out.println("Not in PA Stage: " + res.size());
				for (Integer p: files.keySet())
					System.out.println("\tpriority higher than " + p + ": " + files.get(p));
			}

			res = dao.fromPAStageToPA(days);
			if (res != null) {
				for (Map<String, String> map: res) {
					for (String k: map.keySet()) {
						System.out.print(k + ": " + map.get(k) + " ");
					}
					System.out.println();
				}
				System.out.println("Not in PA: " + res.size());
			}
			
			res = null;
			res = dao.fromPAToSA(days);
			if (res != null) {
//				for (Map<String, String> map: res) {
//					for (String k: map.keySet()) {
//						System.out.print(k + ": " + map.get(k) + " ");
//					}
//					System.out.println();
//				}
				System.out.println("Not in SA: " + res.size());
			}
			
			/*SortedMap<Timestamp, Double> bw = dao.bandwidthUsage(new Timestamp(System.currentTimeMillis()-1000*3600*1), new Timestamp(System.currentTimeMillis()), 10);
			for (Timestamp d: bw.keySet()) {
				System.out.println(d + " " + bw.get(d));
			}*/
			
			res = null;
			//res = dao.transmissionTime(10);
			if (res != null) {
				/*for (Map<String, String> map: res) {
					for (String k: map.keySet()) {
						System.out.print(k + ": " + map.get(k) + " ");
					}
					System.out.println();
				}*/
				System.out.println(res.size());
			}
//			dao.searchInSA("SINFO.2008-09-22T12:00:02.563");
//			dao.searchInSA("UVES.2008-09-15T12:58:59.432");
//			
//			dao.searchInPA("SINFO.2008-09-22T12:00:02.563");
//			dao.searchInPA("UVES.2008-09-15T12:58:59.432");
//			dao.searchInPA("HAWKI.2008-09-16T05:01:20.412");
//			
//			dao.searchInOLAS("INFO.2008-09-22T12:00:02.563");
//			dao.searchInOLAS("UVES.2008-09-15T12:58:59.432");
//			dao.searchInOLAS("HAWKI.2008-09-16T05:01:20.412");
//			
//			dao.searchInOLAS(dao.orignameToFileId("HAWKI_IMG_OBS_AutoJitter260_0021.fits"));
//			
//			dao.searchInStagingGarching("WFI.2008-09-11T19:36:21.319");
//			
//			System.out.println(dao.searchInUSBDisks("WFI.2008-09-11T19:36:21.319"));
			
		} catch (DAOException e) {
			System.out.println(e.getMessage());
		}
	}
}
