package org.eso.util.dal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class Package {
	
	public enum ViewMode {
		STANDARD,
		QC,
		FINAL,
		FILEIDS
	}
		
	/**
	 * package information (run_id, instrument...)
	 */
	private PackageInfo			d_info = null;
	
	/**
	 * package reports
	 */
	private List<PackageReport> d_reports = null;
	
	/**
	 * package files
	 */
	private List<PackageFile>	d_files = null;
	
	/**
	 * associations between fileIds and aux files contained in them 
	 */
	private Map<String, List<String> > d_auxFiles = null;
	
	public PackageInfo getInfo() {
		return d_info;
	}

	public void setInfo(PackageInfo d_info) {
		this.d_info = d_info;
	}

	public List<PackageFile> getFiles() {
		return d_files;
	}

	public void setFiles(List<PackageFile> d_files) {
		this.d_files = d_files;
	}

	public Map<String, List<String>> getAuxFiles() {
		return d_auxFiles;
	}

	public void setAuxFiles(Map<String, List<String>> files) {
		d_auxFiles = files;
	}

	public List<PackageReport> getReports() {
		return d_reports;
	}

	public void setReports(List<PackageReport> d_reports) {
		this.d_reports = d_reports;
	}
	
	public String toString(ViewMode mode) {
		
		String res;
		
		if (mode == ViewMode.FILEIDS) {
			res = this.toFileIds();
		} else if (mode == ViewMode.FINAL) {
			res = this.toFinalNames();
		} else if (mode == ViewMode.QC) {
			res = this.toQCNames();
		} else if (mode == ViewMode.STANDARD) {
			res = this.toStandardView();
		} else {
			res = "Unrecognized mode " + mode.toString();
		}
		
		return res;
	}	

	@Override 
	public String toString() {
		return toStandardView();
	}
	
	public String toStandardView() {
		String res = header();

		List<String> lines = new ArrayList<String>();
		for (PackageFile file: d_files) {
			String tmp = "";
			if (file.getFileType() == PackageFile.Type.AUX) {
				tmp += file.getFilePath() + System.getProperty("file.separator") + file.getFileName() + " [" + file.getFileId() + "]\n";
				for (String auxFile: d_auxFiles.get(file.getFileId())) {
					tmp += file.getFilePath().replaceAll(".", " ") + "-" + auxFile + "\n";
				}
			} else {
				tmp += file.getFilePath() + System.getProperty("file.separator") + file.getFileName() + " [" + file.getFileId() + "]\n"; 
			}
			lines.add(tmp);
		}
		for (PackageReport rep: getActiveReports(true))
			lines.add(rep.getFileUrl() + "\n");
		Collections.sort(lines);
		for (String line: lines)
			res += line;

		return res;
	}
	
	public String toQCNames() {
		String res = header();
		List<String> lines = new ArrayList<String>();
		for (PackageFile file: d_files) {
			lines.add(file.getFilePath() + System.getProperty("file.separator") + file.getFileName() + "\n"); 
		}
		for (PackageReport rep: getActiveReports(true))
			lines.add(rep.getFileUrl() + "\n");
		Collections.sort(lines);
		for (String line: lines)
			res += line;

		return res;
	}
	
	public String toFinalNames() {
		String res = header();
		List<String> lines = new ArrayList<String>();
		for (PackageFile file: d_files) {
			if (file.getFileType() == PackageFile.Type.AUX) {
				for (String auxFile: d_auxFiles.get(file.getFileId())) {
					lines.add(file.getFilePath() + System.getProperty("file.separator") + auxFile + "\n");
				}
			} else {
				lines.add(file.getFilePath() + System.getProperty("file.separator") + file.getFileName() + "\n"); 
			}
		}
		for (PackageReport rep: getActiveReports(true))
			lines.add(rep.getFileUrl() + "\n");
		Collections.sort(lines);
		for (String line: lines)
			res += line;

		return res;
	}

	public String toFileIds() {
		String res = header();
		List<String> lines = new ArrayList<String>();
		for (PackageFile file: d_files) {
			lines.add(file.getFileId() + "\n");
		}
		Collections.sort(lines);
		for (String line: lines)
			res += line;

		return res;
	}
	
	private String header() {
		String res = "\nPackage ID: " + d_info.getPackageId() + "\n";
		res += "Run ID: " + d_info.getRunId() + "\n";
		res += "Instrument: " + d_info.getInstrument() + "\n";
		res += "----------------------------------------------\n";
		res += "Contents:\n\n";
		return res;
	}
	
	public List<PackageReport> getActiveReports(boolean includePreservedDirectories) {
		
		List<PackageReport> res = new ArrayList<PackageReport>();
		
		/* list of all the directories in the package slice */
		Set<String> directories = new HashSet<String>();
		for (PackageFile file: d_files)
			directories.add(file.getFilePath());
		if (includePreservedDirectories)
			directories.add(".");
		
		/* let's retrieve the reports associated to the given directories */
		for (PackageReport report: d_reports) {
			String url = report.getFileUrl();
			
			/* base dir is by default the root dir */
			String baseDir = ".";
			int idx = url.lastIndexOf("/");
			if (idx != -1)
				baseDir = url.substring(0, url.lastIndexOf("/"));
			
			if (includePreservedDirectories) {
				if (directories.contains(baseDir) || baseDir.contains("GEN_INFO")) {
					res.add(report);
				}
			} else {
				if (directories.contains(baseDir)) {
					res.add(report);
				}
			}
		}
		return res;
	}

	public List<String> getFileIds() {
		List<String> res = new ArrayList<String>();
		for (PackageFile file: d_files) {
			res.add(file.getFileId());
		}
		
		return res;
	}
}
