package org.eso.util.dal;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import org.apache.log4j.Logger;

/* TODO */
/* remove all deprecated stuff ASAP:
 * - release date update
 * - get request id
 * - mail related stuff
 */

/**
 * This class provides services for the DAOs related to the package selector, i.e.
 * PackageInfoDAO, PackageFilesDAO, PackageAuxFilesDAO and DataProductsDAO
 * 
 * @author vforchi
 *
 */
public class PackageService {
	
	/** Apache Log4J logger. */
    static final Logger logger = Logger.getLogger(PackageService.class);
    
    /** The name of this class for logging purposes. */
	private static final String classLogName = "PackageService";
	
	/** maximum text size for queries */
	private Integer MAX_TEXT_SIZE = 10000000;
	
	private boolean d_configured = false;
	
	private String d_dbUrl			= null;
    private String d_dbUsername		= null;
    private String d_dbPassword		= null;
    private String d_dbInfoTable	= null;
    private String d_dbReportsTable	= null;
    private String d_dbFilesTable	= null;
    private String d_dbObsDataProductsTable	= null;
    private String d_dbCalibDataProductsTable	= null;
    private String d_dbAuxFilesTable		= null;
    private String d_dbNgasFilesTable = null;
    
    /* used to update release date of special files */
    private Map<String, Integer> d_specialPeriods = null;

	private String d_dbUniqueIdProcedure;

    /**
     * Constructor
     * 
     * @param dbUrl	url of the database server
     * @param dbUsername database username
     * @param dbPassword database password
     */
	public PackageService(String dbUrl, String dbUsername, String dbPassword) {
		d_dbUrl			= dbUrl;
		d_dbUsername	= dbUsername;
		d_dbPassword	= dbPassword;
	}
	
	/**
	 * Configure database tables
	 * 
	 * @param dbInfoTable
	 * @param dbReportsTable
	 * @param dbFilesTable
	 * @param dbObsDataProductsTable
	 * @param dbCalibDataProductsTable
	 * @param dbAuxFilesTable
	 * @param dbNgasfilesTable
	 * @param dbUniqueIdProcedure
	 */
	public void configureTables(String dbInfoTable, String dbReportsTable, String dbFilesTable, 
			String dbObsDataProductsTable, String dbCalibDataProductsTable, String dbAuxFilesTable,
			String dbNgasFilesTable, String dbUniqueIdProcedure) {
		
		d_dbInfoTable	= dbInfoTable;
		d_dbFilesTable	= dbFilesTable;
		d_dbReportsTable	= dbReportsTable;
		d_dbObsDataProductsTable	= dbObsDataProductsTable;
		d_dbCalibDataProductsTable	= dbCalibDataProductsTable;
		d_dbAuxFilesTable		= dbAuxFilesTable;
		d_dbNgasFilesTable		= dbNgasFilesTable;
		d_dbUniqueIdProcedure	= dbUniqueIdProcedure;

		d_configured = true;
	}
	
	/**
	 * Finds the file id associated to a given filename in QC naming convention: this method searches the observations db first 
	 * for raw files, if the file is not there it seatches the calib db for products. 
	 * 
	 * @param filename 
	 * @return the file id
	 * @throws ServiceException
	 */
	public String fileIdFromFilename(String filename) throws ServiceException {
		
		final String thisMethod = classLogName + "::fileIdFromFilename()";
		logger.trace(thisMethod);
		
		DataProductsDAOSyb obsDao = new DataProductsDAOSyb(d_dbUrl, d_dbUsername, d_dbPassword, d_dbObsDataProductsTable);
		DataProductsDAOSyb calibDao = new DataProductsDAOSyb(d_dbUrl, d_dbUsername, d_dbPassword, d_dbCalibDataProductsTable);
		
		/* TODO: parameters embedded for tests: REMOVE */
		//DataProductsDAO obsDao = new DataProductsDAOSyb("acdb:2025", "starcat", "tacrats", d_dbObsDataProductsTable);
		//DataProductsDAO calibDao = new DataProductsDAOSyb("acdb:2025", "starcat", "tacrats", d_dbCalibDataProductsTable);
		
		try {
			logger.debug(thisMethod + " - looking for file id of file [" + filename + "]");
			String filenameNoExt = filename.replace(".fits", "");
			/* search in observation db */
			if (obsDao.findDpId(filenameNoExt))
				return filenameNoExt;
			/* search in calibrations db */
			else
				return calibDao.getDpId(filename);
		} catch (DAOException e) {
			throw new ServiceException(e.getMessage());
		}
	}
	
	public Integer getPackageId(String runId, String instrument) throws ServiceException {
		try {
			Connection conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
			PackageInfoDAO pkgInfoDAO = new PackageInfoDAOSyb(conn, d_dbInfoTable, d_dbUniqueIdProcedure);
			return pkgInfoDAO.getPackageId(runId, instrument);
		} catch (Exception e) {
			throw new ServiceException(e.getMessage());
		}
	}
	
	/**
	 * This method inserts a Package object in the db in the db
	 * 
	 * @param pkg
	 * @throws ServiceException
	 */
	public void insertPackage(Package pkg) throws ServiceException {
		
		final String thisMethod = classLogName + "::insertPackage()";
		logger.trace(thisMethod);
		
		Connection conn = null;
		String message = null;
		try {
			conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
			SybaseConnectionService.executeStatement(conn, "set textsize " + MAX_TEXT_SIZE);
					
			PackageInfoDAO pkgInfoDAO = new PackageInfoDAOSyb(conn, d_dbInfoTable, "archive..opc_get_key");
			PackageReportsDAO pkgReportsDAO = new PackageReportsDAOSyb(conn, d_dbReportsTable);
			PackageFilesDAO pkgFilesDAO = new PackageFilesDAOSyb(conn, d_dbFilesTable);
			PackageAuxFilesDAO pkgAuxFilesDAO = new PackageAuxFilesDAOSyb(conn, d_dbAuxFilesTable);
			
			conn.setAutoCommit(false);
			
			logger.debug(thisMethod + " - insert package info");
			pkgInfoDAO.insertIfNotFound(pkg.getInfo());

			logger.debug(thisMethod + " - insert package reports");
			/* TODO: change PackageReportsDAO API? */
			for (PackageReport report: pkg.getReports())
				pkgReportsDAO.insertOrUpdate(report);
			
			logger.debug(thisMethod + " - insert package files");
			pkgFilesDAO.insertIfNotFound(pkg.getFiles());
			
			logger.debug(thisMethod + " - insert package aux files");
			pkgAuxFilesDAO.insertPackageAuxFilesIfNotFound(pkg.getInfo().getPackageId(), pkg.getAuxFiles());
		
			logger.debug(thisMethod + "commit transaction");
			conn.commit();
					
		} catch (ConnectionServiceException e) {
			message = e.getMessage();
		} catch (SQLException e) {
			message = e.getMessage();
		} catch (DAOException e) {
			try {
				conn.rollback();
				message = e.getMessage();
				logger.error(thisMethod + " - " + e.getMessage());
				logger.error(thisMethod + " - rollback transaction");
			} catch (SQLException ex) {
				logger.error(thisMethod + " - " + ex.getMessage());
				message = ex.getMessage();
			}
		} finally {
			if (conn != null)
			{
				try {
					conn.setAutoCommit(true);
					SybaseConnectionService.closeConnection(conn);
				} catch (ConnectionServiceException ex) {
					logger.error(thisMethod + " - " + ex.getMessage());
					message = ex.getMessage();
				} catch (SQLException ex) {
					logger.error(thisMethod + " - " + ex.getMessage());
					message = ex.getMessage();
				}
			}
			if (message != null)
				throw new ServiceException(message);
		}
	}
	
	/**
	 * This method removes a Package object from the db
	 * 
	 * Note: not completely implemented yet
	 * 
	 * @param pkg
	 * @throws ServiceException
	 */
	public int deletePackage(String run_id, String instrument, Timestamp start, Timestamp stop) throws ServiceException {
		
		final String thisMethod = classLogName + "::deletePackage()";
		logger.trace(thisMethod);
		
		int res = -1;
				
		Package pkg = findPackage(run_id, instrument, start, stop);
		int pkgId = pkg.getInfo().getPackageId();
		
		Connection conn = null;
		String message = null;
		try {
			conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
		
			PackageInfoDAO pkgInfoDAO = new PackageInfoDAOSyb(conn, d_dbInfoTable, "archive..opc_get_key");
			PackageReportsDAO pkgReportsDAO = new PackageReportsDAOSyb(conn, d_dbReportsTable);
			PackageFilesDAO pkgFilesDAO = new PackageFilesDAOSyb(conn, d_dbFilesTable);
			PackageAuxFilesDAO pkgAuxFilesDAO = new PackageAuxFilesDAOSyb(conn, d_dbAuxFilesTable);
			
			conn.setAutoCommit(false);
			
			logger.debug(thisMethod + " - delete aux files");
			pkgAuxFilesDAO.deletePackageAuxFiles(pkgId, pkg.getAuxFiles());
			logger.debug(thisMethod + " - delete package files");
			pkgFilesDAO.delete(pkg.getFiles());
			logger.debug(thisMethod + " - delete active package reports");
			pkgReportsDAO.delete(pkg.getActiveReports(false));
			
			/* check whether we have to delete package info too */
			/* looking for other files related to this package */			
			List<PackageFile> remainingFiles = pkgFilesDAO.find(pkgId, new Timestamp(0), new Timestamp(System.currentTimeMillis()));
			if (remainingFiles.size() == 0) {
				logger.info(thisMethod + " - No more file in the package: removing all reports and package information");
				
				logger.debug(thisMethod + " - delete all package reports");
				pkgReportsDAO.delete(pkg.getReports());
								
				logger.debug(thisMethod + " - delete package info");
				pkgInfoDAO.delete(pkg.getInfo());
			}
					
			conn.commit();		
			res = 0;
			
		} catch (ConnectionServiceException e) {
			message = e.getMessage();
		} catch (SQLException e) {
			message = e.getMessage();
		} catch (DAOException e) {
			try {
				conn.rollback();
				message = "Rollback";
			} catch (SQLException ex) {
				message = ex.getMessage();
			}
		} finally {
			if (conn != null)
			{
				try {
					conn.setAutoCommit(true);
					SybaseConnectionService.closeConnection(conn);
				} catch (ConnectionServiceException ex) {
					message = ex.getMessage();
				} catch (SQLException ex) {
					message = ex.getMessage();
				}
			}
			if (message != null)
				throw new ServiceException(message);
		}
		
		return res;
	}
		
	/**
	 * This methods look for a Package or for a Package slice in the db
	 * 
	 * @param run_id The run_id of the package
	 * @param instrument The instrument of the package
	 * @param start Begin date
	 * @param stop End date
	 * @return 
	 * @throws ServiceException
	 */
	public Package findPackage(String run_id, String instrument, Timestamp start, Timestamp stop) throws ServiceException {
		
		final String thisMethod = classLogName + "::findPackage()";
		logger.trace(thisMethod);
		
		Package result = new Package();
		Connection conn = null;
		String message = null;
		try {
			conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
			SybaseConnectionService.executeStatement(conn, "set textsize " + MAX_TEXT_SIZE);
		
			PackageInfoDAO pkgInfoDAO = new PackageInfoDAOSyb(conn, d_dbInfoTable, "archive..opc_get_key");
			PackageReportsDAO pkgReportsDAO = new PackageReportsDAOSyb(conn, d_dbReportsTable);
			PackageFilesDAO pkgFilesDAO = new PackageFilesDAOSyb(conn, d_dbFilesTable);
			PackageAuxFilesDAO pkgAuxFilesDAO = new PackageAuxFilesDAOSyb(conn, d_dbAuxFilesTable);
		
			logger.debug(thisMethod + " - find package info");
			PackageInfo info = pkgInfoDAO.find(run_id, instrument);
			if (info == null) {
				throw new ServiceException("No information found about run_id " + run_id + " and instrument " + instrument);
			}
			
			logger.debug(thisMethod + " - find package reports");
			List<PackageReport> reports = pkgReportsDAO.find(info.getPackageId());
			
			logger.debug(thisMethod + " - find package files");
			List<PackageFile> files = pkgFilesDAO.find(info.getPackageId(), start, stop);
			
			List<String> fileIds = new ArrayList<String>();
			for (PackageFile file: files) {
				fileIds.add(file.getFileId());
			}
			
			logger.debug(thisMethod + " - find package aux files");
			Map<String, List<String>> auxFiles = pkgAuxFilesDAO.getAuxFiles(info.getPackageId(), fileIds);
			
			result.setInfo(info);
			result.setReports(reports);
			result.setFiles(files);
			result.setAuxFiles(auxFiles);

		} catch (ConnectionServiceException e) {
			message = e.getMessage();
		} catch (DAOException e) {
			message = e.getMessage();
		} finally {
			if (conn != null)
			{
				try {
					SybaseConnectionService.closeConnection(conn);
				} catch (ConnectionServiceException ex) {
					message = ex.getMessage();
				}
			}
			if (message != null)
				throw new ServiceException(message);
		}		
		return result;
	}

	@Deprecated
	public void updateReleaseDates(String run_id, List<PackageFile> files, String reqId, String dbcmMail, String mailSubject, String propPeriodFilename) throws ServiceException {
		
		if (d_specialPeriods == null)
			readSpecialPeriods(propPeriodFilename);

		Connection conn = null;
		try {
			conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
		} catch (ConnectionServiceException e) {
			throw new ServiceException(e.getMessage());
		}
		
		try {
			conn.setAutoCommit(false);
		
			for (PackageFile file: files) {
				updateReleaseDate(conn, run_id, file.getFileId(), dbcmMail, mailSubject);
				logDownload(conn, run_id, file.getFileId());
			}
			
			String query = "INSERT INTO archive..rel_progs SELECT '" + run_id + "', 'DBPI', ";
			query += " datediff(ss, '1-1-1980', getdate()),'" + reqId + "', 'FTP', 1";
			logger.debug(query);
			SybaseConnectionService.executeStatement(conn, query);

			conn.commit();
			
		} catch (Exception e) {
			try {
				logger.error(e.getMessage());
				conn.rollback();
				throw new ServiceException(e.getMessage());
			} catch (SQLException ex) {
				throw new ServiceException(ex.getMessage());
			}
		} finally {
			try {
				SybaseConnectionService.closeConnection(conn);
			} catch (ConnectionServiceException e) {
				throw new ServiceException(e.getMessage());
			}	
		}
	}
	
	@Deprecated
	private void updateReleaseDate(Connection conn, String run_id, String dp_id, String dbcmMail, String mailSubject) throws ServiceException, ConnectionServiceException {

		String query;
		
		/* check if we are in service mode */
		query = "SELECT obs_mode FROM opc70..obs_runs WHERE run_code='" + run_id + "'";
		logger.debug(query);
		try {
			ResultSet res = SybaseConnectionService.executeStatement(conn, query);
			if (res.next()) {
				if (res.getString("obs_mode").toLowerCase().equals("v")) {
					logger.debug("Visitor mode: not updating release date");
					return;
				}
			}
		} catch (SQLException e) {
			throw new ServiceException(e.getMessage());
		}
		
		
		/* check if there is already an entry in rel_files */
		query = "SELECT COUNT(*) FROM observations..rel_files WHERE dp_id='" + dp_id + "'";
		logger.debug(query);
		try {
			ResultSet res = SybaseConnectionService.executeStatement(conn, query);
			if (res.next()) {
				if (res.getInt(1) != 0) {
					logger.debug("Entry already present in rel_files: not updating");
					return;
				}
			}
		} catch (SQLException e) {
			throw new ServiceException(e.getMessage());
		}

		/* read proprietary period */
		Integer propPeriod = d_specialPeriods.get(run_id);
//		Integer propPeriod = null;
//		if (d_specialPeriods.containsKey(run_id)) {
//			if (d_specialPeriods.get(run_id) == 0) {
//				sendMail("vforchi@eso.org", "DBPI", "cannot determine proprietary period of file " + dp_id);
//				update = false;
//			} else {
//				propPeriod = d_specialPeriods.get(run_id);
//			}
//		}
		
		/* check if proprietary period has been modified by hand */
		if (propPeriod == null) {
			query = "SELECT dateadd(yy, 1, exp_start), ";
		} else {
			query = "SELECT dateadd(dd, " + propPeriod + ", exp_start), ";
		}
		query += " rel_date from observations..data_products dp LEFT OUTER JOIN observations..rel_files rf on dp.dp_id = rf.dp_id WHERE dp.dp_id='" + dp_id + "' AND rf.dp_id=NULL AND ( dp.dp_cat like 'SCIENCE%' or dp.dp_cat like 'ACQUISITION%' )";
		
		logger.debug(query);
		try {
			ResultSet res = SybaseConnectionService.executeStatement(conn, query);
			if (res.next()) {
				Date evalRelDate = res.getDate(1);
				Date relDate = res.getDate(2);
				if (Math.abs(evalRelDate.getTime() - relDate.getTime()) > 1000) {
					sendMail(dbcmMail, mailSubject, "Release date of file " + dp_id + " has been modified by hand, not updating it");
					return;
				}
			}
		} catch (SQLException e) {
			throw new ServiceException(e.getMessage());
		}

		/* update release date */
		query =  "UPDATE observations..data_products SET rel_date = dateadd(";
		if (propPeriod == null) {
			query += "yy, 1, getdate()) ";
		} else {
			query += "dd, " + d_specialPeriods.get(run_id) + ", getdate()) ";
		}

		query += " WHERE dp_id = '" + dp_id + "'";
		//query += " AND rel_date > getdate() AND dp_cat IN ('SCIENCE','ACQUISITION' )";
		query += " AND rel_date > getdate() AND ( dp_cat like 'SCIENCE%' or dp_cat like 'ACQUISITION%' )";
		
		query += " AND dp_id NOT IN (select dp_id FROM observations..rel_files";
		query += " where prog_id = '" + run_id + "' AND dp_id = '" + dp_id + "')";

		logger.debug(query);
		SybaseConnectionService.executeStatement(conn, query);
		
	}

	@Deprecated
	private void logDownload(Connection conn, String run_id, String dp_id) throws ServiceException, ConnectionServiceException {

		String query;
		
		/* log the request in rel_files */
		query = "INSERT INTO observations..rel_files SELECT '" + dp_id + "',";
		query += " datediff(ss, '1-1-1980', dateadd(yy, 1, getdate())), '" + run_id + "'";

		logger.debug(query);
		SybaseConnectionService.executeStatement(conn, query);

	}
	
	@Deprecated
	private void readSpecialPeriods(String propPeriodFilename) {
		d_specialPeriods = new HashMap<String, Integer>();
		
		//String propPeriodFilename = "/home/vforchi/Desktop/update_prog.conf";
		//String propPeriodFilename = "/opsw/dbcm/cron/Prog_id/update_prog.conf";
		
		try {
			FileReader specialPeriodsFile = new FileReader(propPeriodFilename);
			BufferedReader buf = new BufferedReader(specialPeriodsFile);
			String line;
			int counter = 0;
			while ((line = buf.readLine()) != null) {
				String[] tokens = line.split("[ \\t]+");
				counter++;
				if (tokens.length == 4) {
					try {
						d_specialPeriods.put(tokens[0], Integer.valueOf(tokens[2]));
					} catch (NumberFormatException e) {
						logger.warn(classLogName + "::readSpecialPeriods: line " + counter + " badly formatted");
					}
				} else {
					logger.warn(classLogName + "::readSpecialPeriods: line " + counter + " badly formatted");
				}
			}
		} catch (FileNotFoundException e) {
			logger.fatal(classLogName + "::readSpecialPeriods: cannot read file " + propPeriodFilename + ", " + e.getMessage());
			System.exit(-1);
		} catch (IOException e) {
			logger.fatal(classLogName + "::readSpecialPeriods: cannot read file " + propPeriodFilename + ", " + e.getMessage());
			System.exit(-1);
		}
	}
	
	private void sendMail(String receiver, String subject, String body) throws ServiceException {
		try {
			
			logger.info("Sending e-mail to: " + receiver);
			logger.info("Subject: " + subject);
			logger.info("Body: " + body);
						
			/* send e-mail */
			Properties props = System.getProperties();
			Session session = Session.getInstance(props);
			session.setDebug(true);			

			MimeMessage message = new MimeMessage(session);

			message.setFrom(new InternetAddress("pipack@eso.org"));
			message.setRecipients(Message.RecipientType.TO,	InternetAddress.parse(receiver, false));
			message.setHeader("X-Mailer", "msgsend");
			message.setSentDate(new Date());

			Transport transport = session.getTransport("smtp");

			// set subject and text to send
			message.setSubject(subject);
			message.setText(body);

			// send the message
			transport.connect("localhost", null, null);
			message.saveChanges();

			transport.sendMessage(message, message.getAllRecipients());
			transport.close();
			
			logger.info("E-Mail successfully sent");

		} catch (AddressException e) {
			throw new ServiceException(e.getMessage());
		} catch (MessagingException e) {
			throw new ServiceException(e.getMessage());
		}
	}
	
	public Long getPackageSize(Package pkg) throws ServiceException {
		
		final String thisMethod = classLogName + "::getPackageSize()";
		logger.trace(thisMethod);
		
		String message = null;
		Long result = null;
		
		try {
				
			NgasFileDAO ngasFilesDAO = new NgasFileDAOSyb(d_dbUrl, d_dbUsername, d_dbPassword, d_dbNgasFilesTable);
			result = ngasFilesDAO.getSize(pkg.getFileIds());
			
		} catch (DAOException e) {
			message = e.getMessage();
		} finally {
			if (message != null)
				throw new ServiceException(message);
		}		
		
		return result;
	}
	
	public boolean findByQCFilename(Integer pkgId, String filePath, String qcFilename) throws ServiceException {
		
		final String thisMethod = classLogName + "::findByQCFilename()";
		logger.trace(thisMethod);
		
		Connection conn = null;
		String message = null;
		
		boolean result = false;
		
		try {
			conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
			PackageFilesDAO pkgFilesDAO = new PackageFilesDAOSyb(conn, d_dbFilesTable);
			result = pkgFilesDAO.findByQCFilename(pkgId, filePath, qcFilename);
		} catch (ConnectionServiceException e) {
			message = e.getMessage();
		} catch (DAOException e) {
			message = e.getMessage();
		} finally {
			if (conn != null)
			{
				try {
					SybaseConnectionService.closeConnection(conn);
				} catch (ConnectionServiceException ex) {
					message = ex.getMessage();
				}
			}
			if (message != null)
				throw new ServiceException(message);
		}		
		return result;		
	}
	
	
	@Deprecated
	public String getRequestId() throws ServiceException {
		try {
			Connection conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword);
			CallableStatement cstmt = null;
			conn.setAutoCommit(true);
			cstmt = conn.prepareCall("{ call " + this.d_dbUniqueIdProcedure + " (?,?)}");

			cstmt.setString(1, "PI_REQ_ID");
			cstmt.registerOutParameter(2, Types.INTEGER);

			cstmt.execute();
			return new DecimalFormat("DBPI00000").format(cstmt.getInt(2));
		} catch (ConnectionServiceException e) {
			throw new ServiceException(e.getMessage());
		} catch (SQLException e) {
			throw new ServiceException(e.getMessage());
		}
	}

	public boolean isConfigured() {
		return d_configured;
	}

	public void setConfigured(boolean configured) {
		this.d_configured = configured;
	}

	public void setMaxTextSize(Integer size) {
		MAX_TEXT_SIZE = size;
	}
	
	public Integer getMaxTextSize() {
		return MAX_TEXT_SIZE;
	}
	
}
