package org.eso.util.dal;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.eso.util.dal.ReplicationRequest.Status;

public class ReplicationRequestDAOSyb implements ReplicationRequestDAO {

	/** Apache Log4J logger. */
    static final Logger logger = Logger.getLogger(ReplicationRequestDAOSyb.class);
    
    /** table columns */
    private static String FILEID_KW			= "file_id";
    private static String FILEVERSION_KW	= "file_version";
    private static String DISKID_KW			= "disk_id";
    private static String SUBMITTERID_KW	= "submitter_id";
    private static String SUBMISSIONTIME_KW	= "submission_time";
    private static String PRIORITY_KW		= "priority";
    private static String STATUSID_KW		= "status_id";
    private static String MESSAGE_KW		= "message";
    private static String LASTUPDATETIME_KW = "last_update_time";
    
    /** database's connection parameters */
    private String d_dbUrl			= null;
    private String d_dbUsername		= null;
    private String d_dbPassword		= null;
    
    /** db table */
    private String d_dbTable		= null;
    
    /** db connection */
    private Connection d_connection	= null;
    
	public ReplicationRequestDAOSyb(String dbUrl, String dbUsername, String dbPassword, String dbTable) {
		d_dbUrl			= dbUrl;
		d_dbUsername	= dbUsername;
		d_dbPassword	= dbPassword;
		d_dbTable		= dbTable;
	}
	
	public void delete(ReplicationRequest request)
			throws DAOException {
		ReplicationRequest.validate(request);
		String query = "delete from " + d_dbTable + " where " + FILEID_KW + "='" + request.getFileId() + "'";
		executeQuery(query);
	}

	public List<ReplicationRequest> findByStatus(Status status)
			throws DAOException {
		String query = "SELECT * from " + d_dbTable + " WHERE " + STATUSID_KW + "= " + status.getCode() + " ORDER BY " + SUBMISSIONTIME_KW + " DESC";
		return executeQuery(query);
	}

	public void insert(ReplicationRequest request)
			throws DAOException {
		ReplicationRequest.validate(request);
		String query = "insert into " + d_dbTable + " " + formatForInsert(request);
		executeQuery(query);
	}

	public void update(ReplicationRequest request)
			throws DAOException {
		ReplicationRequest.validate(request);
		String query = "update " + d_dbTable + " " + formatForUpdate(request) + " where " + FILEID_KW + "='" + request.getFileId() + "'";
		executeQuery(query);
	}

	/**
	 * Converts the result of a database query to a list of ReplicationRequest objects
	 * 
	 * @param rs the result of the query
	 * @return a list of ReplicationRequest objects
	 * @throws DAOException
	 */
	public static List<ReplicationRequest> toReplicationRequestList(ResultSet rs) throws DAOException {
		List<ReplicationRequest> res = new ArrayList<ReplicationRequest>();
		try {
			while (rs.next()) {
				ReplicationRequest rr = new ReplicationRequest();

				rr.setFileId(rs.getString(FILEID_KW));
				rr.setFileVersion(rs.getInt(FILEVERSION_KW)); 
				rr.setDiskId(rs.getString(DISKID_KW)); 
				rr.setSubmitterId(rs.getString(SUBMITTERID_KW)); 
				rr.setSubmissionTime(rs.getDate(SUBMISSIONTIME_KW)); 
				rr.setPriority(rs.getInt(PRIORITY_KW)); 
				rr.setStatusId(ReplicationRequest.Status.get(rs.getInt(STATUSID_KW))); 
				rr.setLastUpdateTime(rs.getDate(LASTUPDATETIME_KW)); 
				rr.setMessage(rs.getString(MESSAGE_KW));
				
				res.add(rr);
			}
		} catch (SQLException e) {
			throw new DAOException(e.getMessage());
		} catch (NullPointerException e) {}		
		return res;
	}

	
	/**
	 * Convenience method used for database access
	 * @param query
	 * @return a List of ReplicationRequest objects
	 * @throws DAOException
	 */
	private List<ReplicationRequest> executeQuery(String query) throws DAOException {
		try {
			List<ReplicationRequest> res = null;
			if (d_connection == null)
			{
				Connection conn = SybaseConnectionService.getConnection(d_dbUrl, d_dbUsername, d_dbPassword); 
				res = toReplicationRequestList(SybaseConnectionService.executeStatement(conn, query));
				SybaseConnectionService.closeConnection(conn);
			}
			else
				res = toReplicationRequestList(SybaseConnectionService.executeStatement(d_connection, query));
			return res;
		} catch (ConnectionServiceException e) {
			throw new DAOException(e.getMessage());
		/* the following catch clauses fix DFS07824 */
		} catch (IllegalArgumentException e) {
			throw new DAOException(e.getMessage());
		} catch (Throwable e) {
			/* we shouldn't be here */
			logger.warn("ReplicationRequestDAOSyb::executeQuery - unexpected exception");
			throw new DAOException(e.getMessage());
		}
	}
	
	/**
	 * Converts an ReplicationRequest object to a string that can be use in an update query 
	 * @param rr
	 * @return
	 */
	public static String formatForUpdate(ReplicationRequest rr)
	{
		String res = "";

		if (rr.getFileId() != null)
			res += " set file_id='" + rr.getFileId() + "'";
		if (rr.getFileVersion() != null)
			res += ", file_version=" + rr.getFileVersion();
		if (rr.getDiskId() != null)
			res += ", disk_id ='" + rr.getDiskId() + "'";
		if (rr.getSubmitterId() != null)
			res += ", submitter_id ='" + rr.getSubmitterId() + "'";
		if (rr.getSubmissionTime() != null)
			res += ", submission_time ='" + rr.getSubmissionTime() + "'";
		if (rr.getPriority() != null)
			res += ", priority =" + rr.getPriority();
		if (rr.getStatusId() != null)
			res += ", status_id=" + rr.getStatusId().getCode();
		if (rr.getLastUpdateTime() != null)
			res += ", last_update_time ='" + rr.getLastUpdateTime() + "'";
		if (rr.getMessage() != null)
			res += ", message ='" + rr.getMessage() + "'";
		
		return res;
	}
	
	/**
	 * Converts an ReplicationRequest object to a String that can be used in an insert query
	 * @param rr
	 * @return
	 */
	public static String formatForInsert(ReplicationRequest rr)
	{
		String first = "(";
		String second = "(";

		if (rr.getFileId() != null)
		{
			first	+= "file_id";
			second	+= "'" + rr.getFileId() + "'";
		}
		if (rr.getFileVersion() != null)
		{
			first	+= "file_version";
			second	+= rr.getFileVersion();
		}
		if (rr.getDiskId() != null)
		{
			first	+= "disk_id";
			second	+= "'" + rr.getDiskId() + "'";
		}
		if (rr.getSubmitterId() != null)
		{
			first	+= ", submitter_id";
			second	+= ", '" + rr.getSubmitterId() + "'"; 
		}
		if (rr.getSubmissionTime() != null)
		{
			first	+= ", submission_time";
			second	+= ", '" + rr.getSubmissionTime() + "'";
		}
		if (rr.getPriority() != null)
		{
			first	+= "priority";
			second	+= rr.getPriority();
		}
		if (rr.getStatusId() != null)
		{
			first	+= "status_id";
			second	+= rr.getStatusId().getCode();
		}
		if (rr.getLastUpdateTime() != null)
		{
			first	+=	", last_update_time";
			second	+= 	", '" + rr.getLastUpdateTime() + "'";
		}
		if (rr.getMessage() != null)
		{
			first	+=	", message";
			second	+= 	", '" + rr.getMessage() + "'";
		}
		
		first	+= ")";
		second	+= ")";
		
		return first + " values " + second;
	}

}
