package org.eso.util.dal;

import java.util.Date;
import java.util.List;

/**
 * DAO Interface for TransferRequest objects
 * 
 * @author cmoins
 * 
 */
public interface TransferRequestDAO {

    /* transfer requests sorting order */
    public enum SortingOrder {
        DESCENDING, ASCENDING
    };

    /**
     * Find all transfer requests in the data repository
     * 
     * @param status
     *            status of the transfer requests to be retrieved
     * @return all the transfer requests in the data repository
     */
    public List<TransferRequest> findByStatus(TransferRequest.Status status)
            throws DAOException;

    /**
     * Updates the status of a specified transfer request
     * 
     * @param transferRequest
     *            transfer request whose status should be updated
     * @param status
     *            new status of the transfer request
     * @return the number of updated transfer requests if succeeded, -1
     *         otherwise
     * @throws DAOException
     */
    public int updateStatus(TransferRequest transferRequest,
            TransferRequest.Status status) throws DAOException;

    /**
     * Saves a transfer request in the data repository
     * 
     * @param TransferRequest
     *            transfer request to be saved
     * 
     * @return the number of inserted transfer requests if succeeded, -1
     *         otherwise
     * @throws DAOException
     */
    public int save(TransferRequest transferRequest) throws DAOException;

    /**
     * Deletes a transfer request from the data repository
     * 
     * @param TransferRequest
     *            transfer request to be saved
     * @return the number of deleted transfer requests if succeeded, -1
     *         otherwise
     * @throws DAOException
     */
    public int delete(TransferRequest transferRequest) throws DAOException;

    /**
     * checks if a transfer request is valid or not
     * 
     * @param transferRequest
     *            transfer request to be validated
     * @return true if the transfer request is valid, false otherwise
     */
    public Boolean isValid(TransferRequest transferRequest);

    /**
     * gets transfer requests associated to a list of instruments, in a given
     * status, starting from a given date ordered according to a given field in
     * a given order. If no instrument is given, all the instruments will be
     * selected
     * 
     * @param startDate
     *            date the research should start from
     * @param instrumentList
     *            List of instruments whose files shall be retrieved.
     *            Instruments will be separated by a space
     * @param status
     *            transfer request status
     * @param orderBy
     *            results sorting criteria
     * @param sortingOrder
     *            sorting order
     * @return a list of files fulfilling the user conditions
     */
    public List<TransferRequest> findByDateGreaterThan(Date startDate,
            String instrumentList, TransferRequest.Status status,
            TransferRequest.FIELDS orderBy, SortingOrder sortingOrder)
            throws DAOException;

    /**
     * gets transfer requests associated to a list of instruments and to a given
     * type of file, of a given observation date, in a given status, ordered
     * according to some given fields in a given order. If no instrument is
     * given, all the instruments will be selected
     * 
     * @param obsDate
     *            observation date of the files to be retrieved, starting from
     *            noon of the given day till noon of the following day
     * @param instrumentList
     *            List of instruments whose files shall be retrieved.
     *            Instruments will be separated by a space. Non mandatory
     * @param statusList
     *            list of transfer request status
     * @param fileType
     *            type of the file to be transfered. Non mandatory
     * @param orderByList
     *            results sorting criterions
     * @param sortingOrder
     *            sorting order
     * @return a list of files fulfilling the user conditions
     */
    public List<TransferRequest> findByDate(Date obsDate,
            String instrumentList, List<TransferRequest.Status> statusList,
            String fileType, List<TransferRequest.FIELDS> orderByList,
            SortingOrder sortingOrder) throws DAOException;

    /**
     * Deletes all transfer requests with the given status
     * 
     * @param status
     *            status of the transfer requests to be deleted
     */
    public void deleteByStatus(TransferRequest.Status status)
            throws DAOException;

}
