package org.eso.util.datatransfer;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.eso.util.stream.Message;
import org.eso.util.stream.ProcessingTask;
import org.eso.util.stream.TerminatingTask;

/**
 * This task maintains a history of processed messages and computes some
 * statistical information (e.g. throughput) at regular intervals.
 * 
 * @author szampier
 * 
 */

public class StatsKeeperTask extends ProcessingTask {

	/** Apache Log4J logger. */
	static final Logger logger = Logger.getLogger(StatsKeeperTask.class);

	/** The name of this class for logging purposes. */
	private static final String classLogName = "StatsKeeperTask";

	/** When this task was created. Used to display the application uptime. */
	final private long creationTime = System.currentTimeMillis();

	/** The total size of processed data, since the application was started. */
	private long totSize = 0;

	/**
	 * The total number of messages processed, since the application was
	 * started.
	 */
	private int totMessages = 0;

	/**
	 * The total number of transfer requests processed, since the application
	 * was started.
	 */
	private int totRequests = 0;

	/**
	 * This class is a storage for some statistical information about each
	 * processed message.
	 */
	private class HistoryElement {
		/** When the message was received by the application. */
		final long startTime;
		/** When the message was processed by the application. */
		final long endTime;
		/** The message size (in bytes) */
		final long size;
		/** The number of transfer requests contained in this message. */
		final int numRequests;

		/** Constructor with startTime and size. */
		public HistoryElement(long startTime, long size) {
			this.startTime = startTime;
			this.endTime = System.currentTimeMillis();
			this.size = size;
			this.numRequests = 1;
		}

		/** Constructor with startTime, size and numRequests. */
		public HistoryElement(long startTime, long size, int numRequests) {
			this.startTime = startTime;
			this.endTime = System.currentTimeMillis();
			this.size = size;
			this.numRequests = numRequests;
		}
	}

	/**
	 * This list maintains statistical information about the last N processed
	 * messages.
	 */
	LinkedList<HistoryElement> history = new LinkedList<HistoryElement>();

	/** The max history size. */
	private final int historyDepth;

	/** The next time statistics should be computed and displayed. */
	private final Date nextComputingTime;

	/** The time in seconds between stats computation. */
	private final int computingInterval;

	/** Used to format compute stats times in log messages. */
	private final SimpleDateFormat dateFormat = new SimpleDateFormat(
			"yyyy-MM-dd'T'HH:mm:ss.SSS");

	// Some useful constants.
	final static int MILLIS_DAY = 24 * 60 * 60 * 1000;
	final static int MILLIS_HOUR = 60 * 60 * 1000;
	final static int MILLIS_MIN = 60 * 1000;
	final static int MILLIS_SEC = 1 * 1000;

	final static int KILOBYTE = 1024;
	final static int MEGABYTE = 1024 * 1024;
	final static int GIGABYTE = 1024 * 1024 * 1024;
	final static long TERABYTE = (long) 1024 * 1024 * 1024 * 1024;

	/**
	 * Constructor.
	 * 
	 * @param config
	 * @throws NullPointerException
	 */
	public StatsKeeperTask(int computingInterval, int historyDepth)
			throws NullPointerException {
		super("StatsKeeperTask", 1, computingInterval);

		final String methodLogName = classLogName + "::StatsKeeperTask()";
		logger.trace(methodLogName);

		nextComputingTime = new Date();
		this.computingInterval = computingInterval;
		this.historyDepth = historyDepth;
	}

	@Override
	protected boolean canProcessMessage(Message message) {
		final String methodLogName = classLogName + "::canProcessMessage()";
		logger.trace(methodLogName);
		
		StatsInfo info = (StatsInfo) message;
		return info.canProcessInfo();
	}

	@Override
	protected void process(List<Message> msgList) {
		final String methodLogName = classLogName + "::process()";
		logger.trace(methodLogName);

		// Check if it is time to compute statistics.
		if (System.currentTimeMillis() >= nextComputingTime.getTime()) {
			computeStats();
			computeRunningStats();

			// Set the next polling time.
			nextComputingTime
					.setTime((System.currentTimeMillis() + (computingInterval * 1000)));

			logger.debug(methodLogName + " - next computing time: "
					+ dateFormat.format(nextComputingTime) + ".");

		}

		for (Iterator<Message> msgIter = msgList.iterator(); msgIter.hasNext();) {
			StatsInfo info = (StatsInfo) msgIter.next();

			// Update history.
			history.add(new HistoryElement(info.getStartTime(),
					info.getSize(),
					info.getNumRequests()));
			// Prevent history to grow indefinitely.
			if (history.size() > historyDepth) {
				history.removeFirst();
			}
			// Update global statistics.
			totMessages++;
			totRequests += info.getNumRequests();
			totSize += info.getSize();
		}
	}

	private void computeRunningStats() {
		final String methodLogName = classLogName + "::computeRunningStats()";
		logger.trace(methodLogName);

		if (history.isEmpty()) {
			return;
		}

		long size = 0;
		for (HistoryElement element : history) {
			size += element.size;
		}
		String messagesStr = "" + history.size() + " msgs";

		String throughputStr = "" + (float) size / KILOBYTE
				/ (history.getLast().endTime - history.getFirst().startTime)
				* 1000 + " KB/s";

		String infoStr = "[running stats] " + messagesStr + ", " + throughputStr;
		logger.info(infoStr);
	}

	private void computeStats() {
		final String methodLogName = classLogName + "::computeStats()";
		logger.trace(methodLogName);

		// Compute uptime and format it as : up <N> days <HH>:<MM>:<SS> 
		long uptime = System.currentTimeMillis() - creationTime;
		int nDays = (int) uptime / MILLIS_DAY;
		int remainder = (int) uptime % MILLIS_DAY;
		int nHours = remainder / MILLIS_HOUR;
		remainder = remainder % MILLIS_HOUR;
		int nMinutes = remainder / MILLIS_MIN;
		remainder = remainder % MILLIS_MIN;
		int nSeconds = remainder / MILLIS_SEC;

		DecimalFormat df = new DecimalFormat("00");

		String uptimeStr = "up " + nDays + " days " + df.format(nHours) + ":"
				+ df.format(nMinutes) + ":" + df.format(nSeconds);
		
		// Format the total number of processed messages.
		String totMessagesStr = totMessages + " msgs";
		// Format the total number of processed requests.
		String totRequestsStr = totRequests + " reqs";

		// Human readable representation of total processed size.
		String totSizeStr;
		if (totSize > TERABYTE) {
			totSizeStr = "" + (float) totSize / TERABYTE + " TB";
		} else if (totSize > GIGABYTE) {
			totSizeStr = "" + (float) totSize / GIGABYTE + " GB";
		} else if (totSize > MEGABYTE) {
			totSizeStr = "" + (float) totSize / MEGABYTE + " MB";
		} else {
			totSizeStr = "" + (float) totSize / KILOBYTE + " KB";
		}

		// Compute overall throughput in KB/s.
		String throughputStr = "" + (float) totSize / KILOBYTE / uptime
				* 1000 + " KB/s";

		// Assemble statistical info in one line message.
		String infoStr = "[overall stats] " + uptimeStr + ", " + totMessagesStr
				+ ", " + totRequestsStr + ", " + totSizeStr + ", "
				+ throughputStr;
		logger.info(infoStr);
	}


	/**
	 * Test program.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		// Basic Log4j config
		BasicConfigurator.configure();
		Logger utilLogger = Logger.getLogger("org.eso.util");
		utilLogger.setLevel(Level.INFO);

		class TestMessage extends Message implements StatsInfo {
			long creationTime = System.currentTimeMillis();
			final long size;
			final int numRequests; 
			
			public TestMessage(int numRequests, long size) {
				super(MSG_USER);
				this.numRequests = numRequests;
				this.size = size;
			}

			public boolean canProcessInfo() {
				return true;
			}

			public int getNumRequests() {
				return numRequests;
			}

			public long getSize() {
				return size;
			}

			public long getStartTime() {
				return creationTime;
			}
		}

		StatsKeeperTask task = new StatsKeeperTask(1, 100);
		task.setNextTask(new TerminatingTask());
		task.open();
		for (int i = 0; i < 1000; i++) {
			TestMessage msg = new TestMessage(i, 100 * i * i);
			task.put(msg);
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
}
