package org.eso.util.datatransfer;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * This class represents a set of properties associated to a file to transfer.
 * The main operations supported by this class are: - create an empty request
 * file (default constructor) - add properties to the request file (setProperty
 * method) - save the request file to disk (save method) - load the request file
 * from disk (constructor with filename) - get type safe property values (using
 * getPropertyAsInt and getPropertyAsString) In case of errors, a
 * TransferRequestException is thrown.
 * 
 * @author szampier <szampier@eso.org>
 */

@SuppressWarnings("serial")
public class TransferRequest extends Hashtable<String, String> {

    /** The TransferRequest filename */
    private String requestFilename;

    /**
     * The TransferRequest validator can be used to check the validity of the
     * TransferRequest
     */
    private TransferRequestValidator validator = null;

    /** The dictionary of allowed keywords and types */
    private static final Hashtable<String, String> TransferRequestDictionary =
            new Hashtable<String, String>();

    /* Private constants */
    private static final String BEGIN = "TRANSFER.REQUEST.BEGIN";
    private static final String END = "TRANSFER.REQUEST.END";
    private static final String TYPE_STRING = "STRING";
    private static final String TYPE_INT = "INT";
    private static final String TYPE_BOOLEAN = "BOOLEAN";
    private static final String EOL = System.getProperty("line.separator");

    /* List of supported keywords */
    public static final String FILEID = "FILEID";
    public static final String FILENAME = "FILENAME";
    public static final String UNCOMPRESSED_FILENAME = "UNCOMPRESSED.FILENAME";
    public static final String TRANSFER_METHOD = "TRANSFER.METHOD";
    public static final String TRANSFER_PRIORITY = "TRANSFER.PRIORITY";
    public static final String COMPRESSION_METHOD = "COMPRESSION.METHOD";
    public static final String SIZE = "SIZE";
    public static final String UNCOMPRESSED_SIZE = "UNCOMPRESSED.SIZE";
    public static final String CHECKSUM = "CHECKSUM";
    public static final String TRANSFER_CATG = "TRANSFER.CATG";
    public static final String USER = "USER";
    public static final String DESTDIR = "DESTDIR";
    public static final String MULTI_FILE_COMPONENTS = "MULTI.FILE.COMPONENTS";
    public static final String FORMAT = "FORMAT";

    /* Public constants */
    public static final String RAW_FILE_CATG = "rawfile";
    public static final String TOO_CATG = "too";
    public static final String MULTI_FILE_CATG = "multifile";
    public static final String HEADER_FILE_CATG = "header";

    public static final String TRANSFER_METHOD_DISK = "disk";
    public static final String TRANSFER_METHOD_NETWORK = "network";
    public static final String COMPRESSION_METHOD_NONE = "none";
    public static final String COMPRESSION_METHOD_UNIXCOMPRESS = "unixcompress";


    /** The Log4j logger */
    static Logger logger = Logger.getLogger(TransferRequest.class);

	/** The name of this class for logging purposes. */
	private static final String classLogName = "TransferRequest";

    static {
        TransferRequestDictionary.put(FILEID, TYPE_STRING);
        TransferRequestDictionary.put(FILENAME, TYPE_STRING);
        TransferRequestDictionary.put(UNCOMPRESSED_FILENAME, TYPE_STRING);
        TransferRequestDictionary.put(TRANSFER_METHOD, TYPE_STRING);
        TransferRequestDictionary.put(TRANSFER_PRIORITY, TYPE_INT);
        TransferRequestDictionary.put(COMPRESSION_METHOD, TYPE_STRING);
        TransferRequestDictionary.put(SIZE, TYPE_INT);
        TransferRequestDictionary.put(UNCOMPRESSED_SIZE, TYPE_INT);
        TransferRequestDictionary.put(CHECKSUM, TYPE_INT);
        TransferRequestDictionary.put(TRANSFER_CATG, TYPE_STRING);
        TransferRequestDictionary.put(USER, TYPE_STRING);
        TransferRequestDictionary.put(DESTDIR, TYPE_STRING);
        TransferRequestDictionary.put(MULTI_FILE_COMPONENTS, TYPE_STRING);
    }

    /** Default constructor */
    public TransferRequest() {
        super();
        final String methodLogName = classLogName + "::TransferRequest()";
        logger.trace(methodLogName);
    }

    /** Copy constructor */
    public TransferRequest(final TransferRequest anotherTransferRequest) {
    	super(anotherTransferRequest);

    	final String methodLogName = classLogName + "::TransferRequest(TransferRequest)";
        logger.trace(methodLogName);

        requestFilename = anotherTransferRequest.requestFilename;
    }
    
    /**
     * This constructor loads the TransferRequest from the specified filename.
     * 
     * @param filename
     *            The name of the file to load.
     * @throws TransferRequestException
     */
    public TransferRequest(String filename) throws TransferRequestException {
        super();
        
    	final String methodLogName = classLogName + "::TransferRequest(String)";
        logger.trace(methodLogName);

        load(filename);
    }

    /**
     * This constructor loads the TransferRequest from the specified filename
     * and invokes given TransferRequestValidator to validate the current
     * object.
     * 
     * @param filename
     *            The name of the file to load.
     * @param validator
     *            The TransferRequestValidator which determines if the
     *            TransferRequest is valid.
     * @throws TransferRequestException
     */
    public TransferRequest(String filename, TransferRequestValidator validator)
            throws TransferRequestException {
        super();
        
    	final String methodLogName = classLogName + "::TransferRequest(String,TransferRequestValidator)";
        logger.trace(methodLogName);

		if (filename == null) {
			String message = methodLogName + " - filename must not be null.";
			logger.fatal(message);
			throw new IllegalArgumentException(message);
		}

        if (validator == null) {
			String message = methodLogName + " - validator must not be null.";
			logger.fatal(message);
			throw new IllegalArgumentException(message);
        }
        
        this.validator = validator;
        load(filename);
        if (!validator.isValid(this)) {
            String message = "TransferRequest file is not valid.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
    }

    /**
     * Set the TransferRequestValidator.
     * 
     * @param validator
     *            the TransferRequestValidator.
     */
    public void setValidator(TransferRequestValidator validator) {
    	final String methodLogName = classLogName + "::setValidator(TransferRequestValidator)";
        logger.trace(methodLogName);

        if (validator == null) {
			String message = methodLogName + " - validator must not be null.";
			logger.fatal(message);
			throw new IllegalArgumentException(message);
        }

    	this.validator = validator;
    }

    /**
     * This method invokes the TransferRequestValidator to check the validity of
     * the current object.
     * 
     * @return true if this TransferRequest is valid. False otherwise.
     */
    public boolean isValid() {
        if (validator == null) {
            logger.error("The TransferRequestValidator is not set.");
            return false;
        }
        return validator.isValid(this);
    }

    /**
     * Returns the TransferRequest filename.
     * 
     * @return the name of the request file.
     */
    public String getRequestFilename() {
        return requestFilename;
    }

    /**
     * Set the TransferRequest filename.
     */
    public void setRequestFilename(String f) {
        requestFilename = f;
    }

    /**
     * Load the TransferRequest from the specified filename
     * 
     * @param filename
     *            The name of the file to load.
     * 
     * @throws TransferRequestException
     *             If an IO error occurs or the TransferRequest is not formatted
     *             correctly.
     */
    private void load(String filename) throws TransferRequestException {
        String thisMethod = "TransferRequest::load(String)";
        logger.trace(thisMethod);

        // Check input parameter
        if (filename == null) {
            throw new IllegalArgumentException(thisMethod
                    + " - filename must not be null");
        }

        BufferedReader trFileReader;
        try {
            logger.debug(thisMethod + " - Opening file " + filename);
            trFileReader = new BufferedReader(new FileReader(filename));
        } catch (FileNotFoundException fnfe) {
            String message = "Could not open file " + filename;
            logger.error(thisMethod + " - " + message);
            throw new TransferRequestException(message, fnfe);
        }

        requestFilename = filename;

        // The current line in the TRQ file.
        String currentLine;
        String[] currentCard;
        boolean endKeywordFound = false;

        // Read and process lines from the TransferRequest file.
        try {
            // Read the first line of the file.
            currentLine = trFileReader.readLine();
            // If the first line is null, the file is empty.
            if (currentLine == null) {
                String message = "File " + requestFilename + " is empty.";
                logger.error(thisMethod + " - " + message);
                throw new TransferRequestException(message);
            }
            // The first line of a TransferRequest must contain the keyword
            // TransferRequest
            if (!currentLine.equals(BEGIN)) {
                String message =
                        "File " + requestFilename
                                + " is not a valid TransferRequest"
                                + " because it doesn't start with '" + BEGIN
                                + "'.";
                logger.error(thisMethod + " - " + message);
                throw new TransferRequestException(message);
            }
            // Parse the TransferRequest contents.
            while ((currentLine = trFileReader.readLine()) != null) {
                logger.debug(thisMethod + " - Parsing line: " + currentLine);
                if (endKeywordFound == true) {
                    String message =
                            "File '" + requestFilename
                                    + "' is not a valid TransferRequest"
                                    + " because it doesn't end with '" + END
                                    + "'.";
                    logger.error(thisMethod + " - " + message);
                    throw new TransferRequestException(message);
                }
                if (currentLine.equals(END)) {
                    endKeywordFound = true;
                    continue;
                }
                if (currentLine.trim().equals("")) {
                    /* Blank line read - skip to the next line. */
                    continue;
                }
                currentCard = currentLine.split("=");
                if (currentCard.length != 2) {
                    String message = "Could not parse line " + currentLine;
                    logger.error(thisMethod + " - " + message);
                    throw new TransferRequestException(message);
                }
                super.put(currentCard[0].trim(), currentCard[1].trim());
            }
            if (endKeywordFound == false) {
                String message =
                        "File '" + requestFilename
                                + "' is not a valid TransferRequest"
                                + " because it doesn't end with '" + END + "'.";
                logger.error(thisMethod + " - " + message);
                throw new TransferRequestException(message);
            }
        } catch (IOException ioe) {
            /* An IO error occurred whilst reading the TRQ file. */
            String message =
                    "An IO error occurred whilst reading from " + "TRQ file '"
                            + requestFilename + "'.";
            logger.error(thisMethod + " - " + message);
            throw new TransferRequestException(message, ioe);
        } finally {
            try {
                logger.debug(thisMethod + " - Closing file " + filename);
                trFileReader.close();
            } catch (IOException ioe) {
                String message = "Could not close file " + filename;
                logger.error(thisMethod + " - " + message);
                throw new TransferRequestException(message, ioe);
            }
        }
    }

    /**
     * Save the TransferRequest to the specified filename.
     * 
     * @param filename
     *            The name of the file where the TransferRequest should be saved
     *            to.
     * 
     * @throws TransferRequestException
     *             If the TransferRequest is empty or an IO error occurs.
     */
    public void save(String filename) throws TransferRequestException {
        String thisMethod = "TransferRequest::save()";

        // Check input parameter
        if (filename == null) {
            throw new IllegalArgumentException(thisMethod
                    + " - filename must not be null");
        }

        if (isEmpty()) {
            String message = "Could not save empty TransferRequest.";
            logger.error(message);
            throw new TransferRequestException(message);
        }

        FileWriter trFileWriter;
        try {
        	trFileWriter = new FileWriter(new File(filename));
        	trFileWriter.write(BEGIN);
        	trFileWriter.write(EOL);
            for (Iterator<String> iter = keySet().iterator(); iter.hasNext();) {
                String key = iter.next();
                String value = super.get(key);
                String currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            trFileWriter.write(END);
            trFileWriter.close();
        } catch (IOException ioe) {
            String message = "Could not open TRQ file " + filename;
            logger.error(message);
            throw new TransferRequestException(message, ioe);
        }
        requestFilename = filename;
    }

    /**
     * Returns the integer value of the given property.
     * 
     * @param key
     *            The property name
     * @return The integer value of the given property or null if the property
     *         is not defined.
     * @throws TransferRequestException
     *             if the property is not in the dictionary or the type does not
     *             match.
     */
    public Integer getPropertyAsInt(String key) throws TransferRequestException {
        if (!TransferRequestDictionary.containsKey(key)) {
            String message = "Keyword '" + key + "' is not in the dictionary.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
        String keyType = TransferRequestDictionary.get(key);
        if (keyType != TYPE_INT) {
            String message =
                    "Requested INT value for keyword '" + key
                            + "', which has type " + keyType;
            logger.error(message);
            throw new TransferRequestException(message);
        }
        if (containsKey(key)) {
            return new Integer(super.get(key));
        } else {
            return null;
        }
    }

    /**
     * Returns the string value of the given property.
     * 
     * @param key
     *            The property name
     * @return The string value of the given property or null if the property is
     *         not defined.
     * @throws TransferRequestException
     *             if the property is not in the dictionary or the type does not
     *             match.
     */
    public String getPropertyAsString(String key)
            throws TransferRequestException {
        if (!TransferRequestDictionary.containsKey(key)) {
            String message = "Keyword '" + key + "' is not in the dictionary.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
        String keyType = TransferRequestDictionary.get(key);
        if (keyType != TYPE_STRING) {
            String message =
                    "Requested STRING value for keyword '" + key
                            + "', which has type " + keyType;
            logger.error(message);
            throw new TransferRequestException(message);
        }
        return super.get(key);
    }

    /**
     * Returns the boolean value of the given property.
     * 
     * @param key
     *            The property name
     * @return The boolean value of the given property or null if the property
     *         is not defined.
     * @throws TransferRequestException
     *             if the property is not in the dictionary or the type does not
     *             match.
     */
    public Boolean getPropertyAsBoolean(String key)
            throws TransferRequestException {
        if (!TransferRequestDictionary.containsKey(key)) {
            String message = "Keyword '" + key + "' is not in the dictionary.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
        String keyType = TransferRequestDictionary.get(key);
        if (keyType != TYPE_BOOLEAN) {
            String message =
                    "Requested BOOLEAN value for keyword '" + key
                            + "', which has type " + keyType;
            logger.error(message);
            throw new TransferRequestException(message);
        }
        if (containsKey(key)) {
            return new Boolean(super.get(key));
        } else {
            return null;
        }
    }

    /**
     * Checks if a property is valid, i.e if it is available inside the transfer
     * request dictionary and if the associated value is consistent with the
     * required type in case of an integer
     * 
     * @param key
     *            key associated to the property to check
     * @return true if the property is valid, false otherwise
     * @throws TransferRequestException
     */
    public Boolean isValidProperty(String key) throws TransferRequestException {
        Boolean isValid = true;
        if (!TransferRequestDictionary.containsKey(key)) {
            String message = "Keyword '" + key + "' is not in the dictionary.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
        String keyType = TransferRequestDictionary.get(key);
        String keyValue = super.get(key);
        if (keyType == TYPE_INT) {
            try {
                Integer.valueOf(keyValue);
            } catch (NumberFormatException e) {
                String message =
                        "NumberFormatException " + e.getMessage() + " " + key
                                + " value is not a valid integer";
                logger.error(message);
                isValid = false;
            }
        }
        return isValid;
    }

    /**
     * This method throws a RuntimeException to indicate that it should never be
     * called by the user. The method setProperty should be used instead.
     */
//    public String put(String key, String value) {
//        String thisMethod = "TransferRequest::put()";
//        String message =
//                thisMethod
//                        + " This method should be never called. Use setProperty instead.";
//        logger.fatal(message);
//        throw new RuntimeException(message);
//    }

    /**
     * This method throws a RuntimeException to indicate that it should never be
     * called by the user. The method getProperty should be used instead.
     */
//    public String get(String key) {
//        String thisMethod = "TransferRequest::get()";
//        String message =
//                thisMethod
//                        + " This method should be never called. Use getProperty instead.";
//        logger.fatal(message);
//        throw new RuntimeException(message);
//    }

    /**
     * Store a string-valued property in the TransferRequest.
     * 
     * @param key
     *            The property name
     * @param value
     *            The property value.
     * @throws TransferRequestException
     *             If the property is not in the dictionary or if the type does
     *             not match.
     */
    public void setProperty(String key, String value)
            throws TransferRequestException {
        setProperty(key, TYPE_STRING, value);
    }

    /**
     * Store an integer-valued property in the TransferRequest.
     * 
     * @param key
     *            The property name
     * @param value
     *            The property value.
     * @throws TransferRequestException
     *             If the property is not in the dictionary or if the type does
     *             not match.
     */
    public void setProperty(String key, int value)
            throws TransferRequestException {
        setProperty(key, TYPE_INT, value);
    }

    /**
     * Store a property in the TransferRequest.
     * 
     * @param key
     *            The property name.
     * @param type
     *            The property type.
     * @param value
     *            The property value.
     * @throws TransferRequestException
     *             If the property is not in the dictionary or if the type does
     *             not match.
     */
    private void setProperty(String key, String type, Object value)
            throws TransferRequestException {
        if (!TransferRequestDictionary.containsKey(key)) {
            String message = "Keyword '" + key + "' is not in the dictionary.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
        String keyType = TransferRequestDictionary.get(key);
        if (keyType != type) {
            String message =
                    "Keyword '" + key + "' must have type '" + keyType
                            + "' not '" + type + "'.";
            logger.error(message);
            throw new TransferRequestException(message);
        }
        super.put(key, value.toString());
    }

    /**
     * Test the basic functionalities of the TransferRequest class.
     * 
     * @param args
     */
    public static void main(String[] args) {
        // Basic Log4j configuration
        BasicConfigurator.configure();

        // Create new TransferRequest for writing
        TransferRequest tr = new TransferRequest();
        String filename = "c:\\temp\\pippo.trq";
        String filename2 = "c:\\temp\\pippo2.trq";
        try {
            tr.setProperty(TransferRequest.FILENAME, "/tmp/pippo");
            tr.setProperty(TransferRequest.FILEID, "pippo");
            tr.setProperty(TransferRequest.SIZE, 999);
            // tr.setProperty(TransferRequest.CHECKSUM, 999);
            // tr.put(TransferRequest.FILENAME, "/tmp/pippo");
            // tr.get(TransferRequest.FILENAME);
            // System.out.println(tr.getPropertyAsInt("FILENAME"));
            // The following operations cause a TransferRequestException to be
            // thrown.

            // System.out.println(tr.getPropertyAsInt("FILENAME"));
            // tr.setProperty("FILENAME",1);
            // tr.setProperty("KEY.NOT.IN.DICTIONARY", 1);

            // Save the TransferRequest.
            tr.save(filename);
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {
            // Load a TransferRequest from the filesystem.
            TransferRequest tr2 = new TransferRequest(filename);
            logger.debug(tr2.getRequestFilename());

            // System.out.println(tr2.get(TransferRequest.FILENAME));
            logger.debug(tr2.getPropertyAsString(TransferRequest.FILENAME));
            logger.debug(tr2.getPropertyAsInt(TransferRequest.SIZE));

            // The following operations cause a TransferRequestException to be
            // thrown.
            // System.out.println(tr2.getPropertyAsInt("FILENAME"));
            // System.out.println(tr2.getPropertyAsInt("KEY.NOT.IN.DICTIONARY"));
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {
            TransferRequest tr3 =
                    new TransferRequest(filename2,
                            new DefaultTransferRequestValidator());
        } catch (TransferRequestException e) {
            e.printStackTrace();
        }
    }
}
