package org.eso.util.misc; 

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;

import java.util.concurrent.Semaphore; 

import org.apache.log4j.Logger;

/** <p>StreamConsumer objects are used to read all available data from an 
 * input stream until end of file (also known as "stream consumption"). 
 * Stream consumption is performed in the background by a dedicated daemon 
 * thread, allowing callers to continue working and collect the string of 
 * consumed characters at a later point.</p> 
 *
 * <p>The stream to consume can be specified at construction time, or by 
 * calling the {@link #consumeStream()} method. Once a stream has been 
 * consumed, the StreamConsumer can be reused by calling consumeStream() 
 * again. The stream consumption thread will be terminated when the 
 * StreamConsumer is garbage collected or the JVM exits.</p> 
 * 
 * <p><b>WARNING:</b> Once a stream has been consumed, the StreamConsumer 
 * does not close the stream. If a stream needs to be closed explicitly, 
 * this should be done by the caller. However, the StreamConsumer does 
 * drop its reference to the stream, allowing it to be garbage collected 
 * after consumption.</p> 
 * 
 * <p><b>WARNING:</b> Due to a defect in the Java API, if a StreamConsumer 
 * has read all of the data available on a stream, and then the process 
 * feeding the stream hangs, the StreamConsumer and any threads waiting 
 * for it will hang. This will not affect other StreamConsumers - each 
 * gets its own stream consumption thread - but it will permanently lock 
 * up system resources until the stream closes or the JVM exits. If that 
 * happens often enough, the JVM will hit system-dependent limits on file 
 * descriptors and/or threads, causing it to crash.</p> 
 * 
 * <p>The only way to guarantee avoiding this problem is to ensure that the 
 * process feeding the stream closes its end, even after encountering the 
 * worst of errors. Although the operating system can (slowly) reclaim a 
 * dangling stream after a process exits, it can do nothing if the process 
 * hangs or otherwise fails to return its stream file descriptors.</p> 
 * 
 * <p>The Sun bug tracker contains more information on the root problem 
 * within the Java API - see ticket number 
 * <a href="http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6192696">6192696</a>. 
 * </p>
 *
 * <p><b>WARNING:</b> Although this class uses a mixture of synchronisation 
 * and semaphores to maintain a consistent internal state, it is still 
 * possible to run into ordering problems when a single StreamConsumer is 
 * shared by several threads. For example, the following sequence of events 
 * is perfectly feasible: 
 * <ul>
 *   <li>thread A calls {@link #consumeStream()} for stream A.</li>
 *   <li>thread A calls {@link #getConsumedString()} for the results.</li>
 *   <li>thread B calls {@link #awaitStreamConsumption()}.</li>
 *   <li>the stream consumption thread consumes stream A.</li>
 *   <li>{@link #awaitStreamConsumption()} returns to thread B.</li>
 *   <li>thread B calls {@link #consumeStream()} for stream B.</li>
 *   <li>the stream consumption thread consumes stream B.</li>
 *   <li>{@link #getConsumedString()} returns the contents of stream B 
 *       to thread A.</li>
 * </ul>
 * Calling threads are responsible for coordinating their actions so that 
 * these situations do not occur.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org>; 
 *         Stefano Zampieri ESO/SDD/DFI/BET. 
 *         Questions should be addressed to Neil in the first instance. 
 */ 
public class StreamConsumer 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(StreamConsumer.class);

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "StreamConsumer"; 

    /** The stream consumption thread for this StreamConsumer. */ 
    private final StreamConsumptionThread thread; 

    /** Constructs a new StreamConsumer object; the stream consumption 
     * thread is given a JVM-generated name and started, ready for the 
     * first {@link #consumeStream()} call. 
     */ 
    public StreamConsumer() 
    {
        final String methodLogName = classLogName + "::StreamConsumer()"; 

        logger.trace(methodLogName); 

        // Create the stream consumption thread; it is started automatically. 
        thread = new StreamConsumptionThread(null); 
    }

    /** Constructs a new StreamConsumer object; the stream consumption 
     * thread is given the specified thread name and started, ready for 
     * the first {@link #consumeStream()} call. 
     *
     * @param threadName the name of the stream consumption thread. 
     * 
     * @throws NullPointerException if threadName is null. 
     */ 
    public StreamConsumer(final String threadName) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::StreamConsumer(String)"; 

        logger.trace(methodLogName); 

        if(threadName == null) 
        {
            final String message = 
                methodLogName + " - threadName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Create the stream consumption thread; it is started automatically. 
        thread = new StreamConsumptionThread(threadName); 
    }

    /** Constructs a new StreamConsumer object; the stream consumption 
     * thread is given a JVM-generated name, started and told to read 
     * from the specified inputStream. 
     *
     * @param inputStream the input stream to consume characters from. 
     * 
     * @throws NullPointerException if inputStream is null. 
     */ 
    public StreamConsumer(final InputStream inputStream) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::StreamConsumer(InputStream)"; 

        logger.trace(methodLogName); 

        // Create the stream consumption thread; it is started automatically. 
        thread = new StreamConsumptionThread(null); 

        // Tell the thread to read from the specified input stream. 
        thread.consumeStream(inputStream); 
    }

    /** Constructs a new StreamConsumer object; the stream consumption 
     * thread is given the specified thread name, started by this constructor 
     * and told to read from the specified inputStream. 
     *
     * @param inputStream the input stream to consume characters from. 
     * @param threadName the name of the stream consumption thread. 
     * 
     * @throws NullPointerException if inputStream or threadName is null. 
     */ 
    public StreamConsumer(final InputStream inputStream, 
                          final String threadName) 
        throws NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::StreamConsumer(InputStream, String)"; 

        logger.trace(methodLogName); 

        if(threadName == null) 
        {
            final String message = 
                methodLogName + " - threadName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Create the stream consumption thread; it is started automatically. 
        thread = new StreamConsumptionThread(threadName); 

        // Tell the thread to read from the specified input stream. 
        thread.consumeStream(inputStream); 
    }

    /** Tells the stream consumption thread to consume characters from the 
     * specified input stream. 
     *
     * @param inputStream the input stream to consume characters from. 
     * 
     * @throws IllegalStateException if this StreamConsumer is currently 
     *                               consuming characters from a stream. 
     * @throws NullPointerException if inputStream is null. 
     */
    public void consumeStream(final InputStream inputStream) 
        throws IllegalStateException, NullPointerException 
    {
        final String methodLogName = classLogName + "::consumeStream()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        thread.consumeStream(inputStream); 
    }
    
    /** Returns true if this StreamConsumer is currently consuming characters 
     * from an input stream, false otherwise. 
     */
    public boolean isConsumingStream()
    {
        final String methodLogName = classLogName + "::isConsumingStream()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        return(thread.isConsumingStream()); 
    }

    /** Returns the input stream this StreamConsumer is consuming characters 
     * from, or null if no input stream is being consumed. 
     */
    public InputStream getInputStream()
    {
        final String methodLogName = classLogName + "::getInputStream()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        return(thread.getInputStream()); 
    }

    /** Returns true if no errors have occurred while this StreamConsumer 
     * was consuming characters from the current input stream. 
     */ 
    public boolean good()
    {
        final String methodLogName = classLogName + "::good()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        return(thread.good()); 
    }

    /** Blocks until the current input stream has been consumed. If no input 
     * stream has been set yet, this method will block until a stream has been
     * set and stream consumption is complete. 
     */ 
    public void awaitStreamConsumption() 
    {
        final String methodLogName = 
            classLogName + "::awaitStreamConsumption()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        thread.awaitStreamConsumption(); 
    }

    /** Returns the full string consumed from the last input stream to be set 
     * (see {@link #consumeStream()}). If no input stream has been set yet, or
     * stream consumption is still in progress, this method will block until 
     * a stream has been set and stream consumption is complete. 
     * 
     * @return the full string consumed from the last input stream to be set. 
     */
    public String getConsumedString() 
    {
        final String methodLogName = classLogName + "::getConsumedString()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        return(thread.getConsumedString()); 
    }

    /** Aborts consumption of the current input stream, if any. Please 
     * note that this method does not wait for stream consumption to 
     * finish - if necessary, use the {@link #isConsumingStream()} and 
     * {@link #awaitStreamConsumption()} methods to ensure that this 
     * StreamConsumer is ready for reuse. 
     */ 
    synchronized void abortStreamConsumption() 
    {
        final String methodLogName = 
            classLogName + "::abortStreamConsumption()"; 

        logger.trace(methodLogName); 

        // Delegate to the equivalent StreamConsumptionThread method. 
        thread.abortStreamConsumption(); 
    }

    /** Finalises this StreamConsumer by terminating its stream consumption 
     * thread; it is not safe to use a StreamConsumer after calling this. 
     */ 
    public void finalize()
    {
        final String methodLogName = classLogName + "::finalize()"; 

        logger.trace(methodLogName); 

        logger.debug(methodLogName + " - finalising."); 

        // Terminate the StreamConsumptionThread. 
        thread.terminate(); 
    }
}

/** <p>Applications should not use this class directly.</p>
 *
 * <p>Every StreamConsumer object uses an instance of this class to consume 
 * characters from the input stream. The separation of StreamConsumer and 
 * StreamConsumptionThread allows a StreamConsumer to be garbage collected 
 * and finalised while its StreamConsumptionThread is still running, so 
 * that finalisation can stop the thread and conserve JVM resources. See 
 * {@link StreamConsumer#finalize()}.</p> 
 * 
 * <p>StreamConsumptionThread cannot be made a private class within the 
 * StreamConsumer class, because instances of a private class always hold 
 * a reference to an instance of their containing class. This would prevent 
 * the garbage collection and finalisation that thread cleanup relies on.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org>; 
 *         Stefano Zampieri ESO/SDD/DFI/BET. 
 *         Questions should be addressed to Neil in the first instance. 
 */ 
class StreamConsumptionThread extends Thread 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = 
        Logger.getLogger(StreamConsumptionThread.class);

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "StreamConsumptionThread"; 

    /** The 'End Of Line' string for this system. */
    private static final String EOL = System.getProperty("line.separator"); 

    /** The maximum number of characters to consume at once. */ 
    private static final int READ_BUFFER_SIZE = 1024; 

    /** The polling period when waiting for more characters from the stream. 
     * Specified in milliseconds. 
     */ 
    private static final int STREAM_POLLING_INTERVAL = 100; 

    /** The input stream to consume characters from. */ 
    private volatile InputStream inputStream = null; 

    /** A string buffer to accumulate consumed characters in. */ 
    private StringBuffer consumedStringBuffer = null; 

    /** A string storing consumed characters for later retrieval. */ 
    private String consumedString = null; 

    /** A binary semaphore, used to ensure that either the stream consumption 
     * thread is updating the consumed string, or a thread calling 
     * {@link #getConsumedString()} is reading it, but not both at once. 
     */ 
    private final Semaphore consumedStringPermit = new Semaphore(0); 

    /** A binary semaphore, used to block the stream consumption thread until 
     * a {@link #consumeStream()} operation has been requested. 
     */ 
    private final Semaphore streamConsumptionPermit = new Semaphore(0); 

    /** True if this StreamConsumptionThread has never been told to consume 
     * a stream before. 
     */ 
    private boolean newStreamConsumptionThread = true; 

    /** True if the stream consumption thread is currently reading from 
     * an input stream. 
     */ 
    private volatile boolean isConsumingStream = false; 

    /** True if the stream consumption thread should abort reading from 
     * the current input stream. 
     */ 
    private volatile boolean abortStreamConsumption = false; 

    /** True if no errors have occurred whilst consuming characters from the 
     * input stream. 
     */
    private volatile boolean good = true; 
    
    /** A binary semaphore, used to prevent the constructor returning until  
     * the stream consumption thread is properly started. This sidesteps a
     * potential finalisation deadlock for temporary objects.
     */ 
    private final Semaphore activationSemaphore = new Semaphore(0); 

    /** True if the stream consumption thread should terminate. */ 
    private volatile boolean terminate = false; 

    /** Constructs a new StreamConsumptionThread object. 
     *
     * @param threadName an optional name for the stream consumption thread. 
     *               If this is null, the name will be generated by the JVM. 
     */ 
    StreamConsumptionThread(final String threadName)
    {
        final String methodLogName = 
            classLogName + "::StreamConsumptionThread()"; 

        logger.trace(methodLogName); 

        // If a name was specified for the stream consumption thread, use 
        // it, otherwise let the JVM generate one automatically. 
        if(threadName != null)
        {
            logger.debug(methodLogName 
                         + " - starting stream consumption thread [" 
                         + threadName + "]."); 

            setName(threadName); 
        }
        else
        {
            logger.debug(methodLogName 
                         + " - starting stream consumption thread."); 
        }

        // Start the thread in daemon mode. 
        setDaemon(true);
        start(); 

        // Wait for the stream consumption thread to start (and release
        // the activation semaphore) before returning. 
        activationSemaphore.acquireUninterruptibly();
    }

    /** Tells the stream consumption thread to consume characters from the 
     * specified input stream. 
     *
     * @param inputStream the input stream to consume characters from. 
     * 
     * @throws IllegalStateException if this StreamConsumptionThread 
     *              is currently consuming characters from a stream. 
     * @throws NullPointerException if inputStream is null. 
     */
    synchronized void consumeStream(final InputStream inputStream) 
        throws IllegalStateException, NullPointerException 
    {
        final String methodLogName = classLogName + "::consumeStream()"; 

        logger.trace(methodLogName); 

        if(inputStream == null) 
        {
            final String message = 
                methodLogName + " - inputStream must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Ensure that this StreamConsumptionThread is not already 
        // consuming characters from a stream. 
        if(isConsumingStream)
        {
            final String message = 
                methodLogName + " - this StreamConsumptionThread"
                + " is currently consuming characters from a stream." 
                + " It is not ready for a new consumeStream() call."; 

            logger.fatal(message); 
            throw new IllegalStateException(message); 
        }

        // Lock out any attempts to get a previously consumed string, 
        // before setting the new input stream. 
        if(newStreamConsumptionThread)
        {
            // This is a special case; when updating the consumed string 
            // buffer for the first time, the permit implicitly belongs to 
            // the stream consumption thread. 
            logger.debug(methodLogName 
                         + " - this is a new StreamConsumptionThread."); 
            newStreamConsumptionThread = false; 
        }
        else 
        {
            // This is the normal case; acquire the permit. 
            consumedStringPermit.acquireUninterruptibly(); 
        }

        // As a binary semaphore, the consumedStringPermit must never have 
        // any permits available at this stage. 
        if(consumedStringPermit.availablePermits() != 0)
        {
            final String message = 
                methodLogName + " - [consumedStringPermit]" 
                + " binary semaphore violation detected."; 

            logger.fatal(message); 
            throw new AssertionError(message); 
        }

        // Set the input stream... 
        logger.debug(methodLogName + " - setting new input stream."); 
        this.inputStream = inputStream; 

        // ...prepare to consume characters from it... 
        abortStreamConsumption = false; 
        consumedStringBuffer = new StringBuffer(); 
        consumedString = null; 
        isConsumingStream = true; 
        good = true; 

        // ...and make a consumption permit available to the stream 
        // consumption thread. 
        streamConsumptionPermit.release(); 
    }

    /** Returns true if this StreamConsumptionThread is currently consuming 
     * characters from an input stream, false otherwise. 
     */
    boolean isConsumingStream()
    {
        final String methodLogName = classLogName + "::isConsumingStream()"; 

        logger.trace(methodLogName); 

        return(isConsumingStream); 
    }

    /** Returns the input stream this StreamConsumptionThread is consuming 
     * characters from, or null if no input stream is being consumed. 
     */
    InputStream getInputStream()
    {
        final String methodLogName = classLogName + "::getInputStream()"; 

        logger.trace(methodLogName); 

        return(inputStream); 
    }

    /** This method is run by the stream consumption thread; it consists 
     * of a loop performing the following actions: 
     * <ul>
     *   <li>if this thread has been interrupted, exit the loop and 
     *       terminate the thread.</li> 
     *   <li>otherwise, await notification that a new input stream has 
     *       been set for consumption.</li>
     *   <li>get a reader for the stream.</li>
     *   <li>consume characters from the stream until end of file.</li>
     *   <li>release the consumed string buffer to other methods.</li>
     * </ul>
     */
    public void run() 
    {
        final String methodLogName = classLogName + "::run()"; 

        logger.trace(methodLogName); 

        logger.debug(methodLogName 
                     + " - stream consumption thread started."); 

        // Notify the constructor that the thread has started. 
        activationSemaphore.release(); 

        // A reader for the current input stream. 
        BufferedReader reader = null; 

        // The current characters from the input stream. 
        final char[] currentChars = new char[READ_BUFFER_SIZE]; 

        // The number of characters last read from the stream. 
        int numCharsRead = 0; 

        // This thread runs until terminate() is called or the JVM exits. 
    THREAD_LOOP: 
        while(!terminate) 
        {
            // Await notification that a new input stream has been set 
            // for consumption. 
            logger.debug(methodLogName 
                         + " - waiting for a new input stream."); 

            while(true) 
            {
                try
                {
                    streamConsumptionPermit.acquire(); 
                    break; 
                }
                catch(InterruptedException ie)
                {
                    if(terminate) 
                    {
                        logger.debug(
                            methodLogName 
                            + " - stream consumption thread interrupted for" 
                            + " termination whilst acquiring semaphore."); 

                        break THREAD_LOOP; 
                    }
                    else if(abortStreamConsumption)
                    {
                        // Quietly ignore this; no stream is being consumed. 
                    }
                    else
                    {
                        // This thread was unexpectedly interrupted - log 
                        // it, drop it and keep waiting for the semaphore. 
                        logger.debug(
                            methodLogName 
                            + " - unexpected interrupt received whilst" 
                            + " waiting for stream consumption permit."); 
                    }
                }
            }

            // As a binary semaphore, the streamConsumptionPermit must 
            // never have any permits available at this stage. 
            if(streamConsumptionPermit.availablePermits() != 0)
            {
                final String message = 
                    methodLogName + " - [streamConsumptionPermit]" 
                    + " binary semaphore violation detected."; 

                logger.fatal(message); 
                throw new AssertionError(message); 
            }

            // This method is now permitted to update the consumed string 
            // buffer; the permit was acquired by consumeStream(). 
            logger.debug(
                methodLogName 
                + " - new input stream set; preparing to consume it."); 

            // Get a reader for the input stream. 
            reader = 
                new BufferedReader(new InputStreamReader(inputStream)); 

            // Consume characters from the stream until end of file. 
            try 
            {
            STREAM_CONSUMPTION_LOOP:
                while(true)
                {
                    do
                    {
                        if(abortStreamConsumption) 
                        {
                            logger.debug(
                                methodLogName 
                                + " - stream consumption aborted."); 
 
                            break STREAM_CONSUMPTION_LOOP; 
                        }
                        else if(terminate) 
                        {
                            logger.debug(
                                methodLogName + " - terminating" 
                                + " stream consumption thread."); 

                            break THREAD_LOOP; 
                        }

                        // Read characters from the stream. 
                        numCharsRead = 
                            reader.read(currentChars, 0, READ_BUFFER_SIZE); 

                        if(numCharsRead < 0)
                        {
                            // The stream has reached end of file. 
                            logger.debug(methodLogName 
                                         + " - stream consumption complete.");

                            break STREAM_CONSUMPTION_LOOP; 
                        }
                        else if(numCharsRead > 0)
                        {
                            consumedStringBuffer.append(currentChars, 
                                                        0, 
                                                        numCharsRead); 
                        }
                    } 
                    while(numCharsRead > 0); 

                    // Avoid hogging the CPU when no data are available 
                    // from the stream. 
                    try
                    {
                        sleep(STREAM_POLLING_INTERVAL); 
                    }
                    catch(InterruptedException ie) 
                    {
                        if(abortStreamConsumption || terminate) 
                        {
                            logger.debug(
                                methodLogName 
                                + " - stream consumption thread interrupted" 
                                + (abortStreamConsumption 
                                   ? " to abort consumption" 
                                   : " for termination") 
                                + " whilst sleeping."); 
                        }
                        else 
                        {
                            // This thread was unexpectedly interrupted - 
                            // log it, drop it and carry on. 
                            logger.debug(
                                methodLogName 
                                + " - unexpected interrupt received whilst" 
                                + " waiting to poll stream for data."); 
                        }
                    } 
                }
            } 
            catch(IOException ioe) 
            {
                logger.error(methodLogName 
                             + " - error [" + ioe.getMessage() 
                             + "] whilst reading from input stream."); 

                good = false; 
            } 

            // Clean up after stream consumption. 
            reader = null; 
            inputStream = null; 
            isConsumingStream = false; 

            logger.debug(methodLogName 
                         + " - finished with input stream."); 

            // Get the consumed string and release it to other methods. 
            consumedString = consumedStringBuffer.toString(); 
            consumedStringPermit.release(); 
        }

        logger.debug(methodLogName 
                     + " - stream consumption thread terminated."); 
    }

    /** Returns true if no errors have occurred while this stream consumption 
     * thread was consuming characters from the current input stream. 
     */ 
    boolean good()
    {
        final String methodLogName = classLogName + "::good()"; 

        logger.trace(methodLogName); 

        return(good); 
    }

    /** Blocks until the current input stream has been consumed. If no input 
     * stream has been set yet, this method will block until a stream has been
     * set and stream consumption is complete. 
     */ 
    void awaitStreamConsumption() 
    {
        final String methodLogName = 
            classLogName + "::awaitStreamConsumption()"; 

        logger.trace(methodLogName); 

        // Wait for the consumed string buffer to become available. 
        consumedStringPermit.acquireUninterruptibly(); 

        // As a binary semaphore, the consumedStringPermit must never have 
        // any permits available at this stage. 
        if(consumedStringPermit.availablePermits() != 0)
        {
            final String message = 
                methodLogName + " - [consumedStringPermit]" 
                + " binary semaphore violation detected."; 

            logger.fatal(message); 
            throw new AssertionError(message); 
        }

        // Release the consumed string buffer. 
        consumedStringPermit.release(); 
    }

    /** Returns the full string consumed from the last input stream to be set 
     * (see {@link #consumeStream()}). If no input stream has been set yet, or
     * stream consumption is still in progress, this method will block until 
     * a stream has been set and stream consumption is complete. 
     * 
     * @return the full string consumed from the last input stream to be set. 
     */
    String getConsumedString() 
    {
        final String methodLogName = classLogName + "::getConsumedString()"; 

        logger.trace(methodLogName); 

        // Lock out any attempts to update the consumed string buffer. 
        consumedStringPermit.acquireUninterruptibly(); 

        // As a binary semaphore, the consumedStringPermit must never have 
        // any permits available at this stage. 
        if(consumedStringPermit.availablePermits() != 0)
        {
            final String message = 
                methodLogName + " - [consumedStringPermit]" 
                + " binary semaphore violation detected."; 

            logger.fatal(message); 
            throw new AssertionError(message); 
        }

        // Get the consumed string. 
        final String result = consumedString; 

        // Release the consumed string buffer and return the consumed string. 
        consumedStringPermit.release(); 
        return(result); 
    }

    /** Aborts consumption of the current input stream, if any. Please 
     * note that this method does not wait for stream consumption to 
     * finish - if necessary, use the {@link #isConsumingStream()} and 
     * {@link #awaitStreamConsumption()} methods to ensure that this 
     * StreamConsumptionThread is ready for reuse. 
     */ 
    synchronized void abortStreamConsumption() 
    {
        final String methodLogName = 
            classLogName + "::abortStreamConsumption()"; 

        logger.trace(methodLogName); 

        if(isConsumingStream)
        {
            // Flag this thread to abort stream consumption. 
            abortStreamConsumption = true; 

            logger.debug(methodLogName + " - thread [" 
                         + getName() + "] aborting stream consumption."); 

            // Interrupt the stream consumption thread, causing it to abort 
            // consumption of the current input stream, if any. 
            interrupt(); 
        }
        else
        {
            logger.debug(methodLogName + " - thread [" 
                         + getName() + "]: no stream is being consumed."); 
        }
    }

    /** Terminates this StreamConsumptionThread. */ 
    synchronized void terminate() 
    {
        final String methodLogName = classLogName + "::terminate()"; 

        logger.trace(methodLogName); 

        // Flag this thread to terminate. 
        terminate = true; 

        logger.debug(methodLogName + " - terminating; interrupting thread [" 
                     + getName() + "]."); 

        // Interrupt the stream consumption thread, causing it to terminate. 
        interrupt(); 
    }
}
