package org.eso.util.misc; 

import org.apache.log4j.Logger;

/** <p>This class represents an immutable string of characters which have 
 * been wrapped to match a given set of formatting parameters. It is often 
 * useful for printing to standard output, or other places where text 
 * should match a particular screen width. One can also specify a string 
 * representing a left margin - for example, an indent or a comment string. 
 * The standard {@link #toString()} method is provided.</p> 
 *
 * <p>NOTE: consider calling {@link java.lang.String#trim() String.trim()} 
 * on the original string, then constructing the WrappedString with the 
 * result. This will help to avoid unsightly blank lines at the start or 
 * end of the WrappedString.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
public final class WrappedString 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(WrappedString.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "WrappedString"; 

    /** A reasonable line width for wrapped text when printing to the 
     * standard output. 
     */ 
    public static final int STANDARD_OUTPUT_LINE_WIDTH = 78; 

    /** The standard length for ISO tabs: 8 ' ' (space) characters. */ 
    public static final int ISO_TAB_LENGTH = 8; 

    /** Characters which are considered to be whitespace in text. */ 
    public static final String WHITESPACE = " \n\r\t\f"; 

    /** Internal storage for the wrapped string. */ 
    private final String wrappedString; 

    /** Constructs a new WrappedString with no margin string, using 
     * the built-in formatting assumptions (see 
     * {@link #STANDARD_OUTPUT_LINE_WIDTH}, {@link #ISO_TAB_LENGTH} 
     * and {@link #WHITESPACE}). 
     *
     * @throws NullPointerException if sourceString is null. 
     */ 
    public WrappedString(final String sourceString) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::WrappedString(String)"; 

        logger.trace(methodLogName); 

        wrappedString = wrapText("", 
                                 sourceString, 
                                 STANDARD_OUTPUT_LINE_WIDTH, 
                                 ISO_TAB_LENGTH, 
                                 WHITESPACE); 
    }

    /** Constructs a new WrappedString using the specified left margin 
     * string and the built-in formatting assumptions (see 
     * {@link #STANDARD_OUTPUT_LINE_WIDTH}, {@link #ISO_TAB_LENGTH} 
     * and {@link #WHITESPACE}). 
     *
     * @throws NullPointerException if leftMarginString or sourceString 
     *                              is null. 
     * @throws IllegalArgumentException if {@link #STANDARD_OUTPUT_LINE_WIDTH} 
     *                                  is not longer than leftMarginString. 
     */ 
    public WrappedString(final String leftMarginString, 
                         final String sourceString) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::WrappedString(String, String)"; 

        logger.trace(methodLogName); 

        wrappedString = wrapText(leftMarginString, 
                                 sourceString, 
                                 STANDARD_OUTPUT_LINE_WIDTH, 
                                 ISO_TAB_LENGTH, 
                                 WHITESPACE); 
    }

    /** Constructs a new WrappedString using the specified left margin 
     * string and formatting assumptions (see 
     * {@link #STANDARD_OUTPUT_LINE_WIDTH}, {@link #ISO_TAB_LENGTH} 
     * and {@link #WHITESPACE} for the standard values). 
     *
     * @throws NullPointerException if leftMarginString, sourceString or  
     *                              whitespace is null. 
     * @throws IllegalArgumentException if lineWidth is not longer than 
     *                                  leftMarginString, or tabLength 
     *                                  is negative. 
     */ 
    public WrappedString(final String leftMarginString, 
                         final String sourceString, 
                         final int lineWidth, 
                         final int tabLength, 
                         final String whitespace) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName 
            + "::WrappedString(String, String, int, int, String)"; 

        logger.trace(methodLogName); 

        wrappedString = wrapText(leftMarginString, 
                                 sourceString, 
                                 lineWidth, 
                                 tabLength, 
                                 whitespace); 
    }

    /** Returns the contents of this WrappedString as a String. */ 
    public String toString()
    {
        final String methodLogName = classLogName + "::toString()"; 

        logger.trace(methodLogName); 

        return(wrappedString); 
    }

    /** Wraps sourceString to the given lineWidth, including leftMarginString 
     * at the left margin. Newlines and copies of leftMarginString will be 
     * inserted as required. Breaks are only made at whitespace characters. 
     * 
     * @param leftMarginString a string which should occupy the left margin 
     *                         of every line. 
     * @param sourceString the text to wrap. 
     * @param lineWidth the line width in characters. 
     * @param tabLength the tab length in characters. 
     * @param whitespace a string of characters which should be considered 
     *                   whitespace. 
     * 
     * @throws NullPointerException if leftMarginString, sourceString or  
     *                              whitespace is null. 
     * @throws IllegalArgumentException if lineWidth is not longer than 
     *                                  leftMarginString, or tabLength 
     *                                  is negative. 
     * 
     * @returns a String containing the justified text. 
     */ 
    private static String wrapText(final String leftMarginString, 
                                   final String sourceString, 
                                   int lineWidth, 
                                   final int tabLength, 
                                   final String whitespace)
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::wrapText()"; 

        logger.trace(methodLogName); 

        if((leftMarginString == null) 
           || (sourceString == null) 
           || (whitespace == null)) 
        {
            final String message = 
                methodLogName + " - leftMarginString [" + leftMarginString 
                + "], sourceString[" + sourceString
                + "] and whitespace [" + whitespace + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(!(lineWidth > leftMarginString.length())) 
        {
            final String message = 
                methodLogName + " - lineWidth [" + lineWidth 
                + "] must be longer than leftMarginString [" 
                + leftMarginString.length() + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        if(tabLength < 0) 
        {
            final String message = 
                methodLogName + " - tabLength [" + tabLength 
                + "] must not be negative."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        logger.debug(methodLogName 
                     + " - wrapping text to width [" + lineWidth + "]."); 

        // Offset the characterWidth to exclude margin text. 
        lineWidth -= leftMarginString.length(); 

        // The length of the free form text. 
        final int sourceStringLength = sourceString.length(); 

        // The new, justified text... 
        StringBuilder wrappedText = 
            new StringBuilder(sourceStringLength
                              + (sourceStringLength / lineWidth
                                 * (leftMarginString.length() + 1))); 

        // ... starting with the first copy of the margin text. 
        wrappedText.append(leftMarginString); 

        // The index of the starting character in the current line. 
        int lineStart = 0; 

        // The current line length, excluding margin text. 
        int lineLength = 0; 

        // The index of the last place a line break made sense. 
        int previousBreak = 0; 

        // True if a new line should be forced. 
        boolean forceNewLine = false; 

        // Loop through the free form text, inserting margin text and 
        // newlines as required. 
        for(int i = 0; i < sourceStringLength; ++i) 
        {
            if(whitespace.indexOf(sourceString.charAt(i)) > -1) 
            {
                // This character represents a new place to break. 
                previousBreak = i; 
            }

            // This character has been inspected - move on. 
            if(sourceString.charAt(i) == '\t')
            {
                // Tabs represent more than one character position. 
                lineLength += tabLength; 
            }
            else if(sourceString.charAt(i) == '\n') 
            {
                // An embedded newline has been found. Force a new line. 
                forceNewLine = true; 
            }
            else
            {
                // This is a normal character - advance by one. 
                ++lineLength; 
            }

            if(forceNewLine 
               || ((lineLength > lineWidth) && (previousBreak != lineStart)))
            {
                // A new line was forced, or it is time to start a new line. 
                // 
                // Add a new justified line using the characters up to the 
                // last sensible break. 
                wrappedText.append(
                    sourceString.substring(lineStart, previousBreak)); 

                // Add a newline and the next margin text. Note that 
                // the newline effectively replaces the break character. 
                wrappedText.append('\n'); 
                wrappedText.append(leftMarginString); 

                // The line now consists of any characters following the 
                // previous break. 
                lineLength = i - previousBreak; 

                // Drop the break character and update the line start. 
                ++previousBreak; 
                lineStart = previousBreak; 
                forceNewLine = false; 
            }
        }

        // Add the final characters of the string. 
        wrappedText.append(sourceString.substring(lineStart)); 

        return(wrappedText.toString()); 
    }
}
