package org.eso.util.requesthandler;

import java.io.Serializable;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.dialect.Dialect;
import org.hibernate.engine.SessionImplementor;
import org.hibernate.engine.TransactionHelper;
import org.hibernate.id.Configurable;
import org.hibernate.id.IdentifierGenerator;
import org.hibernate.id.TableHiLoGenerator;
import org.hibernate.type.Type;
import org.hibernate.util.PropertiesHelper;

/**
 * An <tt>IdentifierGenerator</tt> that uses a database table to store the
 * last generated value. It is not intended that applications use this strategy
 * directly. However, it may be used to build other (efficient) strategies. The
 * returned type is <tt>Integer</tt>.<br>
 * <br>
 * The hi value MUST be fetched in a seperate transaction to the
 * <tt>Session</tt> transaction so the generator must be able to obtain a new
 * connection and commit it. Hence this implementation may not be used when
 * Hibernate is fetching connections when the user is supplying connections.<br>
 * <br>
 * The returned value is of type <tt>integer</tt>.<br>
 * <br>
 * Mapping parameters supported: table, column
 * 
 * @see TableHiLoGenerator
 * @author Tim Canavan
 */

public class IdGenerator extends TransactionHelper implements
        IdentifierGenerator, Configurable {

    // initializes logger
    private final static Logger LOGGER = Logger.getLogger(IdGenerator.class);
    private final static String CLASS_LOG_NAME = "IdGenerator";

    private static final String DEFAULT_VALUE = "no";

    /* COLUMN and TABLE should be renamed but it would break the public API */
    /** The column parameter */
    public static final String COLUMN = "column";
    public static final String DEFAULT_COLUMN_NAME = "next_hi";
    public static final String TABLE = "table";
    public static final String DEFAULT_TABLE_NAME = "keys";

    protected String tableName;
    protected String keyName;
    protected String catalog;

    public void configure(Type type, Properties params, Dialect dialect) {
        this.tableName =
                PropertiesHelper.getString(TABLE, params, DEFAULT_TABLE_NAME);
        this.keyName =
                PropertiesHelper.getString("keyName", params, DEFAULT_VALUE);
        this.catalog =
                PropertiesHelper.getString("catalog", params, DEFAULT_VALUE);

    }

    /*
     * The generated key will be the concatenation of the back end user id
     * associated to the user portal id and of a key generated by a stored
     * procedure
     * 
     * @see org.hibernate.id.IdentifierGenerator#generate(org.hibernate.engine.SessionImplementor,
     *      java.lang.Object)
     */
    public synchronized Serializable generate(SessionImplementor session,
            Object object) throws HibernateException {
        String methodLogName = CLASS_LOG_NAME + "::generate()";

        Connection conn = session.getBatcher().openConnection();
        DatasetRequest dsRequest = (DatasetRequest) object;
        String id;
        if ((session == null) || (object == null)) {
            throw new IllegalArgumentException("Some input parameters are missing");
        }
        try {
            // the user id will be converted to a back end user id
            String userPortalIdToBackEndIdQuery =
                    " SELECT archive_user FROM user_portal..archive_users au, "
                            + "user_portal..accounts ac WHERE ac.account_id = "
                            + "au.account_id AND ac.username='"
                            + dsRequest.getUserId() + "'";
            String result[] =
                    (String[]) doWorkInCurrentTransaction(conn,
                            userPortalIdToBackEndIdQuery);
            // the identifier is a concatenation of the back end user id and of
            // the generated key
            id = result[0] + result[1];
            // The user id is changed to the back end one
            dsRequest.setUserId(result[0]);
        } catch (SQLException e) {
            LOGGER.error(methodLogName + " - SQLException : " + e.getMessage());
            throw new HibernateException(e);
        }
        return id;
    }

    public Object generatorKey() {
        return tableName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.hibernate.engine.TransactionHelper#doWorkInCurrentTransaction(java.sql.Connection,
     *      java.lang.String)
     */
    protected Serializable doWorkInCurrentTransaction(Connection conn,
            String sql) throws SQLException {

        String methodLogName =
                CLASS_LOG_NAME + "::doWorkInCurrentTransaction()";
        final String GET_KEY_STORED_PROCEDURE = "sp_get_key";
        String catalogName = catalog;

        if (catalogName.equals(DEFAULT_VALUE)) {
            catalogName = conn.getCatalog();
        }

        CallableStatement cstmt = null;
        ResultSet rs = null;
        // the first part of the result will be the result of the query given as
        // a parameter
        String resultQuery[] = new String[2];

        try {
            conn.setAutoCommit(true);
            // execution of the user query
            rs = conn.createStatement().executeQuery(sql);
            if (rs != null) {
                rs.next();
                String backEndUserId = rs.getString(1);
                resultQuery[0] = backEndUserId;
            } else {
                LOGGER.error(methodLogName
                        + " - no user portal id to back end id conversion possible");
                throw new SQLException("User id conversion not possible : ",
                        sql);
            }

            // execution of the stored procedure to get the second element of
            // the query
            String queryGetKey =
                    "{? = call " + catalogName + ".."
                            + GET_KEY_STORED_PROCEDURE + "(?)}";
            cstmt = conn.prepareCall(queryGetKey);
            cstmt.registerOutParameter(1, Types.INTEGER);
            cstmt.setString(2, keyName);
            cstmt.execute();
            cstmt.getMoreResults();
            cstmt.getMoreResults();
            rs = cstmt.getResultSet();
            if (rs != null) {
                rs.next();
                Integer id = new Integer(rs.getInt(1));
                resultQuery[1] = id.toString();
            } else {
                LOGGER.error(methodLogName + " - no ID returned");
                throw new SQLException("ID not generated: ", sql);
            }
        } finally {
            conn.close();
        }
        return resultQuery;
    }
}