package org.eso.util.stream; 

import java.util.concurrent.ScheduledThreadPoolExecutor; 
import java.util.concurrent.TimeUnit; 

import org.apache.log4j.Logger; 

import org.eso.util.misc.DefaultDaemonThreadFactory; 

/** <p>DelayTasks are used to insert arbitrary delays in the flow of messages 
 * within a stream. Messages will be held until either the default delay or 
 * the delay specified with the {@link put(Message, long, TimeUnit)} method 
 * has expired, as appropriate. Messages are never modified.</p> 
 *
 * <p>The next task must be set before using an instance of this class.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class DelayTask extends RoutingTask 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(DelayTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "DelayTask"; 

    /** A scheduled execution pool containing one thread to forward Messages 
     * as they become due. 
     */ 
    private static final ScheduledThreadPoolExecutor executor = 
        new ScheduledThreadPoolExecutor(1, new DefaultDaemonThreadFactory()); 

    /** The default length of time this DelayTask should wait before 
     * forwarding a given message. 
     */ 
    private long delay; 

    /** The unit of time that the default delay is specified in. */ 
    private TimeUnit unit; 

    /** Constructs a new DelayTask with the name "DelayTask". 
     *
     * @param delay the default length of time this DelayTask should wait 
     *              before forwarding a given message. 
     * @param unit the unit of time that the delay is specified in. 
     * 
     * @throws IllegalArgumentException if delay is negative. 
     * @throws NullPointerException if unit is null. 
     */ 
    public DelayTask(final long delay, final TimeUnit unit) 
        throws IllegalArgumentException, NullPointerException 
    {
        super("DelayTask");

        final String methodLogName = 
            classLogName + "::DelayTask() [for task DelayTask]"; 

        logger.trace(methodLogName); 

        if(unit == null) 
        {
            final String message = 
                methodLogName + " - unit must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(delay < 0)
        {
            final String message = 
                methodLogName 
                + " - delay [" + delay + "] must not be negative."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        this.delay = delay; 
        this.unit = unit; 
    }

    /** Constructs a new DelayTask with the specified name. 
     *
     * @param taskName the name of this task. 
     * @param delay the default length of time this DelayTask should wait 
     *              before forwarding a given message. 
     * @param unit the unit of time that the delay is specified in. 
     * 
     * @throws IllegalArgumentException if delay is negative. 
     * @throws NullPointerException if taskName or unit is null. 
     */
    public DelayTask(final String taskName, 
                     final long delay, 
                     final TimeUnit unit) 
        throws IllegalArgumentException, NullPointerException 
    {
        super(taskName);

        final String methodLogName = 
            classLogName + "::DelayTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 

        if(unit == null) 
        {
            final String message = 
                methodLogName + " - unit must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(delay < 0)
        {
            final String message = 
                methodLogName 
                + " - delay [" + delay + "] must not be negative."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        this.delay = delay; 
        this.unit = unit; 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get(final long timeout) 
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** {@inheritDoc} Once the default message forwarding delay has 
     * expired, the message will be forwarded to the next task using 
     * a {@link #putNext(Message message)} call. Note that this method 
     * does not block; messages are forwarded asynchronously. 
     */
    public final synchronized void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put(Message) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // Delegate to put(Message, long, TimeUnit). 
        put(message, delay, unit); 
    }

    /** Sends a message to this task. Once the specified number of time 
     * units have passed, the message will be forwarded to the next task 
     * using a {@link #putNext(Message message)} call. Note that this 
     * method does not block; messages are forwarded asynchronously. 
     *
     * @param message a message for this task. 
     * @param delay the length of time this DelayTask should wait before 
     *              forwarding the message. 
     * @param unit the unit of time that the delay is specified in. 
     * 
     * @throws IllegalArgumentException if delay is negative. 
     * @throws NullPointerException if unit is null. 
     */ 
    public final void put(final Message message, 
                          final long delay, 
                          final TimeUnit unit) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::put(Message, long, TimeUnit) [for task " 
            + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        if(unit == null) 
        {
            final String logMessage = 
                methodLogName + " - unit must not be null."; 

            logger.fatal(logMessage); 
            throw new NullPointerException(logMessage); 
        }

        if(delay < 0)
        {
            final String logMessage = 
                methodLogName 
                + " - delay [" + delay + "] must not be negative."; 

            logger.fatal(logMessage); 
            throw new IllegalArgumentException(logMessage); 
        }

        // Schedule the message for forwarding. 
        executor.schedule(new MessageForwarder(message), delay, unit); 

        logger.debug(methodLogName 
                     + " - scheduled message for forwarding in [" 
                     + delay + "][" + unit + "]."); 
    }

    // Sends a message to the task following this task in a stream. 
    // 
    // Documentation is inherited from the Task interface. 
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - forwarding message to next task."); 

        getNextTask().put(message); 
    }

    /** Returns the default delay after which a message will be forwarded 
     * to the next task. 
     */ 
    public final long getDefaultDelay() 
    {
        final String methodLogName = 
            classLogName 
            + "::getDefaultDelay() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(delay); 
    }

    /** Returns the unit of time that the default message forwarding delay 
     * is specified in. 
     */ 
    public final TimeUnit getDefaultUnit() 
    {
        final String methodLogName = 
            classLogName 
            + "::getDefaultUnit() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(unit); 
    }

    /** Sets the default delay after which new messages will be forwarded 
     * to the next task. Messages which have already been submitted to this 
     * DelayTask are not affected. 
     * 
     * @param delay the default length of time this DelayTask should wait 
     *              before forwarding a given message. 
     * @param unit the unit of time that the delay is specified in. 
     * 
     * @throws IllegalArgumentException if delay is negative. 
     * @throws NullPointerException if unit is null. 
     */ 
    public final synchronized void setDefaultDelay(final long delay, 
                                                   final TimeUnit unit) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName 
            + "::setDefaultDelay() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        if(unit == null) 
        {
            final String logMessage = 
                methodLogName + " - unit must not be null."; 

            logger.fatal(logMessage); 
            throw new NullPointerException(logMessage); 
        }

        if(delay < 0)
        {
            final String logMessage = 
                methodLogName 
                + " - delay [" + delay + "] must not be negative."; 

            logger.fatal(logMessage); 
            throw new IllegalArgumentException(logMessage); 
        }

        this.delay = delay; 
        this.unit = unit; 
    }

    /** Instances of this class contain single {@link Message}s which 
     * are forwarded to the next task in a stream when the {@link #run()} 
     * method is called. 
     * 
     * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
     */ 
    private class MessageForwarder implements Runnable 
    {
        /** The name of this class for logging purposes. */ 
        private final String classLogName = "MessageForwarder"; 

        /** The {@link Message} this MessageForwarder will forward when 
         * the {@link #run()} method is called. 
         */ 
        private final Message message; 

        /** Constructs a new MessageForwarder to forward the specified 
         * {@link Message}. 
         * 
         * @param message the {@link Message} to forward when {@link #run()} 
         *                is called. 
         * 
         * @throws NullPointerException if message is null. 
         */
        MessageForwarder(final Message message) throws NullPointerException 
        {
            final String methodLogName = 
                classLogName 
                + "::MessageForwarder() [for task " + getTaskName() + "]"; 

            logger.trace(methodLogName); 

            if(message == null) 
            {
                final String logMessage = 
                    methodLogName + " - message must not be null."; 

                logger.fatal(logMessage); 
                throw new NullPointerException(logMessage); 
            }

            // Set the message to forward. 
            this.message = message; 
        }

        /** Forwards the {@link Message} held by this MessageForwarder to 
         * the next task in the stream. 
         */
        public void run() 
        {
            final String methodLogName = 
                classLogName + "::run() [for task " + getTaskName() + "]"; 

            logger.trace(methodLogName); 

            putNext(message); 
        }
    }
}
