package org.eso.util.stream; 

import org.apache.log4j.Logger;

/** <p>PassThroughTasks act as placeholders in a stream - all messages 
 * received are immediately forwarded to the next task. Messages are never 
 * held or modified.</p> 
 *
 * <p>The next task must be set before using an instance of this class.</p> 
 *
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class PassThroughTask extends RoutingTask 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(PassThroughTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "PassThroughTask"; 

    /** Constructs a new PassThroughTask with the name "PassThroughTask". */ 
    public PassThroughTask()
    {
        super("PassThroughTask");

        final String methodLogName = 
            classLogName + "::PassThroughTask() [for task PassThroughTask]"; 

        logger.trace(methodLogName); 
    }

    /** Constructs a new PassThroughTask with the specified name. 
     *
     * @param taskName the name of this task. 
     */
    public PassThroughTask(final String taskName) 
    {
        super(taskName);

        final String methodLogName = 
            classLogName 
            + "::PassThroughTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get(final long timeout) 
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** {@inheritDoc} Messages are immediately forwarded to the next task 
     * using a {@link #putNext(Message message)} call. 
     */
    public final void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // Delegate to putNext(). 
        putNext(message); 
    }

    // Sends a message to the task following this task in a stream. 
    // 
    // Documentation is inherited from the Task interface. 
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - forwarding message to next task."); 

        getNextTask().put(message); 
    }
}
