package org.eso.util.stream; 

import java.util.Comparator; 
import java.util.List; 
import java.util.LinkedList; 

import java.util.concurrent.BlockingQueue; 
import java.util.concurrent.LinkedBlockingQueue; 
import java.util.concurrent.PriorityBlockingQueue; 
import java.util.concurrent.Semaphore; 
import java.util.concurrent.TimeUnit; 

import org.apache.log4j.Logger; 

/** <p>ProcessingTask is an abstract class providing common infrastructure 
 * for subclasses which need to process the content of messages within a
 * stream. Tasks which only route messages, rather than processing them, 
 * should normally subclass the {@link RoutingTask} class, not this one.</p> 
 * 
 * <p>Every ProcessingTask instance has one or more dedicated processing 
 * threads. Functionality required to deal with things like message queues 
 * and stream commands is also provided, with support for fixed-period 
 * blocking whilst waiting for messages.</p> 
 * 
 * <p>Subclasses need only implement two methods: 
 *   <ul>
 *     <li>{@link #canProcessMessage(Message message)} to filter the message 
 *         stream for the messages they are interested in.</li> 
 *     <li>{@link #process(List msgList)} to process the selected messages. 
 *         </li> 
 *   </ul>
 * </p> 
 *
 * <p>This class offers two types of message queue: 
 *   <ul>
 *     <li>a strict FIFO queue.</li> 
 *     <li>a priority queue, where higher priority messages can overtake 
 *         lower priority ones.</li> 
 *   </ul>
 * Both types of queue implement full locking for concurrent operation.</p> 
 *
 * <p>The next task must be set before using an instance of this class.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public abstract class ProcessingTask implements Task, Runnable 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(ProcessingTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "ProcessingTask"; 

    /** The module containing this task (if any). */
    private Module containingModule = null; 

    /** The name of this task. */ 
    private final String taskName; 

    /** The number of threads used to run this task. */ 
    private final int numTaskThreads; 

    /** An array of threads currently running this task. */ 
    private final Thread[] taskThreads; 

    /** This is used by {@link #open()}, {@link #run()} and {@link #close()} 
     * to track the number of task threads which are running. It is also used 
     * by {@link #isOpen()} - if any threads are running, the task is open. 
     */ 
    private final Semaphore threadPermits = new Semaphore(0); 

    /** True if task closure has been enabled; see {@link #close()}. */ 
    private boolean taskClosureEnabled = false; 

    /** A queue of incoming messages to be processed by this task. */ 
    private final BlockingQueue<Message> msgQueue; 

    /** The maximum period {@link #run()} should block for whilst waiting 
     * for new messages. Specified in seconds. If the period is set to zero, 
     * run() will block forever until a new message arrives. If the period 
     * is negative, run() will not block at all; the {@link process()} 
     * method will be executed as often as possible (this will consume 
     * a lot of CPU time) even if no messages have arrived. 
     */
    private final long maxBlockPeriod; 

    /** The task which should receive any messages forwarded by this task. */ 
    private Task nextTask = null; 

    /** Constructs a new ProcessingTask with a strict FIFO message queue. 
     * 
     * @throws IllegalArgumentException if numTaskThreads is less than 1. 
     * @throws NullPointerException if taskName is null. 
     *
     * @param taskName the name of this task. 
     * @param numTaskThreads the number of threads used to run this task. 
     * @param maxBlockPeriod maximum period {@link #run()} should block for 
     *            whilst waiting for new messages. Specified in seconds. If 
     *            the period is set to zero, run() will block forever until 
     *            a new message arrives. If the period is negative, run() 
     *            will not block at all; the {@link process()} method will 
     *            be executed as often as possible (this will consume a lot 
     *            of CPU time) even if no messages have arrived. 
     */
    public ProcessingTask(final String taskName, 
                          final int numTaskThreads, 
                          final long maxBlockPeriod) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::ProcessingTask() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        if(taskName == null) 
        {
            final String message = 
                methodLogName + " - taskName must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(numTaskThreads < 1) 
        {
            final String message = 
                methodLogName + " - numTaskThreads must not be less than 1."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Store the task parameters. 
        this.taskName = taskName; 
        this.numTaskThreads = numTaskThreads; 
        this.maxBlockPeriod = maxBlockPeriod; 

        // Initialise the thread tracking semaphore by releasing enough 
        // permits for the specified number of threads. 
        threadPermits.release(numTaskThreads); 

        // Create the task thread array. 
        this.taskThreads = new Thread[numTaskThreads]; 

        // This task will use a LinkedBlockingQueue, which is strictly FIFO. 
        this.msgQueue = new LinkedBlockingQueue<Message>(); 
    }

    /** Constructs a new ProcessingTask with a prioritising message queue. 
     * 
     * @throws IllegalArgumentException if numTaskThreads is less than 1. 
     * @throws NullPointerException if taskName or messagePriorityComparator 
     *                              is null. 
     *
     * @param taskName the name of this task. 
     * @param numTaskThreads the number of threads used to run this task. 
     * @param maxBlockPeriod maximum period {@link #run()} should block for 
     *            whilst waiting for new messages. Specified in seconds. If 
     *            the period is set to zero, run() will block forever until 
     *            a new message arrives. If the period is negative, run() 
     *            will not block at all; the {@link process()} method will 
     *            be executed as often as possible (this will consume a lot 
     *            of CPU time) even if no messages have arrived. 
     * @param messagePriorityComparator a comparator for the message queue to 
     *                                  use when prioritising messages. 
     *
     * @see {@link java.util.Comparator} 
     */
    public ProcessingTask(final String taskName, 
                          final int numTaskThreads, 
                          final long maxBlockPeriod, 
                          final Comparator<Message> messagePriorityComparator)
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::ProcessingTask() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        if((taskName == null) || (messagePriorityComparator == null)) 
        {
            final String message = 
                methodLogName + " - taskName [" 
                + taskName + "] and messagePriorityComparator [" 
                + messagePriorityComparator + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(numTaskThreads < 1) 
        {
            final String message = 
                methodLogName + " - numTaskThreads must not be less than 1."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Store the task parameters. 
        this.taskName = taskName; 
        this.numTaskThreads = numTaskThreads; 
        this.maxBlockPeriod = maxBlockPeriod; 

        // Initialise the thread tracking semaphore by releasing enough 
        // permits for the specified number of threads. 
        threadPermits.release(numTaskThreads); 

        // Create the task thread array. 
        this.taskThreads = new Thread[numTaskThreads]; 

        // This task will use a PriorityBlockingQueue, which prioritises 
        // its messages. The initial queue capacity is 11 elements. 
        this.msgQueue = 
            new PriorityBlockingQueue<Message>(11, messagePriorityComparator);
    }

    /** <p>Creates one or more new threads and runs this ProcessingTask's 
     * {@link #run()} method in them. This method blocks until all threads 
     * are running.</p> 
     *
     * <p>NOTE: 
     *   <ul>
     *     <li>Calling this method from a task thread has no effect.</li>
     *     <li>Threads may start in any order. If multiple threads are 
     *         in use, there will be a window of time where some threads 
     *         are running this task already, but not all threads have 
     *         been started yet.</li>
     *   </ul>
     * </p> 
     */
    public final synchronized void open() 
    {
        final String methodLogName = 
            classLogName + "::open() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 
        logger.debug(methodLogName + " - opening task with " + numTaskThreads 
                     + (numTaskThreads == 1 ? " thread." : " threads.")); 

        if(!isOpen()) 
        {
            // No task threads are running. 

            // The future task threads are not allowed to exit until further 
            // notice (this catches unexpected interrupts). 
            logger.debug(methodLogName + " - disabling task closure."); 
            taskClosureEnabled = false; 

            // A name for the thread currently being created. 
            String threadName = null; 

            // Create and start the threads. 
            for(int i = 0; i < numTaskThreads; ++i)
            {
                threadName = 
                    taskName + (numTaskThreads == 1 ? "" : "[" + i + "]"); 

                logger.debug(methodLogName + " - starting task thread [" 
                             + threadName + "]."); 

                taskThreads[i] = new Thread(this, threadName); 
                taskThreads[i].start(); 
            }

            // True if this thread has been interrupted. 
            boolean interruptDetected = Thread.interrupted(); 

            // Wait for all threads to start running. 
            synchronized(threadPermits) 
            {
                while(threadPermits.availablePermits() != 0) 
                {
                    try
                    {
                        threadPermits.wait(); 
                    }
                    catch(InterruptedException ie)
                    {
                        // This thread was interrupted whilst waiting for all 
                        // task threads to exit - note it and keep waiting. 
                        logger.debug(methodLogName 
                                     + " - interrupt received whilst waiting" 
                                     + " for all task threads to exit."); 

                        interruptDetected = true; 
                    }
                }
            }

            // If an interrupt was detected, set this thread's interrupt 
            // status. 
            if(interruptDetected) 
            {
                Thread.currentThread().interrupt(); 
            }

            logger.debug(methodLogName + " - task open."); 
        }
        else
        {
            logger.debug(methodLogName + " - task already open."); 
        }
    }

    // Returns true if this task is open. 
    // 
    // Documentation is inherited from the Task interface. 
    public final synchronized boolean isOpen() 
    {
        final String methodLogName = 
            classLogName + "::isOpen() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        return(threadPermits.availablePermits() != numTaskThreads); 
    }

    /** <p>Interrupts the threads running this task, causing the task to 
     * stop executing. This method blocks until the task has closed.</p> 
     * 
     * <p>NOTE: 
     *   <ul>
     *     <li><p>If this method is called from a task thread, it will only 
     *         block until the other task threads have terminated. This 
     *         allows the calling thread to terminate in turn.</p> 
     *         <p><b>WARNING:</b> In this case, the {@link #open()} method 
     *         will not be able to reopen this task until the calling task 
     *         thread has terminated.</p></li> 
     *     <li>Threads may terminate in any order. If multiple threads are 
     *         in use, there will be a window of time where some threads 
     *         have already terminated, but others are still performing 
     *         their last processing cycles.</li>
     *   </ul> 
     * </p> 
     */ 
    public final synchronized void close() 
    {
        final String methodLogName = 
            classLogName + "::close() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 
        logger.debug(methodLogName + " - closing task."); 

        if(isOpen()) 
        {
            // Inform the task threads that they are allowed to exit. 
            logger.debug(methodLogName + " - enabling task closure."); 
            taskClosureEnabled = true; 

            // True if the calling thread is a task thread. 
            boolean callerIsTaskThread = false; 

            // Interrupt the task threads, so that they can exit. 
            logger.debug(methodLogName + " - interrupting task threads.");

            final Thread currentThread = Thread.currentThread(); 

            for(int i = 0; i < numTaskThreads; ++i) 
            {
                if(currentThread == taskThreads[i]) 
                {
                    callerIsTaskThread = true; 
                }

                taskThreads[i].interrupt(); 
            }

            // True if this thread has been interrupted. 
            boolean interruptDetected = Thread.interrupted(); 

            // The number of task thread permits which must become available 
            // before task closure will be considered complete. If close() 
            // was called by a task thread, that thread's permit is exempt, 
            // so that it can continue out of this method and exit in turn. 
            int requiredPermits = 
                (callerIsTaskThread ? numTaskThreads - 1 : numTaskThreads); 

            // Wait for the (other) task threads to relinquish their permits. 
            synchronized(threadPermits)
            {
                while(threadPermits.availablePermits() != requiredPermits) 
                {
                    try
                    {
                        threadPermits.wait(); 
                    }
                    catch(InterruptedException ie)
                    {
                        // This thread was interrupted whilst waiting for 
                        // the (other) task threads to exit - flag it and 
                        // keep waiting. 
                        logger.debug(methodLogName 
                                     + " - interrupt received whilst waiting" 
                                     + " for all task threads to exit."); 

                        interruptDetected = true; 
                    }
                }
            }

            // If an interrupt was detected, set this thread's interrupt 
            // status. 
            if(interruptDetected) 
            {
                currentThread.interrupt(); 
            }

            logger.debug(methodLogName + 
                         (callerIsTaskThread ? 
                          " - task closing." : " - task closed.")); 
        }
        else
        {
            logger.debug(methodLogName + " - task already closed."); 
        }
    }

    /** <p>Restarts this task by calling close(), then open(). This method 
     * blocks until the restart is complete.</p> 
     *
     * <p><b>WARNING:</b> It is impossible for this method to work if it 
     * is called by a task thread, because that thread would be closed as 
     * part of the restart, leaving no thread to reopen the task.</p> 
     * 
     * @throws AssertionError if called by a task thread. 
     */
    public final synchronized void restart()
    {
        final String methodLogName = 
            classLogName + "::restart() [for task " + taskName + "]"; 

        logger.trace(methodLogName + ".");

        // True if the calling thread is a task thread. 
        boolean callerIsTaskThread = false; 

        // Check whether the caller is a task thread. 
        for(int i = 0; i < numTaskThreads; ++i) 
        {
            if(Thread.currentThread() == taskThreads[i]) 
            {
                callerIsTaskThread = true; 
            }
        } 

        if(callerIsTaskThread) 
        {
            final String logMessage = 
                methodLogName 
                + " - this method must not be called from a task thread."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // Restart the task. 
        logger.debug(methodLogName + " - restarting task."); 

        close(); 
        open(); 
    }

    /** Generic service method run by each task thread. This method consists 
     * of a loop performing the following actions: 
     * <ul>
     *   <li>receive new messages.</li>
     *   <li>for messages representing relevant stream commands, such as the 
     *       {@link Message#MSG_SHUTDOWN MSG_SHUTDOWN} message type, perform 
     *       the command.</li> 
     *   </li>
     *   <li>for all messages: 
     *     <ul>
     *       <li>run {@link #process()} on messages for which 
     *           {@link #canProcessMessage()} returns true.</li> 
     *       <li>forward processed messages to the next task in the stream. 
     *           Note that {@link #process()} is permitted to generate, 
     *           modify, store and delete messages as it sees fit.</li> 
     *       <li>forward messages for which {@link #canProcessMessage()} 
     *           returns false to the next task in the stream.</li> 
     *     </ul>
     *   </li>
     *   <li>if this thread has been interrupted, exit the loop and 
     *       terminate the thread, otherwise continue.</li> 
     * </ul> 
     * 
     * @throws IllegalStateException if an unexpected interrupt is detected. 
     */ 
    public final void run() throws IllegalStateException 
    {
        final String methodLogName = 
            classLogName + "::run() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 
        logger.debug(methodLogName + " - task thread started."); 

        // The message currently being considered. 
        Message message = null;

        // A list of messages to process. 
        final LinkedList<Message> msgList = new LinkedList<Message>(); 

        // Acquire a task thread permit for this thread. 
        synchronized(threadPermits)
        {
            if(!threadPermits.tryAcquire()) 
            {
                // No thread permits are available - this is a bug. 
                final String logMessage = 
                    methodLogName + " - unable to acquire thread permit.";

                logger.fatal(logMessage); 
                throw new AssertionError(logMessage); 
            }

            // Tell the open() method call which created this thread that 
            // the thread is running. 
            threadPermits.notify(); 
        }

        // Task thread loop. 
        while(true)
        {
            // The message list must be empty at the start of the task thread 
            // loop. 
            if(!msgList.isEmpty()) 
            {
                final String logMessage = 
                    methodLogName + " - message list not empty [" 
                    + msgList.size() 
                    + " elements] at start of task thread loop."; 

                logger.fatal(logMessage); 
                throw new AssertionError(logMessage); 
            }

            // Wait for new messages to arrive; time out if necessary. 
            logger.debug(methodLogName + " - waiting on message queue."); 

            message = get(maxBlockPeriod); 

            if(message == null) 
            {
                // The get() call timed out whilst waiting for a message; 
                // leave the message queue empty and continue. 
                logger.debug(methodLogName + " - no messages available."); 
            }
            else
            {
                // A message was returned from the queue. 
                if(message.getMessageType() == Message.MSG_SHUTDOWN)
                {
                    // This message is a stream shutdown command; deal with 
                    // it here. 
                    logger.debug(methodLogName 
                                 + " - stream shutdown command received;" 
                                 + " initiating task closure."); 
                    
                    // Close this task; threads will exit when they reach 
                    // the end of this loop. 
                    close(); 
                }

                if(canProcessMessage(message)) 
                {
                    // This message should be processed in some task-specific 
                    // way - put it in the message list. 
                    logger.debug(
                        methodLogName 
                        + " - message received for this task to process."); 

                    msgList.add(message); 

                    // This method is no longer responsible for the message. 
                    message = null; 
                }
            }

            // Call the task-specific processing method. 
            logger.debug(methodLogName 
                         + " - starting task-specific processing."); 

            process(msgList); 

            // Task-specific processing complete. 
            logger.debug(methodLogName 
                         + " - task-specific processing complete."); 

            // Forward any messages in the message list to the next task. 
            if(!msgList.isEmpty()) 
            {
                logger.debug(
                    methodLogName 
                    + " - forwarding processed messages to next task."); 
            }

            while(!msgList.isEmpty()) 
            {
                // Forward the first message in the list. 
                putNext(msgList.getFirst()); 

                // Remove the newly forwarded message. 
                msgList.removeFirst(); 
            } 

            // Ensure that the last message received (if any) has been dealt 
            // with appropriately. 
            if(message != null) 
            {
                // A message is pending, but it did not require task-specific 
                // processing, so it was not forwarded from the message list. 
                // 
                // Forward the message to the next task now. 
                logger.debug(
                    methodLogName 
                    + " - forwarding unprocessed message to next task."); 

                putNext(message); 
            }

            // Check for interrupts. 
            if(Thread.interrupted())
            {
                if(taskClosureEnabled) 
                {
                    // This thread was interrupted as part of a task closure; 
                    // clean up and exit normally. 
                    logger.debug(
                        methodLogName + " - task thread interrupted;" 
                        + " task closure in progress. Terminating now."); 

                    // Release this thread's task thread permit. 
                    synchronized(threadPermits)
                    {
                        threadPermits.release(); 

                        if(threadPermits.availablePermits() > numTaskThreads) 
                        {
                            // Too many thread permits have been released - 
                            // this is a bug. 
                            final String logMessage = 
                                methodLogName 
                                + " - too many thread permits released."; 

                            logger.fatal(logMessage); 
                            throw new AssertionError(logMessage); 
                        }

                        // Tell the close() method call which enabled this 
                        // task closure that this thread is finished. 
                        threadPermits.notify(); 
                    }

                    // Break out of the task thread loop and exit. 
                    break; 
                }
                else
                {
                    // This thread was interrupted, but no task closure 
                    // is in progress. 
                    final String logMessage = 
                        methodLogName 
                        + " - task thread received an unexpected interrupt."; 

                    logger.fatal(logMessage); 
                    throw new IllegalStateException(logMessage); 
                }
            }
        }
    }

    /** {@inheritDoc} Delegates to {@link #get(long timeout)} with the timeout 
     * set to zero. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        // Delegate to get(long). 
        return(get(0)); 
    }

    /** Removes the next message from this task's message queue and returns 
     * it. If this thread is interrupted whilst waiting for a message, its 
     * interrupt status will be set and null will be returned. 
     *
     * @param timeout the time this method should wait for a message to arrive 
     *                on the message queue. Specified in seconds. Set to zero 
     *                if this method should block until a message arrives. 
     *
     * @return the next message from this task's message queue, or null if the 
     *         queue was empty and the timeout has expired, or null if this 
     *         thread was interrupted whilst waiting for a message. 
     */
    public final Message get(final long timeout)
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + taskName + "]"; 

        logger.trace(methodLogName + ".");

        Message message = null;         // A newly received message. 

        try
        {
            if(timeout == 0) 
            {
                // Block until a message arrives. 
                message = msgQueue.take(); 
            }
            else
            {
                // Block until a message arrives or the timeout expires. 
                message = msgQueue.poll(timeout, TimeUnit.SECONDS); 
            }
        }
        catch(InterruptedException ie)
        {
            // This thread was interrupted whilst waiting for a message; 
            // set its interrupt status. 
            logger.debug(
                methodLogName 
                + " - interrupt received whilst waiting for message."); 
            
            Thread.currentThread().interrupt(); 
        }

        if(message == null)
        {
            logger.debug(methodLogName + " - no message retrieved."); 
        }
        else
        {
            logger.debug(methodLogName + " - message retrieved."); 
        }

        return(message); 
    }

    /** <p>Sends a message to this task. If the message is relevant to this 
     * task, or the message does not permit fast forwarding, it will be 
     * queued for this task. Otherwise, it will be forwarded directly to 
     * the next task.</p> 
     * 
     * <p>This method will not return until the message has been queued for 
     * this task or accepted by a subsequent task. Thread interrupts will 
     * be noted, but ignored until the message has been dealt with. If an 
     * interrupt occurs, this thread's interrupt status will be set before 
     * this method returns.</p> 
     *
     * @param message a message for this task. 
     */ 
    public final void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        // Check whether the message should be queued for this task. 
        if(canProcessMessage(message) || !message.permitsFastForwarding()) 
        {
            // The message is relevant to this task, or it does not permit 
            // fast forwarding. Add it to the queue. 

            // True if this thread has been interrupted. 
            boolean interruptDetected = Thread.interrupted(); 

            // The message must be successfully queued. 
            while(true)
            {
                try
                {
                    // Try to queue the message. 
                    msgQueue.put(message); 

                    logger.debug(
                        methodLogName + " - added message to queue."); 
                }
                catch(InterruptedException ie)
                {
                    // This thread was interrupted whilst waiting to queue 
                    // the message - note the interrupt and try again. 
                    logger.debug(methodLogName 
                                 + " - interrupt received whilst waiting" 
                                 + " to queue a message."); 

                    interruptDetected = true; 
                    continue; 
                }

                // The message was successfully queued. 
                break; 
            }

            // If an interrupt was detected, set this thread's interrupt 
            // status. 
            if(interruptDetected) 
            {
                Thread.currentThread().interrupt(); 
            }
        }
        else 
        {
            // The message is not relevant to this task, and it permits fast 
            // forwarding; forward it to the next task in the stream. 
            // 
            // Delegate to putNext(). 
            putNext(message); 
        }
    }

    // Sends a message to the task following this task in a stream. 
    // 
    // Documentation is inherited from the Task interface. 
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + taskName + "]"; 

        logger.trace(methodLogName + ".");
        logger.debug(methodLogName + " - forwarding message to next task."); 

        nextTask.put(message); 
    }

    // Returns the name of this task. 
    // 
    // Documentation is inherited from the Task interface. 
    public final String getTaskName() 
    {
        final String methodLogName = 
            classLogName + "::getTaskName() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        return(taskName); 
    }

    /** Returns the number of threads used to run this task. 
     * 
     * @return the number of threads an {@link #open()} call will create to 
     *         run this task. 
     */ 
    public final int getNumTaskThreads()
    {
        final String methodLogName = 
            classLogName 
            + "::getNumTaskThreads() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        return(numTaskThreads); 
    }

    /** Returns the maximum blocking period of this task when waiting for a 
     * message to arrive. 
     *
     * @return the maximum period {@link #run()} should block for whilst 
     *         waiting for new messages. Specified in seconds. If the period 
     *         is zero, run() will block forever until a new message arrives. 
     *         If the period is negative, run() will not block at all; the 
     *         {@link process()} method will be executed as often as possible 
     *         (this will consume a lot of CPU time) even if no messages have 
     *         arrived. 
     */
    public final long getMaxBlockPeriod() 
    {
        final String methodLogName = 
            classLogName 
            + "::getMaxBlockPeriod() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        return(maxBlockPeriod); 
    }

    // Returns a reference to the task following this one in a stream. 
    // 
    // Documentation is inherited from the Task interface. 
    public final Task getNextTask() 
    {
        final String methodLogName = 
            classLogName + "::getNextTask() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        return(nextTask); 
    }
    
    /** {@inheritDoc} This method may not be called while this task is open. 
     *
     * @throws AssertionError if called while this task is open. 
     */ 
    public final Task setNextTask(final Task task) 
    {
        final String methodLogName = 
            classLogName + "::setNextTask() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        // Ensure that the task is closed. 
        if(isOpen()) 
        {
            final String logMessage = 
                methodLogName 
                + " - task must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // Store the old next task to return, then set the new next task. 
        final Task oldTask = nextTask; 
        nextTask = task; 

        logger.debug(methodLogName + " - next task set [previously: "
                     + (oldTask == null 
                        ? "null][now: " : oldTask.getTaskName() + "][now: ") 
                     + (nextTask == null 
                        ? "null]." : nextTask.getTaskName() + "].")); 

        return(oldTask); 
    }

    // Returns a reference to the module containing this task. 
    // 
    // Documentation is inherited from the Task interface. 
    public final Module getContainingModule()
    {
        final String methodLogName = 
            classLogName 
            + "::getContainingModule() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        return(containingModule); 
    }

    /** {@inheritDoc} This method may not be called while this task is open. 
     *
     * @throws AssertionError if called while this task is open. 
     */ 
    public final Module setContainingModule(final Module module) 
    {
        final String methodLogName = 
            classLogName 
            + "::setContainingModule() [for task " + taskName + "]"; 

        logger.trace(methodLogName + "."); 

        // Ensure that the task is closed. 
        if(isOpen()) 
        {
            final String logMessage = 
                methodLogName 
                + " - task must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        // Store the old containing module to return, then set the new 
        // containing module. 
        final Module oldModule = containingModule; 
        containingModule = module; 

        logger.debug(methodLogName + " - containing module set."); 

        return(oldModule); 
    }

    /** <p>Task-specific method, called by the {@link #put()} and 
     * {@link #run()} methods to filter messages.</p> 
     *
     * <p><b>WARNING:</b> This method must not change the state of the 
     * message being considered, if doing so would cause another call of 
     * this method on the updated message to return a different result.</p> 
     *
     * @param message The message to consider. 
     *
     * @return true if this task can process the message, or false if the 
     *         message should be passed to another task. 
     */
    protected abstract boolean canProcessMessage(final Message message); 

    /** <p>Task-specific processing method, called by {@link #run()}.</p> 
     *
     * <p>NOTE: the process() method is not permitted to throw checked 
     * exceptions; these must be handled internally.</p> 
     *
     * <p><b>WARNING:</b> The process() method is solely responsible for all 
     * messages passed to it, including system messages. If a message should 
     * be forwarded to the next task, this method must return it in the 
     * message list or explicitly forward it. System messages almost always 
     * need to be forwarded.</p> 
     *
     * @param msgList a list of zero or more messages for this task to 
     *                process. The process() method may modify the list as 
     *                it sees fit. Any messages found in the list after 
     *                process() returns will be forwarded to the next task 
     *                in the stream. 
     */
    protected abstract void process(final List<Message> msgList); 
}
