package org.eso.util.stream; 

import java.util.LinkedList; 
import java.util.Iterator; 

import org.apache.log4j.Logger;

/** <p>The Stream class is the highest-level element of the stream package. 
 * When fully configured, a Stream instance holds one or more {@link Module} 
 * objects, which in turn hold two {@link Task} instances each. The stream 
 * model used in this package was inspired by the ACE Stream framework (as 
 * implemented in C++).</p> 
 *
 * <p>A stream contains a set of modules, modelled as a stack - only the 
 * head module is directly visible. Other modules are connected to it in the 
 * same sequence as they appear in the stack. Each module has a downstream 
 * task and an upstream task, with downstream being defined as the direction 
 * pointing away from the head. The downstream tasks are connected to each 
 * other in the same order as their containing modules, and the upstream 
 * tasks are connected to each other in the reverse order.</p> 
 * 
 * <p>In normal use, a stream-based application sends messages to the first 
 * downstream task, which is the one held by the head module. That task 
 * can process the message and/or route it to the next task downstream. Any 
 * messages that should be sent back to the application are transferred to 
 * an upstream task - each downstream task can discover the corresponding 
 * upstream task of its module. Once on the upstream side, messages can 
 * again be processed and/or routed to the next task upstream. Finally, 
 * messages arriving at the upstream task of the head module can be held 
 * for collection by the application. All messages are instances of the 
 * Message class or its subclasses.</p> 
 *
 * <p>There are many other ways to structure a stream. To aid in building the 
 * most usual structures, the following task classes are provided:</p> 
 *
 * <ul>
 *   <li>{@link PassThroughTask} - received messages are sent straight to the 
 *       next task on that side of the stream.</li>
 *   <li>{@link SystemPassThroughTask} - system messages are sent straight to 
 *       the next task on that side of the stream. User-defined message types 
 *       are discarded.</li>
 *   <li>{@link CrossOverTask} - received messages are sent straight to the 
 *       task's counterpart on the other side of the stream.</li>
 *   <li>{@link HoldingTask} - received messages are held for collection.</li>
 *   <li>{@link TerminatingTask} - received messages are discarded.</li> 
 * </ul> 
 *
 * <p>These are all subclasses of {@link RoutingTask}. Applications also have 
 * the option of defining any custom routing tasks they might need.</p> 
 * 
 * <p>In the stream model, the primary purpose of a routing task is to route 
 * messages around a stream without processing any data they might be 
 * carrying. If the content of a message should be processed, a subclass 
 * of {@link ProcessingTask} should be used, as that class provides a 
 * dedicated processing thread for every instance, along with hooks for 
 * defining which messages will be processed in what way.</p> 
 *
 * <p>There is no particular requirement to adhere to the routing/processing 
 * division of tasks, although this structure is suggested as a relatively 
 * easy way to think of things. The only concrete requirement on a task is 
 * that it implement the {@link Task} interface. It is even possible to use 
 * a task on a standalone basis - without any other stream infrastructure - 
 * as long as all methods are implemented in a way that supports this.</p> 
 *
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class Stream
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(Stream.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "Stream"; 

    /** True if this stream has been opened by calling open(), and close() 
     * has not been called since the last time open() was called. 
     */ 
    private boolean streamOpen = false; 

    /** The list of modules which form this stream. */ 
    private final LinkedList<Module> modules = new LinkedList<Module>(); 

    /** Constructs a new Stream. */
    public Stream()
    {
        final String methodLogName = classLogName + "::Stream()"; 

        logger.trace(methodLogName); 
    }

    /** Performs any operations required to start/open/prepare this stream 
     * for use after it has been constructed. For example, open() will be 
     * called on each module in the stream. 
     */
    public final synchronized void open()
    {
        final String methodLogName = classLogName + "::open()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - opening stream."); 

        // Open each module in the module list. 
        final Iterator<Module> currentModule = modules.iterator(); 

        while(currentModule.hasNext())
        {
            currentModule.next().open(); 
        }

        streamOpen = true; 
    }

    /** Returns true if this stream is open. 
     *
     * @return true if this stream is open, false otherwise. */
    public final boolean isOpen()
    {
        final String methodLogName = classLogName + "::isOpen()"; 

        logger.trace(methodLogName); 

        return(streamOpen); 
    }

    /** Performs any operations required to stop/close/prevent use of this 
     * stream. For example, close() will be called on each module in the 
     * stream. 
     */
    public final synchronized void close()
    {
        final String methodLogName = classLogName + "::close()"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName + " - closing stream."); 

        // Close each module in the module list. 
        final Iterator<Module> currentModule = modules.iterator(); 

        while(currentModule.hasNext())
        {
            currentModule.next().close(); 
        }

        streamOpen = false; 
    }

    /** Restarts this stream by calling close(), then open(). */ 
    public final synchronized void restart()
    {
        final String methodLogName = classLogName + "::restart()"; 

        logger.trace(methodLogName);
        logger.debug(methodLogName + " - restarting stream."); 

        close(); 
        open(); 
    }

    /** Pushes the specified module into the head position of the stream. Any 
     * existing modules are pushed downstream to make space for the new one. 
     *
     * @throws AssertionError if called while this stream is open. 
     * 
     * @param module the module to push. 
     */
    public final synchronized void push(final Module module)
    {
        final String methodLogName = classLogName + "::push()"; 

        logger.trace(methodLogName); 

        // Ensure that the stream is closed. 
        if(streamOpen)
        {
            final String logMessage = 
                methodLogName 
                + " - stream must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        if(!modules.isEmpty()) 
        {
            // There is already at least one module in the stream; get the 
            // current head module. 
            final Module oldHead = modules.getFirst(); 

            // Connect the new head module upstream of the old one. 
            logger.debug(methodLogName + " - connecting new head module" 
                         + " upstream of existing head module."); 

            module.connectDownstreamModule(oldHead); 
            oldHead.connectUpstreamModule(module); 
        }

        // Push the new head module into the module list. 
        logger.debug(methodLogName 
                     + " - pushing new head module into stream."); 

        modules.addFirst(module); 
    }

    /** Pops the head module out of the stream. The module following it, if 
     * any, will become the new head. 
     *
     * @throws AssertionError if called while this stream is open. 
     * 
     * @return the popped module, or null if there were no modules in the 
     *         stream. 
     */
    public final synchronized Module pop()
    {
        final String methodLogName = classLogName + "::pop()"; 

        logger.trace(methodLogName); 

        // Ensure that the stream is closed. 
        if(streamOpen)
        {
            final String logMessage = 
                methodLogName 
                + " - stream must not be open when calling this method."; 

            logger.fatal(logMessage); 
            throw new AssertionError(logMessage); 
        }

        if(!modules.isEmpty()) 
        {
            // The stream contains at least one module; remove the head 
            // module from the module list. 
            logger.debug(methodLogName 
                         + " - popping old head module from stream."); 

            final Module oldHead = modules.removeFirst(); 

            if(!modules.isEmpty()) 
            {
                // The stream still contains at least one module; disconnect 
                // the old and new head modules from each other. 
                logger.debug(methodLogName + " - disconnecting" 
                             + " old head module from new head module."); 

                oldHead.disconnectDownstreamModule(); 
                modules.getFirst().disconnectUpstreamModule(); 
            }
            
            // Return the former head module. 
            return(oldHead); 
        }

        // There were no modules in the stream. 
        logger.debug(methodLogName + " - stream empty; no module to pop."); 

        return(null); 
    }

    /** Sends the specified message to the head module's downstream task. 
     * At least one module must be pushed into the stream before calling 
     * this method. 
     *
     * @param message the message to send. 
     */
    public void put(final Message message)
    {
        final String methodLogName = classLogName + "::put()"; 

        logger.trace(methodLogName); 

        modules.getFirst().getDownstreamTask().put(message); 

        logger.debug(methodLogName + " - message sent."); 
    }

    /** Gets a message from the head module's upstream task, if available. 
     * At least one module must be pushed into the stream before calling 
     * this method. 
     *
     * @return the next message, or null if no message was available. 
     */
    public Message get()
    {
        final String methodLogName = classLogName + "::get()"; 

        logger.trace(methodLogName); 

        final Message message = modules.getFirst().getUpstreamTask().get(); 

        if(message == null)
        {
            logger.debug(methodLogName + " - no message retrieved."); 
        }
        else
        {
            logger.debug(methodLogName + " - message retrieved."); 
        }

        return(message); 
    }

    /** Gets a message from the head module's upstream task, if available. 
     * This method will not wait beyond the specified timeout for a message. 
     * At least one module must be pushed into the stream before calling 
     * this method. 
     * 
     * @param timeout the maximum time this method should wait for a message 
     *                to become available. Specified in seconds. Set to zero 
     *                if this method should block until a message arrives. 
     *
     * @return the next message, or null if no message was available, or 
     *         null if the timeout expired. 
     */
    public Message get(final long timeout)
    {
        final String methodLogName = classLogName + "::get(long)"; 

        logger.trace(methodLogName); 

        final Message message = 
            modules.getFirst().getUpstreamTask().get(timeout); 

        if(message == null)
        {
            logger.debug(methodLogName + " - no message retrieved."); 
        }
        else
        {
            logger.debug(methodLogName + " - message retrieved."); 
        }

        return(message); 
    }
}
