package org.eso.util.stream; 

/** The Task interface describes task objects which are suitable for use in 
 * stream-based applications. 
 *
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public interface Task
{
    /** Performs any operations required to start/open/prepare this task for 
     * use after it has been constructed. For example, a task thread may be 
     * started to process messages. 
     */
    public void open(); 

    /** Returns true if this task is open, false otherwise. 
     */
    public boolean isOpen(); 

    /** Performs any operations required to stop/close/prevent use of this 
     * task. For example, a running task thread may be interrupted and stop 
     * processing messages. 
     */
    public void close(); 

    /** Restarts this task by calling close(), then open(). */ 
    public void restart(); 

    /** Gets a message from this task, if available. 
     *
     * @return the next message from this task, or null if no message was 
     *         available. 
     */
    public Message get(); 

    /** Gets a message from this task, if available. This method will not 
     * wait beyond the specified timeout for a message. 
     *
     * @param timeout the maximum time this method should wait for a message 
     *                to become available. Specified in seconds. Set to zero 
     *                if this method should block until a message arrives. 
     *
     * @return the next message from this task, or null if no message was 
     *         available, or null if the timeout expired. 
     */
    public Message get(final long timeout); 

    /** Sends a message to this task. 
     *
     * @param message a message for this task. 
     */
    public void put(final Message message); 

    /** Sends a message to the task following this task in a stream. 
     *
     * @param message the message for the next task. 
     */
    public void putNext(final Message message); 

    /** Returns a String containing the name of this task. 
     */
    public String getTaskName(); 

    /** Returns a reference to the task following this one in a stream, or 
     * null if no next task is set. 
     */
    public Task getNextTask(); 

    /** Updates the task following this one in a stream to be the specified 
     * task. The former next task is returned. 
     *
     * @param task the task that will now follow this one in the stream. Set 
     *             this to null if this task should not have a next task. 
     *
     * @return The former next task, or null if no next task was set. 
     */
    public Task setNextTask(final Task task); 

    /** Returns a reference to the module containing this task, or null if 
     * no containing module is set. 
     */
    public Module getContainingModule(); 

    /** Updates this task to have the specified containing module. The former 
     * containing module is returned. 
     * 
     * @param module the module that will now contain this task. Set this to 
     *               null if this task should not have a containing module. 
     *
     * @return The module formerly containing this task, or null if no 
     *         containing module was set. 
     */ 
    public Module setContainingModule(final Module module); 
}
