package org.eso.dfs.gui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;

/**
 * 
 * @author Nick Kornweibel
 *
 * This is the ComboBoxModel used by the RecallingComboBox widget. It is basically a list
 * that can have elements added or removed while maintaining the currently selected item.
 * 
 * The model is capable of maintaining its items in sorted order. Also there is a maximum
 * number of items it will store. This is given by the maxCapacity member, which is by
 * default 20. If a new item is added, beyond the maximum, the oldest item entered will
 * be removed.
 * 
 */
public class RecallingComboBoxModel extends DefaultComboBoxModel implements ActionListener {
    
        // true if the list is to be maintained in sorted order.
    private boolean sortList = false;
    
        // the maximum number of items the list can hold.
    private int maxCapacity = 20;

        // This is a list that maintains a history of the items selected.
        // index 0 is the latest item selected. When an item is selected, it
        // removed from its current index and placed at index 0. There are most
        // likely better implementations of this model, but the lists should be
        // of small capacity only and therefore should not be a performance problem.
    private List selectionHistory = new ArrayList();

    /**
     * Create a RecallingComboBoxModel where elements are not
     * maintained in sorted order.
     * @param sort
     */
    public RecallingComboBoxModel() {
        this(false);
    }

    /**
     * Create a RecallingComboBoxModel specifying if the elements are to be
     * maintianed sorted or not.
     * @param sort
     */
    public RecallingComboBoxModel(boolean sort) {
        sortList = sort;
    }
 
    /**
     * Constructor for RecallingComboBoxModel.
     * @param items
     */
    public RecallingComboBoxModel(Object[] items) {
        super(items);
    }

    /**
     * Constructor for RecallingComboBoxModel.
     * @param v
     */
    public RecallingComboBoxModel(Vector v) {
        super(v);
    }

    /**
     * Add a new element and record the selection of an element. This callback does
     * nothing if the event source is not a JComboBox.
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     */
    public void actionPerformed(ActionEvent e) {
        if (e.getSource() instanceof JComboBox) {
            JComboBox cb = (JComboBox) e.getSource();
            // add the combo box item to the list, if not already present
            Object o = cb.getSelectedItem();
            updateModel(o);
            cb.setSelectedItem(o);
        }
    }

    /**
     * @see javax.swing.MutableComboBoxModel#addElement(java.lang.Object)
     */
    public void addElement(Object obj) {
        updateModel(obj);
    }

    /**
     * Perform the various house keeping routines on the lists
     * after a selection has been made or the model has been 
     * modified.
     * If the newSelection is null or has a toString() value of "", the
     * newSelection is not integrated into the model.
     * @param newSelection the new item selected, possibly to be added to the model.
     */
    protected void updateModel(Object newSelection) {
        if (newSelection != null && newSelection.toString().length() > 0) {
            boolean newItem = recordSelection(newSelection);
            if (newItem) {
                trimToCapacity();
                sort();
            }
        }
    }
    
    /**
     * Update the selectionHistory list with the
     * current selected item. This means moving the
     * selected item to index 0, if already in the list,
     * or inserting the item at index 0 if not in the list.
     * @return true if the item is a new item in the list, else false.
     */
    protected boolean recordSelection(Object newSelection) {
        int idx = selectionHistory.indexOf(newSelection);
        boolean newItem = idx < 0;
        if (idx > 0) {
            // item which is in the list but
            // not at idx 0
            selectionHistory.remove(idx);
            // place it at the front of the list
            // as the most recent selection.
            selectionHistory.add(0, newSelection);
        } else if (idx < 0) {
            // item which is not currently in the list
            // place the item at the front of the list
            // as the most recent selection.
            selectionHistory.add(0, newSelection);
        }
        if (newItem) {
            // add the item to the combobox model
            super.addElement(newSelection);
        }
        return newItem;
    }
    
    /**
     * Trim the lists to the current maxCapacity value.
     */
    protected void trimToCapacity() {
        while (selectionHistory.size() > maxCapacity) {
            Object removedItem = selectionHistory.remove(selectionHistory.size()-1);
            // remove the item from the combobox model.
            if (removedItem != null) {
                super.removeElement(removedItem);
            }
        }
    }
    
    /**
     * Sort the elements in the model, if sorting is true.
     * Does nothing is sortList is false.
     */
    protected void sort() {
        if (sortList) {
            // build a list
            List l = new ArrayList();
            for (int i = 0; i < getSize(); i++) {
                l.add(this.getElementAt(i));
            }
            // sort the elements.
            Collections.sort(l, Collator.getInstance());
            // replace the elements in the combo box list
            super.removeAllElements();
            for (int i = 0; i < l.size(); i++) {
                super.addElement(l.get(i));
            }           
        }
    }

    /**
     * Returns the sortList. This is set in the constructor.
     * @return boolean
     */
    public boolean isSortList() {
        return sortList;
    }

    /**
     * Returns the maxCapacity.
     * @return int
     */
    public int getMaxCapacity() {
        return maxCapacity;
    }

    /**
     * Sets the maxCapacity. Does nothing if
     * max less than 0.
     * @param maxCapacity The maxCapacity to set
     */
    public void setMaxCapacity(int max) {
        if (max >= 0) {
            maxCapacity = max;
            trimToCapacity();
            sort();
        }
    }

    /**
     * @see javax.swing.MutableComboBoxModel#insertElementAt(java.lang.Object, int)
     */
    public void insertElementAt(Object obj, int index) {
        super.insertElementAt(obj, index);
    }

    /**
     * @see javax.swing.DefaultComboBoxModel#removeAllElements()
     */
    public void removeAllElements() {
        super.removeAllElements();
        selectionHistory.clear();
    }

    /**
     * @see javax.swing.MutableComboBoxModel#removeElement(java.lang.Object)
     */
    public void removeElement(Object obj) {
        super.removeElement(obj);
        selectionHistory.remove(obj);
    }

    /**
     * @see javax.swing.MutableComboBoxModel#removeElementAt(int)
     */
    public void removeElementAt(int index) {
        Object obj = getElementAt(index);
        super.removeElementAt(index);
        selectionHistory.remove(obj);
    }
}
