/**
 * File for Gasgano
 */

package org.eso.gasgano;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import org.eso.dfs.services.data.AssociationService;
import org.eso.gasgano.datamodel.BasicFileDescription;
import org.eso.gasgano.datamodel.BasicObservationDescription;
import org.eso.gasgano.datamodel.BasicProgramDescription;
import org.eso.gasgano.datamodel.ObsStatusDescription;
import org.eso.gasgano.datamodel.ResultsModel;
import org.eso.gasgano.datamodel.database.DBDataModel;
import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.gasgano.datamodel.filesystem.FitsFile;
import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.gui.GUIUtils;
import org.eso.gasgano.keyword.Keyword;
import org.eso.gasgano.keyword.RuleTable;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.DirUtils;
import org.eso.oca.parser.ParseException;

public class Gasgano {

	// NOTE: if you update this version number, remember to update also
	//       the one in "configure.ac".
    public static final String VERSION = "GASGANO Version: 2.4.2-beta1";
    
    private DFSDataModel dataModel = null;
    private DBDataModel dbDataModel = null;
    private ResultsModel resultsModel = null;
    private GasganoDisplay gDisplay = null;
    private String userName = null;
    private String osName = null;
    private String userHome = null;
    private boolean safeEnvironment = false;

    public void printUsageAndExit() {
        System.out.println(VERSION);
        //System.out.println("Usage: [-v] [-nodb] [<preferences_file>]");
        //System.out.println("    -v    : print this screen.");
        //System.out.println("    -nodb : turn off database access.");
        System.exit(0);
    }

    /**
     * all decisions concerning user preferences 
     * and external environment variables captured throught 
     * the launching script should be handled in this method
     **/

    private void analyzeProperties() {
        int debugLevel = 0;
        try {
            debugLevel =
                Integer.parseInt(PropertyDB.getInstance().getProperty("VERBOSE_LEVEL"));
        } catch (Exception e) {
            debugLevel = 0;
        }

        PropertyDB.getInstance().setDebugLevel(debugLevel);

        if (PropertyDB.getInstance()
            .getProperty("SHORT_FILENAME")
            .toString()
            .equals("true")) {
            OBComponent.setShortName(true);
        } else {
            OBComponent.setShortName(false);
        }

        if (PropertyDB.getInstance().getProperty("RADEC_CONVERSION").equals("true")) {
            FitsFile.setRadecConversion(true);
        } else {
            FitsFile.setRadecConversion(false);
        }
        if (safeEnvironment) {
            if (PropertyDB.getInstance()
                .getProperty("OB_STATUS_FROM_DB_REQUIRED")
                .equals("true")) {
                ObsStatusDescription obsDesc = new ObsStatusDescription();
                obsDesc.setPasswordSource(gDisplay.getPasswordSource());
                dataModel.setDefaultObservationDescription(obsDesc);
            }
            if (PropertyDB.getInstance().getProperty("DBOBS_REQUIRED").equals("true")) {
                dbDataModel.setPasswordSource(gDisplay.getPasswordSource());
            }
        } else {
            dataModel.setDefaultObservationDescription(
                new BasicObservationDescription());
        }
    }

    public Gasgano(String[] args) {
        RuleTable ruleTable = null;
        GasProp properties = null;
        StringBuffer windowTitle = new StringBuffer(VERSION);
        String ruleFile = null;
        String prefFilePath = null;
        Vector displayCols = null;
        Keyword[] cols = null;
        JFrame silentFrame = new JFrame();

        userName = System.getProperties().getProperty("user.name");
        userHome = System.getProperties().getProperty("user.home");
        osName = System.getProperties().getProperty("os.name");

        prefFilePath = userHome.concat(File.separator + ".gasganorc");

        if (userName != null) {
            windowTitle.append(" " + userName);
        }
        if (osName != null) {
            windowTitle.append(" / " + osName);
        }

        // always set the environment to "unsafe" - the "safe" mode is to be abandonded.
        GasProp.setEnvironmentUnsafe();
        for (int i = 0; i < args.length; i++) {
            
            if (args[i].equals("-v")) {
                printUsageAndExit();
            //} else if (args[i].equals("-nodb")) {
            //    GasProp.setEnvironmentUnsafe();
            } else {
                prefFilePath = args[i];
            }
        }
        
        //
        // load the preferences/properties file passed on the command line
        //
        // The checks introduced/changed below are for DFS07633
        //
        // note: the checks below are not completely error-proof as loadProperties
        //       does not really check if the file is a valid properties file.
        //       For instance, Gasgano runs fine even with an empty properties
        //       file or with a file containing ASCII garbage ! this is because
        //       there's no header or no keyword that can be used to check if
        //       the file is a valid one.
        //
        properties = new GasProp();
        final File propertiesFile = new File(prefFilePath);
        if(propertiesFile.exists()) {
            if (properties.loadProperties(prefFilePath) == false) {
    			GUIUtils.userMessage(silentFrame,
					"File " + prefFilePath + "\n" +
					"is not a valid preferences file or cannot be loaded.\n" +
					"Please check file type and access permissions.\n\n" +
					"Loading aborted.");
                System.exit(1);
            }
        } else {
			GUIUtils.userMessage(silentFrame,
				"Preferences file " + prefFilePath + " does not exist.\n" +
				"Creating " + prefFilePath + " with default values.");
			if (!properties.saveProperties(prefFilePath)) {
				GUIUtils.userMessage(silentFrame,
					"Failed to save preferences file " + prefFilePath + "\n" +
					"Please check path and directory access permissions."
					);
                System.exit(2);
			}
        }

        GasProp.setProperties(properties);
        
        safeEnvironment = GasProp.isEnvironmentSafe();

        //dataModel = new DFSDataModel();
        dataModel = DFSDataModel.getDataModel();
        // register this as an association handler.
        AssociationService.getInstance().addAssociationHandler(dataModel);
        if (safeEnvironment) {
            dbDataModel = new DBDataModel();
        }
        resultsModel = new ResultsModel();

        ruleFile = PropertyDB.getInstance().getProperty("CLASSRULE_FILE");

        // try to force the HP-UX to get the metal look and feel
        if (System.getProperty("os.name").equals("HP-UX")) {
            try {
                // String laf = UIManager.getSystemLookAndFeelClassName();
                // System.out.println("LAF: " + laf);

                UIManager.setLookAndFeel(
                    "javax.swing.plaf.metal.MetalLookAndFeel");

                // other Look and Feel classes
                // "javax.swing.plaf.metal.MetalLookAndFeel"
                // Java 1.4.2 only: "com.sun.java.swing.plaf.gtk.GTKLookAndFeel"
                // "com.sun.java.swing.plaf.windows.WindowsLookAndFeel" - JRE 1.4.2+ has XP L&F
            } catch (UnsupportedLookAndFeelException e) {
                System.out.println("Exception: " + e);
            } catch (ClassNotFoundException e) {
                System.out.println("Exception: " + e);
            } catch (InstantiationException e) {
                System.out.println("Exception: " + e);
            } catch (IllegalAccessException e) {
                System.out.println("Exception: " + e);
            }
        }

        try {
			dataModel.setClassifier(
					DFSDataModel.createClassifier(
							new FileInputStream(ruleFile)
						)
					);
		} catch (FileNotFoundException e) {
            GUIUtils.userMessage(
                    silentFrame,
                    "Unable to access rules file [" + ruleFile + "].\n" + e.getMessage());
		} catch (ParseException e) {
			GUIUtils.userError(
					silentFrame,
					"Rule file [" + ruleFile + "] contains errors: \n\n" + e.getMessage() +
					"\nPlease fix the errors and restart Gasgano.");
            System.exit(1);
		} catch(Exception e) {
            GUIUtils.userMessage(
                    silentFrame,
                    "Unexpected exception while accessing rules file [" + ruleFile + "].\n" + e.getMessage());
		} catch (Error e) {
			GUIUtils.userMessage(silentFrame,
					"Unexpected error while accessing rules file [" + ruleFile
							+ "].\n" + e.getMessage());
		}
        
        dataModel.setDefaultProgramDescription(new BasicProgramDescription());

        displayCols = GasProp.getProperties().fsDisplayColumns;
        cols = new Keyword[displayCols.size()];
        for (int i = 0; i < cols.length; i++) {
            cols[i] = new Keyword((String) displayCols.elementAt(i), null);
        }
        dataModel.setDefaultFileDescription(new BasicFileDescription(cols));

        if (safeEnvironment) {
            displayCols = GasProp.getProperties().dbDisplayColumns;
            cols = new Keyword[displayCols.size()];
            for (int i = 0; i < cols.length; i++) {
                cols[i] = new Keyword((String) displayCols.elementAt(i), null);
            }
            dbDataModel.setDefaultFileDescription(
                new BasicFileDescription(cols));
        }

        dataModel.update();

        gDisplay =
            new GasganoDisplay(
                windowTitle.toString(),
                dataModel,
                dbDataModel,
                resultsModel);
        // register as a listener for property instance changes
        GasProp.addChangeListener(gDisplay);
        analyzeProperties();
        gDisplay.initializeFileSelections();
    }

    public static void main(String args[]) {
        Gasgano g = new Gasgano(args);
    }
}
// end of file Gasgano.java
