/**
 * @author N. Kornweibel
 * @version 1.0
 * An interface to filtering OBComponents based on a KeywordExpression.
 */

package org.eso.gasgano.datamodel;

import org.eso.gasgano.datamodel.filesystem.DFSFile;
import org.eso.gasgano.datamodel.filesystem.FitsFile;
import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.keyword.ExpressionEvaluationException;
import org.eso.gasgano.keyword.KeywordExpression;
import org.eso.gasgano.properties.PropertyDB;

public class ExprFilter implements OBComponentFilter {

    /**
     * The Keyword expression to use for filtering OBComponent objects.
     */
    private KeywordExpression expr = null;

    /**
     * Create an ExprFilter based on the KeywordExpression given.
     * @param exp
     */
    public ExprFilter(KeywordExpression exp) {
        expr = exp;
    }

    /**
     * @param f The OBComponent to be filtered.
     * @return true if the OBComponent passes the filter, else false.
     */
    public boolean accept(OBComponent f) {
        boolean rc = false;
        DFSFile dfsFile = f.getSourceFile();
        
        if (dfsFile instanceof FitsFile) {
            FitsFile fitsFile = (FitsFile) dfsFile;
            boolean freeMemory = true;
            if (fitsFile.headerInMemory()) {
                freeMemory = false;
            } else {
                fitsFile.readFitsHeader();
                fitsFile.addMetaKeywords();
            }
            try {
                rc = expr.evaluate(fitsFile);
            } catch (ExpressionEvaluationException e) {
                if (PropertyDB
                    .getInstance()
                    .isDebug(PropertyDB.debugExpressionAndRules)) {
                    System.out.println(e.toString() + " with: " + f.toString());
                }

                rc = false;
            }
            if (freeMemory) {
                fitsFile.freeHeaderMemory();
            }
        }
        return rc;
    }

    /**
     * @return A description of the filter.
     */
    public String getDescription() {
        return expr.toString();
    }
}
