/**
 * @author N. Kornweibel
 * @version 1.0
 * The FitsFile class for reading and displaying Fits files.
 */
package org.eso.gasgano.datamodel.filesystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.eso.dfs.util.FitsTokenizer;
import org.eso.dfs.util.KeywordListener;
import org.eso.dfs.util.KeywordToken;
import org.eso.gasgano.keyword.ClassificationRule;
import org.eso.gasgano.keyword.Keyword;
import org.eso.gasgano.keyword.KeywordBlock;
import org.eso.gasgano.keyword.KeywordSet;
import org.eso.gasgano.keyword.RuleTable;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.RADECConverter;
import org.eso.oca.fits.DataTransportFormatHandler;
import org.eso.oca.fits.FITSHandler;
import org.eso.oca.fits.FileHandlerException;
import org.eso.oca.fits.HDRHandler;
import org.eso.oca.fits.KeywordNotFoundException;
import org.eso.oca.fits.OCAFile;
import org.eso.oca.fits.OCAFileFactory;
import org.eso.oca.fits.OCAFileFilter;
import org.eso.oca.fits.TypedHeaderCard;
import org.eso.oca.fits.TypedHeaderCardException;
import org.eso.oca.fits.UndefinedValueException;
import org.eso.oca.parser.ASTStart;
import org.eso.oca.parser.InterpretationException;
import org.eso.oca.parser.OcaParser;
import org.eso.oca.parser.ParseException;

/**
 * A class for reading in FITS key words and generating a
 * Hash table for storage.
 */

public class FitsFile extends DFSFile implements KeywordSet {

	// list of keywords found in the FITS header
    private Vector keywordArray = null;

    // maps FITS keyword names to keyword values
    private Hashtable hTable = null;

    // the header extensions, if in memory
    private List extensions = null;

    private String dataAddress = null;
    private String classification = null;
    private String instrument = null;
    private String telescope = null;
    private String signature = null;
    private String mjdObs = null;

    private boolean isPipeProduct = false;
    private boolean isCompressed = false;
    private boolean hasRbInfo = false;

    private static Vector keywordFilter;

    public static long classificationTime = 0;
    public static long readHeaderTime = 0;
    public static int classificationNumber = 0;
    public static int readHeaderNumber = 0;

    private static boolean radecConvert = true;
    
    /**
     * The grouping "value" for this file.
     */
    private String groupingValue = null;
    
    /**
     * is this file a calib frame or not.
     * Note that this only determines if the file is a reduced
     * calibration product, it sets raw calib products to false.
     */
    private boolean calibFrame = false;

    /**
     * The length of a single keyword in the header.
     */
    public static final int FITS_REC_LENGTH = 80;

    /**
     * The length of a single record in a fits file. Keyword blocks
     * always start on these record boundaries. This member could also
     * have been named "BLOCK" or "BLOCKSIZE"...
     */
    public static final int BOUNDARY = 2880;
    /**
     * The length of "YYYY-MM-DD".
     */
    public static final int DATE_LENGTH = 10;

    /**
     * Used to mark the end of keyword block
     */
    public static final String endToken = "END";

    /**
     * Constructor for the FitsFile object. The file <i>address</i> is required while
     * the <i>classRules</i> and <i>requestedData</i> may be null.<p>
     * The constructor calls indirectly readData(), which determines the file status.
     * If valid and the classRules is not null, the file is classified. <p>
     * If requestedData is not null, the required keywords  are located and the values 
     * in the array set from the header data. The header data is release from memory
     * upon completion of the constructor.<p>
     *
     * @param a The file path of the Fits file.
     * @param requestedData Keyword array to be filled with values from
     * the fits file header.
     */
    public FitsFile(String address, Keyword[] requestedData) {
        init(address, requestedData);
    }

    /**
     * @param a The file path of the Fits file.
     */
    public FitsFile(String a) {
        init(a, null);
    }

    /**
     * This method calls the methods:<p>
     * determineInstrument();<br>
     * determineTelescope();<br>
     * determineGrouping();<br>
     * determineSignature();<br>
     * determineRbInfo();<br>
     * determineMjdObs();
     *<p>
     * All these methods store some properties of the 
     * FitsFile into some locale variables which can then be
     * queried by getter methods<p>
     *
     * @param a The file path of the Fits file.
     * @param t The RuleTable to be used for classifying the file.
     * @param requestedData Keyword array to be filled with values from
     * the fits file header.
     */
    private void init(String a, /*RuleTable t,*/ Keyword requestedData[]) {
        dataAddress = a;
        // set all other settings.
        readFitsHeader();
        if (getStatus() == FILE_VALID) {
            classify(DFSDataModel.getDataModel().getClassifier());
            addMetaKeywords();
            determineInstrument();
            determineTelescope();
            determineGrouping();
            determineIsCalib();
            determineSignature();
            determineRbInfo();
            determineMjdObs();
            if (requestedData != null) {
                for (int i = 0; i < requestedData.length; i++) {
                    Keyword k = getKeyword(requestedData[i].getName());
                    if (k != null) {
                        requestedData[i].setValue(k.getValue());
                    }
                }
            }
        }
        freeHeaderMemory();
    }

    /**
     * sets the application-wide flag for converting
     * RA/DEC into sessagesimal notation
     */

    static public void setRadecConversion(boolean t) {
        radecConvert = t;
    }

    /**
     * get the application wide setting as to whether RA and DEC
     * values are converted for display purposes.
     * @return boolean
     */
    static public boolean getRadecConversion() {
        return radecConvert;
    }

    /**
     * The Vector returned with getVector() will be the full header unless a Keyword filter is set. 
     * A keyword filter is a Vector of Keyword objects which determines the set of keywords to 
     * be returned from the getVector() method. It is used to retrieve a sub-set of the header keyword set.
     * This is a static method and thus the filter is set for all FitsFile objects. 
     * @param kwv The set of Keywords to be displayed in the 
     * graphic component.
     */
    public static void setKeywordFilter(Vector kwv) {
        keywordFilter = kwv;
    }

    /**
     * Get the current Keyword filter. This is a static method and therefore the filter is set
     * for all FitsFile objects.
     * @return The set of Keywords to be displayed in the 
     * graphic component.
     */
    public static Vector getKeywordFilter() {
        return keywordFilter;
    }

    /**
     * refresh the values for the keywords to be displayed.
     * @param requestedData Keyword array to be filled with values from
     * the fits file header.
     */
    public void updateKeywords(Keyword requestedData[]) {
        if (!headerInMemory()) {
            readFitsHeader();
        }
        addMetaKeywords();
        if (requestedData != null) {
            for (int i = 0; i < requestedData.length; i++) {
                Keyword k = getKeyword(requestedData[i].getName());
                if (k != null) {
                    requestedData[i].setValue(k.getValue());
                }
            }
        }
        freeHeaderMemory();
    }

    /**
     * Returns true if the header is currently in memory. The readFitsHeader() method
     *  is used to read the header into memory and freeHeaderMemory() is used to 
     * release the resources.<p>
     * @return True of the fits header is currently read into memory.
     */
    public boolean headerInMemory() {
        return (hTable != null && keywordArray != null);
    }

    /**
     * @return boolean true if the extensions are currently in memory.
     */
    public boolean extensionsInMemory() {
        return extensions != null;
    }

    private void determineSignature() {
        Keyword tmpKey = getKeyword(Keyword.MD5SIG);
        if (tmpKey == null) {
            signature = new String("");
        } else {
            signature = new String(tmpKey.getValue());
        }
    }

    private void determineInstrument() {
        Keyword tmpKey = getKeyword(Keyword.INSTRUMENT);
        if (tmpKey == null) {
            instrument = new String("");
        } else {
            instrument = new String(tmpKey.getValue());
        }
    }

    private void determineTelescope() {
        Keyword tmpKey = getKeyword(Keyword.TELESCOP);
        if (tmpKey == null) {
            telescope = new String("");
        } else {
            telescope = new String(tmpKey.getValue());
        }
    }

    /**
     * Run through the grouping keyword array and build a string representing
     * the grouping value from the keywords.
     */
    private void determineGrouping() {
        groupingValue = null;
        StringBuffer gValue = new StringBuffer();
        Keyword resKey = null;
        Keyword[] groupKey =
            GasProp.getProperties().getInstrumentGrouping(getInstrument());
        if (groupKey == null) {
            return;
        }
        for (int i = 0; i < groupKey.length; i++) {
            resKey = getKeyword(groupKey[i].getName());
            if (resKey != null) {
                if (gValue.length() > 0) {
                    gValue.append(", ");
                }
                gValue.append(groupKey[i].getName() + " = " + resKey.getValue());
            }
        }
        groupingValue = gValue.toString();
        return;
    }

    /**
     * deteremine if this file has RB info.
     */
    private void determineRbInfo() {
        String keyName = Keyword.RBPRODUCT_NAME;
        if (isPipeProduct && getKeyword(keyName) != null) {
            hasRbInfo = true;
        }
    }

    private void determineMjdObs() {
        Keyword tmpKey = getKeyword("MJD-OBS");
        if (tmpKey != null) {
            mjdObs = tmpKey.getValue();
            if (mjdObs == null) {
                mjdObs = "";
            }
        } else {
            mjdObs = "";
        }

    }
    
    /**
     * Note that this only determines if the file is a reduced
     * calibration product, it sets raw calib products to false.
     */
    private void determineIsCalib() {
        calibFrame = false;
        if (isPipeProduct) {
            Keyword tmpKey = getKeyword(Keyword.DRPCATG);
            if (tmpKey != null && tmpKey.getValue() != null
                && tmpKey.getValue().equals("CALIB")) {
                    calibFrame = true;
            }
        }
    }

    /**
     * Free the resources associated with storing the FITS file header in
     * memory. Also frees up fits extensions, if read in.
     */
    public void freeHeaderMemory() {
        // free the memory now that we have finished with it.
        hTable = null;
        keywordArray = null;
        extensions = null;
    }
    
    
    /**
     * Classify the file using the OCA rules classifier. The result of the
     * classification is placed in 'classification' and is retrievable through
     * getClassification().
     * @param classifier The OCA rules classifier.
     */
    public void classify(ASTStart classifier) {

    	// Check if the classifier is null : this can happen when the OCA rules
    	// file is invalid or could not be loaded.
    	// If it's not valid, just leave the file UNCLASSIFIED.
    	if(classifier == null) {
    		classification = "UNCLASSIFIED";
    		return;
    	}
    	
        long startTime = 0;
        boolean freeMemory = true;
        
        if (headerInMemory()) {
            freeMemory = false;
        } else {
            readFitsHeader();
        }
        
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            startTime = System.currentTimeMillis();
        }
        
        classification = null;

        // Use the OCA file factory to create the OCA file for this FITS.
        //
    	final ArrayList<OCAFile> validOCAFiles = new ArrayList<OCAFile>();
    	final ArrayList<File> skippedFiles = new ArrayList<File>(); 
        final OCAFileFactory rawFileFactory = new OCAFileFactory(dataAddress);
        // an all-pass filter for the OCA file factory: since we're passing
        // a single FITS file at a time, the filter always returns true.
        class GasganoOCAFileFilter implements OCAFileFilter {
            public boolean filter(OCAFile file) {
            	return true;
            }
        }
        final OCAFileFilter filter = new GasganoOCAFileFilter(); 
        /* DFS07535 let's add PRO.CATG to the list of required keywords */
        List<String> keywords = new ArrayList<String>(Arrays.asList(classifier.getRequestedData()));
        keywords.add("PRO.CATG");
        rawFileFactory.createOCAFiles(
			0, 
			keywords.toArray(new String[keywords.size()]),
			filter, 
            validOCAFiles, 
            skippedFiles );
		OCAFile ocaFiles[] = new OCAFile[] { validOCAFiles.get(0) };
    	
		// classify this FITS by invoking the classifier on the OCA file
		try {
			classifier.classify(ocaFiles);
		} catch (InterpretationException e) {
			System.out.println("Error during classification: " +
					e.getMessage() + "\nCause: " + e.getCause());
		}
		
		//
		// the 'DO.CATG' keyword contains the result of the classification
		//
		// ** for DFS07535:
		// in order to classify also pipeline products, the PRO.CATG keyword must be
		// taken in consideration together with DO.CATG
		//
		try {
			classification = ocaFiles[0].getKeywordValue("DO.CATG");
			if(classification == null || classification.length() == 0) {
				classification = ocaFiles[0].getKeywordValue("PRO.CATG");
			}
		} catch (Exception e) {
			//System.out.println("Classification result not available for " + dataAddress + ": " + e.getMessage());
			// DFS07535: if DO.CATG is undefined we get a KeywordNotFoundException
			try {
				classification = ocaFiles[0].getKeywordValue("PRO.CATG");
			} catch (Exception e1) {
				//System.out.println("Classification result not available for " + dataAddress + ": " + e.getMessage());
			}
		}
        
        // if something went wrong, mark the file as UNCLASSIFIED
        if (classification == null || classification.length() == 0) {
            classification = "UNCLASSIFIED";
        }
        
        if (freeMemory) {
            freeHeaderMemory();
        }
        
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            classificationTime += (System.currentTimeMillis() - startTime);
            classificationNumber++;
        }
    }

    /**
     * A product from the DFS Pipeline system is characterized by the
     * presence of the PRO.CATG keyword.
     * @return True if this is a reduced file.
     */
    public boolean isPipelineProduct() {
        return isPipeProduct;
    }

    /**
     * @return The classification of this fits file, from the
     * last execution of the classify() method
     */
    public String getClassification() {
        return classification;
    }

    /**
     * @return The instrument  this fits file was taken with
     */
    public String getInstrument() {
        return instrument;
    }

    /**
     * @return The telescope  this fits file was taken from
     */
    public String getTelescope() {
        return telescope;
    }

    /**
     * @return The Hashtable of header Keywords, if currently in memory,
     * else null.
     */
    private Hashtable getTable() {
        return hTable;
    }

    /**
     * @return The Vector of all keywords from the fits header, if
     * currently in memory, else null.
     */

    private Vector getVector() {
        return keywordArray;
    }

    /**
     * @return The status of this file after reading the header.
     * The header is released from memory after reading.
     * @see org.eso.gasgano.DFSFile
     */
    public int readData() {
        int rc = readFitsHeader();
        freeHeaderMemory();
        return rc;
    }

    /**
     * this method physically accesses the filesystem, opens
     * the file for reading and puts each line into a buffer<p>
     * The keyword END marks the end of the stream.<p>
     * Here compressed and gzipped files are taken care of.
     * @return The status of this file after reading the header.
     * The header is not released from memory after reading.
     * @see org.eso.gasgano.DFSFile
     */
    public int readFitsHeader() {
        return readFitsHeaderAndExtensions(false);
    }

    /**
     * Read the FITS header and extensions if requested. The extensions will only
     * be read if isUseMemoryMapping() is true (reading extensions without the use
     * of memory mapping is far too slow for the user).
     * @param readExtensions true if the extensions are to be read. Note that
     * even if set to true, extensions for compressed files will NOT be read in.
     * @return int the status of the file after reading. (DFSFile status).
     */
    public int readFitsHeaderAndExtensions(boolean readExtensions) {

        long startTime = 0;

        Reader br = null;

        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            startTime = System.currentTimeMillis();
        }

        Keyword metaKeyword = null;
        dataRefreshed();
        setStatus(FILE_VALID);
        boolean endParsed = false;

        // release the old header memory, if any
        freeHeaderMemory();
        // create the new header.
        hTable = new Hashtable();
        keywordArray = new Vector();

        FitsTokenizer ft = new FitsTokenizer(dataAddress, readExtensions);
        FitsFileKeywordListener kwl = new FitsFileKeywordListener(this);
        ft.addKeywordListener(kwl);

        try {
            ft.read();
            // set the names of all the extensions now.
            for (int i = 0; i < getExtensionCount(); i++) {
                setExtensionName(i);
            }
        } catch (IOException e) {
            // error reading from the Reader.
            setStatus(FILE_INVALID);
            System.out.println(
                "Error reading file: "
                    + getFileName()
                    + ", read not completed. "
                    + e.toString());
        }

        Keyword k = getKeyword(Keyword.PIPELINE_IDENT);
        if (k != null) {
            isPipeProduct = true;
        }
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            readHeaderTime += (System.currentTimeMillis() - startTime);
            readHeaderNumber++;
        }

        return getStatus();
    } // end of method readFitsData().

    /** 
     * Set the name of the specified extension, based on the standard
     * algorithm in FitsExtKeywordBlock.
     */
    private void setExtensionName(int ext) {
        FitsExtKeywordBlock kwb = (FitsExtKeywordBlock) getExtension(ext);

        String name = FitsExtKeywordBlock.generateExtensionName(kwb, ext);
        kwb.setName(name);
    }

    /**
     * adds the metakeywords to the KeywordSet
     * @see FitsFileReport and ExprFilter
     */
    public boolean addMetaKeywords() {
        boolean rc = true;
        if (headerInMemory()) {
            // add the meta keywords - these go at the end of the keywordArray
            Keyword metaKeyword =
                new Keyword(Keyword.CLASSIFICATION_METAKEYWORD, classification);
            addKeyword(metaKeyword);
            metaKeyword =
                new Keyword(Keyword.FILE_PATH_METAKEYWORD, getPathString());
            addKeyword(metaKeyword);
            metaKeyword = new Keyword(Keyword.PIPE_PRODUCT_METAKEYWORD, null);
            if (isPipelineProduct()) {
                metaKeyword.setValue("T");
            } else {
                metaKeyword.setValue("F");
            }
            addKeyword(metaKeyword);
        } else {
            rc = false;
        }
        return rc;
    }

    /**
     * @return A File object for this file.
     */
    public File getPath() {
        return new File(dataAddress);
    }

    /**
     * @return the original address with which the object was created.
     * The Path object returned by getPath() will return the resolved path
     * name, with file system links resolved.
     */
    public String getAddress() {
        return dataAddress;
    }

    /**
     * Return a Keyword object from the header, or null if not located.
     * Implementation of the KeywordSet interface.
     * @param s The name of the Keyword to find.
     * @return The Keyword located, or null if not located.
     */
    public Keyword getKeyword(String name) {
        Keyword rc = null;
        if (hTable == null) {
            readFitsHeader();
        }
        rc = (Keyword) hTable.get(name);
        return rc;
    }

    /**
     * Add a Keyword object to the set of keywords representing the header.
     * Implementation of the KeywordSet interface.
     * @param k The Keyword to be added.
     */
    public void addKeyword(Keyword k) {
        if (k != null) {
            try {
                hTable.put(k.getName(), k);
            } catch (NullPointerException ex) {
                System.out.println("error:" + ex.toString());
            }
            keywordArray.addElement(k);
        }
    }

    /**
     * Implementation of the KeywordSet interface.
     * @param k The Keyword to be deleted. This method does nothing.
     */
    public void removeKeyword(String k) {
        return;
    }

    /**
     * Implementation of the KeywordSet interface.
     * @param k The Keyword to be replaced. This method does nothing.
     */
    public void replaceKeyword(Keyword k) {
        return;
    }

    /**
     * @param key The name of the Keyword to be printed.
     */
    public void printKeyword(String key) {
        Keyword v = getKeyword(key);
        if (v != null) {
            System.out.println(v.getName() + ":\t\t\t" + v.getValue());
        } else {
            System.out.println(key + ":\t\t\tundefined");
        }
    }

    /**
     * returns the value compute in the <p>
     * determineGrouping() method, which can be null
     **/
    public String getIntermediateGrouping() {
        return groupingValue;
    }

    /** 
     * Note that this only determines if the file is a reduced
     * calibration product, it sets raw calib products to false.
     *
     * @return boolean true if this frame is a calib frame.
     * i.e. has the keyword PRO.CATG or has DPR.CATG with a value
     * "CALIB".
     */
    public boolean isCalibFrame() {
        return calibFrame;
    }

    /**
     * Add an extension to the file.
     */
    public void addExtension() {
        if (extensions == null) {
            extensions = new Vector();
        }
        FitsExtKeywordBlock b = new FitsExtKeywordBlock();
        // add the meta keyword to mark the start of the block
        int extNum = getExtensionCount() + 1;
        extensions.add(b);
    }

    /**
     * Get the primary header as a keyword blocl
     * @return KeywordBlock
     */
    public KeywordBlock getHeaderKeywordBlock() {
        return new KeywordBlock(keywordArray, hTable);
    }

    /**
     * @return int the number of extensions read in, or -1 if not in memory.
     */
    public int getExtensionCount() {
        return extensions == null ? -1 : extensions.size();
    }

    /**
     * @param ext the extension number (starting at 0).
     * @return KeywordBlock the requested extension or null if not found or not 
     * in memory.
     */
    public FitsExtKeywordBlock getExtension(int ext) {
        FitsExtKeywordBlock rc;
        if (extensions == null || ext < 0 || ext >= extensions.size()) {
            rc = null;
        } else {
            rc = (FitsExtKeywordBlock) extensions.get(ext);
        }
        return rc;
    }

    /**
     * Add the given keyword to the specified extension. If the extension does not
     * exist, nothing is done.
     * @param ext
     * @param k
     */
    public void addKeywordToExtension(int ext, Keyword k) {
        if (k == null
            || extensions == null
            || ext < 0
            || ext >= extensions.size()) {
            return;
        }
        KeywordBlock kwb = (KeywordBlock) extensions.get(ext);
        kwb.addKeyword(k);
    }

    /**
     * The signature is a MD5 checksum of the data part of a
     * Fits file<p>
     * @return the signature string read-in from the header.
     */

    public String getSignature() {
        return signature;
    }

    public String getMjdObs() {
        return mjdObs;
    }

    public boolean containsRbInfo() {
        return hasRbInfo;
    }

    /**
     * Returns the useMemoryMapping.
     * @return boolean
     */
    public static boolean isUseMemoryMapping() {
        return FitsTokenizer.isUseMemoryMapping();
    }

    /**
     * Test application to open a single FITS file.
     * @see org.eso.gasgano.DFSFile#main(String[])
     */
    public static void main(String[] args) {
        boolean summaryOnly = false;
        boolean headerOnly = false;
        if (args.length == 0) {
            System.out.println(
                "Please list fits files on command line.\n"
                    + " -summary for summary.\n"
                    + " -headeronly to avoid reading extensions.\n"
                    + " -memmap for memory mapped files.\n");
        } else {
            for (int i = 0; i < args.length; i++) {
                if (args[i].equalsIgnoreCase("-summary")) {
                    summaryOnly = true;
                    continue;
                }
                if (args[i].equalsIgnoreCase("-memmap")) {
                    FitsTokenizer.setMemoryMappingEnabled(true);
                    continue;
                }
                if (args[i].equalsIgnoreCase("-headeronly")) {
                    headerOnly = true;
                    continue;
                }
                Date start = new Date();
                FitsFile f = new FitsFile(args[i]);

                if (f.getStatus() == DFSFile.FILE_VALID) {
                    System.out.println(
                        f.getFileName() + ": " + f.getStringStatus());
                    System.out.println(
                        "Read header in "
                            + (new Date().getTime() - start.getTime())
                            + " ms.");
                    start = new Date();
                    f.readFitsHeaderAndExtensions(!headerOnly);
                    System.out.println(
                        "Read header"
                            + (headerOnly ? "" : "+extensions")
                            + " in "
                            + (new Date().getTime() - start.getTime())
                            + " ms.");
                    if (summaryOnly) {
                        System.out.println(
                            "Found " + f.getExtensionCount() + " extensions.");
                    } else {
                        Vector header = f.getVector();
                        if (header == null) {
                            System.out.println("Header not read in.");
                        } else {
                            for (Enumeration e = header.elements();
                                e.hasMoreElements();
                                ) {
                                Keyword kw = (Keyword) e.nextElement();
                                System.out.println(
                                    ""
                                        + kw.getName()
                                        + "\t\t\t"
                                        + kw.getValue());
                            }
                        }
                        if (headerOnly || f.getExtensionCount() < 0) {
                            System.out.println("Extensions not read in.");
                        } else {
                            for (int x = 0; x < f.getExtensionCount(); x++) {
                                System.out.println("Extension " + (x + 1));
                                KeywordBlock kwb = f.getExtension(x);
                                List kws = kwb.getList();
                                for (Iterator e = kws.iterator();
                                    e.hasNext();
                                    ) {
                                    Keyword kw = (Keyword) e.next();
                                    System.out.println(
                                        ""
                                            + kw.getName()
                                            + "\t\t\t"
                                            + kw.getValue());
                                }
                            }
                        }
                    }
                } else {
                    System.out.println(
                        f.getFileName() + ": " + f.getStringStatus());
                }
            }
        }
    }
} // end of class FitsFile.java

class FitsFileKeywordListener implements KeywordListener {

    // the idx of the current extension being built (primary header is idx -1)
    private int currentExtension = -1;

    /**
     * The FitsFile object manipulated by this listener.
     */
    private FitsFile ff;

    /**
     * Constructor
     * @param f the fits file to be updated by the listener.
     */
    protected FitsFileKeywordListener(FitsFile f) {
        ff = f;
    }

    /**
     * @see org.eso.dfs.util.KeywordListener#tokenRead(org.eso.dfs.util.KeywordToken)
     */
    public boolean tokenRead(KeywordToken kwt) {
        Keyword k = new Keyword(kwt);
        if (FitsFile.getRadecConversion()) {
            if (k.getName().equals("RA")) {
                k.setValue(RADECConverter.RAconvert(k.getValue()));
            }
            if (k.getName().equals("DEC")) {
                k.setValue(RADECConverter.DECconvert(k.getValue()));
            }
        }
        // note:
        // getExtension() ranges from 0..x while currentExtension goes
        // from -1..(x-1).
        if (kwt.getExtension() != (currentExtension + 1)) {
            // this is a new extension - assume it is the next
            // extension number.
            currentExtension++;
            ff.addExtension();
        }
        if (currentExtension == -1) {
            ff.addKeyword(k);
        } else {
            ff.addKeywordToExtension(currentExtension, k);
        }

        return false;
    }
}

// end of file FitsFile.java
