/**
 * @author N. Kornweibel
 * @version 1.0
 * TableModel for display of Fits Headers
 * Given a Vector of Keywords from a FitsFile object.
 * The table has three columns:
 *  0: Keyword name
 *  1: Keyword value
 *  2: Keyword extension (the extension that this keyword belongs too, starting
 *          at 0 if it is a part of the primary header). Typically not displayed.
 */
package org.eso.gasgano.datamodel.gui;

import java.util.Vector;

import javax.swing.table.AbstractTableModel;

import org.eso.gasgano.datamodel.filesystem.FitsExtKeywordBlock;
import org.eso.gasgano.datamodel.filesystem.FitsFile;
import org.eso.gasgano.keyword.Keyword;
import org.eso.gasgano.keyword.KeywordBlock;
import org.eso.gasgano.keyword.KeywordSet;

public class FitsTableModel extends AbstractTableModel {

    private String pathString;
    private DBOBComponent dComp;

    private Vector keywordList;
    private boolean filtered = false;
    
    // the fits primary header.
    private KeywordBlock header;
    // the fits extensions.
    private FitsExtKeywordBlock[] extensions;
    
    private static final String[] columns = {
        "Keyword", "Value", "Extension"
    };

    /**
     * Build a model from the fits file.
     * @param f The FitsFile to be used as the data source for this table.
     */
    public FitsTableModel(FitsFile f) {
        header = f.getHeaderKeywordBlock();
        pathString = f.getPathString();
        keywordList = FitsFile.getKeywordFilter();
        if (f.extensionsInMemory()) {
            extensions = new FitsExtKeywordBlock[f.getExtensionCount()];
            for (int i = 0; i < extensions.length; i++) {
                extensions[i] = f.getExtension(i);
            }
        }
    }

    /**
     * @param f The FitsFile to be used as the data source for this table.
     */
    public FitsTableModel(DBOBComponent d) {
        dComp = d;
        header = d.getHeaderKeywordBlock();
        keywordList = DBOBComponent.getKeywordFilter();
        pathString = dComp.getId();
    }

    /**
     * @return The file path of the fits file.
     */
    public String getFileName() {
        return pathString;
    }

    /**
     * @param c The column number, from 0.
     * @return The name of the column.
     */
    public String getColumnName(int c) {
        String rc = null;
        if (c >= 0 && c < columns.length) {
            rc = columns[c];
        }
        return rc;
    }

    /**
     * @return The number of culomns in the model (always 2).
     */
    public int getColumnCount() {
        return columns.length;
    }

    /**
     * @return The number of data rows in this model.
     */
    public int getRowCount() {
        int rc = getExtensionStartRow();
        if (extensions != null) {
            for (int i = 0; i < extensions.length; i++) {
                rc += getExtensionSize(i);
            }
        }
        return rc;
    }

    /**
     * @return the length of the extension. If filtered, this will be
     * the size of the filter, plus 1 for the extension name, otherwise it
     * is the length of the extension (which includes the exension name).
     * @param ext the extension number.
     */
    private int getExtensionSize(int ext) {
        int rc;
        if (extensions == null || extensions.length <= ext ||  ext < 0) {
            rc = 0;
        } else {
            if (isFiltered()) {
                rc = keywordList.size() + 1;
            } else {
                rc = extensions[ext].size();
            }
        }
        return rc;
    }

    /**
     * @param f Set to true if the model is to be filtered. Will only
     * filter the model if a filter has been set in the FitsFile data source.
     */
    public void setFiltered(boolean f) {
        if (f) {
            keywordList = FitsFile.getKeywordFilter();
        }
        filtered = (keywordList != null) && f;
        fireTableDataChanged();
    }

    /**
     * @return true if filtering has been set.
     */
    public boolean isFiltered() {
        return filtered;
    }

    /**
     * The last column ("Extension") will return the extension number, starting
     * at 0 for the primary header.
     * @param r The row number.
     * @param c The column number.
     * @return The Object value of the requested cell.
     */
    public Object getValueAt(int r, int c) {
        Object rc = null;
        int xsr = getExtensionStartRow();
        if (r >= xsr && extensions != null) {
            // requesting a keyword in an extension
            // r is transformed into a value relative to the
            // start of the extension blocks.
            r -= xsr;
            for (int i = 0; i < extensions.length; i++) {
                if (r < getExtensionSize(i)) {
                    if (c == 2) {
                        // return the extension number
                        rc = new Integer(i+1);  
                    } else {
                        if (r == 0 || !isFiltered()) {
                            Keyword k = extensions[i].getKeywordAt(r);
                            if (c == 0) {
                                rc = k.getName();
                            } else if (c == 1) {
                                rc = k.getValue();
                            }
                        } else {
                            // get a filtered value (r-1 is used as r == 0
                            // represents the extension name, which can not be
                            // filtered out.
                            String name = (String) keywordList.elementAt(r-1);
                            if (c == 0) {
                                rc = name;
                            } else {
                                Keyword k = extensions[i].getKeyword(name);
                                if (k == null) {
                                    rc = "<not found>";
                                } else {
                                    rc = k.getValue();
                                }
                            }
                        }
                    }
                    break;                  
                } else {
                    r -= getExtensionSize(i);
                }
            }
        } else {
            // requesting a keyword in the header.
            if (!isFiltered()) {
                if (c == 2) {
                    // return the extension number (-1 for primary header).
                    rc = new Integer(0);
                } else {
                    Keyword k = null;
                    k = header.getKeywordAt(r);
                    if (c == 0) {
                        rc = k.getName();
                    } else {
                        rc = k.getValue();
                    }
                }
            } else {
                if (c == 2) {
                    // return the extension number (-1 for primary header).
                    rc = new Integer(0);
                } else {          
                    String name = (String) keywordList.elementAt(r);
                    if (c == 0) {
                        rc = name;
                    } else {
                        Keyword k = header.getKeyword(name);
                        if (k == null) {
                            rc = "<not found>";
                        } else {
                            rc = k.getValue();
                        }
                    }
                }
            }
        }
        return rc;
    }
    
    /**
     * @return int the number of extensions, or -1 if not read in.
     */
    public int getExtensionCount() {
        return extensions == null ? -1 : extensions.length;
    }
    
    /**
     * @param x the extensions to read, starting at 0
     * @return String the name of the specified extension, or null if
     * there are no extensions of the index is out of range.
     */
    public String getExtensionName(int x) {
        String rc;
        if (extensions != null && x >=0 && x < getExtensionCount()) {
             rc = extensions[x].getName();
        } else {
            rc = null;
        }
        return rc;
    }

    /**
     * @return The row number of the start of the first extension.
     */
    private int getExtensionStartRow() {
        int rc;
        if (isFiltered()) {
            rc = keywordList.size();
        } else {
            rc = (header.size());
        }
        return rc;
    }
    
    
    /**
     * Get the start row of an extension or primary header.
     * @param i the requested block, -1 for primary header.
     * @return int the row number of the forst row of the specified block,
     * with -1 for the primary header and 0... for extensions. Return -1 if
     * invalid block requested.
     */
    public int getExtensionStartRow(int i) {
        int rc = -1;
        if (i == -1) {
            rc = 0;
        } else if (extensions != null && i >= 0 && i < extensions.length) {
            rc = getExtensionStartRow();
            i--;
            while (i >= 0) {
                rc += getExtensionSize(i);
                i--;
            }
        }
        return rc;
    }
    
    /**
     * @param r The row number of the cell to be edited.
     * @param c The column number of the cell to be edited.
     * @return Always false as the model is not editable.
     */
    public boolean isCellEditable(int r, int c) {
        return false;
    }

} // end of class FitsTableModel

// end of file FitsTableModel.java
