/**
 * 
 */
package org.eso.gasgano.gui;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JFrame;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.Element;
import javax.swing.JLabel;

import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.oca.parser.ParseException;

/**
 * @author mpadovan
 * 
 */
@SuppressWarnings("serial")
public class ClassificationRuleFileEditor extends JDialog implements ActionListener {

	private static final String ACTION_SAVE = "save";
	private static final String ACTION_CLOSE = "close";
	
	private final JTextArea textArea;
	private final JScrollPane scrollableTextArea;
	private final JButton saveButton;
	private final JButton closeButton;
	private final JLabel caretCoords;
	
	private final String ruleFile;
	private final String title;
	
	/**
	 * Used internally to change the GUI when the file is edited.
	 */
	private boolean edited = false;
	
	/**
	 * Set to true when the rule file has been edited and saved to disk.
	 */
	private boolean ruleFileChanged = false;
	

	public ClassificationRuleFileEditor(String ruleFile) throws Exception {

		this.title = "Rule File Editor - " + ruleFile;
		this.ruleFile = ruleFile;

		this.setTitle(title);
		
		String fileContent;
		try {
			fileContent = loadFile(ruleFile);
		} catch (Exception e) {
			throw new Exception("Unable to load rule file [" + ruleFile + "]: "
					+ e.getMessage());
		}
		
		this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		
		// invoked when the form is closed
		this.addWindowListener(
			new WindowAdapter() {
				public void windowClosing(WindowEvent w) {
					if( saveAndExit() ) {
						dispose();
					}
				}
			}
		);

		// Create a TextArea to display the contents of the file in
		textArea = new JTextArea(fileContent);
		scrollableTextArea = new JScrollPane(textArea);
		
		saveButton = new JButton("Save");
		closeButton = new JButton("Close");

		final boolean writable = new File(ruleFile).canWrite();
		if (!writable) {
			GUIUtils.userMessage(
				this,
				"Rule file [" + ruleFile + "] is not writable.\n" +
				"Opening in read only mode.\n" +
				"Editing and saving will not be allowed."
				);
		}
		textArea.setEditable(writable);
		
		if(writable) {
			textArea.getDocument().addDocumentListener(
				new DocumentListener() {
					@Override
					public void changedUpdate(DocumentEvent e) {
						setEdited(true);
					}
					@Override
					public void insertUpdate(DocumentEvent e) {
						setEdited(true);
					}
					@Override
					public void removeUpdate(DocumentEvent e) {
						setEdited(true);
					}
				}
			);
		}

		saveButton.setEnabled(writable);

		// add a caret listener to display the caret row and column
		caretCoords = new JLabel("Row 1, Column 1");
		textArea.addCaretListener( new CaretListener() {
			public void caretUpdate(CaretEvent e) {
				final int caretPosition = textArea.getCaretPosition();
				final Element root = textArea.getDocument().getDefaultRootElement();
				final int row = root.getElementIndex( caretPosition );
				final int column = caretPosition - root.getElement( row ).getStartOffset();
				caretCoords.setText("Row " + (row+1) + ", Column " + (column+1));
			}
		});
		
		// put the textarea in the middle
		this.add("Center", scrollableTextArea);

		// create a bottom panel to hold the buttons
		JPanel bottomPanel = new JPanel();
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.LINE_AXIS));
		this.add(bottomPanel, "South");

		// set up the label to display the caret coordinates
		bottomPanel.add(caretCoords);
		
		// add a stretchable invisible component to keep the label aligned
		// to the left and the buttons aligned to the right
		bottomPanel.add(Box.createHorizontalGlue());

		// set up the buttons
		saveButton.addActionListener(this);
		saveButton.setActionCommand(ACTION_SAVE);
		closeButton.addActionListener(this);
		closeButton.setActionCommand(ACTION_CLOSE);
		bottomPanel.add(saveButton);
		bottomPanel.add(closeButton);
		
		setEdited(false);
		
		// make the window big enough
		this.setPreferredSize(new Dimension(800, 600));
		// and don't let the user make it too small
		this.setMinimumSize(new Dimension(320,240));
		this.pack();
		
		this.isRuleFileCorrect();
	}

	public void actionPerformed(ActionEvent e) {
		final String action = e.getActionCommand();
		if(action.equals("save")) {
			if(isRuleFileCorrect()) {
				saveFile();
			}
		} else if(action.equals(ACTION_CLOSE)) {
			if( saveAndExit() ) {
				// close the window
				this.dispose();
			}
		}
	}

	/**
	 * @return True if the window has to be closed.
	 */
	private boolean saveAndExit() {
		
		boolean exit = false;
		
		if(edited) {
			int result = JOptionPane.showConfirmDialog(this, "The rule file has been changed.\n\nWould you like to save the changes?");
			if(result == JOptionPane.YES_OPTION) {
				if(isRuleFileCorrect()) {
					final boolean successfullySaved = saveFile();
					if(successfullySaved) {
						exit = true;
					}
				}
			} else if(result == JOptionPane.NO_OPTION) {
				exit = true;
			}
		} else {
			exit = true;
		}
		
		return exit;
	}
	
	private static String loadFile(String filename)
			throws FileNotFoundException, IOException {
		File file = new File(filename);
		FileReader fileReader = new FileReader(filename);
		BufferedReader bufferedReader = new BufferedReader(fileReader);
		StringBuilder fileContent = new StringBuilder((int) file.length());
		String line;
		while ((line = bufferedReader.readLine()) != null) {
			fileContent.append(line + "\n");
		}
		bufferedReader.close();
		return fileContent.toString();
	}
	
	private boolean saveFile() {
		try {
			Writer writer = new BufferedWriter(new FileWriter(this.ruleFile));
			writer.write(this.textArea.getText());
			writer.close();
			this.setEdited(false);
			ruleFileChanged = true;
			return true;
		} catch (IOException ex) {
			GUIUtils.userMessage(
					this,
					"Unable to save file [" + this.ruleFile + "].\n\n" +
					ex.getMessage()
					);
			return false;
		}
	}
	
	private void setEdited(boolean edited) {
		this.edited = edited;
		this.setTitle(title + (edited ? "*" : ""));
		this.saveButton.setEnabled(edited);
	}
	
	private boolean isRuleFileCorrect() {
		
		boolean isCorrect = false;
		
		try {
			DFSDataModel.createClassifier(new ByteArrayInputStream(textArea.getText().getBytes()));
			isCorrect = true;
		} catch (ParseException e) {
			GUIUtils.userError(this, "The rule file contains errors: \n\n" + e.getMessage());
		} catch(Exception e) {
            GUIUtils.userMessage(
                    this,
                    "Unexpected exception while accessing rules file [" + ruleFile +
                    "].\n" + e.getMessage());
		} catch (Error e) {
			GUIUtils.userMessage(
					this,
					"Unexpected error while accessing rules file [" + ruleFile
					+ "].\n" + e.getMessage());
		}
		
		return isCorrect;
	}

	public boolean isRuleFileChanged() {
		return ruleFileChanged;
	}

}
