package org.eso.cpl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * A constraint on a value which requires it to assume one of a fixed
 * set of values.
 *
 * @author   Mark Taylor
 */
public class EnumConstraint implements ParameterConstraint {

    private List enum_;
    private String description_;

    /**
     * Constructs a new EnumConstraint from a Collection.
     *
     * @param  enum  a collection of objects in which a value must appear
     *         in order to count as legal (the <tt>equals</tt> method is 
     *         used to assess equivalence)
     */
    public EnumConstraint(Collection myEnum) {
        enum_ = new ArrayList(myEnum);

        // Assemble a user-readable description of this constraint
        StringBuffer descrip = new StringBuffer();
        for (Iterator it = myEnum.iterator(); it.hasNext();) {
            descrip.append(it.next());
            if (it.hasNext()) {
               descrip.append(", ");
            }
        }
        description_ = descrip.toString();
    }

    /**
     * Constructs a new EnumConstraint from an array.
     *
     * @param  enum  an array of objects in which a value must appear
     *         in order to count as legal (the <tt>equals</tt> method is 
     *         used to assess equivalence)
     */
    public EnumConstraint(Object[] myEnum) {
        this(Arrays.asList(myEnum));
    }

    /**
     * Returns allowed options for this constraint.
     *
     * @return  an unmodifiable list of the enumerated options 
     */
    public List getOptions() {
        return Collections.unmodifiableList(enum_);
    }

    public void checkValue(Object val) throws ParameterValueException {
        if (!enum_.contains(val)) {
            throw new ParameterValueException(val + 
                                              " is not one of the options");
        }
    }

    public String getDescription() {
        return description_;
    }

    public String toString() {
        return description_;
    }
}
