package org.eso.cpl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.eso.cpl.jni.JNIRecipe;
import org.eso.cpl.jni.LibraryLoader;

/**
 * Represents a request to execute a recipe with given parameters etc.
 * It holds state
 * including the recipe object itself, its parameters with a given set
 * of values, and a product root location.
 * <p>
 * A Request can be interrogated, serialized and deserialized from XML,
 * and serialized to a shell script.
 *
 * @author   Mark Taylor
 */
public class Request {

    /**
     * The current state of the request. This is stored in the
     * state property.
     */
    public static final int UNKNOWN  = -1;
    public static final int PENDING   = 0;
    public static final int EXECUTING = 1;
    public static final int SUCCESS   = 2;
    public static final int FAILURE   = 3;
    
    
    
    private final Recipe recipe_;
    private final Parameter[] params_;
    private final MarkedFrameList inFrames_;
    private File productRoot_;
    private NamingScheme namer_;
    
    private int state = PENDING;
    
    /**
     * The default value for logging messages property.
     */
    private static boolean defaultLogMessages_ = true;
    
    /**
     * Should the request generate a text file
     */
    private boolean logMessages_ = defaultLogMessages_;

    /**
     * Constructs a new Request.  Deep copies of the constructor parameters
     * are made, so that the resulting Request has no dependence on any
     * other mutable object.
     *
     * @param  recipe  the recipe to execute
     * @param  params  the parameter list
     * @param  inFrames  the input frame list on which <tt>recipe</tt> will 
     *         operate
     * @param  productRoot  the directory for output frames
     * @param  namer  naming scheme used for output frames
     * @throws IllegalArgumentException if the parameter list isn't of a 
     *         suitable type for the recipe
     */
    public Request(Recipe recipe, Parameter[] params, MarkedFrameList inFrames, 
                   File productRoot, NamingScheme namer) {

        // Take a copy of the recipe.
        recipe_ = recipe;

        // Take a copy of the product root directory.
        productRoot_ = productRoot;

        // Take a copy of the product naming scheme.
        namer_ = namer;

        // Make a deep copy of the supplied parameter list.
        if (recipe != null) {
            params_ = recipe_.getDefaultParameters();
            int nparam = params.length;
            for (int i = 0; i < nparam; i++) {
                if (!params_[i].getName().equals(params[i].getName())) {
                    throw new IllegalArgumentException(
                        "Parameter list mismatch at " + i + ": " + params[i] +
                        " != " + params_[i]);
                }
                try {
                    params_[i].setValue(params[i].getValue());
                }
                catch (ParameterValueException e) {
                    throw (IllegalArgumentException) 
                          new IllegalArgumentException("Must be a parameter " +
                                                       "list mismatch")
                         .initCause(e);
                }
            }
        }
        else {
            params_ = null;
        }

        // Make a deep copy of the supplied input frame list.
        if (inFrames != null) {
            inFrames_ = new MarkedFrameList();
            for (Iterator it = inFrames.iterator(); it.hasNext(); ) {
                Frame frm = (Frame) it.next();
                Frame newFrm = new Frame(frm);
                inFrames_.add(newFrm);
                inFrames_.setActive(newFrm, inFrames.isActive(frm));
                inFrames_.setRetained(newFrm, inFrames.isRetained(frm));
            }
        }
        else {
            inFrames_ = null;
        }
    }

    /**
     * Constructs a default request to execute a recipe.
     * Following construction the parameters have their default values
     * and there are no input frames.
     *
     * @param  recipe  recipe to execute
     */
    public Request(Recipe recipe) {
        this(recipe, recipe.getDefaultParameters(), new MarkedFrameList(), 
             null, NamingScheme.DEFAULT);
    }

    /**
     * Clone constructor - constructs a new request with the same
     * characteristics as an exising one.  Deep copies of all the mutable
     * characteristics of <tt>req</tt> are made so that the resulting
     * object has no dependence on any other object.
     *
     * @param   req   request to copy
     */
    public Request(Request req) {
        this(req.getRecipe(), req.getParameters(), req.getInputFrames(),
             req.getProductRoot(), req.getNamingScheme());
    }

    /**
     * Constructs a new Request from an input stream containing a
     * serialized state as written by the {@link #write} method.
     * The input stream is not closed.
     *
     * @param  istrm  input stream to read state from
     * @param  libLoader  a LibraryLoader object which can be used to
     *         locate PluginLibraries which may contain recipes referenced
     *         in the serialized state
     * @throws  IOException  if there is an I/O error
     */
    public Request(InputStream istrm, LibraryLoader libLoader)
            throws IOException, CPLException, ParameterValueException {
        StateLoader loader = new StateLoader(istrm, libLoader);
        recipe_ = loader.getRecipe();
        Parameter[] params = recipe_.getDefaultParameters();
        loader.setParameterValues(params);
        params_ = params;
        inFrames_ = new MarkedFrameList(loader.getInputFrames());
        String proot = loader.getProductRoot();
        productRoot_ = proot == null ? null : new File(proot);
        namer_ = loader.getNamingScheme();
    }

    /**
     * Returns the recipe which this request will execute.
     *
     * @return  recipe  the recipe
     */
    public Recipe getRecipe() { 
        return recipe_;
    }

    /**
     * Returns the parameter list used by this request.
     *
     * @return  parameter list
     */
    public Parameter[] getParameters() {
        return params_;
    }

    /**
     * Returns the input frames on which execution will operate.
     *
     * @return  input frame list
     */
    public MarkedFrameList getInputFrames() {
        return inFrames_;
    }

    /**
     * Returns the directory into which files resulting from the execution
     * will be written.
     *
     * @return  procuct root directory
     */
    public File getProductRoot() {
        return productRoot_;
    }

    /**
     * Sets the directory into which files resulting from the execution
     * will be written.
     *
     * @param  dir  product root directory
     */
    public void setProductRoot(File dir) {
        productRoot_ = dir;
    }

    /**
     * Returns the naming scheme used to name product files.
     *
     * @return  product naming scheme
     */
    public NamingScheme getNamingScheme() {
        return namer_;
    }

    /**
     * Sets the naming scheme used to name product files.
     *
     * @param  namer  naming scheme
     */
    public void setNamingScheme(NamingScheme namer) {
        namer_ = namer;
    }

    /**
     * Serializes the current state of this request to an output stream. 
     * The stream is not closed.
     *
     * @param  ostrm  stream to serialize to
     * @param  useParams  whether to write parameter values out
     * @throws  IOException  if there is I/O trouble
     */
    public void write(OutputStream ostrm) throws IOException {
        new StateSaver(recipe_, params_, inFrames_, 
                       productRoot_ == null ? null : productRoot_.toString(),
                       namer_)
            .write(ostrm);
    }

    /**
     * Appends a sequence of commands to the given command sequence object
     * which correspond to the execution of this Request object.
     *
     * @param  script  the script object to append lines to
     */
    public void appendToScript(CommandSequence script) {
        if (! (recipe_ instanceof JNIRecipe)) {
            throw new IllegalStateException("No JNI recipe");
        }
        script.blankLine();
        String recipeName = recipe_.getName();

        // Find the library directory if there is one.
        String libdir = null;
        RecipeLibrary lib = recipe_.getLibrary();
        if (lib != null) {
            String loc = lib.getLocation();
            if (loc != null) {
                File file = new File(loc);
                if (file.exists()) {
                    File dir = file.isDirectory() ? file : file.getParentFile();
                    libdir = dir.toString();
                }
            }
        }

        // Find the product root directory if there is one.
        String proot = null;
        if (productRoot_ != null && 
            productRoot_.toString().trim().length() > 0) {
            proot = productRoot_.toString();
        }

        // Set up a macro for arguments to cplexec itself.
        List cplParams = new ArrayList();
        if (libdir != null) {
            cplParams.add("--plugin-dir=" + script.quoteWord(libdir));
        }
        if (proot != null) {
            cplParams.add("--output-dir=" + script.quoteWord(proot));
        }
        script.defineMacro("cpl_params", cplParams);

        // Set up a macro for parameter values.
        List recParams = new ArrayList();
        String paramMacro = recipeName + "_params";
        for (int i = 0; i < params_.length; i++) {
            Parameter param = params_[i];
            String pname = param.getName();
            String pval = param.getValue().toString();
            recParams.add("--" + pname + "=" + script.quoteWord(pval));
        }
        script.defineMacro(paramMacro, recParams);

        // Set up a macro for the the active input frames.
        List calibFrames = new ArrayList();
        List rawFrames = new ArrayList();
        for (Iterator it = inFrames_.selectActive().iterator(); it.hasNext(); ) {
            Frame frm = (Frame) it.next();
            FrameGroup grp = frm.getGroup();
            if (grp == FrameGroup.RAW) {
                rawFrames.add(script.quoteWord(frm.getFile().toString()));
            } 
            else if (grp == FrameGroup.CALIB) {
                calibFrames.add(script.quoteWord(frm.getFile().toString()));
            }
        }
        script.defineMacro("raw_frames", rawFrames);
        script.defineMacro("calib_frames", calibFrames);

        // Write the cplexec command.
        script.invokeCommand("cplexec " +
                             script.useMacro("cpl_params") + " " +
                             recipeName + " " +
                             script.useMacro("calib_frames") + " " +
                             script.useMacro("raw_frames") + " " +
                             script.useMacro(paramMacro));
    }
    
    
    /**
     * Create a text report of the request, including parameters, product
     * root and input frames.
     *
     * @param  script  the script object to append lines to
     */
    public String toReportString() {
        StringBuffer rc = new StringBuffer();
        rc.append("Recipe: "+recipe_.getName()+"\n");

        // Find the library directory if there is one.
        String libdir = null;
        RecipeLibrary lib = recipe_.getLibrary();
        if (lib != null) {
            String loc = lib.getLocation();
            if (loc != null) {
                File file = new File(loc);
                if (file.exists()) {
                    File dir = file.isDirectory() ? file : file.getParentFile();
                    libdir = dir.toString();
                }
            }
        }

        // Find the product root directory if there is one.
        String proot = null;
        if (productRoot_ != null && 
            productRoot_.toString().trim().length() > 0) {
            proot = productRoot_.toString();
        }

        // add all the parameters
        if (libdir != null) {
            rc.append("Recipe location: "+libdir+"\n");
        }
        if (proot != null) {
            rc.append("Output directory: "+proot+"\n");
        }
        rc.append("Parameters:\n");
        for (int i = 0; i < params_.length; i++) {
            Parameter param = params_[i];
            String pname = param.getName();
            String pval = param.getValue().toString();
            rc.append("  "+pname+"="+pval+"\n");
        }
        rc.append("Input Frames:\n");
        for (Iterator it = inFrames_.selectActive().iterator(); it.hasNext(); ) {
            rc.append("  "+((Frame)it.next()).toString()+"\n");
        }
        return rc.toString();
    }
    
    /**
     * Returns the current state of the request.
     * The state is a transient property, and is not saved in the
     * write() method.
     * @return int
     */
    public int getState() {
        return state;
    }

    /**
     * Sets the current state of the request.
     * The state is a transient property, and is not saved in the
     * write() method.
     * @param state The state to set
     */
    public void setState(int state) {
        this.state = state;
    }

}
