package org.eso.cpl;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.eso.cpl.CPLException;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameLevel;
import org.eso.cpl.FrameList;
import org.eso.cpl.FrameType;
import org.eso.cpl.Parameter;
import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Class which deals with storing and serializing the state of a
 * recipe execution.  The output format is XML.
 * <p>
 * In general, not all state of the items is serialized into the XML,
 * only things which might be useful in order to instantiate/configure
 * an object when it is deserialized.  So, for instance, the description
 * and default values of a Parameter are not serialized.
 * <p>
 * The format of the written XML errs slightly on the side of too much
 * structure rather than too little; for instance the &lt;parameter&gt;
 * elements are stored inside a grouping &lt;parameters&gt; element rather
 * than bare inside the &lt;recipe&gt;.  This is intended to allow for
 * future generalisation - for instance you might want a number of
 * different (alternative) sets of parameters, each identified by
 * some sort of tag, to be associated with the same recipe.
 * <p>
 * Doing it all in one place is not the only way of tackling this problem.
 * Alternatives would be to let the objects themselves handle
 * their own (de)serialization in a recursive fashion, or to use the
 * existing <tt>XMLEncoder/XMLDecoder</tt> classes in the <tt>java.beans</tt>
 * package.
 * However in view of the fact that the objects to (de)serialise are not
 * all that numerous or complicated, and for reasons connected with
 * the fact that several of the classes do not (for good reasons)
 * have public constructors, I think it's less trouble doing it this way.
 *
 * @author   Mark Taylor
 * @see    StateLoader
 * @see    "cpl-state.dtd"
 */
class StateSaver {

    private final Document doc_;

    /**
     * Constructs a new saver and initializes it from given state elements.
     *
     * @param  recipe  the recipe 
     * @param  params  array of Parameter objects
     * @param  inFrames  the input frame list
     * @param  productRoot  the product root location
     * @param  namer   the naming scheme
     */
    public StateSaver(Recipe recipe, Parameter[] params, FrameList inFrames,
                      String productRoot, NamingScheme namer) {
        try {

            // Construct a new DOM Document.
            doc_ = DocumentBuilderFactory.newInstance()
                                         .newDocumentBuilder()
                                         .newDocument();

            // Populate it with the given state.
            Element el = storeRequest(recipe, params, inFrames, productRoot,
                                      namer);
            doc_.appendChild(el);
        }

        // Shouldn't happen.
        catch (ParserConfigurationException e) {
            throw new RuntimeException("Unexpected XML exception", e);
        }
    }

    /**
     * Outputs the state with which this object was initialized to a 
     * file, serialized as in XML.
     *
     * @param ostrm  the stream to write to
     * @throws  IOException  if there is output trouble
     */
    public void write(OutputStream ostrm) throws IOException {
        try {
            Transformer trans = TransformerFactory.newInstance()
                                                  .newTransformer();
            trans.setOutputProperty(OutputKeys.METHOD, "xml");
            trans.setOutputProperty(OutputKeys.INDENT, "yes");
            trans.setOutputProperty("{http://xml.apache.org/xslt}indent-amount",
                                    "2");
            trans.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            trans.transform(new DOMSource(doc_), new StreamResult(ostrm));
        }

        // Shouldn't happen.
        catch (TransformerException e) {
            throw new RuntimeException("Unexpected XML error", e);
        }
    }

    /**
     * Serializes a given execution request into a new DOM Element.
     *
     * @param  doc  DOM document which will own the new element
     * @param  recipe   recipe
     * @param  inFrames  input frame list
     * @param  productRoot   output frame product root string
     * @param  namer   naming scheme
     * @return  new DOM Element containing the serialized request.  This element
     *          has not been inserted into <tt>doc</tt>
     * @throws  CPLException  if there is trouble interrogating the recipe
     *          for its parameters
     */
    private Element storeRequest(Recipe recipe, Parameter[] params,
                                 FrameList inFrames, String productRoot,
                                 NamingScheme namer) {

        // Create and configure a library element if we can.
        Element libraryEl;
        if (recipe != null) {
            RecipeLibrary rlib = recipe.getLibrary();
            if (rlib != null) {
                libraryEl = doc_.createElement("library");
                setAttributeValue(libraryEl, "name", rlib.getName());
                setAttributeValue(libraryEl, "class", 
                                  rlib.getClass().getName());
                setAttributeValue(libraryEl, "location", rlib.getLocation());
            }
            else {
                libraryEl = null;
            }
        }
        else {
            libraryEl = null;
        }

        // Create and configure an input parameter list element.
        Element paramsEl;
        if (params != null) {
            paramsEl = doc_.createElement("parameters");
            for (int i = 0; i < params.length; i++ ) {
                paramsEl.appendChild(storeParameter(params[i]));
            }
        }
        else {
            paramsEl = null;
        }

        // Create and configure an input framelist element.
        Element inframesEl;
        if (inFrames != null) {
            inframesEl = doc_.createElement("frames");
            for (Iterator it = inFrames.iterator(); it.hasNext();) {
                inframesEl.appendChild(storeFrame((Frame) it.next()));
            }
        }
        else {
            inframesEl = null;
        }

        // Create and configure a productroot element if necessary.
        Element prootEl;
        if ((productRoot != null && !productRoot.equals(".")) ||
            namer != null) {
            prootEl = doc_.createElement("products");
            setAttributeValue(prootEl, "directory", productRoot);
            setAttributeValue(prootEl, "naming", namer.getName());
        }
        else {
            prootEl = null;
        }

        // Put them all together inside a recipe element.
        Element recipeEl;
        if (recipe != null) {
            recipeEl = doc_.createElement("recipe");
            setAttributeValue(recipeEl, "name", recipe.getName());
            setAttributeValue(recipeEl, "version",
                              Long.toString(recipe.getVersion()));
            if (libraryEl != null) {
                recipeEl.appendChild(libraryEl);
            }
            if (paramsEl != null) {
                recipeEl.appendChild(paramsEl);
            }
        }
        else {
            recipeEl = null;
        }

        // Wrap the whole lot in a request element and return.
        Element requestEl = doc_.createElement("request");
        if (recipeEl != null) {
            requestEl.appendChild(recipeEl);
        }
        if (inframesEl != null) {
            requestEl.appendChild(inframesEl);
        }
        if (prootEl != null) {
            requestEl.appendChild(prootEl);
        }
        return requestEl;
    }

    /**
     * Serializes a parameter object into a DOM Element
     *
     * @param  doc   the document
     * @param  param  the parameter to serialize
     * @return  a new element representing <tt>param</tt>
     */
    public Element storeParameter(Parameter param) {
        Element el = doc_.createElement("parameter");
        setAttributeValue(el, "name", param.getName());
        setAttributeValue(el, "value", param.getValue().toString());
        return el;
    }

    /**
     * Serializes a given Frame into a new DOM Element.
     *
     * @param  doc  the DOM document which will own the new element
     * @param  frame the Frame to serialize
     * @return  new DOM element containing the serialized frame
     */
    public Element storeFrame(Frame frame) {
        Element el = doc_.createElement("frame");
        String filename = frame.getFile().getPath();
        String tag = frame.getTag();
        FrameLevel level = frame.getLevel();
        FrameGroup group = frame.getGroup();
        FrameType type = frame.getType();
        setAttributeValue(el, "filename", filename);
        setAttributeValue(el, "tag", tag);
        setAttributeValue(el, "level", level == null ? null : level.getName());
        setAttributeValue(el, "group", group == null ? null : group.getName());
        setAttributeValue(el, "type", type == null ? null : type.getName());
        return el;
    }

    /**
     * Sets the value of an element's attribute.
     * This does two jobs additional to
     * <tt>el.setAttribute(attname,attvalue)</tt>:
     * <ol>
     * <li>It does nothing (does not add a new attribute) if
     *      <tt>attvalue</tt> is <tt>null</tt>
     * <li>It escapes any characters in <tt>attvalue</tt> which need escaping
     * </ol>
     *
     * @param  el  the element whose attribute is to be set
     * @param  attname  the name of the attribute
     * @param  attvalue  the value to which the attribute will be set, or null
     */
    private static void setAttributeValue(Element el, String attname,
                                          String attvalue) {
        if (attvalue != null) {
            el.setAttribute(attname, escapeAttributeValue(attvalue));
        }
    }

    /**
     * Escapes a string so that it is in a form suitable for use as
     * an XML attribute value.
     *
     * @param  text  the raw string
     * @return  the escaped string
     */
    private static String escapeAttributeValue(String text) {
        int leng = text.length();
        StringBuffer sbuf = new StringBuffer(leng);
        for (int i = 0; i < leng; i++) {
            char c = text.charAt(i);
            switch (c) {
                case '<':
                    sbuf.append("&lt;");
                    break;
                case '>':
                    sbuf.append("&gt;");
                    break;
                case '&':
                    sbuf.append("&amp;");
                    break;
                case '\'':
                    sbuf.append("&apos;");
                    break;
                case '"':
                    sbuf.append("&quot;");
                    break;
                default:
                    sbuf.append(c);
            }
        }
        return sbuf.toString();
    }
}
