package org.eso.cpl.gui;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.SwingUtilities;

import org.eso.cpl.Frame;
import org.eso.cpl.FrameList;
import org.eso.cpl.FrameType;
import org.eso.cpl.NamingScheme;
import org.eso.cpl.Parameter;
import org.eso.cpl.Recipe;
import org.eso.cpl.Request;
import org.eso.cpl.jni.CPLControl;
import org.eso.cpl.jni.MessageListener;

/**
 * Performs execution of an Request object with event notification.
 * It implements the MessageListener interface to listen for error messages
 * when the recipe is being executed. If an error message is recorded, the
 * request status is marked as failed.
 *
 * @author   Mark Taylor
 */
public class RequestExecutor implements MessageListener {

    private List executionListeners_ = new ArrayList();
    
    /**
     * The boolean is used to record an error during execution
     * of a recipe. This is used to mark the status of the request
     * as failed.
     */
    private boolean errorMessageRecorded;

    /**
     * Adds a listener that wants to be informed about execution events.
     *
     * @param  listener  listener to add
     */
    public void addExecutionListener(ExecutionListener listener) {
        executionListeners_.add(listener);
    }

    /**
     * Removes a listener that no longer wants to be informed about
     * execution events.
     *
     * @param  listener  listener to remove
     */
    public void removeExecutionListener(ExecutionListener listener) {
        executionListeners_.remove(listener);
    }

    /**
     * Perfores the execution of a given request.
     * When the execution is finished started and finished, 
     * listeners will be informed by receiving an ExecutionEvent.
     * The events are fired in the event dispatch thread, but this 
     * (synchronous) method may be called from any thread.
     *
     * @param  request  the request to execute
     */
    public void execute(Request request) {

        // Obtain and validate request components.
        Recipe recipe = request.getRecipe();
        Parameter[] params = request.getParameters();
        // take only active request frames.
        FrameList inFrames = request.getInputFrames().selectActive();
        File productRoot = request.getProductRoot();
        if (recipe == null) {
            request.setState(Request.FAILURE);
            throw new IllegalStateException("No recipe");
        }
        if (params == null) {
            request.setState(Request.FAILURE);
            throw new IllegalStateException("No parameters");
        }
        if (inFrames == null) {
            request.setState(Request.FAILURE);
            throw new IllegalStateException("No input frames");
        }

        request.setState(Request.EXECUTING);
        // Asynchronously notify listeners that the execution is about to start.
        final ExecutionEvent startEvt =
            new ExecutionEvent(
                this,
                request,
                null,
                ExecutionEvent.Status.STARTED,
                null);
        /**
         * All updates to the GUI should be done through the AWT thread,
         * The fireExecutionStarted() is called synchronously from this
         * thread if it is the event dispatch thread, or through invokeAndWait().
         * An exception should only occur is we are already in
         * the AWT thread (and were unable to detect the fact!).
         */
        boolean inEventThread = SwingUtilities.isEventDispatchThread();
        try {
            if (inEventThread) {
                fireExecutionStarted(startEvt);
            } else {
                SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        fireExecutionStarted(startEvt);
                    }
                });
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
        // Execute the request.
        ExecutionEvent endEvt;
        
        // set the error flag to false and set the message listener.
        errorMessageRecorded = false;
        CPLControl.addMessageListener(this);
        try {
            FrameList outFrames = recipe.execute(params, inFrames, productRoot);
            endEvt =
                new ExecutionEvent(
                    this,
                    request,
                    outFrames,
                    // dsforna DFS05101
                    // // (errorMessageRecorded ? ExecutionEvent.Status.FAILURE : ExecutionEvent.Status.SUCCESS),
                    // If the pipeline fails an exception is thrown (see JNIRecipe.execute), 
                    // so, when here, the status must be always SUCCESS:
                    ExecutionEvent.Status.SUCCESS,
                    null);

            // Try to rename the products if required.
            // Note that the ".fits" extension is "forced" if not already 
            // present - this needs to be done more intelligently.

            NamingScheme namer = request.getNamingScheme();
            for (Iterator it = outFrames.iterator(); it.hasNext();) {
                renameFrame((Frame) it.next(), namer);
            }
            // set the request status based on the reciept of an error message.
            // dsforna DFS05101
            // // request.setState(errorMessageRecorded ? Request.FAILURE: Request.SUCCESS);
            request.setState(Request.SUCCESS);
        } catch (Throwable e) {
            request.setState(Request.FAILURE);
            endEvt =
                new ExecutionEvent(
                    this,
                    request,
                    null,
                    ExecutionEvent.Status.FAILURE,
                    e);
        }
        CPLControl.removeMessageListener(this);
        /**
         * All updates to the GUI should be done through the AWT thread,
         * The fireExecutionEnded() is called synchronously from this
         * thread if it is the event dispatch thread, or through invokeAndWait().
         * An exception should only occur is we are already in
         * the AWT thread (and were unable to detect the fact!).
         */
        final ExecutionEvent endEvt1 = endEvt;
        try {
            if (inEventThread) {
                fireExecutionEnded(startEvt);
            } else {
                SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        fireExecutionEnded(endEvt1);
                    }
                });
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
    }

    /**
     * Rename the frame according to the frame and namer given. The new name
     * is set into the Frame given.
     * If the file has no extension, then a ".fits" extension is added. This is
     * a hack that needs improving.
     * A file is assumed to have no extension if it does not contain a "." 
     * in the name or the part of the filename after the "." is 6 characters
     * or longer.
     * The final solution should be to look at the type of the Frame and determine
     * the correct extension from that.
     * @param frm
     * @throws Exception on error (should a specific exception be used here?).
     */
    private void renameFrame(Frame frm, NamingScheme namer) throws Exception {
        File file1 = frm.getFile();
        File file2 = file1;
        int extidx = file1.getName().lastIndexOf(".");
        // add an extension if there is not one.
        if (extidx < 0 || (file1.getName().length()-extidx > 6)) {
            // this file has no extension - give it an extension if we can
            String ext = "";
            if (frm.getType() == FrameType.TABLE) {
                ext = ".tfits";
            } else if (frm.getType() == FrameType.IMAGE){
                ext = ".fits";
            }
            System.out.println("Product frame (proposed ext: "+ext+"): "+frm.toString());
            file2 = new File(file1.getPath() + ext);
        } else {
            System.out.println("Product frame with ext: "+frm.toString());
        }
        if (namer != null) {
            file2 = namer.transform(file2);
        }
        if (!file1.equals(file2)) {
            if (!file1.renameTo(file2)) {
                // should we use a specific exception type here?
                throw new Exception(
                    "Failed to rename product "
                        + file1.getPath()
                        + " to "
                        + file2.getPath());
            }
        }
        file2 = file2.getAbsoluteFile();
        frm.setFile(file2);
    }

    /**
     * Transmits an execution ended event to all registered listeners.
     *
     * @param   evt  the event to transmit
     */
    private void fireExecutionEnded(ExecutionEvent evt) {
        for (Iterator it = executionListeners_.iterator(); it.hasNext();) {
            ExecutionListener listener = (ExecutionListener) it.next();
            listener.executionEnded(evt);
        }
    }

    /**
     * Transmits an execution started event to all registered listeners.
     *
     * @param   evt  the event to transmit
     */
    private void fireExecutionStarted(ExecutionEvent evt) {
        for (Iterator it = executionListeners_.iterator(); it.hasNext();) {
            ExecutionListener listener = (ExecutionListener) it.next();
            listener.executionStarted(evt);
        }
    }
    /**
     * MessageListener interface.
     * @see org.eso.cpl.jni.MessageListener#logMessage(java.lang.String, java.lang.String, int)
     */
    public void logMessage(String domain, String msg, int flags) {
    }
    /**
     * MessageListener interface
     * @see org.eso.cpl.jni.MessageListener#outMessage(java.lang.String)
     */
    public void outMessage(String msg) {
    }
    /**
     * MessageListener interface
     * @see org.eso.cpl.jni.MessageListener#errMessage(java.lang.String)
     */
    public void errMessage(String msg) {
        // we got an error message.
        errorMessageRecorded = true;
    }
}
