package org.eso.cpl.jni;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eso.cpl.CPLException;

/**
 * Provides static control utilities associated with the CPL shared library.
 *
 * @author   Mark Taylor
 */
public class CPLControl {

    /** The name of the native library on which these classes depend. */
    /* public static final String NATIVE_LIBRARY = "javacpl"; */
    public static final String NATIVE_LIBRARY = "cplgasgano";

    /** Version number of the native library for which library is written. */
    public static final String JAVACPL_VERSION = "1.0";

    /**
     * Object used for synchronization of native calls.
     * Since the underlying native libraries are (probably) not written
     * with any notion of thread-safety in mind, we have to take care not
     * to invoke them from different threads simultaneously (it's static
     * variables which are the problem).  The errors thus caused by not
     * taking care in this way are most likely going to result in core dumps.
     * <p>
     * Therefore any time native code using the CPL-related libraries, is
     * invoked, it should wrap the invocation in a synchronize block
     * synchronized on this object.
     * <p>
     * I think LTDL could be used in a thread-safe way by using its 
     * mutex hooks, but the native code doesn't currently do this.
     * If the JVM itself is using LTDL that could cause problems I suppose.
     * I wonder if it is?
     */
    static final Object LOCK = new Object();

    private static List messageListeners_ = new ArrayList();

    /**
     * Static initializer loads and initializes the native library.
     */
    static {
        synchronized (CPLControl.LOCK) {
            System.loadLibrary(NATIVE_LIBRARY);
            try {
            	nativeEnsureSetup();
			} catch (CPLException e) {
				System.err.println(
						"Fatal error. Cannot ensure setup: " +
						e.getMessage() +
						"\nExecution terminated."
						);
				System.exit(1);
			}
            String nativeVersion = nativeGetVersion();
            if (!JAVACPL_VERSION.equals(nativeVersion)) {
				System.err.println(
						"Fatal error. Version mismatch: " +
						"java version = " +
                        JAVACPL_VERSION +
                        ", native version = " +
                        nativeVersion + "." +
                        "\nExecution terminated."
                        );
				System.exit(1);
            }
        }
    }

    /**
     * Ensures that the native libraries required for these classes 
     * is present and initalized.  May be called multiple times.
     */
    public static void ensureSetup() {
        // no action required - it's done in the static initializer
    }

    /**
     * Registers a new listener to be notified of messages from the CPL system.
     *
     * @param  listener  the listener to add
     */
    public static void addMessageListener(MessageListener listener) {
        messageListeners_.add(listener);
    }

    /**
     * Unregisters a listener so it will no longer be notified of messages
     * from the CPL system.
     *
     * @param  listener  the listener to remove
     */
    public static void removeMessageListener(MessageListener listener) {
        messageListeners_.remove(listener);
    }

    /**
     * Disposes of a normal (standard output) message from the CPL system
     * by offering it to all registered listeners.
     *
     * @param  msg  message text
     */
    private static void outMessage(String msg) {
        for (Iterator it = messageListeners_.iterator(); it.hasNext();) {
            MessageListener listener = (MessageListener) it.next();
            listener.outMessage(msg);
        }
    }

    /**
     * Disposes of an error message from the CPL system
     * by offering it to all registered listeners.
     *
     * @param  msg  message text
     */
    private static void errMessage(String msg) {
        for (Iterator it = messageListeners_.iterator(); it.hasNext();) {
            MessageListener listener = (MessageListener) it.next();
            listener.errMessage(msg);
        }
    }

    /**
     * Disposes of a log message from the CPL system
     * by offering it to all registered listeners.
     *
     * @param   domain  the domain of the message
     * @param   msg     the message to log
     * @param   flags   an integer flag pertaining to this message;
     *          the format of this is currently undefined
     */
    private static void logMessage(String domain, String msg, int flags) {
        for (Iterator it = messageListeners_.iterator(); it.hasNext();) {
            MessageListener listener = (MessageListener) it.next();
            listener.logMessage(domain, msg, flags);
        }
    }

    /**
     * Performs initialisation of many static constants required for
     * all the native methods in this package to work.  If there is
     * any mismatch between the native library's view of the java classes
     * and their actual state (version mismatch or programming error)
     * this method is likely to throw a <tt>LinkageError</tt> of some sort.
     *
     * @throws  LinkageError if the native library {@link #NATIVE_LIBRARY}
     *          does not match the class structure in this package
     */
    private static native void nativeEnsureSetup() throws CPLException;

    private static native String nativeGetVersion();
}
