package org.eso.cpl.jni;

/**
 * A plugin found in a library which does not correspond to a usable
 * JNIRecipe.  It has two attributes, the name under which it is stored
 * in the DLL, and a comment describing why it doesn't qualify as a recipe.
 * The reason for this is likely to be either that it is a non-recipe
 * plugin, or that its API version does not match the one that the
 * java/cpl system knows about.  Since in this latter case we cannot
 * be sure of any facts about the interface, and querying it might 
 * in principle lead to serious problems (core dump) this class
 * does not attempt to supply any more information.
 *
 * @author   Mark Taylor
 */
public class OtherPlugin {
    private final String name_;
    private final String comment_;

    /**
     * Constructs a new OtherPlugin object.
     *
     * @param  name
     * @param  comment
     */
    OtherPlugin(String name, String comment) {
        name_ = name;
        comment_ = comment;
    }

    /**
     * Returns the name under which the plugin is stored in the library.
     *
     * @return  name
     */
    public String getName() {
        return name_;
    }

    /**
     * Returns a comment describing why this plugin is not a usable recipe.
     *
     * @return  comment
     */
    public String getComment() {
        return comment_;
    }

    public String toString() {
        return name_ + " (" + comment_ + ")";
    }

    /**
     * A plugin which has an API version incompatible with this library.
     */
    static class BadAPIPlugin extends OtherPlugin {
        BadAPIPlugin(String name, int libAPI, int pluginAPI) {
            super(name, "API version mismatch: " + pluginAPI + " != " + libAPI);
        }
    }

    /**
     * A plugin which is not a recipe.
     */
    static class NonRecipePlugin extends OtherPlugin {
        NonRecipePlugin(String name) {
            super(name, "Not a recipe-type plugin");
        }
    }
}
