package org.eso.cpl.test;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.textui.TestRunner;

import org.eso.cpl.CPLException;
import org.eso.cpl.EnumConstraint;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameLevel;
import org.eso.cpl.FrameList;
import org.eso.cpl.FrameType;
import org.eso.cpl.MarkedFrameList;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterConstraint;
import org.eso.cpl.ParameterType;
import org.eso.cpl.ParameterValueException;
import org.eso.cpl.RangeConstraint;
import org.eso.cpl.Recipe;
import org.eso.cpl.Request;
import org.eso.cpl.jni.CPLControl;
import org.eso.cpl.jni.LTDLException;
import org.eso.cpl.jni.LibraryLoader;
import org.eso.cpl.jni.MessageListener;
import org.eso.cpl.jni.PluginLibrary;

/**
 * Loads a DLL, gets a recipe from it, and executes the recipe.
 * This tests most of the JNI/CPL system.
 * 
 * NOTE: as the recipes change (gimasterbias.so and gimasterflat.so,
 * some of the parameters will change and thus this test will need 
 * modification.
 * 
 * <p>
 * $Id: JNICPLTest.java,v 1.26 2008/10/01 15:49:11 mpadovan Exp $
 * @author   Mark Taylor
 */
public class JNICPLTest extends TestCase {

    public final static String DLL_NAME = "libtestgimasterbias.so";
    public final static String DLL_SUBDIR = "recipes";
    public final static String DATA_DIR = "../testdata";
    public final static String RECIPE_NAME = "gimasterbias";

    public JNICPLTest(String name) {
        super(name);
    }

    public void testCPL() throws LTDLException, CPLException, 
                                 ParameterValueException, IOException {
        LibraryLoader loader = new LibraryLoader();

        // Try and fail to load library with an invalid name.
        try {
            loader.getLibrary("libneverwas.so");
            fail();
        }
        catch (LTDLException e) {
        }

        // Try and succeed to load library with the path.
        PluginLibrary giraffeLib = getGiraffeLibrary(loader);
        assertNotNull(giraffeLib);

        // Test the library instance.
        exercisePluginLibrary(giraffeLib);
    }

    private static PluginLibrary getGiraffeLibrary(LibraryLoader loader) 
            throws LTDLException, CPLException {
        String[] pathEls = System.getProperty("java.library.path").split(":");
        for (int i = 0; i < pathEls.length; i++) {
            loader.getPath().add(pathEls[i] + File.separator + DLL_SUBDIR);
        }
        PluginLibrary giraffeLib = loader.getLibrary(DLL_NAME);
        return giraffeLib;
    }

    public static PluginLibrary getGiraffeLibrary() 
            throws LTDLException, CPLException {
        return getGiraffeLibrary(new LibraryLoader());
    }

    private void exercisePluginLibrary(PluginLibrary lib)
            throws CPLException, ParameterValueException, IOException {

        // Get the recipe list and check it looks like it should.
        Recipe[] giraffeRecipes = lib.getRecipes();
        assertNotNull(giraffeRecipes);
        assertEquals(1, giraffeRecipes.length);

        // Get the first recipe.
        Recipe rec = giraffeRecipes[0];
        assertEquals(RECIPE_NAME, rec.getName());

        // Test the recipe instance.
        exerciseRecipe(rec);
    }

    private void exerciseRecipe(Recipe recipe)
            throws CPLException, ParameterValueException, IOException {

        // Get the recipe metadata.
        String author = recipe.getAuthor();
        String copyright = recipe.getCopyright();
        String description = recipe.getDescription();
        String email = recipe.getEmail();
        String name = recipe.getName();
        String synopsis = recipe.getSynopsis();
        long version = recipe.getVersion();
        
        // Check that it looks right.
        assertEquals("Michael Kiesgen", author);
        assertTrue(copyright.indexOf("(C)") >= 0);
        assertEquals("TBD", description);
        assertTrue(email.indexOf("@eso.org") > 0);
        assertTrue(synopsis.length() > 0);
        assertEquals(105, version);

        // Get the recipe parameters.
        Parameter[] params = recipe.getDefaultParameters();

        // Examine the parameters.
        assertEquals(8, params.length);
        Set ptypes = new HashSet();
        Set pconstraints = new HashSet();
        for (int i = 0; i < params.length; i++) {
            Parameter param = params[i];
            assertEquals(params[i].getDefault(), params[i].getValue());
            exerciseParameter(param, ptypes, pconstraints);
        }

        // Check that we have found examples of the parameter types
        // we know are present in this parameter list.
        ParameterType[] knownTypes = new ParameterType[] {
            ParameterType.BOOL,
            ParameterType.DOUBLE,
            ParameterType.INT,
            ParameterType.STRING,
        };
        assertEquals(new HashSet(Arrays.asList(knownTypes)), 
                     new HashSet(ptypes));

        // Check that we have found examples of the constraint types 
        // we know are present in this parameter list.
        Class[] knownConstraints = new Class[] {
            EnumConstraint.class,
            null,
        };
        assertEquals(new HashSet(Arrays.asList(knownConstraints)),
                     new HashSet(pconstraints));

        // Reset parameter values.
        for (int i = 0; i < params.length; i++) {
            Parameter param = params[i];
            param.setValue(param.getDefault());
        }

        // Execute the recipe, checking its message output.
        TestMessageListener ml1 = new TestMessageListener("ML1");
        TestMessageListener ml2 = new TestMessageListener("ML2");
        CPLControl.addMessageListener(ml1);
        CPLControl.addMessageListener(ml2);
        executeMasterBiasRecipe(recipe);

        assertEquals(0, ml1.logMessageCount);
        assertEquals(3, ml1.outMessageCount);
        assertEquals(3, ml1.errMessageCount);
        assertEquals(ml1.logMessageCount, ml2.logMessageCount);
        assertEquals(ml1.outMessageCount, ml2.outMessageCount);
        assertEquals(ml1.errMessageCount, ml2.errMessageCount);
    }

    private void exerciseParameter(Parameter param, Set usedTypes, 
                                   Set usedConstraints)
            throws ParameterValueException {
        ParameterType type = param.getType();
        ParameterConstraint constraint = param.getConstraint();
        String context = param.getContext();
        String help = param.getHelp();
        String name = param.getName();
        String tag = param.getTag();
        Object initValue = param.getValue();

        usedTypes.add(type);
        Class constraintClass = constraint == null ? null
                                                   : constraint.getClass();
        usedConstraints.add(constraintClass);

        // Try and fail to set a null value.  Don't really care if this
        // fails with a NullPointerException or ParameterValueException.
        try {
            param.setValue(null);
            fail();
        }
        catch (NullPointerException e) {
        }
        catch (ParameterValueException e) {
        }
       
        if (constraintClass == null) {
            if (type == ParameterType.BOOL) {
                parameterSet(param, Boolean.TRUE);
                parameterSet(param, Boolean.FALSE);
                parameterSetString(param, "true", Boolean.TRUE);
                parameterSetString(param, "FaLsE", Boolean.FALSE);
                parameterBad(param, new Integer(100));
                parameterBadString(param, "not a boolean");
            }
            else if (type == ParameterType.DOUBLE) {
                parameterSet(param, new Double(-23.0));
                parameterSet(param, new Double(Double.NaN));
                parameterSetString(param, "-23", new Double(-23.0));
                parameterSetString(param, "0.25e-33", new Double(0.25e-33));
                parameterBad(param, new Integer(101));
                parameterBadString(param, "not a double");
            }
            else if (type == ParameterType.INT) {
                parameterSet(param, new Integer(5));
                parameterSetString(param, "-5", new Integer(-5));
                parameterBad(param, new Double(Math.PI));
                parameterBadString(param, "3.1415");
                parameterBadString(param, "not an integer");
            }
            else if (type == ParameterType.STRING) {
                parameterSet(param, "");
                parameterSet(param, "value");
                parameterSetString(param, "", "");
                parameterSetString(param, "value", "value");
            }
            else {
                fail();
            }
        }

        else if (constraintClass == EnumConstraint.class) {
            EnumConstraint econs = (EnumConstraint) constraint;
            for (Iterator it = econs.getOptions().iterator(); it.hasNext();) {
                parameterSet(param, it.next());
            }
            if (type == ParameterType.STRING) {
                parameterBad(param, "not an option");
            }
            else if (type == ParameterType.INT) {
                parameterBad(param, new Integer(9998));
            }
            else if (type == ParameterType.DOUBLE) {
                parameterBad(param, new Double(Math.E + 0.1));
            }
            else {
                fail();
            }
        }

        else if (constraintClass == RangeConstraint.class) {
            RangeConstraint rcons = (RangeConstraint) constraint;
            if (type == ParameterType.INT) {
                int min = ((Integer) rcons.getMinimum()).intValue();
                int max = ((Integer) rcons.getMaximum()).intValue();
                parameterSet(param, new Integer(min));
                parameterSet(param, new Integer(min + 1));
                parameterSet(param, new Integer(max - 1));
                parameterSet(param, new Integer(max));
                parameterBad(param, new Integer(min - 1));
                parameterBad(param, new Integer(max + 1));
            }
            else if (type == ParameterType.DOUBLE) {
                double min = ((Double) rcons.getMinimum()).doubleValue();
                double max = ((Double) rcons.getMaximum()).doubleValue();
                parameterSet(param, new Double(min));
                parameterSet(param, new Double(min + 0.1));
                parameterSet(param, new Double(max - 0.1));
                parameterSet(param, new Double(max));
                parameterBad(param, new Double(min - 0.1));
                parameterBad(param, new Double(min + 0.1));
            }
            else {
                fail();
            }
        }
    }

    private void executeMasterBiasRecipe(Recipe recipe)
            throws CPLException, ParameterValueException, IOException {
        Parameter[] params = recipe.getDefaultParameters();

        Parameter[] params2 = recipe.getDefaultParameters();
        assertTrue(params != params2);
        params2[0].setValue(params2[0].getValue());
        FrameList inFrames = getMasterBiasInputFrames();

        // Try to execute writing products to a non-existent directory.
        File dir = new File("Sir_Not-Appearing-In-This-Filesystem");
        assertTrue(!dir.exists());
        try {
            recipe.execute(params, inFrames, dir);
            fail();
        }
        catch (CPLException e) {
        }

        // And a non-writable one.
        dir = new File("/");
        assertTrue(!dir.canWrite());
        try {
            recipe.execute(params, inFrames, dir);
            fail();
        }
        catch (CPLException e) {
        }

        CPLDebug logger = new CPLDebug();
        File productName = new File("master_bias.fits");
        productName.delete();
        assertTrue(!productName.exists());
        FrameList outFrames = recipe.execute(params, inFrames, null);
        //
        // The execute does nothing, so there is not need to check the results.
        //
        //assertTrue(productName.exists());

        //assertEquals(outFrames,
        //             outFrames.selectByGroup(FrameGroup.PRODUCT));
        //assertEquals(outFrames.size(), 1);
        //Frame ofrm = outFrames.getFrame(0);
        //assertEquals(new File(".").getCanonicalFile(),
        //             ofrm.getFile().getCanonicalFile().getParentFile());
        //assertEquals(FrameGroup.PRODUCT, ofrm.getGroup());
        //assertEquals(FrameType.IMAGE, ofrm.getType());
        //assertEquals(null, ofrm.getLevel());
        //assertEquals("MASTER_BIAS", ofrm.getTag());

        // Try it in a different directory.
        File newDir = new File("./tmp_JNICPLTest");
        if (newDir.exists()) {
            File[] files = newDir.listFiles();
            for (int i = 0; i < files.length; i++) {
                assertTrue(files[i].delete());
            }
            assertTrue(newDir.delete());
        }
        assertTrue(newDir.mkdir());
        FrameList outFrames2 = recipe.execute(params, inFrames, newDir);
        //
        // The execute does nothing, so there is not need to check the results.
        //       
        //Frame ofrm2 = outFrames2.getFrame(0);
        //assertEquals(newDir.getCanonicalFile(),
        //             ofrm2.getFile().getCanonicalFile().getParentFile());
        //assertTrue(!newDir.delete());
        //assertTrue(ofrm2.getFile().delete());
        //assertTrue(newDir.delete());

        // Try it with different parameters.
        params = recipe.getDefaultParameters();
        Parameter cbpParam = params[7];
        assertEquals("giraffe.masterbias.crebadpixmap", cbpParam.getName());
        assertEquals(ParameterType.BOOL, cbpParam.getType());
        assertEquals(Boolean.FALSE, cbpParam.getValue());
        cbpParam.setValue(Boolean.TRUE);
        FrameList outFrames3 = recipe.execute(params, inFrames, null);
        //assertEquals(2, outFrames3.size());
    }

    public static MarkedFrameList getMasterBiasInputFrames() {
        File[] files = new File(DATA_DIR).listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.startsWith("small_GIRAF") && name.endsWith(".fits");
            }
        } );
        MarkedFrameList inFrames = new MarkedFrameList();
        for (int i = 0; i < files.length; i++ ) {
            File file = files[i];
            Frame frm = new Frame(file);
            inFrames.add(frm);
            frm.setGroup(FrameGroup.RAW);
            frm.setLevel(FrameLevel.INTERMEDIATE);
            frm.setType(FrameType.IMAGE);
            frm.setTag("BIAS");
        }
        return inFrames;
    }

    public static Request getMasterBiasRequest() {
        try {
            Request req = new Request(getGiraffeLibrary().getRecipes()[0]);
            req.getInputFrames().addAll(getMasterBiasInputFrames());
            return req;
        }
        catch (LTDLException e) {
            throw new RuntimeException(e);
        } catch (CPLException e) {
            throw new RuntimeException(e);
		}
    }

    private void parameterSet(Parameter param, Object value)
            throws ParameterValueException {
        param.setValue(value);
        assertEquals(value, param.getValue());
    }

    private void parameterSetString(Parameter param, String str, Object value)
            throws ParameterValueException {
        param.setValueFromString(str);
        assertEquals(value, param.getValue());
    }

    private void parameterBad(Parameter param, Object value) {
        try {
            param.setValue(value);
            fail(param + " erroneously accepted value " + value);
        }
        catch (ParameterValueException e) {
        }
    }

    private void parameterBadString(Parameter param, String str) {
        try {
            param.setValueFromString(str);
            fail(param + " erroneously accepted value " + str);
        }
        catch (ParameterValueException e) {
        }
    }

    public static TestSuite suite() {
        return new TestSuite(JNICPLTest.class);
    }

    public static void main(String[] args) {
        TestRunner.run(suite());
    }

    private static class TestMessageListener implements MessageListener {

        int outMessageCount;
        int errMessageCount;
        int logMessageCount;
        String name;
        
        public TestMessageListener(String nm) {
            name = nm;
        }

        public void logMessage(String domain, String msg, int flags) { 
            logMessageCount++;
            System.out.println("["+name+"] LOGMSG "+logMessageCount+": "+domain+" :: "+msg);
        }

        public void outMessage(String msg) {
            outMessageCount++;
            System.out.println("["+name+"] OUTMSG "+outMessageCount+": "+msg);
        }

        public void errMessage(String msg) {
            errMessageCount++;
            System.out.println("["+name+"] ERRMSG "+errMessageCount+": "+msg);
        }
    }
}
